////////////////////////////////////// c++ /////////////////////////////////////
//
//  Module           : Clique.cpp
//  Description      : Store cliques for exact clique separation.
//  Author           : Steffen Rebennack
//  Email            : srebenac@ix.urz.uni-heidelberg.de;
//                     steffen.rebennack@web.de
//  Copyright        : (C) 2006 by the University of Heidelberg
//  Created on       : Mon Apr 03 20:11:31 2006
//  Last modified by : -
//  Last modified on : -
//  Update count     : 0
//
////////////////////////////////////////////////////////////////////////////////
//
//  Date	Name		Changes/Extensions
//  ----	----		------------------
//
////////////////////////////////////////////////////////////////////////////////


#include "Clique.hh"
#include "Graph.hh"
#include "StableSetMaster.hh"



// -----------------------------------------------------------------------------
// --------------- M e t h o d s  ( p u b l i c ) ------------------------------
// -----------------------------------------------------------------------------


// -----------------------------------------------------------------------------
// C o n s t r u c t o r
//
// The main part of the consructor is the initialisation of the ABA_BUFFERs to
// store the clique and its hash key.
// -----------------------------------------------------------------------------
Clique::Clique(ABA_MASTER *master, Graph *theGraph, int sizeSep):
    master_(master),
    itsGraph(theGraph),
    allCliquesFound(false),
    lastClique(NULL),
    lastCliqueNew(false)
{
    maxCliquesSep    = new ABA_BUFFER<int*> (master, sizeSep);
    keyOfCliqueSep   = new ABA_BUFFER<long> (master, sizeSep);
}


// -----------------------------------------------------------------------------
// D e s t r u c t o r
// -----------------------------------------------------------------------------
Clique::~Clique() {

    for (int i=0; i<maxCliquesSep->number(); i++) {
	delete [] (*maxCliquesSep)[i];
	(*maxCliquesSep)[i] = NULL;
    }
    maxCliquesSep->clear();
    delete maxCliquesSep;

    keyOfCliqueSep->clear();
    delete keyOfCliqueSep;

    if (lastCliqueNew) {
	delete [] lastClique;
    }

} // end: destructor


// -----------------------------------------------------------------------------
// a d d M a x C l i q u e
//
// This function adds a new maximal clique to the ABA_BUFFER of the clique
// constraints of the max clique separation. It first checks if this
// clique is really a new one before it is stored.
// -----------------------------------------------------------------------------
bool Clique::addMaxClique(vector<int> *clique) {

    if (lastCliqueNew) {
	delete [] lastClique;
	lastCliqueNew = false;
    }

    int *newClique = new int[clique->size()+1];
    newClique[0]   = clique->size();
    for (int i=0; i<clique->size(); i++) {
	newClique[i+1] = (*clique)[i];
    }

    // Check if this a really a new clique
    if (partOfMaximalClique(clique)) {
        // Store the last found maximal clique.
	lastClique    = newClique;
	lastCliqueNew = true;
	return false;
    }

    // Increase the buffer if necessary.
    if (maxCliquesSep->full()) {
	maxCliquesSep->realloc(maxCliquesSep->size()*2);
	keyOfCliqueSep->realloc(maxCliquesSep->size()*2);
    }
    // Add new maximal clique.
    maxCliquesSep->push(newClique);
    // Store the last found maximal clique.
    lastClique = newClique;
    // Calculate key of this constraint and store it.
    keyOfCliqueSep->push(calculateKey(clique));

    // New maximal clique added.
    return true;

}// end: bool addMaxCliqueSep(..)


// -----------------------------------------------------------------------------
// g e t L a s t M a x C l i q u e
//
// Return last clique found by the exact max clique separation routine.
// -----------------------------------------------------------------------------
int *Clique::getLastMaxClique() const {
    return lastClique;
}


// -----------------------------------------------------------------------------
// a l l M a x i m a l C l i q u e s F o u n d
// -----------------------------------------------------------------------------
bool Clique::allMaximalCliquesFound() const {
    return allCliquesFound;
}


// -----------------------------------------------------------------------------
// s e t A l l C l i q u e s F o u n d
// -----------------------------------------------------------------------------
void Clique::setAllCliquesFound() {
    allCliquesFound = true;
}


// -----------------------------------------------------------------------------
// c h e c k M a x C l i q u e s
// 
// This function checks all stored cliques if one is violated or not.
// It returns the number of violated clique inequalities. The cliques are
// available through the pointer to a vector with name maxClique.
// -----------------------------------------------------------------------------
int Clique::checkMaxCliques(vector<int*> *maxClique, double *xVal_) const {
//cout << "Start max clique" << endl;
    int    i, j;
    int    size;
    int    numberOfMaxCliques = 0;		// Return value.
    double weight;

    for (i=0; i<maxCliquesSep->number(); i++) {
	size = (*maxCliquesSep)[i][0];
	weight = 0;
	for (j=0; j<size; j++) {
	    weight += xVal_[(*maxCliquesSep)[i][j+1]];
	}
	if (weight >= 1 + ((StableSetMaster *)master_)->m_minAbsViolationClique) {
	    maxClique->push_back((*maxCliquesSep)[i]);
	    numberOfMaxCliques++;
	}

    }

    return numberOfMaxCliques;
}// end: int checkMaxCliques();



// -----------------------------------------------------------------------------
// --------------- M e t h o d s  ( p r i v a t e ) ----------------------------
// -----------------------------------------------------------------------------


// -----------------------------------------------------------------------------
// c a l c u l a t e K e y
//
// This function calculates a key for a clique. One clique is NOT part of an
// other clique, if there is no natural number that the multiplication of one
// key is equal to the otherkey.
// -----------------------------------------------------------------------------
unsigned long Clique::calculateKey(vector<int> *clique) const {

    // The key to be calculated.
    unsigned long key = 1;

    // Calculate the key.
    for (int i=0; i<(*clique).size(); i++) {
	// Ignore the node which is zero.
	if ((*clique)[i] == 0) {
	    continue;
	}
	// Update key if possible.
	if (MAX_KEY_SIZE / (*clique)[i] > key) {
	    key *= (*clique)[i];
	}
	else {
	    key = 0;
	    break;
	}
    }

    return key;

}// end: unsigned long calculateKey(..)


// -----------------------------------------------------------------------------
// p a r t O f M a x i m a l C l i q u e
//
// This function checks if a clique is part of a maximal clique allready found.
// Therefore it is necessary that the array clique is sorted!
// -----------------------------------------------------------------------------
bool Clique::partOfMaximalClique(vector<int> *clique) const {

    int  i, j, k;		// Loop variables.
    int  numberOfCliques;	// Number of max cliques allready computed.
    bool found;			// If a vertex could not be found this has value
    				// false.

    // Calculate a key and compare it with all others.
    // If the key is a fraction of one other, than compare all
    // elements of the clique.
    unsigned long key = calculateKey(clique);
    
    // Check all cliques from the clique separation.
    numberOfCliques = maxCliquesSep->number();

    for (i=0; i<numberOfCliques; i++) {
        // If key > keyOfCliqueCycle[i] or the number of elements
	// of the clique is greater than the clique to be compared:
	// continue with next clique.
	if (((key == 0) && ((*keyOfCliqueSep)[i] != 0))
	     || clique->size() > (*maxCliquesSep)[i][0]
	    ) {
	   continue;
	}
	if ((key == 0)
	    || ((*keyOfCliqueSep)[i] == 0)
	    || (((*keyOfCliqueSep)[i] % key) == 0)
	   ) {
	    // Compare each element.
	    found = true;
	    j     = 0;
	    k     = 1;
	    while(j < clique->size()) {
   	        if(!isMemberOfClique(i, k, (*clique)[j])) {
		    found = false;
		    break;
		}
	    }
	    j++;
	}
	if (found) {
	    return true;
	}
    }

    return false;

}// end: bool partOfMaximalClique(..)


// -----------------------------------------------------------------------------
// i s M e m b e r O f C l i q u e
//
// Check if a node is member of a special clique.
// This function uses, that the cliques are sorted.
// index has to start with 1!!
// -----------------------------------------------------------------------------
bool Clique::isMemberOfClique(int index, int &i, int node) const {

    bool isElement = false;
    int  size;

    size = (*maxCliquesSep)[index][0];
    while (i < (size + 1)) {
        if (node <= (*maxCliquesSep)[index][i]) {
    	    // The clique is sorted and so we know that this node is deciding.
  	    if (node == (*maxCliquesSep)[index][i]) {
    	        isElement = true;
	    }
	    break;
        }
	i++;
    }
 
    return isElement;

}// end: bool isMemberOfClique(..)

