////////////////////////////////////// c++ /////////////////////////////////////
//
//  Module           : Clique.hh
//  Description      :
//  Author           : Steffen Rebennack
//  Email            : srebenac@ix.urz.uni-heidelberg.de;
//                     steffen.rebennack@web.de
//  Copyright        : (C) 2006 by the University of Heidelberg
//  Created on       : Mon Apr 03 20:17:39 2006
//  Last modified by : -
//  Last modified on : -
//  Update count     : 0
//
////////////////////////////////////////////////////////////////////////////////
//
//  Date	Name		Changes/Extensions
//  ----	----		------------------
//
////////////////////////////////////////////////////////////////////////////////


#ifndef CLIQUE_HH
#define CLIQUE_HH

#include "abacus/buffer.h"
#include "abacus/master.h"
#include <vector>

using namespace std;

class Graph;



/**
 * This class stores all computed maximal cliques by the separation routine
 * for maximal cliques. These maximal cliques are stored in the ABA_BUFFER
 * maxCliquesSep.
 */
class Clique {


public:

    // -------------------------------------------------------------------------
    // ------------- M e t h o d s  ( p u b l i c ) ----------------------------
    // -------------------------------------------------------------------------

    /**
     * Constructor.
     *
     * @param *master Pointer to the master of the problem.
     * @param *theGraph Pointer to the graph.
     * @param sizeSep The size of the ABA_BUFFER maxCliqueSep.
     */
    Clique(ABA_MASTER *master, Graph *theGraph, int sizeSep);

    /**
     * Destructor.
     */
    ~Clique();

    /**
     * Add a maximal clique.
     *
     * @param *clique Pointer to a vector which stores a clique.
     * @return true If this clique could be added.
     * @return false If the clique is part of a max clique allready found.
     */
    bool addMaxClique(vector<int> *clique);

    /**
     * Get last max clique found by separation.
     *
     * @param void
     * @return Pointer to the last maximal clique.
     */
    int* getLastMaxClique() const;

    /**
     * Have all maximal cliques been found?
     *
     * @param void
     * @return true If all cliques have been found.
     * @return false Otherwise.
     */
    bool allMaximalCliquesFound() const;

    /**
     * Set bool allCliquesFound to value true.
     * @param void
     * @return void
     */
    void setAllCliquesFound();

    /**
     * Check all maximal cliques if they are violated.
     *	
     * @param *maxCliques This is a pointer to a vector storing all 
     *	      violated cliques.
     * @param *xVal_ Pointer to the current LP solution.
     * @return The number of violated cliques. 
     */
    int checkMaxCliques(vector<int*> *maxClique, double *xVal_) const;


private:

    // -------------------------------------------------------------------------
    // ------------- M e t h o d s  ( p r i v a t e ) --------------------------
    // -------------------------------------------------------------------------

    /**
     * Calculate a key of a clique.
     *
     * @param *clique Pointer to a vector which stores a clique.
     * @return The key.
     */
    unsigned long calculateKey(vector<int> *clique) const;

    /**
     * Check if clique is part of a max clique.
     *
     * @param clique Pointer to a vector storing a clique.
     * @return true If clique is part of a max clique.
     * @return false Otherwise.
     */
    bool partOfMaximalClique(vector<int> *clique) const;

    /**
     * Check if node is member of index-th clique.
     *
     * @param index Index of an element of ABA_BUFFER.
     * @param &i Index of the vertex node in the clique.
     * @param node Node which is looked for.
     * @return true Node is part of this clique.
     * @return false Otherwise.
     */
    bool isMemberOfClique(int index, int &i, int node) const;

 
    // -------------------------------------------------------------------------
    // ------------- V a r i a b l e s  ( p r i v a t e ) ----------------------
    // -------------------------------------------------------------------------

    /**
     * Maximal number for the key of a clique.
     */
    const static long MAX_KEY_SIZE = 40000000;

    /**
     * Pointer to the master of the problem.
     */
    ABA_MASTER *master_;

    /**
     * Pointer to the corresponding graph of the problem instance of class
     * StableSetMaster.
     */
    Graph *itsGraph;

    /**
     * Store max cliques found by the exact max clique separation.
     * The first element of each clique stores the size of the clique. So the
     * size of the array is the size of the clique plus one.
     */
    ABA_BUFFER<int*> *maxCliquesSep;

    /**
     * Store the key of each clique of the buffer maxCliquesSep.
     */
    ABA_BUFFER<long> *keyOfCliqueSep;

    /**
     * If all max cliques have been computed, this variable is true.
     */
    bool allCliquesFound;

    /**
     * Store the last clique which has been found by the exact max
     * clique separation.
     * The size of this clique is the first element of the array which
     * is pointed at.
     */
    int *lastClique;

    /**
     * A flag which shows if pointer lastClique is initialized by the new
     * operator and therefore if it has to be deleted with [].
     */
    bool lastCliqueNew;

};

#endif


