////////////////////////////////////// c++ /////////////////////////////////////
//
//  Module           : CliqueConstraint.cpp
//  Description      : Constraint class for clique inequalities.
//  Author           : Steffen Rebennack
//  Email            : srebenac@ix.urz.uni-heidelberg.de;
//                     steffen.rebennack@web.de
//  Copyright        : (C) 2006 by the University of Heidelberg
//  Created on       : Tue Apr 04 13:45:46 2006
//  Last modified by : -
//  Last modified on : -
//  Update count     : 0
//
////////////////////////////////////////////////////////////////////////////////
//
//  Date	Name		Changes/Extensions
//  ----	----		------------------
//
////////////////////////////////////////////////////////////////////////////////

#include "abacus/row.h"
#include "CliqueConstraint.hh"
#include "Graph.hh"
#include "Node.hh"



// -----------------------------------------------------------------------------
// --------------- M e t h o d s  ( p u b l i c ) ------------------------------
// -----------------------------------------------------------------------------


// -----------------------------------------------------------------------------
// C o n s t r u c t o r
//
// This class is derived from the abstract base class ABA_CONSTRAINT.
//
// Because a clique constraint is uniquely determined by its nodes, this
// constructor saves all nodes of the clique. The main part is
// the call of the constructor of class ABA_CONSTRAINT with the following
// arguments:
// 1.) A pointer to the corresponding master of the optimization.
// 2.) A pointer to the subproblem associated with this constraint because the
//     constraints should be only active in this sub problem and its children.
// 3.) This constraint is of kind less (or equal).
// 4.) The ride hand side of this constraint is 1.
// 5.) This constraint can be removed from the active constraint set.
// 6.) This constraint is global valid and not only local.
// 7.) There is no lifting procedure implemented.
// -----------------------------------------------------------------------------
CliqueConstraint::CliqueConstraint(ABA_MASTER *master, vector<int> *clique,
				   Graph *theGraph):
    ABA_CONSTRAINT(master, NULL, ABA_CSENSE::Less, 1.0, true, false, false),
    nodesOfClique(clique->size()),
    itsGraph(theGraph)
{
    // Store this inequality that it can be checked.
    ofstream fout(theGraph->getFileName(), ios::app);

    // Save the clique inequality.
    for (int i=0; i<clique->size(); i++) {
        nodesOfClique[i] = (*clique)[i];
	fout << "  1 x_" << itsGraph->translateNode(nodesOfClique[i]);
    }
    fout << "  <  1" << endl;
    fout.close();

    // Calculate the key.
    calculateHashKey();
}


// -----------------------------------------------------------------------------
// D e s t r u c t o r
// -----------------------------------------------------------------------------
CliqueConstraint::~CliqueConstraint() {
}


// -----------------------------------------------------------------------------
// g e n R o w
//
// This function redefines the pure virtual function of the base class
// ABA_CONSTRAINT. It returns the constraint in sprase format. 
// -----------------------------------------------------------------------------
int CliqueConstraint::genRow(ABA_ACTIVE<ABA_VARIABLE,ABA_CONSTRAINT> *var, 
			     ABA_ROW &row)
{
//    row.realloc(nodesOfClique.size() + 1);
//    row.clear();
    
    // Store the indices of the nodes of the clique.
    for (int i=0; i<nodesOfClique.size(); i++) {
        row.insert(nodesOfClique[i],1.0);
    }
    // The rhs of the clique.
    row.rhs(1);
    // Sense of the inequality.
    row.sense(ABA_CSENSE::Less);
    
    return row.nnz();
}


// -----------------------------------------------------------------------------
// c o e f f
// 
// This function computes the coefficient of the variable var. It redefines the
// pure virtual function of the base class ABA_CONSTRAINT. The coefficient of
// a variable of a clique constraint is 1 if the variable is part of the clique 
// and otherwise its coefficient is 0.
// -----------------------------------------------------------------------------
double CliqueConstraint::coeff(ABA_VARIABLE *var) {

    // Cast ABA_VARIABLE* to NODE*.
    Node *node = (Node *) var;
    // search for this variable in the clique
    for (int i=0; i<nodesOfClique.size(); i++) {
        if (node->nodeNumber() == nodesOfClique[i]) {
            return 1.0;
        }
    }
    return 0.0;
}


// -----------------------------------------------------------------------------
// g e t S i z e
// -----------------------------------------------------------------------------
int CliqueConstraint::getSize() {
    return nodesOfClique.size();
}


// -----------------------------------------------------------------------------
// i s I n C o n s t r a i n t
//
// This function checks if a node is member of this constraint. It return the
// value true if the node is a memebr and otherwise false.
// This method is used by the function equal(..).
// -----------------------------------------------------------------------------
bool CliqueConstraint::isInConstraint(int node) {

    for(int i=0; i<nodesOfClique.size(); i++) {
        if (nodesOfClique[i] == node) {
            return true;
	}
    }

    return false;
}


// -----------------------------------------------------------------------------
// e q u a l
//
// This function compares if this constraint is identical with the constraint
// cv in a mathematical sense. This function is required becuase the constraint
// is stored in a pool of the class ABA_NONDUPLPOOL.
// It returns true if the constraints are equal and otherwise false.
// -----------------------------------------------------------------------------
bool CliqueConstraint::equal(ABA_CONVAR *cv) {

    bool isEqual = true;
    if (thisHashKey != ((CliqueConstraint *)cv)->hashKey()) {
        isEqual = false;
    }
    if (nodesOfClique.size() != ((CliqueConstraint *)cv)->getSize()) {
        isEqual = false;
    }

    for(int i=0; i<nodesOfClique.size(); i++) {
        if (!((CliqueConstraint *)cv)->isInConstraint(nodesOfClique[i])) {
	    isEqual = false;
	}
    }

    return isEqual;
}


// -----------------------------------------------------------------------------
// n a m e
//
// This function returns the name of the constraint. It is required if the
// constraint is stored in a pool of class ABA_NONDUPLPOOL.
// -----------------------------------------------------------------------------
const char* CliqueConstraint::name() {
    return "CliqueConstraint";
}


// -----------------------------------------------------------------------------
// h a s h K e y
//
// This method provides a key for the constraint that can be used to insert it
// into a hash table. In this case it is not required that any two items have
// different keys. This function is required to store the constraint in a pool
// of the class ABA_NIONDUPLPOOL.
// -----------------------------------------------------------------------------
unsigned CliqueConstraint::hashKey() {
    return thisHashKey;
}



// -----------------------------------------------------------------------------
// --------------- M e t h o d s  ( p r i v a t e ) ----------------------------
// -----------------------------------------------------------------------------


// -----------------------------------------------------------------------------
// c a l c u l a t e H a s h K e y
// -----------------------------------------------------------------------------
void CliqueConstraint::calculateHashKey() {

    thisHashKey = 1;
    for (int i=0; i<nodesOfClique.size(); i++) {
        thisHashKey *= (nodesOfClique[i] + 1265);
	thisHashKey = thisHashKey % 9973;
    }
}

