////////////////////////////////////// c++ /////////////////////////////////////
//
//  Module           : CliqueConstraint.hh
//  Description      : Constraint class for clique inequalities.
//  Author           : Steffen Rebennack
//  Email            : srebenac@ix.urz.uni-heidelberg.de;
//                     steffen.rebennack@web.de
//  Copyright        : (C) 2006 by the University of Heidelberg
//  Created on       : Tue Apr 04 14:00:24 2006
//  Last modified by : -
//  Last modified on : - 
//  Update count     : 0
//
////////////////////////////////////////////////////////////////////////////////
//
//  Date	Name		Changes/Extensions
//  ----	----		------------------
//
////////////////////////////////////////////////////////////////////////////////


#ifndef CLIQUE_CONSTRAINT_HH
#define CLIQUE_CONSTRAINT_HH


#include "abacus/constraint.h"
#include <vector>

using namespace std;

class Graph;


/**
 * The class CliqueConstraint is derived from the abstract ABACUS constraint
 * class ABA_CONSTRAINT. It stores a maximal clique and computes the 
 * corresponding clique constraint.
 */
class CliqueConstraint: public ABA_CONSTRAINT {


public:

    // -------------------------------------------------------------------------
    // ------------- M e t h o d s  ( p u b l i c ) ----------------------------
    // -------------------------------------------------------------------------

    /**
     * Constructor.
     *
     * @param *master Pointer to the corresponding master of the optimization.
     * @param *clique Pointer to a vector which stores a clique.
     * @param *theGraph Pointer to the object representing the graph of the
     *                  problem.
     */
    CliqueConstraint(ABA_MASTER *master, vector<int> *clique,  Graph *theGraph);

    /**
     * Destructor.
     */
    virtual ~CliqueConstraint();

    /**
     * Get constraint in sparse format.
     *
     * @param *var Pointer to a variable set.
     * @param &row Reference to abacus object storing the inequality.
     * @return Number of non-zeros in the inequality.
     */
    int genRow(ABA_ACTIVE<ABA_VARIABLE,ABA_CONSTRAINT> *var, ABA_ROW &row);

    /**
     * Get coefficient of a variable.
     *
     * @param *var Pointer to a variable.
     * @return The coefficient of the variable var in this clique constraint.
     */
    virtual double coeff(ABA_VARIABLE *var);

    /**
     * Get size of clique.
     *
     * @param void
     * @return Size of vector nodesOfClique.
     */
    int getSize();

    /**
     * Checks if node is member of this constraint.
     *
     * @param node A node.
     * @return true Node is member of this constraint.
     * @return false Otherwise.
     */
    bool isInConstraint(int node);

    /**
     * Compare if constraint cv is equal to this constraint.
     *
     * @param *cv Pointer to the constraint to be compared with this one.
     * @return true If both constraints are equal.
     * @return false Otherwise.
     */
    virtual bool equal(ABA_CONVAR *cv);

    /**
     * This method returns the name of this constraint.
     *
     * @param void
     * @return Pointer to the name of this constraint
     */
    virtual const char* name();

    /**
     * This method returns the hashkey of the constraint.
     *
     * @param void
     * @return The hashkey of the constraint.
     */
    virtual unsigned hashKey();


private:

    // -------------------------------------------------------------------------
    // ------------- M e t h o d s  ( p r i v a t e ) --------------------------
    // -------------------------------------------------------------------------

    /**
     * This method calculates the haskey of the constraint.
     *
     * @param void
     * @return void
     */
    void calculateHashKey();


    // -------------------------------------------------------------------------
    // ------------- V a r i a b l e s  ( p r i v a t e ) ----------------------
    // -------------------------------------------------------------------------

    /**
     * Node set of the clique.
     */
    vector<int> nodesOfClique;

    /**
     * Store a pointer to the object representing the graph of the problem
     * instance.
     */
    Graph *itsGraph;

    /**
     * The hashkey of this constraint.
     */
    unsigned thisHashKey;
};

#endif




