////////////////////////////////////// c++ /////////////////////////////////////
//
//  Module           : CliqueHeuristicsSeparator.cpp
//  Description      : Separate the clique inequalities with heuristics.
//  Author           : Steffen Rebennack
//  Email            : srebenac@ix.urz.uni-heidelberg.de;
//                     steffen.rebennack@web.de
//  Copyright        : (C) 2006 by the University of Heidelberg
//  Created on       : Tue Apr 04 14:16:53 2006 
//  Last modified by : -
//  Last modified on : -
//  Update count     : 0
//
////////////////////////////////////////////////////////////////////////////////
//
//  Date	Name		Changes/Extensions
//  ----	----		------------------
//
////////////////////////////////////////////////////////////////////////////////


#include "CliqueHeuristicsSeparator.hh"
#include "CliqueConstraint.hh"
#include "Graph.hh"
#include "StableSetMaster.hh"
#include "StableSetLPSolution.hh"



// -----------------------------------------------------------------------------
// C o n s t r u c t o r
// -----------------------------------------------------------------------------
CliqueHeuristicsSeparator::CliqueHeuristicsSeparator(
				       StableSetLPSolution *solution,
				       Graph *theGraph):
    ABA_SEPARATOR<ABA_CONSTRAINT, ABA_VARIABLE> (solution, true, 300),
    itsGraph(theGraph),
    lpSolution(solution)
{
}


// -----------------------------------------------------------------------------
// D e s t r u c t o r
// -----------------------------------------------------------------------------
CliqueHeuristicsSeparator::~CliqueHeuristicsSeparator ()
{
}


// -----------------------------------------------------------------------------
// s e p a r a t e
//
// Construct randomly maximAL cliques. Start with an edge and lift ist to a
// maximAL clique. The procedure is stopped until the maximal iteration number
// is reached or enough inequalities have been computed.
// -----------------------------------------------------------------------------
void CliqueHeuristicsSeparator::separate() {

    int i, j;		// Loop counter.
    int index;		// Index to a node.
    double value;	// The value of the clique in the current LP-solution.
    vector< vector<int> > computedCliques; // Store all computed cliques.
    vector<int> clique(2);      // This stores the maximAL clique.
    vector<int> adjacentNodes;  // Set of nodes adjacent to the current clique.
    int size = itsGraph->numberOfNodes(); // Size of the graph.
    int cnt = 0;	// Counter.
    bool found;		//
    bool newOne;	// Indicate if a clique is new or have been computed
			// earlier.
    int theCnt = 0;     // Counter.
    int cutCnt = 0;     // Counter for the generated cuts.
    int begin = 0;      // Index to check in adjacent list.
    double *solution = lpSolution->zVal();  // Current LP-solution.
    CliqueConstraint *cliqueConstr;  	    // Pointer to a clique sontraint
					    // To store each computed clique in
					    // Buffer cutfound().
    int level = lpSolution->sub()->level();


    //
    // Parameter.
    //
    static const int    MAX_ITER_SEP  = level==1 ? size * 1.5 : size / 3;
    static const int    MAX_STEP      = level==1 ? size / 4 : (int)( 5 + 0.01 * size );
    static const int    MAX_CUTS      = level==1 ? size : size / 5;
    static const double MIN_VIOLATION = 0.0001;


    //
    // Construct maximAL cliques.
    //
    while ( ((cnt < MAX_ITER_SEP) || theCnt < MAX_STEP)
            && (cutCnt < MAX_CUTS) ) {

        //
        // Select edge to start from.
        //
        found = false;
        while (!found) {
            clique[0] = getRandomNumber(size);
            clique[1] = getRandomNumber(size);
            if (clique[0] != clique[1]) {
                if (clique[0] > clique[1]) {
		    begin = 0;
                    if (itsGraph->isMemberOfAdjacentList(clique[0], begin,
		                                         clique[1])) {
                        found = true;
		    }
                }// if
            }
        }// while

        //
        // Get list of adjacent nodes to this edge.
        //
        adjacentNodes = (*itsGraph->getAdjacentList(clique[0]));
        for (i=adjacentNodes.size()-1; i>=0; i--) {
            if (clique.back() == adjacentNodes[i]) {
                adjacentNodes.erase(adjacentNodes.begin()+i);
                continue;
            }
	    begin = 0;
            if (!itsGraph->isMemberOfAdjacentList(clique.back(), begin,
	    					  adjacentNodes[i])) {
                    adjacentNodes.erase(adjacentNodes.begin()+i);
            }
        }// for


        //
        // Construct a maximAL clique containing this egde.
        //
        while(!adjacentNodes.empty()) {
            index = getRandomNumber(adjacentNodes.size());
            clique.push_back(adjacentNodes[index]);
            adjacentNodes.erase(adjacentNodes.begin() + index);
            // Update adjacent nodes.
            for (i=adjacentNodes.size()-1; i>=0; i--) {
                begin = 0;
                if (!itsGraph->isMemberOfAdjacentList(clique.back(), begin,
	     					      adjacentNodes[i])) {
		    adjacentNodes.erase(adjacentNodes.begin()+i);
                }
            }
        }// while


        //
        // Check if this inequality is violated.
        //
        theCnt++;
        value = 0;
        for (i=0; i<clique.size(); i++) {
            value += solution[clique[i]];
        }
        if (value > 1 + MIN_VIOLATION) {
            // Store this inequality if it is a new one.
            sort(clique.begin(), clique.end());

	    newOne = true;
            for (i=0; i<computedCliques.size(); i++) {
                if (computedCliques[i].size() != clique.size()) {
                    continue;
                }
		j=0;
		while (j<computedCliques[i].size()) {
                    if (clique[j] != computedCliques[i][j]) {
                        break;
                    }
		    j++;
                }
                if (j == computedCliques[i].size()) {
                    newOne = false;
                    break;
                }
            }

            if (newOne) {
	        computedCliques.push_back(clique);

		cliqueConstr = new CliqueConstraint(master_, &clique, itsGraph);
		cutFound(cliqueConstr);

		theCnt = 0;
		cutCnt++;

 	    }

        }// if

	// Clean up.
        clique.clear();
        adjacentNodes.clear();
        clique.resize(2);
        cnt++;

    }// while

} // end: separate()


// -----------------------------------------------------------------------------
// g e t R a n d o m N u m b e r
// -----------------------------------------------------------------------------
int CliqueHeuristicsSeparator::getRandomNumber(int bound) const {
    int    aNumber = rand();
    double d       = (double(aNumber)/RAND_MAX);

    return (int)(d*bound);
}



