////////////////////////////////////// c++ /////////////////////////////////////
//
//  Module           : EdgeProjection.cpp
//  Description      : Separation through edge projection.
//  Author           : Steffen Rebennack
//  Email            : srebenac@ix.urz.uni-heidelberg.de;
//                     steffen.rebennack@web.de
//  Copyright        : (C) 2006 by the University of L'Aquvila and
//                     University of Heidelberg
//  Created on       : Wed Apr 05 08:35:12 2006
//  Last modified by : -
//  Last modified on : - 
//  Update count     : 0
//
////////////////////////////////////////////////////////////////////////////////
//
//  Date	Name		Changes/Extensions
//  ----	----		------------------
//
////////////////////////////////////////////////////////////////////////////////

#include "EdgeProjection.hh"
#include <cstdlib>                   // rand()

#include "abacus/master.h"
#include "StableSetMaster.hh"



// -----------------------------------------------------------------------------
// --------------- M e t h o d s  ( p u b l i c ) ------------------------------
// -----------------------------------------------------------------------------


// -----------------------------------------------------------------------------
// C o n s t r u c t o r
//
// Copy adjacent list of the graph.
// Initialize the adjacent matrix as a lower triangular matrix storing if
// an edge is in the reduced graph or not.
// Initilaize the tree.
// -----------------------------------------------------------------------------
EdgeProjection::EdgeProjection(vector< vector<int> > *graph):
currentNode(NULL)
{

    int i, j;   // Loop variables.
    int node;   // Store a node.

    // Number of nodes of the graph.
    graphSize = graph->size();

    // Allocate memory for the adjacent list size counter.
    adjacentListSize = new int[graphSize];

    // Allocate memory for the adjacent matrix.
    lowerAdjacentMatrix.resize(graphSize - 1);
    for (i=1; i<graphSize; i++) {
        lowerAdjacentMatrix[i-1].resize(i,0);
    }

    // Copy graph and store each size of the adjacent list.
    for (i=0; i<graphSize; i++) {
        // Update the size of each adjacent list.
        adjacentListSize[i] = (*graph)[i].size();
        for (j=0; j<adjacentListSize[i]; j++) {
            node = (*graph)[i][j];
            // Store this edge in the adjacent matrix.
            if (i > node) {
                lowerAdjacentMatrix[i-1][node] = 1;
            }
        }// for
    }// for


    // Check each edge if it is strong projectable.
//    checkStrongProjectability();
//    cout << "Number of strongly projectable edges: ";
//    cout << projectable.size() << endl;

    // Initialize random counter.
    srand(1);
    
    // Print information of the root of the tree.    
    //printRoot();

}// end constructor


// -----------------------------------------------------------------------------
// D e s t r u c t o r
// -----------------------------------------------------------------------------
EdgeProjection::~EdgeProjection() {

    delete []adjacentListSize;

    if (!lhs.empty()) {
        // Delete the inequalities.
        for (int i=lhs.size()-1; i>=0; i--) {
            delete lhs[i];
            lhs[i] = NULL;
        }
        lhs.clear();
    }
    
}// end destructor


// -----------------------------------------------------------------------------
// s e p a r a t e
//
// This is the heart of the edge projection.
// Until a stop criteria, it selects an edge and computes its projection.
// Now, we have the freedom to call separation routines or to start again. 
// -----------------------------------------------------------------------------
int EdgeProjection::separate(double *lpValue) {

    int i;          // Loop variable.
    int numCut;     // The number of violated cuts found.
    int reducedGraphSize = graphSize;
    vector< vector<int> > theLhs;  // Store violated inequalities (lhs).
    vector<int> theRhs;            // Store violated inequalities (rhs).
    bool condition = true;  // Stop criteria for projection.
    bool separate  = true;  // Call separation or not.
    
    //
    // Parameter
    //
    const int ENOUGH_INEQUALITIES = (int) .2 * graphSize;
    
    // 
    // Clear all stored inequalities.
    //
    for (i=lhs.size()-1; i>=0; i--) {
        delete lhs[i];
        lhs[i] = NULL;
    }
    lhs.clear();
    rhs.clear();
    slack.clear();
   

    //
    // Do edge projection until the graph is too small or enough cuts have
    // have been computed.
    //    
    while (condition) {

        //
        // Select a new projectable edge
        //
        // Includes:
        // + check if the edge is projectable
        // + store all edges to be deleted that the edge is strongly projectable
        // + compute all nodes to be removed
        //
        bool edgeFound = edgeSelection(lpValue);

	if (!edgeFound) {
 	    break;
	}

        //
        // Projection: update reduction graph
        //  
        projection();

	//
        // Update size of reduced graph to check if the graph is too small.
	// 
        reducedGraphSize = 0;
        for (i=0; i<graphSize; i++) {
            if (adjacentListSize[i] != 0) {
                reducedGraphSize++;
            }
        }
        if (reducedGraphSize <= 5) {
            separate = false;
            condition = false;
        }

        //
        // Deside whether to start separation rountines or to project again.
        //   
        if (separate) {
            
            //
            // Call separation routines on the reduced graph.
            // theLhs and theRhs will store the information of the inequalities.
            //
            numCut = separation(lpValue, &theLhs, &theRhs);

            //
            // Check if a violated inequality has been found.
            //
            if (numCut) {
                //
                // Lift these inequalities.
                //
                lifting(&theLhs, &theRhs);

                //
                // Store these inequalities.
                //
                storeInequalities(&theLhs, &theRhs, lpValue);
	
	        //
		// Check if enough inequalities have been computed.
		//
		if (lhs.size() >= ENOUGH_INEQUALITIES) {
    		    condition = false;
		}
            }
        }// if

        //
        // Clean up
        //
        edge[0] = edge[1] = -1;
        deleteNodes.clear();
        for (i=deleteEdges.size()-1; i>=0; i--) {
            delete [] deleteEdges[i];
        }
        deleteEdges.clear();
        theLhs.clear();
        theRhs.clear();

    }// while
    
    //
    // Select only the best inequalities.
    //
//    cleanUpInequalities(lpValue);
    
    //
    // Clean up tree.
    //
    cleanUp();

    // Return: no error.
    return 0;

}// end separate(..)


// -----------------------------------------------------------------------------
// c o m p u t e L a r g e I n e q u a l i t y
//
// Compute a large rank inequality. Project the graph until it is small enough
// to solve a maximum stable set problem fast. Call the stable set solver for
// that small graph and receive a large rank inequality.
// -----------------------------------------------------------------------------
int EdgeProjection::computeLargeInequality(double *lpValue,
					   StableSetMaster *theMaster) {

    int i;  // Loop counter.
    int reducedGraphSize = graphSize;
    bool condition = true;	// Stop criteria for the projection.
    vector<int> theLhs;		// Store the rank inequality (lhs)
    int theRhs;			// Store the rank inequality (rhs)

    //
    // Parameter.
    //
    const static int MAX_SIZE_OF_GRAPH = 45; //(int) graphSize / 5 ;  
  
    // 
    // Clear all stored inequalities.
    //
    for (i=lhs.size()-1; i>=0; i--) {
        delete lhs[i];
        lhs[i] = NULL;
    }
    lhs.clear();
    rhs.clear();
    slack.clear();
   
    //
    // Project until the graph is small enough to solve a stable set problem
    // exact. After one cut has been computed stop the projection.
    //    
    while (condition) {

        //
        // Select a new projectable edge
        //
        // Includes:
        // + check if the edge is projectable
        // + store all edges to be deleted that the edge is strongly projectable
        // + compute all nodes to be removed
        //
        bool edgeFound = edgeSelection(lpValue);

	if (!edgeFound) {
 	    break;
	}

        //
        // Projection: update reduction graph
        //  
        projection();

	//
        // Update size of reduced graph.
	//        
	reducedGraphSize = 0;
        for(i=0; i<graphSize; i++) {
            if (adjacentListSize[i] != 0) {
        	reducedGraphSize++;
     	    }
	}

	//
        // Deside whether to start solve stable set or to project again.
        //   
	if (reducedGraphSize <= MAX_SIZE_OF_GRAPH) {
	
	    //
 	    // Solve the stable set problem for that small graph exactly.
	    //            
	    solveStableSet(&theLhs, theRhs, theMaster);

            //
            // Lift these inequalities.
            //
            liftAndStore(&theLhs, theRhs);
	
	    //
	    // One cut has been computed: stop projection
	    //
	    condition = false;
        }// if

        //
        // Clean up
        //
        edge[0] = edge[1] = -1;
        deleteNodes.clear();
        for (i=deleteEdges.size()-1; i>=0; i--) {
            delete [] deleteEdges[i];
        }
        deleteEdges.clear();
        theLhs.clear();
 
    }// while
    
    //
    // Clean up tree.
    //
    cleanUp();

    // Return status.
    return 0;

}// computeLargeInequality(..)


// -----------------------------------------------------------------------------
// g e t L H S
//
// Return lhs of all computed inequalities.
// -----------------------------------------------------------------------------
vector< vector<int>* >* EdgeProjection::getLHS() {

    if (lhs.empty()) {
	return NULL;
    }
    else {
        return &lhs;
    }
}


// -----------------------------------------------------------------------------
// g e t R H S
//
// Return ride hand side of all inequalities.
// -----------------------------------------------------------------------------
vector<int>* EdgeProjection::getRHS() {
    return &rhs;
}



// -----------------------------------------------------------------------------
// -------------- M e t h o d s  ( p r i v a t e ) -----------------------------
// -----------------------------------------------------------------------------


// -----------------------------------------------------------------------------
// e d g e S e l e c t i o n
//
// Choose randomly two nodes and and go through the adjacent matrix until a
// "good" edge is found. Check if it is projectable. If the neighborhood of the 
// endnode with smaller degree is not a clique, contruct a maximUM clique in
// its neighborhood. 
// -----------------------------------------------------------------------------
bool EdgeProjection::edgeSelection(double *lpValue) {
    
    int i, j, k;      // Loop counter.
    int size;         // Store a size.

    // 
    // Select an edge as a candidate for the projection edge.
    //
    int stack;        // Store something temporarily.
    int startNodeOne; // First start node for search.
    int startNodeTwo; // Second start node for search.
    int nodeOne;
    int nodeTwo;
    double weight;    // Weight of an edge.
    double bestWeight = 0; // Highest weight of an edge.


    //
    // Parameter.
    //
    // If the sum of the weights of the endnodes of an edge are
    // greater than this number then it is a direct candidate for an
    // projection edge.
    static const double ACCEPT_EDGE = .99;
    // Don not accept an edge as projectable edge which as smaller weight
    // than NOT_ACCEPT_EDGE. 
    static const double NOT_ACCEPT_EDGE = .2;

    
    // Choose two random start nodes.
    startNodeOne = getRandomNumber(graphSize);
    startNodeTwo = getRandomNumber(graphSize);
    while (startNodeOne == startNodeTwo) {
        startNodeTwo = getRandomNumber(graphSize);
    }
    // Sort for the adjacent matrix.    
    if (startNodeOne < startNodeTwo) {
        stack = startNodeOne;
        startNodeOne = startNodeTwo;
        startNodeTwo = stack;
    }    
    // Walk through the adjacent matrix.
    nodeOne = startNodeOne;
    nodeTwo = startNodeTwo;
    do { 
        // Check if this two nodes are adjacent.
        if (lowerAdjacentMatrix[nodeOne-1][nodeTwo]) {
            // Calculate weight of edge.
            weight = lpValue[nodeOne] + lpValue[nodeTwo];
            // Higher weight?
            if (weight > bestWeight) {
                // Already an acceptable weight?
                if (weight > ACCEPT_EDGE) {
                   // This is our projection edge.
                   bestWeight = weight;
                   edge[0] = nodeOne;
                   edge[1] = nodeTwo;
                   break;
                }
                bestWeight = weight;
                edge[0] = nodeOne;
                edge[1] = nodeTwo;
            }
        }// if

        // Select the next edge.
        nodeTwo++;
        if (nodeTwo == nodeOne) {
            nodeOne++;
            nodeTwo = 0;
            if (nodeOne == graphSize) {
                nodeOne = 1;
            }
        }
    } while ((nodeOne != startNodeOne) || (nodeTwo != startNodeTwo)); 
    
    // Check if there was an edge with acceptable weight.
    if (bestWeight < NOT_ACCEPT_EDGE) {
	return 0;
    }
     
    // 
    // Check if this edge is strongly projectable   
    //
    // Find node with smaller degree.
    int index = 0;
    if (adjacentListSize[edge[0]] > adjacentListSize[edge[1]]) {
        index = 1;
    }
    // -> edge[index] has smaller degree.
    
    // Get neighborhood of edge[index]
    vector<int> neighborhood;
    getNeighborhood(edge[index], &neighborhood);

    // Delte edge[(index+1)%2] from the neighborhood
    i = 0;
    size = neighborhood.size();
    while (i < size) {
        if (neighborhood[i] == edge[(index+1)%2]) {
            neighborhood.erase(neighborhood.begin()+i);
            size--;
            break;
        }
        i++;
    }

    // Check if this node was found.
    if(i == size+1) {
        cout << "ERROR" << endl;
        exit(1);
    }

    bool clique = true;
    for (i=0; i<size-1; i++) {
        for (j=i+1; j<size; j++) {
            if (!lowerAdjacentMatrix[neighborhood[j]-1][neighborhood[i]]) {
                clique = false;
                break;
            }
        }// for
        if (!clique) {
            break;
        }
    }// for

    if (clique) {
        // We have found a strongly projectable edge!
        for (i=0; i<size; i++) {
            if (neighborhood[i] > edge[(index+1)%2]) {
                if (lowerAdjacentMatrix[neighborhood[i]-1][edge[(index+1)%2]]) {
                    deleteNodes.push_back(neighborhood[i]);
                }
            }
            else {
                if (lowerAdjacentMatrix[edge[(index+1)%2]-1][neighborhood[i]]) {
                    deleteNodes.push_back(neighborhood[i]);
                }
            }
        }
    }
    else {
        //
        // Make it a projectable edge.
        //
        // Construct induced subgraph of neighborhood.
        vector< vector<int> > adjacentList(size);
        for (i=0; i<size-1; i++) {
            for (j=i+1; j<size; j++) {
                if (lowerAdjacentMatrix[neighborhood[j]-1][neighborhood[i]]) {
                    adjacentList[i].push_back(j);
                    adjacentList[j].push_back(i);
                }
            }                
        }        

        // Compute maximAL clique in induced graph of neighborhood.
        vector<int> *maxClique;
        // This can be any maximAL/maximUM clique generator.
        maxClique = computeMaximalClique(&adjacentList, &neighborhood);

        if (maxClique == NULL) {
            cout << "ERROR with maximal clique computation." << endl;
            exit(1);
        }            

	// Tranlsate clique and fill vectors deleteNodes, deleteEdges
        int cliqueSize = maxClique->size();
        for (i=0; i<cliqueSize; i++) {
            (*maxClique)[i] = neighborhood[(*maxClique)[i]];
        }
        sort(maxClique->begin(), maxClique->end());
        deleteEdges.resize(size - cliqueSize);
        int edgesCnt = 0;
        int *anEdge;
        j = 0;
        for (i=0; i<size; i++) {
            // Check if neighborhood[i] is in the clique. If it is contained
            // this node will be deleted. Otherwise we have to delete an edge.
            while (j < cliqueSize) {
                if ((*maxClique)[j] >= neighborhood[i]) {
                    if ((*maxClique)[j] == neighborhood[i]) {
                        if (neighborhood[i] > edge[(index+1)%2]) {
                            if (lowerAdjacentMatrix[neighborhood[i]-1][edge[(index+1)%2]]) {
                                deleteNodes.push_back(neighborhood[i]);
                            }
                        }// if
                        else {
                            if (lowerAdjacentMatrix[edge[(index+1)%2]-1][neighborhood[i]]) {
                                deleteNodes.push_back(neighborhood[i]);
                            }
                        }// else
                    }
                    else {
                        anEdge = new int[2];
                        anEdge[0] = edge[index];
                        anEdge[1] = neighborhood[i];
                        deleteEdges[edgesCnt++] = anEdge;
                    }
                    break;
                }
                j++;
            }// while
            if (j == cliqueSize) {
                anEdge = new int[2];
                anEdge[0] = edge[index];
                anEdge[1] = neighborhood[i];
                deleteEdges[edgesCnt++] = anEdge;
            }
        }// for

        // Clean up.
        maxClique->clear();
        delete maxClique;
    }// else

    // Return status.
    return 1;

}// end edgeSelection()


// -----------------------------------------------------------------------------
// p r o j e c t i o n
//
// Generate new node of the tree and update the graph.
// Compute all false edges.
// -----------------------------------------------------------------------------
void EdgeProjection::projection() {

    // Store all false edges and pass them to the new node.
    vector<int*> falseEdges;

    //
    // Update graph:
    // Delete projection edge and all nodes to be deleted.
    // The endnodes of the projection node will stay in the graph.
    //
    updateGraph();

    //
    // Compute all false edges.
    //
    computeFalseEdges(&falseEdges);

    //
    // Update graph:
    // Delete endnodes of the projection edge and add false edges.
    //
    updateGraph(&falseEdges);

    //
    // Add new node to the tree.
    //
    addNode(&falseEdges);

    // Clean up.
    for (int i=falseEdges.size()-1; i>=0; i--) {
        delete [] falseEdges[i];
        falseEdges[i] = NULL;
    }
    falseEdges.clear();

}// end projection()


// -----------------------------------------------------------------------------
// s e p a r a t i o n
//
// Construct the graph corresponding to that node of the tree and call
// separation routines for that graph.
// !
// ! The entries of each inequality have to be upwards sorted!!
// !
// -----------------------------------------------------------------------------
int EdgeProjection::separation(double *lpValue, vector< vector<int> > *theLhs,
                                vector<int> *theRhs) {

    int i;	// Loop counter.
    int size;

    //
    // Get reduced graph.
    //
    vector< vector<int> > reducedGraph;
    getReducedGraph(&reducedGraph);

    //
    // Adjust lpValue.
    //
    size = reducedGraph.size();
    double theLPValue[size];
    for (i=0; i<size; i++) {
        theLPValue[i] = lpValue[translator[i]];
    }
 
    //
    // Call separation routines...
    //
    int numberOfCuts = cliqueSeparation(&reducedGraph, theLPValue, theLhs);

    //
    // Update rhs for each inequality.
    //
    for (i=0; i<numberOfCuts; i++) {
        theRhs->push_back(1);
    }

    return theLhs->size();

}// end separation(..)


// -----------------------------------------------------------------------------
// l i f t i n g
//
// This function computes the inequalities for the original graph.
// The process is called anti-projection.
// -----------------------------------------------------------------------------
void EdgeProjection::lifting(vector< vector<int> > *theLhs,
                             vector<int> *theRhs) {

    int i, j, k;  // Loop variables.
    int size;     // Store a size.
    int nodeOne;  // One endnode of the false edges.
    int nodeTwo;  // The other endnode of the false edges.
    int stack;    // Store something temporarily.
    int n_inequalities = theLhs->size();  // Number of inequalities.
    vector<int*>* theFalseEdges;    // Get the false edges for each node.
    vector<int>* theDeletedNodes;   // Get the deleted nodes for each node.
    int deletedNodesSize;    // Size of vector 'theDeletedNodes'.
    ProjectionNode* aNode;   // A node of the tree.

    //
    // Translate this inequality to the labeling of the original graph.
    //
    for (i=0; i<n_inequalities; i++) {
        size = (*theLhs)[i].size();
        for (j=0; j<size; j++) {
            (*theLhs)[i][j] = translator[(*theLhs)[i][j]];
        }
    }

    //
    // 'Lift' this inequality.
    // Therefore check for each node of the tree all false edges
    // if they are contained in this inequality.
    //
    aNode = currentNode;
    // Go the way back to the root.
    while (aNode != NULL) {
        // Get all false edges of tree node 'aNode'
        theFalseEdges = aNode->getFalseEdges();

        // Check for each inequality if one of this false edges are contained.
        for (i=0; i<n_inequalities; i++) {
            j = 0;
            size = (*theLhs)[i].size();
            while (j < theFalseEdges->size()) {
                nodeOne = (*theFalseEdges)[j][0];
                nodeTwo = (*theFalseEdges)[j][1];
                j++;
                // Check if this two nodes are part of the inequalities
                if (nodeOne > nodeTwo) {
                    stack = nodeOne;
                    nodeOne = nodeTwo;
                    nodeTwo = stack;
                }
                k = 0;
                while (k < size) {
                    if( (*theLhs)[i][k] >= nodeOne ) {
                        break;
                    }
                    k++;
                }// while
                if (k == size) {
                    // This node is not contained in this inequality.
                    continue;
                }
                if ((*theLhs)[i][k] != nodeOne) {
                    // This node is not contained in this inequality.
                    continue;
                }
                k++;
                while (k < size) {
                    if( (*theLhs)[i][k] >= nodeTwo ) {
                        break;
                    }
                    k++;
                }
                if (k == size) {
                    // This node is not contained in this inequality.
                    continue;
                }
                if ((*theLhs)[i][k] != nodeTwo) {
                    // This node is not contained in this inequality.
                    continue;
                }

                //
                // This false edge is conained in the inequality
                // -> 'lift' this inequality
                //
                // Add all deleted node.
                theDeletedNodes = aNode->getDeletedNodes();
                deletedNodesSize = theDeletedNodes->size();
                for (k=0; k<deletedNodesSize; k++) {
                    (*theLhs)[i].push_back((*theDeletedNodes)[k]);
                }
                // Add the projection edge.
                (*theLhs)[i].push_back(aNode->getEdge(0));
                (*theLhs)[i].push_back(aNode->getEdge(1));
                // Sort the inequality.
                sort((*theLhs)[i].begin(), (*theLhs)[i].end());
                // Update the right hand side.
                (*theRhs)[i] += 1;
                                
                // The update of this inequality is finished for this node
                // of the tree. Each edge projection step increases the rhs
                // only by value 1.
                break;
            }// end while
        }// for
        // Get previous node in the tree.
        aNode = aNode->getPreviousNode();
    }// while

}// end lifting(..)


// -----------------------------------------------------------------------------
// s t o r e I n e q u a l i t i e s
//
// It checks if the new inequalities are still violated. The violated ones are
// stored in the vectors lhs and rhs.
// -----------------------------------------------------------------------------
void EdgeProjection::storeInequalities(vector< vector<int> > *theLhs,
                             vector<int> *theRhs, double* lpValue) {

    int i, j, k;          // Loop variables.
    int cnt = lhs.size(); // Count the number of inequalities.
    int size;             // Store a size.
    int numberOfInequalities = theLhs->size();  // Number of inequalities.
    double value;         // Value of the LP solution for an inequality.
    vector<int> *inequality;  // Store a violated inequality.
    bool newOne;          // Indicate if an inequality is a new one.

    //
    // Parameter.
    //
    static const double MIN_VIOLATION = .000001;


    //
    // Overestimate the number of inequalities.
    //
    lhs.resize(numberOfInequalities + cnt);
    rhs.resize(numberOfInequalities + cnt);
    slack.resize(numberOfInequalities + cnt);

    //
    // Check each inequality and store it.
    //
    for (i=0; i<numberOfInequalities; i++) {
        value = 0;
        size = (*theLhs)[i].size();
        for (j=0; j<size; j++) {
            value += lpValue[(*theLhs)[i][j]];
        }
        // Check if it is violated.
        if (value > ((*theRhs)[i] + MIN_VIOLATION)) {
            //
            // Check if this inequality was already computed during this
            // separation process.
            //
            newOne = true;
            for (j=0; j<cnt; j++) {
                if ((*theLhs)[i].size() != lhs[j]->size()) {
                    continue;
                }
		k = 0;
                while (k < lhs[j]->size()) {
                    if ((*theLhs)[i][k] != (*lhs[j])[k]) {
                       break;
                    }
		    k++;
                }
                if (k == lhs[j]->size()) {
                   newOne = false;
                   if (rhs[j] > (*theRhs)[i]) {
                       rhs[j] = (*theRhs)[i];
                   }
		   break;
                }
            }// for

            //
            // Store it if it is a new inequality.
            //
            if (newOne) {
                inequality = new vector<int>;
                inequality->resize(size);
                for (j=0; j<size; j++) {
                    (*inequality)[j] = (*theLhs)[i][j];
                }
                lhs[cnt] = inequality;
                rhs[cnt] = (*theRhs)[i];
                slack[cnt] = value;
		cnt++;
            }// if
        }
    }// for

    // Correct the size of the vectors.
    lhs.resize(cnt);
    rhs.resize(cnt);
    slack.resize(cnt);

}// end storeInequalities(..)


//
// The struct inequalitiesAndWeights stores an index of an inequality and its
// weight.
//
struct InequalitiesAndWeights {
    int    iinequality; // index of an inequality
    double weight;      // the weight of an inequality
};


//
// This operator is used by the sort function in cleanUpInequalities to get
// a sorted vector of inequalities with decreasing weight.
//
bool operator<(const InequalitiesAndWeights &a, const InequalitiesAndWeights &b) {
    return a.weight < b.weight;
}


// -----------------------------------------------------------------------------
// c l e a n U p I n e q u a l i t i e s
//
// Select only the best computed inequalities to be stored.
// -----------------------------------------------------------------------------
void EdgeProjection::cleanUpInequalities(double *lpValue) {
        
    //
    // Parameters.
    //
    const int MIN_NUMBER = 10;
    const int MAX_NUMBER = graphSize;
    const int MULTIPLICATOR = 10; 
    
    int size = lhs.empty() ? 0 : lhs.size();
             

    //
    // Do not delete inequalities if the number of inequalities is too small.
    //
    if (size >= MIN_NUMBER) {
    
        int i;               // Loop counter.
        int selectionNumber; // Number of inequalities to be selected.
        double m;            // Slop of line.
        double c;            // y-intersect of line.
        vector<InequalitiesAndWeights> inequalities;
    
        //
        // Calculate weight for each inequality.
        //    
        for (i=0; i<size; i++) {
            InequalitiesAndWeights a;
    		a.iinequality = i;
    		a.weight      = ( (double)(lhs[i]->size()) / rhs[i] ) * slack[i];
            inequalities.push_back(a);
        }
    
        //
        // Sort the inequalities.
        //
        sort(inequalities.begin(), inequalities.end());
    
        //
        // Calculate the number of the inequalities to be selected.
        //    
        if (size >= (MAX_NUMBER * MULTIPLICATOR)) {
            selectionNumber = MAX_NUMBER;
        }
        else {
            // Linear interpolation.
            m = ((double)(MAX_NUMBER - MIN_NUMBER)) / (MAX_NUMBER * MULTIPLICATOR - MIN_NUMBER);
            c = MIN_NUMBER - m * MIN_NUMBER;
            selectionNumber = (int) (m * size + c);
        }
    
        //
        //
        vector< vector<int>* > lhsCopy(selectionNumber);
	vector< int > rhsCopy(selectionNumber);
        for (i=0; i<selectionNumber; i++) {
            lhsCopy[i] = lhs[inequalities[i].iinequality];
	    rhsCopy[i] = rhs[inequalities[i].iinequality];
        }

        lhs.clear();
	rhs.clear();
        lhs = lhsCopy;
	rhs = rhsCopy;
    }// if

}// end cleanUpInequalities()


// -----------------------------------------------------------------------------
// c h e c k S t r o n g P r o j e c t a b i l i t y
//
// Check each edge of the graph if it is strongy projectable.
// All strongly projectable edges are stored in vector 'projectable'.
// -----------------------------------------------------------------------------
void EdgeProjection::checkStrongProjectability() {
   
    int i, j, k, l, m;// Loop counters.
    int size;         // A size of a vector.
    int node;         // A node.
    int nodeU;        // One endnode u of a candidate for a strongly
                      // projectable edge.   
    int nodeV;        // The other endnode v of a candidate for a strongly
                      // projectable edge.
    vector<int> neighborhoodU;  // The neighborhood of node u: N(u). 
    int sizeU;        // The size of N(u).
    vector<int> neighborhoodV;  // The neighborhood of node v: N(v). 
    int sizeV;        // The size of N(v).
    vector<int> nUwnV;          // N(u) \ ( N(v) unified v )
    int sizeUV;       // The size of N(u) \ ( N(v) unified v ).
    vector<int> nVwnU;          // N(v) \ ( N(u) unified u )
    int sizeVU;       // The size of N(v) \ ( N(u) unified u ).
    vector<int> nUcutnV;        // N(u) cut N(v)
    int sizeUCV;      // The size of N(u) cut N(v).
    bool clique;      // Indicate if a set is a clique.
    int *projectableEdge;     // Pointer to a projectable edge. 
    bool bull = false;        // Indicate if a set of nodes defines a bull.
    bool diamond = false;     // Indicate if a set of nodes defines a diamond.
    
    //
    // Check each edge if it is strongly projectable.
    // 
    for (i=1; i<graphSize; i++) {
        for (j=0; j<i; j++) {
            // Check if ij is an edge in the graph
            if (!lowerAdjacentMatrix[i-1][j]) {
                // Consider the next pair of nodes and check if they are
                // adjacent.
                continue;
            }
            
            //
            // Check strong projectability of edge uv.
            //
            // We have to check that edge uv is not an central edge of an 
            // induced subgraph isomorphic to a
            // ~ diamond
            // ~ double fork
            // ~ bull
            //
            
            nodeU = i; 
            nodeV = j;
            
            getNeighborhood(nodeU, &neighborhoodU);
            getNeighborhood(nodeV, &neighborhoodV);
            sizeU = neighborhoodU.size();
            sizeV = neighborhoodV.size();

            //
            // Diamond
            //
            // Edge uv is NOT the central edge of a diamond, iff there are 
            // two nodes a,b which are adjacent to u and v and edge ab is 
            // contained in the graph.
            //             
            
            // Construct: N(nodeU) cut N(nodeV)
            for (k=0; k<sizeU; k++) {
                node = neighborhoodU[k];
                // Check if this node is contained in neighborhoodV.
                l = 0;
                while (l < sizeV) {
                    if (node <= neighborhoodV[l]) {
                        break;
                    }
                    l++;
                }
                if (l == sizeV) {
                    break;
                }
                if (node == neighborhoodV[l]) {
                    // Node is a candidate for the set nUcutnV.
                    // Check if 'node' is not adjacent to all other nodes of
                    // set nUcutnV.
                    size = nUcutnV.empty() ? 0 : nUcutnV.size();
                    for (m=0; m<size; m++) {
                        // nUcutnV[m] < node for all index m since
                        // neighborhoodU is sorted
                        if (!lowerAdjacentMatrix[node-1][nUcutnV[m]]) {
                            // Edge uv is a central edge of a diamond.
                            diamond = true;
                            break;
                        }
                    }
                    if (diamond) {
                        break;
                    }
                    // This node is part of N(u) cut N(v) and does not induce
                    // a diamond with central edge uv. 
                    // -> Store it.
                    nUcutnV.push_back(node);
                }// if
            }// for

            if (diamond) {
                // Edge uv is a central edge of a diamond.
                // Clean up.
                diamond = false;
                nUcutnV.clear();
                neighborhoodU.clear();
                neighborhoodV.clear();
                // Consider the next edge and check if it 
                // strongly projectable.
                continue; 
            }
            // Update size.
            sizeUCV = !nUcutnV.empty() ? nUcutnV.size() : 0;
            
            // At this point we know that edge uv is not the central edge of
            // a diamond. Therefore check the next condition.
                                                            
                                                              
            //
            // Double fork.
            //
            // Edge uv is NOT the central edge of a double fork, iff
            // one of the two sets 
            //   N(u) \ ( N(v) unified v ) or
            //   N(v) \ ( N(u) unified u )
            // defines a clique.
            //
                                        
            // Construct N(u) \ ( N(v) unified v )
            l = 0; 
            for (k=0; k<sizeU; k++) {
                node = neighborhoodU[k];
                if (node == nodeV) {
                    // Check the next node.
                    continue;
                }
                // Check if this node is contained in neighborhoodV.
                while (l < sizeV) {
                    if (node <= neighborhoodV[l]) {
                        break;
                    }
                    l++;
                }
                if (l == sizeV) {
                    nUwnV.push_back(node);
                    continue;
                }
                if (node != neighborhoodV[l]) {
                    nUwnV.push_back(node);
                }
            }// for
            // Update size.
            sizeUV = !nUwnV.empty() ? nUwnV.size() : 0;

            clique = true;
            if (sizeUV) {
                // Check if the set of nodes nUwnV is a clique.
                for (k=0; k<sizeUV-1; k++) {
                    for (l=k+1; l<sizeUV; l++) {
                        if (!lowerAdjacentMatrix[nUwnV[l]-1][nUwnV[k]]) {
                            clique = false;
                            break;                            
                        }
                    }
                    if (!clique) {
                        break;
                    }
                }
            }

            // Now we have to check if set
            //   N(v) \ ( N(u) unified u ) 
            // is a clique.
            // Construct N(v) \ ( N(u) unified u )
            l = 0; 
            for (k=0; k<sizeV; k++) {
                node = neighborhoodV[k];
                if (node == nodeU) {
                    // Check the next node.
                    continue;
                }
                // Check if this node is contained in neighborhoodU.
                while (l < sizeU) {
                    if (node <= neighborhoodU[l]) {
                        break;
                    }
                    l++;
                }
                if (l == sizeU) {
                    nVwnU.push_back(node);
                    continue;
                }
                if (node != neighborhoodU[l]) {
                    nVwnU.push_back(node);
                }
            }// for
            // Update size.
            sizeVU = !nVwnU.empty() ? nVwnU.size() : 0;            

            // If N(u) \ ( N(v) unified v ) is a clique, then we know that
            // this edge is not a central edge of a double fork. 
            // Therefore we have to check the next of the three graphs.   
            if (!clique) {
                clique = true;
                if (sizeVU) {
                    // Check if the set of nodes nVwnU is a clique.
                    for (k=0; k<sizeVU-1; k++) {
                        for (l=k+1; l<sizeVU; l++) {
                            if (!lowerAdjacentMatrix[nVwnU[l]-1][nVwnU[k]]) {
                                clique = false;
                                break;                            
                            }
                        }
                        if (!clique) {
                            break;
                        }
                    }
                    if (!clique) {

                        // At this point we know that edge uv is not
                        // strongly projectable since it is a central edge
                        // of a double fork.
                        // Clean up.
                        nUwnV.clear();
                        nVwnU.clear();
                        nUcutnV.clear();
                        neighborhoodU.clear();
                        neighborhoodV.clear();
                        // Consider the next edge and check if it 
                        // strongly projectable.
                        continue;
                    }
                }
            }// if    
                
            // At this point we know that edge uv is not a central edge of a
            // double fork. Therefor check the last criterion.
            
            //
            // bull
            //
            // Edge uv is NOT a central edge of a bull, iff there are NOT three 
            // nodes k,l,m with 
            //   k in N(u) \ ( N(v) unified v )
            //   l in N(v) \ ( N(u) unified u )
            //   m in N(u) cut N(v)
            // and all three edges
            //   kl, km, lm
            // are not contained in the graph.
            // 
            
            // Check if one of the three sets is empty.            
            if (!nUcutnV.empty() && !nUwnV.empty() && !nVwnU.empty()) {
                // Check each combination of the three sets.
                for (k=0; k<sizeUV; k++) {
                    for (l=0; l<sizeVU; l++) {
                        for (m=0; m<sizeUCV; m++) {
                            // Check kl 
                            if (nUwnV[k] > nVwnU[l]) {
                                if (lowerAdjacentMatrix[nUwnV[k]-1][nVwnU[l]]) {
                                    break;
                                }
                            }
                            else {
                                if (lowerAdjacentMatrix[nVwnU[k]-1][nUwnV[l]]) {
                                    break;
                                }
                            }
                            // Check km
                            if (nUwnV[k] > nUcutnV[m]) {
                                if (lowerAdjacentMatrix[nUwnV[k]-1][nUcutnV[m]]) {
                                    break;
                                }
                            }
                            else {
                                if (lowerAdjacentMatrix[nUcutnV[m]-1][nUwnV[k]]) {
                                    break;
                                }
                            }
                            // Check lm
                            if (nVwnU[l] > nUcutnV[m]) {
                                if (lowerAdjacentMatrix[nVwnU[l]-1][nUcutnV[m]]) {
                                    break;
                                }
                            }
                            else {
                                if (lowerAdjacentMatrix[nUcutnV[m]-1][nVwnU[l]]) {
                                    break;
                                }
                            }
                            if (m == (sizeUCV - 1)) {
                                bull = true;
                                break;
                            }
                        } // for
                        if (bull) {
                           break;
                        }
                    } // for
                    if (bull) {
                        break;
                    }
                }// for
                if (bull) {
                    // At this point we know that edge uv is not
                    // strongly projectable since it is a central edge
                    // of a bull.
                    // Clean up.
                    bull = false;                
                    nUwnV.clear();
                    nVwnU.clear();
                    nUcutnV.clear();
                    neighborhoodU.clear();
                    neighborhoodV.clear();
                    // Consider the next edge and check if it 
                    // strongly projectable.
                    continue;
                }
            } // if
            
            //
            // All the computational effort was not wasted and we have indeed
            // found a strongly projectable edge!
            //
            projectableEdge = new int[2];
            projectableEdge[0] = nodeU;
            projectableEdge[1] = nodeV;
            projectable.push_back(projectableEdge);
            
            // Clean up.   
            nUwnV.clear();
            nVwnU.clear();
            nUcutnV.clear();
            neighborhoodU.clear();
            neighborhoodV.clear();
            
        }// for
    }// for 
    
}// end checkStrongProjectability()


// -----------------------------------------------------------------------------
// c o m p u t e M a x i m a l C l i q u e
//
// This function computes a maximAL clique. The graph is given through the 
// parameters. This function is udes to make an edge strongly projectable.
// -----------------------------------------------------------------------------
vector<int>* EdgeProjection::computeMaximalClique(
            vector< vector<int> > *theAdjacentList, vector<int> *theTranslator) {

    int i;		        // Loop counter.
    int cnt = 0;
    int size = theAdjacentList->size();
    vector<int> clique(2);      // This will be increased to a maximAL clique.
    vector<int> adjacentNodes;  // All adjacent nodes to the clique.
    vector<int> *maxClique;     // Store a maximAL clique.
                                // This will be the return value.
    maxClique = new vector<int>;
    int maxCliqueSize = 0;      // Size of maximAL clique.
    bool found;

    //
    // Parameter.
    //
    const int MAX_ITER = size * 2;


    //
    // Avoid degenerated cases.
    //
    if (theAdjacentList->empty()) {
        return NULL;
    }
    // Check if there is at least one edge.
    for (i=0; i<size; i++) {
        cnt += (*theAdjacentList)[i].size();
        if (cnt >= 1) {
            break;
        }
    }
    if (cnt < 1) {
        maxClique->push_back(0);
        return maxClique;
    }

    //
    // Construct maximal cliques.
    //
    cnt = 0;
    while (cnt < MAX_ITER) {

	//
        // Select edge to start from.
        //
        found = false;
        while (!found) {
            clique[0] = getRandomNumber(size);
            clique[1] = getRandomNumber(size);
            if (clique[0] != clique[1]) {
                if (clique[0] > clique[1]) {
		    // theTranslator is sorted!
                    if (lowerAdjacentMatrix[(*theTranslator)[clique[0]]-1][(*theTranslator)[clique[1]]]) {
                        found = true;
                    }
                }
                else {
                    if (lowerAdjacentMatrix[(*theTranslator)[clique[1]]-1][(*theTranslator)[clique[0]]]) {
                        found = true;
                    }
                }
            }// if
        }// while

        //
        // Get list of adjacent nodes to this edge.
        //
        adjacentNodes = (*theAdjacentList)[clique[0]];
        for (i=adjacentNodes.size()-1; i>=0; i--) {
            if (clique[1] == adjacentNodes[i]) {
                adjacentNodes.erase(adjacentNodes.begin()+i);
                continue;
            }
            if (clique[1] > adjacentNodes[i]) {
                if (!lowerAdjacentMatrix[(*theTranslator)[clique[1]]-1][(*theTranslator)[adjacentNodes[i]]]) {
                    adjacentNodes.erase(adjacentNodes.begin()+i);
                }
            }
            else {
                if (!lowerAdjacentMatrix[(*theTranslator)[adjacentNodes[i]]-1][(*theTranslator)[clique[1]]]) {
                    adjacentNodes.erase(adjacentNodes.begin()+i);
                }
            }
        }// for
        
        
        //
        // Construct maximal clique
        //
        while (!adjacentNodes.empty() && (adjacentNodes.size() + clique.size()) > maxCliqueSize) {
            clique.push_back(adjacentNodes.back());
            adjacentNodes.pop_back();
            // Update adjacent nodes.
            for (i=adjacentNodes.size()-1; i>=0; i--) {
                if (clique.back() > adjacentNodes[i]) {
                    if (!lowerAdjacentMatrix[(*theTranslator)[clique.back()]-1][(*theTranslator)[adjacentNodes[i]]]) {
                        adjacentNodes.erase(adjacentNodes.begin()+i);
                    }
                }
                else {
                   if(!lowerAdjacentMatrix[(*theTranslator)[adjacentNodes[i]]-1][(*theTranslator)[clique.back()]]) {
                        adjacentNodes.erase(adjacentNodes.begin()+i);
                    }

                }
            }
        }

        //
        // Update max clique.
        //
        if ((adjacentNodes.size() + clique.size()) > maxCliqueSize) {
           // Update maximUM clique.
           maxClique->clear();
           (*maxClique) = clique;
           maxCliqueSize = maxClique->size();
        }

        clique.clear();
        adjacentNodes.clear();
        clique.resize(2);
        cnt++;
    }// while

    return maxClique;

}// end computeMaximalClique(..)


// -----------------------------------------------------------------------------
// c o m p u t e F a l s e E d g e s
//
// Let u and v denote the endnodes of the projection edge.
// False edges of the projection of edge uv are between each two nodes of the
// neighborhood of u and the neighborhood of v in the reduced graph.
// This reduced graph contains nodes u and v but not the deleted nodes and not
// the projection edge uv.
// -----------------------------------------------------------------------------
void EdgeProjection::computeFalseEdges(vector<int*> *falseEdges) {

    // Store the neighbothood of node u and of node v.
    vector<int> neighborhoodU;
    vector<int> neighborhoodV;

    //
    // Compute the neighborhoods of node u and v.
    //
    getNeighborhood(edge[0], &neighborhoodU);
    getNeighborhood(edge[1], &neighborhoodV);

    //
    // Now we are ready to compute all false edges.
    //
    int i, j;        // Loop variables.
    int nodeOne;
    int nodeTwo;
    int *falseEdge;  // Store a false edge.
    int sizeU = neighborhoodU.size();
    int sizeV = neighborhoodV.size();
    // Compute all combinations.
    for (i=0; i<sizeU; i++) {
        for (j=0; j<sizeV; j++) {
	    nodeOne = neighborhoodU[i];
	    nodeTwo = neighborhoodV[j];
	    if (nodeOne > nodeTwo) {
                if (!lowerAdjacentMatrix[nodeOne-1][nodeTwo]) {
            	    falseEdge = new int[2];
                    falseEdge[0] = nodeOne;
                    falseEdge[1] = nodeTwo;
                    falseEdges->push_back(falseEdge);
		}
  	    }
	    else {
		if (!lowerAdjacentMatrix[nodeTwo-1][nodeOne]) {
            	    falseEdge = new int[2];
                    falseEdge[0] = nodeTwo;
                    falseEdge[1] = nodeOne;
                    falseEdges->push_back(falseEdge);
		}
	    }
        }
    }// for

}// end computeFalseEdges(..)


// -----------------------------------------------------------------------------
// g e t N e i g h b o r h o o d
//
// This function returns the neighborhood of node 'node' in the reduction graph.
// Therefore it goes in the adjacent matrix trough the list of adjacent
// nodes to 'node' and stores them.
// -----------------------------------------------------------------------------
void EdgeProjection::getNeighborhood(int node, vector<int> *neighborhood) {

    int i, j;    // Loop variable.
    int vertex;  // Store a node.

    //
    // Compute the neighborhood of node 'node'.
    //
    for (i=0; i<node; i++) {
        if (lowerAdjacentMatrix[node-1][i]) {
            neighborhood->push_back(i);
        }
    }// for
    for (j=i+1; j<graphSize; j++) {
        if (lowerAdjacentMatrix[j-1][node]) {
            neighborhood->push_back(j);
        }
    }// for

}// end getNeighborhood(..)


// -----------------------------------------------------------------------------
// u p d a t e G r a p h
//
// This is an overloaded function.
// Delete projection edge and all nodes to be deleted.
// The endnodes of the projection node will stay in the graph. They are deleted
// in the other function 'updateGraph'.
// !
// ! Important: The first node in vector deleteEdges is node u or v.
// !
// -----------------------------------------------------------------------------
void EdgeProjection::updateGraph() {

    int i, j, k, l; // Loop variables.
    int node;       // Store a node.
    int vertex;     // Store another node.
    int size;       // Store a size.

    //
    // Delete the projection edge.
    //
    if (edge[0] > edge[1]) {
        lowerAdjacentMatrix[edge[0]-1][edge[1]] = 0;
    }
    else {
        lowerAdjacentMatrix[edge[1]-1][edge[0]] = 0;
    }

    //
    // Delete all edges of vector 'deleteEdges'.
    // This is done in order to get a strongly projectable edge.
    //
    size = deleteEdges.size();
    for (i=0; i<size; i++) {
        node   = deleteEdges[i][0];
        vertex = deleteEdges[i][1];
        if (node > vertex) {
            lowerAdjacentMatrix[node-1][vertex] = 0;
        }
        else {
            lowerAdjacentMatrix[vertex-1][node] = 0;
        }
        // Update size of the adjacent list.
        adjacentListSize[vertex] -= 1;
    }// for

    //
    // Delete all nodes of vector 'deleteNodes'.
    //
    size = deleteNodes.size();
    for (i=0; i<size; i++) {
        node = deleteNodes[i];
        adjacentListSize[node] = 0;
        // Go through the neighborhood of node 'node'.
        for (j=0; j<node; j++) {
            if (lowerAdjacentMatrix[node-1][j]) {
                lowerAdjacentMatrix[node-1][j] = 0;
                // Update size of adjacent list
                adjacentListSize[j] -= 1;
                if ((j!=edge[0]) && (j!=edge[1])) {
                    for (k=0; k<i; k++) {
                        if (j == deleteNodes[k]) {
                            break;
                        }
                    }
                    if (k == i) {
                        // Add this edge to the deleted edges.
                        int *anEdge = new int[2];
                        anEdge[0] = node;
                        anEdge[1] = j;
                        deleteEdges.push_back(anEdge);
                    }
                }// if
            }// if
        }// for
        for (k=j+1; k<graphSize; k++) {
            if (lowerAdjacentMatrix[k-1][node]) {
                lowerAdjacentMatrix[k-1][node] = 0;
                // Update size of adjacent list
                adjacentListSize[k] -= 1;
                if ((k!=edge[0]) && (k!=edge[1])) {
                    for (l=i+1; l<size; l++) {
                        if (k == deleteNodes[l]) {
                            break;
                        }
                    }
                    if (l == size) {
                        // Add this edge to the deleted edges.
                        int *anEdge = new int[2];
                        anEdge[0] = node;
                        anEdge[1] = k;
                        deleteEdges.push_back(anEdge);
                    }
                }// if
            }
        }// for
    }// for

}// end updateGraph()


// -----------------------------------------------------------------------------
// u p d a t e G r a p h
//
// This is an overloaded function.
// Delete the two endnodes of the projection edge and add all false edges.
// -----------------------------------------------------------------------------
void EdgeProjection::updateGraph(vector<int*> *falseEdges) {

    int i, j, k; // Loop variables.
    int node;    // Store a node.
    int vertex;  // Store another node.
    int size;    // Store a size.
    int *anEdge; // A deleted egde.

    //
    // Delete both endnodes of the projection edge from the graph.
    //
    for (i=0; i<2; i++) {
        node = edge[i];
        adjacentListSize[node] = 0;
        // Go through the neighborhood of node 'node'.
        for (j=0; j<node; j++) {
            if (lowerAdjacentMatrix[node-1][j]) {
                lowerAdjacentMatrix[node-1][j] = 0;
                // Update size of adjacent list
                adjacentListSize[j] -= 1;
                // Add this edge to the deleted edges.
                anEdge = new int[2];
                anEdge[0] = node;
                anEdge[1] = j;
                deleteEdges.push_back(anEdge);
            }
        }// for
        for (k=j+1; k<graphSize; k++) {
            if (lowerAdjacentMatrix[k-1][node]) {
                lowerAdjacentMatrix[k-1][node] = 0;
                // Update size of adjacent list
                adjacentListSize[k] -= 1;
                // Add this edge to the deleted edges.
                anEdge = new int[2];
                anEdge[0] = k;
                anEdge[1] = node;
                deleteEdges.push_back(anEdge);
            }
        }// for
    }// for


    //
    // Add false edges.
    //
    size = falseEdges->size();
    for (i=0; i<size; i++) {
        node   = (*falseEdges)[i][0];
        vertex = (*falseEdges)[i][1];
        if (vertex > node) {
            lowerAdjacentMatrix[vertex-1][node] = 1;
        }
        else {
            lowerAdjacentMatrix[node-1][vertex] = 1;
        }
        adjacentListSize[node]   += 1;
        adjacentListSize[vertex] += 1;
    }// for

}// end updateGraph(..)


// -----------------------------------------------------------------------------
// a d d N o d e
//
// Add a new node to the tree.
// -----------------------------------------------------------------------------
void EdgeProjection::addNode(vector<int*> *falseEdges) {

    //
    // If currentNode == NULL
    // -> tree is empty
    // -> make new node
    //
    if (currentNode == NULL) {
        // New node.
        ProjectionNode *pN;
        pN = new ProjectionNode(edge, &deleteNodes, &deleteEdges, falseEdges,
                                NULL);

        // Store it in the vector.
        itsNodes.push_back(pN);
        // Update pointer to current node.
        currentNode = pN;
    }
    else {
        // Make new node in the correct part of the tree
        // and update current node.
        currentNode = currentNode->addNode(edge, &deleteNodes, &deleteEdges,
                                           falseEdges);
    }

}// end addNode(..)


// -----------------------------------------------------------------------------
// g e t R e d u c e d G r a p h
//
// Compute the reduced graph from the adjacent matrix.
// Ignore isolated nodes!!
// -----------------------------------------------------------------------------
void EdgeProjection::getReducedGraph(vector< vector<int> > *reducedGraph) {

    int i, j;     // Loop variables.
    int cnt = 0;  // Loop counter.
    int cnt2 = 0; // Another loop counter.
    int reducedGraphSize; // Size of the new reduced graph.

    //
    // Translation for the construction of the graph
    //
    int translateNode[graphSize];
    // Clean this vector.
    if (!translator.empty()) {
        translator.clear();
    }
    // Overestimate size of graph
    translator.resize(graphSize);
    for (i=0; i<graphSize; i++) {
        if (adjacentListSize[i] != 0) {
            translateNode[i] = cnt;
            // Translator for restranslation of reduced graph to original graph
            translator[cnt] = i;
            cnt++;
        }
        else {
            translateNode[i] = -1;
        }
    }
    // Correct the vector size.
    translator.resize(cnt);

    //
    // Construct reduced graph
    //
    reducedGraphSize = cnt;
    (*reducedGraph).resize(reducedGraphSize);

    for (i=0; i<graphSize; i++) {
        if (adjacentListSize[i] != 0) {
            cnt = 0;
            (*reducedGraph)[cnt2].resize(adjacentListSize[i]);
            for (j=0; j<i; j++) {
                if (lowerAdjacentMatrix[i-1][j]) {
                    (*reducedGraph)[cnt2][cnt] = translateNode[j];
                    cnt++;
                }
            }
            for (j=i+1; j<graphSize; j++) {
                if (lowerAdjacentMatrix[j-1][i]) {
                    (*reducedGraph)[cnt2][cnt] = translateNode[j];
                    cnt++;
                }
            }
            cnt2++;
        }
    }// for

}// end getReducedGraph(..)


// -----------------------------------------------------------------------------
// c l i q u e S e p a r a t i o n
// -----------------------------------------------------------------------------
int EdgeProjection::cliqueSeparation(vector< vector<int> > *graph,
                                     double *lpValue,
                                     vector< vector<int> > *theLhs) {

    int i, j;     // Loop counter.
    int index;
    double value; // Value of a computed clique.
    int numberOfCutsBeginning = theLhs->size();
    int numberOfCuts = numberOfCutsBeginning;
    vector<int> clique(2);     // Store the maximAL clique.
    vector<int> adjacentNodes; // All adjacent nodes to that clique.
    int size = graph->size();  // size of the current problem.
    vector<int*> *falseEdges = currentNode->getFalseEdges();
    int falseEdgesCnt = 0;     // Number of false edges in that node of the
			       // tree.
    int cnt = 0;
    bool found;
    bool newOne;	       // Indicate if that inequality has not been
			       // generated.
    bool STOP = false;
    int theCnt = 0;
    int cutCnt = 0;

    //
    // Parameter.
    // 
    static const int    MAX_ITER_SEP  = size * 2;
    static const int    MAX_STEP      = size / 5;
    static const int    MAX_CUTS      = size;
    static const double MIN_VIOLATION = .000001;

    //
    // Construct maximal cliques.
    //
    while ( (cnt < MAX_ITER_SEP) ) { // || theCnt < MAX_STEP) && (cutCnt < MAX_CUTS) ) {

        //
        // Select edge to start from.
        //
        if (falseEdgesCnt < falseEdges->size()) {
            clique[0] = (*falseEdges)[falseEdgesCnt][0];
            // Translate this nodes:
            i = 0;
            while(i < size) {
                if (clique[0] == translator[i]) {
                    clique[0] = i;
                    break;
                }
                i++;
            }
            if (i == size) {
                cout << "ERROR" << endl;
                exit(1);
            }

            clique[1] = (*falseEdges)[falseEdgesCnt++][1];
            // Translate this nodes:
            i = 0;
            while(i < size) {
                if (clique[1] == translator[i]) {
                    clique[1] = i;
                    break;
                }
                i++;
            }
            if (i == size) {
                cout << "ERROR" << endl;
                exit(1);
            }
        }
        else {
            found = false;
            while (!found) {
                clique[0] = getRandomNumber(size);
                clique[1] = getRandomNumber(size);
                if (clique[0] != clique[1]) {
                    if (clique[0] > clique[1]) {
                        if (lowerAdjacentMatrix[translator[clique[0]]-1][translator[clique[1]]]) {
                            found = true;
                        }
                    }
                    else {
                        if (lowerAdjacentMatrix[translator[clique[1]]-1][translator[clique[0]]]) {
                            found = true;
                        }
                    }
                }// if
            }// while
        }

        //
        // Get list of adjacent nodes to this edge.
        //
        adjacentNodes = (*graph)[clique[0]];
        for (i=adjacentNodes.size()-1; i>=0; i--) {
            if (clique.back() == adjacentNodes[i]) {
                adjacentNodes.erase(adjacentNodes.begin()+i);
                continue;
            }
            if (clique.back() > adjacentNodes[i]) {
                if (!lowerAdjacentMatrix[translator[clique.back()]-1][translator[adjacentNodes[i]]]) {
                    adjacentNodes.erase(adjacentNodes.begin()+i);
                }
            }
            else {
                if (!lowerAdjacentMatrix[translator[adjacentNodes[i]]-1][translator[clique.back()]]) {
                    adjacentNodes.erase(adjacentNodes.begin()+i);
                }
            }
        }

        //
        // Construct a maximAL clique containing this egde.
        //
        while(!adjacentNodes.empty()) {
            index = getRandomNumber(adjacentNodes.size());
            clique.push_back(adjacentNodes[index]);
            adjacentNodes.erase(adjacentNodes.begin() + index);
            // Update adjacent nodes.
            for (i=adjacentNodes.size()-1; i>=0; i--) {
                if (clique.back() > adjacentNodes[i]) {
                    if (!lowerAdjacentMatrix[translator[clique.back()]-1][translator[adjacentNodes[i]]]) {
                        adjacentNodes.erase(adjacentNodes.begin()+i);
                    }
                }
                else {
                    if (!lowerAdjacentMatrix[translator[adjacentNodes[i]]-1][translator[clique.back()]]) {
                        adjacentNodes.erase(adjacentNodes.begin()+i);
                    }
                }
            }
        }

        //
        // Check if this inequality is violated.
        //
        theCnt++;
        value = 0;
        for (i=0; i<clique.size(); i++) {
            value += lpValue[clique[i]];
        }
        if (value > 1 + MIN_VIOLATION) {
            // Store this inequality if it is a new one.
            sort(clique.begin(), clique.end());
            newOne = true;
            for (i=numberOfCutsBeginning; i<numberOfCuts; i++) {
                if ((*theLhs)[i].size() != clique.size()) {
                    continue;
                }
		j=0;
		while (j<(*theLhs)[i].size()) {
                    if (clique[j] != (*theLhs)[i][j]) {
                        break;
                    }
		    j++;
                }
                if (j == (*theLhs)[i].size()) {
                    newOne = false;
                    break;
                }
            }
            if (newOne) {
                numberOfCuts++;
                theLhs->resize(numberOfCuts);
                for (i=0; i<clique.size(); i++) {
                    (*theLhs)[numberOfCuts-1].push_back(clique[i]);
                }
                theCnt = 0;
	        cutCnt++;
            }// if
        }// if

        clique.clear();
        adjacentNodes.clear();
        clique.resize(2);
        cnt++;
    }

    return cutCnt;

}// end cliqueSeperation(..)


// -----------------------------------------------------------------------------
// s o l v e S t a b l e S e t
//
// Solve the maximum-cardinality stable set problem for that graph exact.
// Call therefore the stable set solver and store the corresponding rank
// inequality.
// -----------------------------------------------------------------------------
void EdgeProjection::solveStableSet(vector<int> *theLhs,
                                    int &theRhs, StableSetMaster *theMaster) {

    int i, j;
    char *graph = theMaster->getSmallGraphName();
    string theGraph("./");
    theGraph.append(graph);
    string inequalities(graph);
    inequalities.append("Ineq.txt"); 

    //
    // Write file with information about the graph.
    //
    int numberOfNodes = 0;
    int numberOfEdges = 0;
    translator.clear();
    translator.resize(graphSize, -1);

    // Calculate number of nodes and number of edges.
    for (i=0; i<graphSize; i++) {
	if (adjacentListSize[i] != 0) {
	    translator[i] = numberOfNodes;
	    numberOfNodes++;
	    numberOfEdges += adjacentListSize[i];
	}
    }
	
    ofstream fout(graph);
    fout << "NNODES : " << numberOfNodes << endl;
    fout << "NEDGES : " << numberOfEdges/2 << endl;
    fout << "WEIGHTS : NO\n";
    fout << "\nEDGE_SECTION :" << endl;
    for (i=0; i<graphSize-1; i++) {
        for (j=i; j<graphSize-1; j++) {
	    if (lowerAdjacentMatrix[j][i]) {
		fout << translator[i] << " " << translator[j+1] << endl;
	    }	    
	}
    }
    fout.close();


    //
    // Generate new stable set object and solve the maximum
    // cardinality stable set problem for that small graph.
    //
    StableSetMaster *stableSet;  
    stableSet = new StableSetMaster(theGraph.c_str(), inequalities.c_str(),//
				    "DUMMY");
    // No output on screan.
    stableSet->outLevel(ABA_MASTER::Silent);
    ABA_STRING theAbaString(stableSet, "00:00:01");
    stableSet->maxCpuTime(theAbaString);
    // Optimize.    
    ABA_MASTER::STATUS error = stableSet->optimize();

    //
    // Check optimality.
    //
    if (error) {
	exit(1);
    }
/*
    string check("./CheckInequalities.exe ");
    check.append("~/Diplomarbeit/Abacus/Stable_Set/");
    check.append(graph);
    check.append(" ");
    check.append(inequalities);

    theMaster->stopTimer();
    theMaster->out() << "\n\nCheck inequalities:" << endl;
    theMaster->out() << "----------------------------";
    theMaster->out() << "------------------------\n" << endl;

    // Call program to check all inequalities.
    int notValid = system(check.c_str());
    check.clear(); 

    theMaster->out() << "----------------------------";
    theMaster->out() << "------------------------\n" << endl;
    theMaster->startTimer();

    if (notValid) {
        exit(1);
    }
*/
	
    //
    // Construct inequality.
    //    
    for (i=0; i<graphSize; i++) {
	if (adjacentListSize[i] != 0) {
	    theLhs->push_back(i);
	}
    }    
    theRhs = (int) stableSet->dualBound() + stableSet->fixedNodesToOne();

    // Clean up.
    delete stableSet;    

}// end solveStableSet(..)


// -----------------------------------------------------------------------------
// l i f t A n d S t o r e
//
// Make the rank inequality valid for the whole graph which is called anti-
// projection. Store the inequality after that.
// -----------------------------------------------------------------------------
void EdgeProjection::liftAndStore(vector<int> *theLhs, int theRhs) {

    int i, j, k;  // Loop variables.
    int size;     // Store a size.
    int nodeOne;  // One endnode of the false edges.
    int nodeTwo;  // The other endnode of the false edges.
    int stack;    // Store something temporarily.
    vector<int*>* theFalseEdges;    // Get the false edges for each node.
    vector<int>* theDeletedNodes;   // Get the deleted nodes for each node.
    int deletedNodesSize;    // Size of vector 'theDeletedNodes'.
    ProjectionNode* aNode;   // A node of the tree.
    
    rhs.resize(1);
    rhs[0] = theRhs;	


    //
    // 'Lift' this inequality.
    // Therefore check for each node of the tree all false edges
    // if they are contained in this inequality.
    //
    aNode = currentNode;
    // Go the way back to the root.
    while (aNode != NULL) {

        // Get all false edges of tree node 'aNode'
        theFalseEdges = aNode->getFalseEdges();

        // Check if one of this false edges are contained.
        j = 0;
        size = theLhs->size();
        while (j < theFalseEdges->size()) {
            nodeOne = (*theFalseEdges)[j][0];
            nodeTwo = (*theFalseEdges)[j][1];
            j++;
            // Check if this two nodes are part of the inequalities
            if (nodeOne > nodeTwo) {
                stack = nodeOne;
                nodeOne = nodeTwo;
                nodeTwo = stack;
            }
            k = 0;
            while (k < size) {
                if( (*theLhs)[k] >= nodeOne ) {
                    break;
                }
                k++;
            }// while
            if (k == size) {
                // This node is not contained in this inequality.
                continue;
            }
            if ((*theLhs)[k] != nodeOne) {
                // This node is not contained in this inequality.
                continue;
            }
            k++;
            while (k < size) {
                if( (*theLhs)[k] >= nodeTwo ) {
                    break;
                }
                k++;
            }
            if (k == size) {
                // This node is not contained in this inequality.
                continue;
            }
            if ((*theLhs)[k] != nodeTwo) {
                // This node is not contained in this inequality.
                continue;
            }
            //
            // This false edge is conained in the inequality
            // -> 'lift' this inequality
            //
            // Add all deleted node.
            theDeletedNodes = aNode->getDeletedNodes();
            deletedNodesSize = theDeletedNodes->size();
            for (k=0; k<deletedNodesSize; k++) {
                theLhs->push_back((*theDeletedNodes)[k]);
            }
            
	    // Add the projection edge.
            theLhs->push_back(aNode->getEdge(0));
            theLhs->push_back(aNode->getEdge(1));
            
            // Sort the inequality.
            sort(theLhs->begin(), theLhs->end());
                // Update the right hand side.
                rhs[0] += 1;
                                
            // The update of this inequality is finished for this node
            // of the tree. Each edge projection step increases the rhs
            // only by value 1.
            break;
        }// end while
        
        // Get previous node in the tree.
        aNode = aNode->getPreviousNode();
    }// while

    //
    // Store this inequality
    //
    vector<int> *inequality = new vector<int>;
    inequality->resize(theLhs->size());
    for (j=0; j<theLhs->size(); j++) {
        (*inequality)[j] = (*theLhs)[j];
    }
    lhs.push_back(inequality);

}// end lifting(..)


// -----------------------------------------------------------------------------
// g e t R a n d o m N u m b e r
// -----------------------------------------------------------------------------
int EdgeProjection::getRandomNumber(int bound) const {
    int    aNumber = rand();
    double d       = (double(aNumber)/RAND_MAX);

    return (int)(d*bound);
}


// -----------------------------------------------------------------------------
// C l e a n U p
// -----------------------------------------------------------------------------
void EdgeProjection::cleanUp() {

    int i, j;               // Loop variables.
    int nodeOne;            // One endnode of the projection edge.
    int nodeTwo;            // The other endnode of the projection edge.
    int size;               // Store the size of a vector.
    vector<int> *nodes;     // Get some nodes.
    vector<int*> *edges;    // Get some edges.

    //
    // Reconstruct graph.
    // At the end of this loop is currentNode = NULL.
    //
    while (currentNode != NULL) {

        //
        // Add projection edge.
        //
        nodeOne = currentNode->getEdge(0);
        nodeTwo = currentNode->getEdge(1);
        if (nodeOne > nodeTwo) {
            lowerAdjacentMatrix[nodeOne-1][nodeTwo] = 1;
        }
        else {
            lowerAdjacentMatrix[nodeTwo-1][nodeOne] = 1;
        }
        adjacentListSize[nodeOne] += 1;
        adjacentListSize[nodeTwo] += 1;

        //
        // Add clique build by all deleted nodes including the two endnodes
        // of the projection edge.
        //
        nodes = currentNode->getDeletedNodes();
        size = nodes->size();
        for (i=0; i<size; i++) {
            // Add edge to nodeOne.
            if (nodeOne > (*nodes)[i]) {
                lowerAdjacentMatrix[nodeOne-1][(*nodes)[i]] = 1;
            }
            else {
                lowerAdjacentMatrix[(*nodes)[i]-1][nodeOne] = 1;
            }
            adjacentListSize[nodeOne] += 1;
            adjacentListSize[(*nodes)[i]] += 1;

            // Add edge to node two.
            if (nodeTwo > (*nodes)[i]) {
                lowerAdjacentMatrix[nodeTwo-1][(*nodes)[i]] = 1;
            }
            else {
                lowerAdjacentMatrix[(*nodes)[i]-1][nodeTwo] = 1;
            }
            adjacentListSize[nodeTwo] += 1;
            adjacentListSize[(*nodes)[i]] += 1;

            // Add edges to all other deleted nodes.
            for (j=i+1; j<size; j++) {
                if ((*nodes)[j] > (*nodes)[i]) {
                    lowerAdjacentMatrix[(*nodes)[j]-1][(*nodes)[i]] = 1;
                }
                else {
                    lowerAdjacentMatrix[(*nodes)[i]-1][(*nodes)[j]] = 1;
                }
                adjacentListSize[(*nodes)[i]] += 1;
                adjacentListSize[(*nodes)[j]] += 1;
            }// for
        }// for

        //
        // Add all other deleted edges.
        //
        edges = currentNode->getDeletedEdges();
        size = edges->size();
        for (i=0; i<size; i++) {
            if ((*edges)[i][1] > (*edges)[i][0]) {
                lowerAdjacentMatrix[(*edges)[i][1]-1][(*edges)[i][0]] = 1;
            }
            else {
                lowerAdjacentMatrix[(*edges)[i][0]-1][(*edges)[i][1]] = 1;
            }
            adjacentListSize[(*edges)[i][0]] += 1;
            adjacentListSize[(*edges)[i][1]] += 1;
        }// for

        //
        // Delete all false edges.
        //
        edges = currentNode->getFalseEdges();
        size = edges->size();
        for (i=0; i<size; i++) {
            if ((*edges)[i][1] > (*edges)[i][0]) {
                lowerAdjacentMatrix[(*edges)[i][1]-1][(*edges)[i][0]] = 0;
            }
            else {
                lowerAdjacentMatrix[(*edges)[i][0]-1][(*edges)[i][1]] = 0;
            }
            adjacentListSize[(*edges)[i][0]] -= 1;
            adjacentListSize[(*edges)[i][1]] -= 1;
        }// for

        //
        // Go one step back in the tree.
        //
        currentNode = currentNode->getPreviousNode();
    }// while

    // Call the destructor of each node.
    for (int i=itsNodes.size()-1; i>=0; i--) {
        delete itsNodes[i];
        itsNodes[i] = NULL;
    }
    itsNodes.clear();

}// end cleanUp()


// -----------------------------------------------------------------------------
// p r i n t R o o t
// -----------------------------------------------------------------------------
void EdgeProjection::printRoot() {

    cout << "\n\nReduction Graph:\n";
    for (int i=0; i<lowerAdjacentMatrix.size(); i++) {
        cout << "\t";
        for (int j=0; j<lowerAdjacentMatrix[i].size(); j++) {
            cout << lowerAdjacentMatrix[i][j] << " ";
        }
        cout << endl;
    }

    cout << "\nAdjacent List sizes:\n\t";
    for (int i=0; i<graphSize; i++) {
        cout << adjacentListSize[i] << " ";
    }
    cout << endl << endl << endl;

}// end printRoot()




// *****************************************************************************
//
// Old code fragments.
//
// *****************************************************************************

/*
// -----------------------------------------------------------------------------
// e d g e S e l e c t i o n
//
// Select a projctable edge. Consider first the strongly projectable edges. 
// If there is no good one, check randomly other edges and ???
// -----------------------------------------------------------------------------
void EdgeProjection::edgeSelection(double *lpValue) {
cout << "Edge selection:" << endl;
    
    int i, j, k;      // Loop counter.
    int size;         // Store a size.

    int stack;        // Store something temporarily.
    int startNodeOne; // First start node for search.
    int startNodeTwo; // Second start node for search.
    int nodeOne;
    int nodeTwo;
    double weight;    // Weight of an edge.
    double bestWeight = 0; // Highest weight of an edge.
    bool edgeFound = true;
    
    
    //
    // Consider first the strongly projectable edges.
    //
//    if (phase == 1) {
    if (!projectable.empty()) {
        // -> !projectable.empty()
        size = projectable.size();
        for (i=0; i<size; i++) {
            weight = lpValue[projectable[i][0]] + lpValue[projectable[i][1]];
            // Higher weight?
            if (weight > bestWeight) {
                // Update edge and weight.
                bestWeight = weight;
                edge[0] = projectable[i][0];
                edge[1] = projectable[i][1];

                // Already an acceptable weight?
                if (weight > ACCEPT_EDGE) {
                    break;
                }
            }// if
        }// for

        if (weight > NOT_ACCEPT_EDGE) {
            // Delete this edge from vector 'projectable'.        
            delete[] projectable[i];
            projectable[i] = NULL;
            projectable.erase(projectable.begin() + i);
//            projectableEdge = true;
        }
        else {
            edgeFound = false;
        }
    }// if
    
    
//    if ((phase != 1) || !edgefound) {
    if (projectable.empty() || !edgeFound) {
        // Choose two random start nodes.
        startNodeOne = getRandomNumber(graphSize);
        startNodeTwo = getRandomNumber(graphSize);
        while (startNodeOne == startNodeTwo) {
            startNodeTwo = getRandomNumber(graphSize);
        }
        // Sort for the adjacent matrix.    
        if (startNodeOne < startNodeTwo) {
            stack = startNodeOne;
            startNodeOne = startNodeTwo;
            startNodeTwo = stack;
        }    
        // Walk through the adjacent matrix.
        nodeOne = startNodeOne;
        nodeTwo = startNodeTwo;
        do { 
            // Check if this two nodes are adjacent.
            if (lowerAdjacentMatrix[nodeOne-1][nodeTwo]) {
                // Calculate weight of edge.
                weight = lpValue[nodeOne] + lpValue[nodeTwo];
                // Higher weight?
                if (weight > bestWeight) {
                    // Already an acceptable weight?
                    if (weight > ACCEPT_EDGE) {
                       // This is our projection edge.
                       bestWeight = weight;
                       edge[0] = nodeOne;
                       edge[1] = nodeTwo;
                       break;
                    }
                    bestWeight = weight;
                    edge[0] = nodeOne;
                    edge[1] = nodeTwo;
                }
            }// if
            // Select the next edge.
            nodeTwo++;
            if (nodeTwo == nodeOne) {
                nodeOne++;
                nodeTwo = 0;
                if (nodeOne == graphSize) {
                    nodeOne = 1;
                }
            }
        } while ((nodeOne != startNodeOne) || (nodeTwo != startNodeTwo)); 
    
        // Check if there was an edge with acceptable weight.
        if (bestWeight < NOT_ACCEPT_EDGE) {
            cout << "No projection edge could be found.";
            exit(1);
        }    
     
     
        // 
        // Check if this edge is strongly projectable.
        // In this case we only check if its neighborhood, without the other 
        // endnode of the projection edge, is a clique.   
        //
        
        // Find node with smaller degree.
        int index = 0;
        if (adjacentListSize[edge[0]] > adjacentListSize[edge[1]]) {
            index = 1;
        }
        // -> edge[index] has smaller degree.
    
        // Get neighborhood of edge[index]
        vector<int> neighborhood;
        getNeighborhood(edge[index], &neighborhood);
        // Delte edge[(index+1)%2] from the neighborhood
        i = 0;
        size = neighborhood.size();
        while (i < size) {
            if (neighborhood[i] == edge[(index+1)%2]) {
                neighborhood.erase(neighborhood.begin()+i);
                size--;
                break;
            }
            i++;
        }
//cout << "neighborhood of " << edge[index] << endl << "\t";
//for (j=0; j<neighborhood.size();j++) cout << neighborhood[j] << " ";
//cout << endl;

        // Check if neighborhood is a clique.
        bool clique = true;
        for (i=0; i<size-1; i++) {
            for (j=i+1; j<size; j++) {
                if (!lowerAdjacentMatrix[neighborhood[j]-1][neighborhood[i]]) {
//cout << "nodes " << neighborhood[j] << " and " << neighborhood[i] << " are not adjacent" << endl;
                    clique = false;
                    break;
                }
            }// for
            if (!clique) {
                break;
            }
        }// for

//cout << "\tProjection edge:\t" << edge[0] << " " << edge[1] << endl;
        if (clique) {
            // We have found a strongly projectable edge!
            for (i=0; i<size; i++) {
                if (neighborhood[i] > edge[(index+1)%2]) {
                    if (lowerAdjacentMatrix[neighborhood[i]-1][edge[(index+1)%2]]) {
                        deleteNodes.push_back(neighborhood[i]);
                    }
                }
                else {
                    if (lowerAdjacentMatrix[edge[(index+1)%2]-1][neighborhood[i]]) {
                        deleteNodes.push_back(neighborhood[i]);
                    }
                }
            }
cout << "\tNeighborhood of node " << edge[index] << " is a clique." << endl;
        }
        else {
            //
            // Make it a projectable edge.
            //
cout << "\tNeighborhood of node " << edge[index] << " is NO clique." << endl;
            
            // Construct induced subgraph of neighborhood.
            vector< vector<int> > adjacentList(size);
            for (i=0; i<size-1; i++) {
                for (j=i+1; j<size; j++) {
                    if (lowerAdjacentMatrix[neighborhood[j]-1][neighborhood[i]]) {
                        adjacentList[i].push_back(j);
                        adjacentList[j].push_back(i);
                    }
                }                
            }        

            // Compute maximAL clique in induced graph of neighborhood.
            vector<int> *maxClique;
            // This can be any maximAL/maximUM clique generator.
            maxClique = computeMaximalClique(&adjacentList, &neighborhood);
if (maxClique == NULL) {
    cout << "ERROR with maximal clique computation." << endl;
    exit(0);
}            
            // Tranlsate clique and fill vectors deleteNodes, deleteEdges
            int cliqueSize = maxClique->size();
            for (i=0; i<cliqueSize; i++) {
                (*maxClique)[i] = neighborhood[(*maxClique)[i]];
            }
            sort(maxClique->begin(), maxClique->end());
cout << "\tMaximum clique:\t";
for(i=0; i<maxClique->size(); i++) cout << (*maxClique)[i] << " ";
cout << endl;
            deleteEdges.resize(size - cliqueSize);
            int edgesCnt = 0;
            int *anEdge;
            j = 0;
            for (i=0; i<size; i++) {
                // Check if neighborhood[i] is in the clique. If it is contained
                // this node will be deleted. Otherwise we have to delete an edge.
                while (j < cliqueSize) {
                    if ((*maxClique)[j] >= neighborhood[i]) {
                        if ((*maxClique)[j] == neighborhood[i]) {
                            if (neighborhood[i] > edge[(index+1)%2]) {
                                if (lowerAdjacentMatrix[neighborhood[i]-1][edge[(index+1)%2]]) {
                                    deleteNodes.push_back(neighborhood[i]);
                                }
                            }// if
                            else {
                                if (lowerAdjacentMatrix[edge[(index+1)%2]-1][neighborhood[i]]) {
                                    deleteNodes.push_back(neighborhood[i]);
                                }
                            }// else
                        }
                        else {
                            anEdge = new int[2];
                            anEdge[0] = edge[index];
                            anEdge[1] = neighborhood[i];
                            deleteEdges[edgesCnt++] = anEdge;
                        }
                        break;
                    }
                    j++;
                }// while
                if (j == cliqueSize) {
                    anEdge = new int[2];
                    anEdge[0] = edge[index];
                    anEdge[1] = neighborhood[i];
                    deleteEdges[edgesCnt++] = anEdge;
                }
            }// for

            // Clean up.
            maxClique->clear();
        }// else
    }// if

cout << "\tDelete nodes:\t";
for(i=0;i<deleteNodes.size();i++) cout << deleteNodes[i] << " ";
cout << "\n\tDelete edges:\t";
for(i=0;i<deleteEdges.size();i++) cout << deleteEdges[i][0] << " " << deleteEdges[i][1] << "; ";
cout << endl << endl << endl;

}// end edgeSelection()



// -----------------------------------------------------------------------------
// e d g e S e l e c t i o n
//
// Choose randomly two nodes and and go through the adjacent matrix until a "good"
// edge is found. Check if it is projectable. If the neighborhood of the 
// endnode with smaller degree is not a clique, contruct a maximUM clique in
// its neighborhood. 
// -----------------------------------------------------------------------------
void EdgeProjection::edgeSelection(double *lpValue) {
cout << "Edge selection:" << endl;
    
    int i, j, k;      // Loop variables.
    int size;         // Store a size.

    // 
    // Select an edge as a candidate for the projection edge.
    //
    int stack;        // Store something temporarily.
    int startNodeOne; // First start node for search.
    int startNodeTwo; // Second start node for search.
    int nodeOne;
    int nodeTwo;
    double weight;    // Weight of an edge.
    double bestWeight = 0; // Highest weight of an edge.
    
    // Choose two random start nodes.
    startNodeOne = getRandomNumber(graphSize);
    startNodeTwo = getRandomNumber(graphSize);
    while (startNodeOne == startNodeTwo) {
        startNodeTwo = getRandomNumber(graphSize);
    }
    // Sort for the adjacent matrix.    
    if (startNodeOne < startNodeTwo) {
        stack = startNodeOne;
        startNodeOne = startNodeTwo;
        startNodeTwo = stack;
    }    
    // Walk through the adjacent matrix.
    nodeOne = startNodeOne;
    nodeTwo = startNodeTwo;
    do { 
//cout << nodeOne << " " << nodeTwo << " ";
        // Check if this two nodes are adjacent.
        if (lowerAdjacentMatrix[nodeOne-1][nodeTwo]) {
            // Calculate weight of edge.
            weight = lpValue[nodeOne] + lpValue[nodeTwo];
//cout << weight;
            // Higher weight?
            if (weight > bestWeight) {
                // Already an acceptable weight?
                if (weight > ACCEPT_EDGE) {
                   // This is our projection edge.
                   bestWeight = weight;
                   edge[0] = nodeOne;
                   edge[1] = nodeTwo;
//cout << endl;
                   break;
                }
                bestWeight = weight;
                edge[0] = nodeOne;
                edge[1] = nodeTwo;
            }
        }// if
//cout << endl;
        // Select the next edge.
        nodeTwo++;
        if (nodeTwo == nodeOne) {
            nodeOne++;
            nodeTwo = 0;
            if (nodeOne == graphSize) {
                nodeOne = 1;
            }
        }
    } while ((nodeOne != startNodeOne) || (nodeTwo != startNodeTwo)); 
//cout << edge[0] << " " << edge[1] << " " << bestWeight << endl;
    
    // Check if there was an edge with acceptable weight.
    if (bestWeight < NOT_ACCEPTABLE) {
        cout << "No projection edge could be found.";
        exit(1);
    }    
     
    // 
    // Check if this edge is strongly projectable   
    //
    // Find node with smaller degree.
    int index = 0;
    if (adjacentListSize[edge[0]] > adjacentListSize[edge[1]]) {
        index = 1;
    }
    // -> edge[index] has smaller degree.
    
    // Get neighborhood of edge[index]
    vector<int> neighborhood;
    getNeighborhood(edge[index], &neighborhood);
//cout << "neighborhood of " << edge[index] << endl << "\t";
//for (j=0; j<neighborhood.size();j++) cout << neighborhood[j] << " ";
//cout << endl;   
    // Delte edge[(index+1)%2] from the neighborhood
    i = 0;
    size = neighborhood.size();
    while (i < size) {
        if (neighborhood[i] == edge[(index+1)%2]) {
            neighborhood.erase(neighborhood.begin()+i);
            size--;
            break;
        }
        i++;
    }
//cout << "neighborhood of " << edge[index] << endl << "\t";
//for (j=0; j<neighborhood.size();j++) cout << neighborhood[j] << " ";
//cout << endl;
    // Check if this node was found.
    if(i == size+1) {
        cout << "ERROR" << endl;
        exit(1);
    }

    // Check if neighborhood is a clique.
    bool clique = true;
    for (i=0; i<size-1; i++) {
        for (j=i+1; j<size; j++) {
            if (!lowerAdjacentMatrix[neighborhood[j]-1][neighborhood[i]]) {
//cout << "nodes " << neighborhood[j] << " and " << neighborhood[i] << " are not adjacent" << endl;
                clique = false;
                break;
            }
        }// for
        if (!clique) {
            break;
        }
    }// for

cout << "\tProjection edge:\t" << edge[0] << " " << edge[1] << endl;
    if (clique) {
        // We have found a strongly projectable edge!
        for (i=0; i<size; i++) {
            if (neighborhood[i] > edge[(index+1)%2]) {
                if (lowerAdjacentMatrix[neighborhood[i]-1][edge[(index+1)%2]]) {
                    deleteNodes.push_back(neighborhood[i]);
                }
            }
            else {
                if (lowerAdjacentMatrix[edge[(index+1)%2]-1][neighborhood[i]]) {
                    deleteNodes.push_back(neighborhood[i]);
                }
            }
        }
cout << "\tNeighborhood of node " << edge[index] << " is a clique." << endl;
    }
    else {
        //
        // Make it a projectable edge.
        //
cout << "\tNeighborhood of node " << edge[index] << " is NO clique." << endl;
        // Construct induced subgraph of neighborhood.
        vector< vector<int> > adjacentList(size);
        for (i=0; i<size-1; i++) {
            for (j=i+1; j<size; j++) {
                if (lowerAdjacentMatrix[neighborhood[j]-1][neighborhood[i]]) {
                    adjacentList[i].push_back(j);
                    adjacentList[j].push_back(i);
                }
            }                
        }        

        // Compute maximUM clique in induced graph of neighborhood.
        vector<int> *maxClique;
        // This can be any exact maximum clique generator!
        maxClique = computeMaximumClique(&adjacentList, &neighborhood);
//cout << "maxClique:\t";
//for(i=0; i<maxClique->size(); i++) cout << (*maxClique)[i] << " ";
//cout << endl;
        // Tranlsate clique and fill vectors deleteNodes, deleteEdges
        int cliqueSize = maxClique->size();
        for (i=0; i<cliqueSize; i++) {
            (*maxClique)[i] = neighborhood[(*maxClique)[i]];
        }
        sort(maxClique->begin(), maxClique->end());
cout << "\tMaximum clique:\t";
for(i=0; i<maxClique->size(); i++) cout << (*maxClique)[i] << " ";
cout << endl;
//        deleteNodes.resize(cliqueSize);
//cout << size - cliqueSize << endl;
        deleteEdges.resize(size - cliqueSize);
//        int nodesCnt = 0;
        int edgesCnt = 0;
        int *anEdge;
        j = 0;
        for (i=0; i<size; i++) {
            // Check if neighborhood[i] is in the clique. If it is contained
            // this node will be deleted. Otherwise we have to delete an edge.
            while (j < cliqueSize) {
                if ((*maxClique)[j] >= neighborhood[i]) {
                    if ((*maxClique)[j] == neighborhood[i]) {
                        if (neighborhood[i] > edge[(index+1)%2]) {
                            if (lowerAdjacentMatrix[neighborhood[i]-1][edge[(index+1)%2]]) {
                                deleteNodes.push_back(neighborhood[i]);
                            }
                        }// if
                        else {
                            if (lowerAdjacentMatrix[edge[(index+1)%2]-1][neighborhood[i]]) {
                                deleteNodes.push_back(neighborhood[i]);
                            }
                        }// else
                    }
                    else {
                        anEdge = new int[2];
                        anEdge[0] = edge[index];
                        anEdge[1] = neighborhood[i];
                        deleteEdges[edgesCnt++] = anEdge;
                    }
                    break;
                }
                j++;
            }// while
            if (j == cliqueSize) {
                anEdge = new int[2];
                anEdge[0] = edge[index];
                anEdge[1] = neighborhood[i];
                deleteEdges[edgesCnt++] = anEdge;
            }
        }// for

        // Clean up.
        delete maxClique;

    }// else

cout << "\tDelete nodes:\t";
for(i=0;i<deleteNodes.size();i++) cout << deleteNodes[i] << " ";
cout << "\n\tDelete edges:\t";
for(i=0;i<deleteEdges.size();i++) cout << deleteEdges[i][0] << " " << deleteEdges[i][1] << "; ";
cout << endl << endl << endl;

}// end edgeSelection()

//
// EXACT COMPUTATION:
//
// -----------------------------------------------------------------------------
// c o m p u t e M a x i m u m C l i q u e
// -----------------------------------------------------------------------------
vector<int>* EdgeProjection::computeMaximumClique(
            vector< vector<int> > *theAdjacentList, vector<int> *theTranslator) {

    int i, j;
    int node;
    int vertex;
    vector<int> clique;         //
    vector<int> adjacentNodes;  //
    vector<int> *maxClique;     // Store a maximUM clique.
                                // This will be the return value.
    int maxCliqueSize = 1;      // Size of maximUM clique.

    clique.push_back(0);
    maxClique = new vector<int>;
    maxClique->push_back(0);
    adjacentNodes = (*theAdjacentList)[0];
    bool stop = false;

    while (!stop) {
//cout << "Clique:\t";
//for(int l=0;l<clique.size();l++) cout << clique[l] << " ";
//cout << endl;
//cout << "Adjacent nodes:\t";
//for(int l=0;l<adjacentNodes.size();l++) cout << adjacentNodes[l] << " ";
//cout << endl;
       while(!adjacentNodes.empty() && (adjacentNodes.size() + clique.size()) > maxCliqueSize) {
           clique.push_back(adjacentNodes.back());
           adjacentNodes.pop_back();
           // Update adjacent nodes.
           for (i=adjacentNodes.size()-1; i>=0; i--) {
               if ((*theTranslator)[clique.back()] > (*theTranslator)[adjacentNodes[i]]) {
                   if(!lowerAdjacentMatrix[(*theTranslator)[clique.back()]-1][(*theTranslator)[adjacentNodes[i]]]) {
                       adjacentNodes.erase(adjacentNodes.begin()+i);
                   }
               }
               else {
                   if(!lowerAdjacentMatrix[(*theTranslator)[adjacentNodes[i]]-1][(*theTranslator)[clique.back()]]) {
                       adjacentNodes.erase(adjacentNodes.begin()+i);
                   }

               }
           }
       }
//cout << "maximal Clique:\t";
//for(int l=0;l<clique.size();l++) cout << clique[l] << " ";
//cout << endl;

       if ((adjacentNodes.size() + clique.size()) > maxCliqueSize) {
//cout << "New max clique:\t";
//for(int l=0;l<clique.size();l++) cout << clique[l] << " ";
//cout << endl;
           // Update maximUM clique.
           maxClique->clear();
           (*maxClique) = clique;
           maxCliqueSize = maxClique->size();
       }
//cout << "vor Clique:\t";
//for(int l=0;l<clique.size();l++) cout << clique[l] << " ";
//cout << endl;
       node = clique.back();
       clique.pop_back();
//cout << "Clique:\t";
//for(int l=0;l<clique.size();l++) cout << clique[l] << " ";
//cout << endl;

       if (clique.empty()) {
           if ((node+1) >= theAdjacentList->size()) {
               stop = true;
               continue;
           }
           clique.push_back(node +1);
           node = theAdjacentList->size();
       }
       adjacentNodes.clear();

       // Get adjacent list.
       i = 0;
       while (i<(*theAdjacentList)[clique[0]].size()){
           vertex = (*theAdjacentList)[clique[0]][i];
           if ( vertex >= node) {
               break;
           }
           j = 1;
           while (j < clique.size()) {
               if (clique[j] == vertex) {
                   break;
               }
               if ((*theTranslator)[vertex] > (*theTranslator)[clique[j]]) {
                   if (!lowerAdjacentMatrix[(*theTranslator)[vertex]-1][(*theTranslator)[clique[j]]]) {
                       break;
                   }
               }
               else {
                   if (!lowerAdjacentMatrix[(*theTranslator)[clique[j]]-1][(*theTranslator)[vertex]]) {
                       break;
                   }
               }
//cout << "trans:\t" << (*theTranslator)[vertex] << "   " << (*theTranslator)[clique[j]] << endl;
               j++;
           }
           if (j == clique.size()) {
//cout << "push " << vertex << endl;
               adjacentNodes.push_back(vertex);
           }
           else {
               if ((!adjacentNodes.size() + (*theAdjacentList)[clique[0]].size() - i + clique.size() + 1) < maxCliqueSize) {
                   break;
               }
           }
           i++;
       }// while
    }// while

    return maxClique;

}// end computeMaximumClique(..)

*/



