////////////////////////////////////// c++ /////////////////////////////////////
//
//  Module           : EdgeProjection.hh
//  Description      : Separation through edge projection.
//  Author           : Steffen Rebennack
//  Email            : srebenac@ix.urz.uni-heidelberg.de;
//                     steffen.rebennack@web.de
//  Copyright        : (C) 2006 by the University of L'Aquvila and
//                     University of Heidelberg
//  Created on       : Wed Apr 05 08:37:06 2006
//  Last modified by : -
//  Last modified on : -
//  Update count     : 0
//
////////////////////////////////////////////////////////////////////////////////
//
//  Date	Name		Changes/Extensions
//  ----	----		------------------
//
////////////////////////////////////////////////////////////////////////////////


#ifndef EDGE_PROJECTION_HH
#define EDGE_PROJECTION_HH

#include "ProjectionNode.hh"
#include <iostream>
#include <fstream>
#include <vector>

using namespace std;

class StableSetMaster;


/**
 * This class provides the edge projection for the stable set problem.
 * After an edge has been selected, its neighborhood is updated to get
 * a strongly projectable edge. The next step is the projection of the
 * graph and the separation of rank inequalities. This is done via
 * maximal cliques. The anti-projection steps make the inequalities 
 * valid for the whole graph. 
 */
class EdgeProjection  {

public:

    // -------------------------------------------------------------------------
    // ------------- M e t h o d s  ( p u b l i c ) ----------------------------
    // -------------------------------------------------------------------------

    /**
     * Constructor.
     *
     * @param *adjacentList Information of the graph.
     */
    EdgeProjection(vector< vector<int> > *adjacentList);
    
    /**
     * Destructor.
     */  
    ~EdgeProjection();
    
    /**
     * This function handles the hole edge projection.
     *
     * @param *lpValue Pointer to a solution of the current LP.
     * @return 0 Program treminated without error.
     * @return 1 There was an error during the separation. 
     */
    int separate(double *lpValue);

    /**
     * This function computes one large rank inequality.
     *
     * @param *lpValue Pointer to a solution of the current LP.
     * @param *theMaster Pointer to the master of the problem.
     * @return 0 Program treminated without error.
     * @return 1 There was an error during the separation. 
     */
    int computeLargeInequality(double *lpValue, StableSetMaster *theMaster); 

    /**
     * This function returns the left hand side of all computed inequalities.
     *
     * @param void
     * @return Pointer to the lhs of the inequalities.
     */  
    vector< vector<int>* >* getLHS();
    
    /**
     * This function returns the right hand side of all computed inequalities.
     *
     * @param void
     * @return Pointer to the rhs of the inequalities.
     */
    vector<int>* getRHS();
 

private:

    // -------------------------------------------------------------------------
    // ------------- M e t h o d s  ( p r i v a t e ) --------------------------
    // -------------------------------------------------------------------------
    
    /**
     * Select an projection edge.
     *
     * @param *lpValue Pointer to the current LP solution.
     * @return 1 An projectable edge has been found.
     * @return 0 No projectable edge has been found.
     */
    bool edgeSelection(double *lpValue);
    
    /**
     * Projection of the new graph.
     *
     * @param void
     * @return void
     */
    void projection();
   
    /**
     * Call separation routines for the reduced graph.
     *
     * @param *lpValue Pointer to the LP solution value. 
     * @param *theLhs Pointer to the inequalities separated by this function.
     * @param *theRhs Pointer to the ride hand side of the inequalities.
     * @return The number of computed inequalities.
     */
    int separation(double *lpValue, vector< vector<int> > *theLhs,
                    vector<int> *theRhs);
     
    /**
     * 'Lift' the inequalities.
     *
     * @param *theLhs Pointer to the lhs of the inequalities to be lifted.
     * @param *theRhs Pointer to the rhs of the inequalities to be lifted.
     * @return void
     */
    void lifting(vector< vector<int> > *theLhs, vector<int> *theRhs);
    
    /**
     * Check if the new inequalities are violated and store them.
     *
     * @param *theLhs Pointer to the lhs of the lifted inequalities.
     * @param *theRhs Pointer to the rhs of the lifted inequalities.
     * @param *lpValue Pointer to the LP solution.
     * @return void
     */   
    void storeInequalities(vector< vector<int> > *theLhs, vector<int> *theRhs,
                           double *lpValue); 
    
    /**
     * Delete inequalities which might not bring an improvement. This is done
     * to avoid a slowdown of the simplex method.
     *
     * @param *lpValue Pointer to the current LP value.
     * @return void
     */
    void cleanUpInequalities(double *lpValue);

    /**
     * Compute all edges which are strongly projectable.
     *
     * @param void
     * @retutn void
     */
    void checkStrongProjectability();

    /**
     * Compute a maximal clique in the graph given through the adjacent list.
     *
     * @param theAdjacentList Pointer to the adjacent list of the graph.
     * @param theTranslator Pointer to a translator for the nodes of the 
     *        adjacent list to the nodes of the original graph.
     * @return Pointer to a vector storing a maximal clique.
     */
    vector<int>* computeMaximalClique(vector< vector<int> > *theAdjacentList,
                                      vector<int> *theTranslator);
                                      
    /**
     * Compte the false edges.
     *
     * @param *falseEdges This vector contains the false edges after their 
     *                    computation.
     * @return void
     */    
    void computeFalseEdges(vector<int*> *falseEdges);
  
    /**
     * Compute the neighborhood of node 'node'. It is stored in 'neighborhood'.
     *
     * @param node We look for the neighborhood of this node.
     * @param *neighborhood Pointer to the neighborhood of 'node'. 
     * @return void
     */
    void getNeighborhood(int node, vector<int> *neighborhood);
    
    /**
     * Delete projection edge and all nodes to of vector deleteNodes.
     *
     * @param void
     * @return void
     */
    void updateGraph();

    /**
     * Delete the two endnodes of the projection edge and add all false edges. 
     *
     * @param *falseEdges Pointer to all false edges been added to the graph.
     * @return void
     */
    void updateGraph(vector<int*> *falseEdges);    
    
    /**
     * Add new node to the tree.
     *
     * @param *falseEdges Pointer to a vector storing all false edges.
     * @return void
     */
    void addNode(vector<int*> *falseEdges);
                            
    /**
     * Construct reduced graph.
     *
     * @param *reducedGraph Pointer to the adjacent list of the reduced graph.
     * @return void
     */
    void getReducedGraph(vector< vector<int> > *reducedGraph);
             
    /**
     * Separate the clique inequalities.
     *
     * @param *graph A pointer to an adjacent List storing the graph.
     * @param *lpValue A pointer to the lpValue cooresponding to the graph.
     * @param *theLhs Pointer to the left hand side of the computed inequalities.
     *                This vector will be filled during the separation and may
     *                contain already inequalities at the beginning. 
     * @return Number of computed violated inequalities.
     */                                               
    int cliqueSeparation(vector< vector<int> > *graph, double *lpValue, 
                         vector< vector<int> > *theLhs);

    /**
     * Solve a stable set problem for a small graph.
     *
     * @param *theLhs The left hand side of the computed rank inequality.
     * @param &theRhs The right hand side of the computed rank inequality.
     * @param *theMaster Pointer to the master of the problem.
     * @return void
     */
    void solveStableSet(vector<int> *theLhs, int &theRhs,
		       StableSetMaster *theMaster);

    /**
     * Lift and store the rank inequality. 
     *
     * @param *theLhs The left hand side of the rank inequality to be lifted.
     * @param &theRhs The right hand side of the rank inequality to be lifted.
     * @return void
     */
    void liftAndStore(vector<int> *theLhs, int theRhs);

    /**
     * Get random number.
     *
     * @param bound Upper bound of random number.
     * @return Random number.
     */
    int getRandomNumber(int bound) const;

    /**
     * Clean up tree after each call of separate.
     *
     * @param void.
     * @return void.
     */
    void cleanUp();
                                                                                                
    /**
     * Print the infrmation of the root of the tree.
     *
     * @param void.
     * @return void.
     */
    void printRoot();


    // -------------------------------------------------------------------------
    // ------------- V a r i a b l e s  ( p r i v a t e ) ----------------------
    // -------------------------------------------------------------------------
    
    /**
     * Translator of nodes of the reduced graph to the nodes of the original
     * graph.
     */
    vector<int> translator;
    
    /**
     * Store the left hand side of the violated inequalities.
     */    
    vector< vector<int>* > lhs;
    
    /**
     * Store the ride hand side of the violated inequalities.
     */ 
    vector<int> rhs;

    /**
     * Store the slack of each inequality.
     */
    vector<double> slack;

    /**
     * Store all computed inequalities in that file that they can be checked
     * after the termination of the program.
     */
    char *storeInequalitiesFile;    
    
    /**
     * The projection edge we are considering at the moment.
     */
    int edge[2];
    
    /**
     * Nodes to be deleted from the graph.
     */
    vector<int> deleteNodes;
    
    /**
     * Edges to be deleted from the graph.
     */
    vector<int*> deleteEdges;
        
    /**
     * Number of nodes in the original graph.
     */
    int graphSize;
    
    /**
     * Lower adjacent matrix.
     */
    vector< vector<bool> > lowerAdjacentMatrix;
    
    /**
     * Store strongly projectable edges.
     */
    vector< int* > projectable;
       
    /**
     * Size of the adjacent list of the reduction graph.
     */
    int *adjacentListSize;
         
    /**
     * Store all sons of this root.
     */
    vector< ProjectionNode* > itsNodes;
    
    /**
     * Store a pointer to the node of the tree we are current working on.
     */
    ProjectionNode* currentNode; 
 
};

#endif
