////////////////////////////////////// c++ /////////////////////////////////////
//
//  Module           : GeneralConstraint.cpp
//  Description      : Store a general integer constraint.
//  Author           : Steffen Rebennack
//  Email            : srebenac@ix.urz.uni-heidelberg.de;
//                     steffen.rebennack@web.de
//  Copyright        : (C) 2006 by the University of Heidelberg
//  Created on       : Wed Apr 05 11:04:46 2006
//  Last modified by : -
//  Last modified on : -
//  Update count     : 0
//
////////////////////////////////////////////////////////////////////////////////
//
//  Date	Name		Changes/Extensions
//  ----	----		------------------
//
////////////////////////////////////////////////////////////////////////////////

#include "abacus/row.h"
#include "GeneralConstraint.hh"
#include "Graph.hh"
#include "Node.hh"

#include <iostream>

using namespace std;


// -----------------------------------------------------------------------------
// C o n s t r u c t o r
//
// This class is derived from the abstract base class ABA_CONSTRAINT.
//
// This constraint is uniquely determined by the set of nodes, their
// corresponding multipliers and the right hand side. These data are stored in
// this class. The main part is the call of the constructor of class
// ABA_CONSTRAINT with the following arguments:
// 1.) A pointer to the corresponding master of the optimization.
// 2.) A pointer to the subproblem associated with this constraint because the
//     constraints should be only active in this sub problem and its children.
// 3.) This constraint is of kind less (or equal).
// 4.) The right hand side of this constraint is rhs.
// 5.) This constraint can be removed from the active constraint set.
// 6.) This constraint is global valid and not only local.
// 7.) There is no lifting procedure implemented.
// -----------------------------------------------------------------------------
GeneralConstraint::GeneralConstraint(ABA_MASTER *master, vector<int> *vertices,
				     vector<int> *dualCoeff, int rhs,
				     Graph *theGraph):
    ABA_CONSTRAINT(master, NULL, ABA_CSENSE::Less, rhs, true, false, false),
    nodes(vertices->size()),
    coefficients(vertices->size()),
    itsRhs(rhs),
    itsGraph(theGraph)
{
    // Store this inequality that it can be checked.
    ofstream fout(theGraph->getFileName(), ios::app);

    // Save the inequality.
    for (int i=0; i<vertices->size(); i++) {
        nodes[i] = (*vertices)[i];
	coefficients[i] = (*dualCoeff)[i];
	fout << "  " << coefficients[i] << " x_";
        fout << itsGraph->translateNode(nodes[i]);
    }
    fout << "  <  " << itsRhs << endl;
    fout.close();

    // Calculate the key.
    calculateHashKey();
}


// -----------------------------------------------------------------------------
// D e s t r u c t o r
// -----------------------------------------------------------------------------
GeneralConstraint::~GeneralConstraint() {
}


// -----------------------------------------------------------------------------
// g e n R o w
//
// This function redefines the pure virtual function of the base class
// ABA_CONSTRAINT. It returns the constraint in sprase format. 
// -----------------------------------------------------------------------------
int GeneralConstraint::genRow(ABA_ACTIVE<ABA_VARIABLE,ABA_CONSTRAINT> *var, 
			     ABA_ROW &row)
{

    for (int i=0; i<nodes.size(); i++) {
        row.insert(nodes[i],coefficients[i]);
    }
    row.rhs(itsRhs);
    row.sense(ABA_CSENSE::Less);

    return row.nnz();
}


// -----------------------------------------------------------------------------
// c o e f f
//
// This function computes the coefficient of the variable var. It redefines the
// pure virtual function of the base class ABA_CONSTRAINT. The coefficient of
// a variable of this constraint is 0 if it is not a member of it and otherwise
// it is the corresponding multiplier.
// -----------------------------------------------------------------------------
double GeneralConstraint::coeff(ABA_VARIABLE *var) {

    // Cast ABA_VARIABLE* to NODE*.
    Node *node = (Node *) var;

    // Search for this variable in vector nodes.
    // This vector is sorted.
    int i = 0;
    int vertex = node->nodeNumber();

    while (vertex > nodes[i]) {
	i++;
	if (i == nodes.size()) {
	    return 0.0;
	}
    }
    if (vertex == nodes[i]) {
	return coefficients[i];
    }
    return 0.0;
}


// -----------------------------------------------------------------------------
// g e t S i z e
//
// This method is used by function equal(..).
// -----------------------------------------------------------------------------
int GeneralConstraint::getSize() {
    return nodes.size();
}


// -----------------------------------------------------------------------------
// i s I n C o n s t r a i n t
//
// This function checks if a node is member of this constraint. It return the
// value true if the node is a member and otherwise false.
// This method is used by the function equal(..).
// -----------------------------------------------------------------------------
bool GeneralConstraint::isInConstraint(int node) {

    for(int i=0; i<nodes.size(); i++) {
        if (nodes[i] == node) {
            return true;
	}
    }

    return false;
}


// -----------------------------------------------------------------------------
// e q u a l
//
// This function compares if this constraint is identical with the constraint
// cv in a mathematical sense. This function is required becuase the constraint
// is stored in a pool of the class ABA_NONDUPLPOOL.
// It returns true if the constraints are equal and otherwise false.
// -----------------------------------------------------------------------------
bool GeneralConstraint::equal(ABA_CONVAR *cv) {

    bool isEqual = true;
    if (thisHashKey != ((GeneralConstraint *)cv)->hashKey()) {
        isEqual = false;
    }
    if (nodes.size() != ((GeneralConstraint *)cv)->getSize()) {
        isEqual = false;
    }

    for(int i=0; i<nodes.size(); i++) {
        if (!((GeneralConstraint *)cv)->isInConstraint(nodes[i])) {
	    isEqual = false;
	}
    }

    return isEqual;
}


// -----------------------------------------------------------------------------
// n a m e
//
// This function returns the name of the constraint. It is required if the
// constraint is stored in a pool of class ABA_NONDUPLPOOL.
// -----------------------------------------------------------------------------
const char* GeneralConstraint::name() {
    return "GeneralConstraint";
}


// -----------------------------------------------------------------------------
// h a s h K e y
//
// This method provides a key for the constraint that can be used to insert it
// into a hash table. In this case it is not required that any two items have
// different keys. This function is required to store the constraint in a pool
// of the class ABA_NIONDUPLPOOL.
// -----------------------------------------------------------------------------
unsigned GeneralConstraint::hashKey() {
    return thisHashKey;
}


// -----------------------------------------------------------------------------
// c a l c u l a t e H a s h K e y
// -----------------------------------------------------------------------------
void GeneralConstraint::calculateHashKey() {

    thisHashKey = 1;
    for (int i=0; i<nodes.size(); i++) {
        thisHashKey *= (nodes[i] + 1265);
	thisHashKey = thisHashKey % 9973;
    }
}


