////////////////////////////////////// c++ /////////////////////////////////////
//
//  Module           : GeneralConstraint.hh
//  Description      : Store a general integer constraint.
//  Author           : Steffen Rebennack
//  Email            : srebenac@ix.urz.uni-heidelberg.de;
//                     steffen.rebennack@web.de
//  Copyright        : (C) 2005 by the University of Heidelberg
//  Created on       : Wed Apr 05 11:09:11 2006
//  Last modified by : -
//  Last modified on : -
//  Update count     : 0
//
////////////////////////////////////////////////////////////////////////////////
//
//  Date	Name		Changes/Extensions
//  ----	----		------------------
//
////////////////////////////////////////////////////////////////////////////////


#ifndef GENERAL_CONSTRAINT_HH
#define GENERAL_CONSTRAINT_HH


#include "abacus/constraint.h"
#include <vector>

using namespace std;

class Graph;


/**
 * The class GeneralConstraint is derived from the abstract ABACUS constraint
 * class ABA_CONSTRAINT. It stores a constraint comeing from a local cut.
 */
class GeneralConstraint: public ABA_CONSTRAINT {


public:

    // -------------------------------------------------------------------------
    // ------------- M e t h o d s  ( p u b l i c ) ----------------------------
    // -------------------------------------------------------------------------

    /**
     * Constructor.
     *
     * @param *master Pointer to the corresponding master of the optimization.
     * @param *vertices Pointer to a vector storing the nodes of the inequality.
     * @param *dualCoeff Pointer to a vector storing the coefficients for
     *                   each node of the inequality.
     * @param rhs The ride hand side of the inequality.
     * @param *theGraph Pointer to the object storing the graph corresponding
     *                  to the problem. 
     */
    GeneralConstraint(ABA_MASTER *master, vector<int> *vertices,
    		      vector<int> *dualCoeff, int rhs, Graph *theGraph);

    /**
     * Destructor.
     */
    virtual ~GeneralConstraint();

    /**
     * Get constraint in sparse format.
     *
     * @param *var Pointer to a variable set.
     * @param &row Reference to abacus object storing the inequality.
     * @return Number of non-zeros in the inequality.
     */
    int genRow(ABA_ACTIVE<ABA_VARIABLE,ABA_CONSTRAINT> *var, ABA_ROW &row);

    /**
     * Get coefficient of a variable.
     *
     * @param *var Pointer to a variable.
     * @return The coefficient of the variable var in this clique constraint.
     */
    virtual double coeff(ABA_VARIABLE *var);

    /**
     * Get number of nodes part of this constraint.
     *
     * @return Size of vector nodes.
     */
    int getSize();

    /**
     * Checks if node is member of this constraint.
     *
     * @param node A node.
     * @return true Node is member of this constraint.
     * @return false Otherwise.
     */
    bool isInConstraint(int node);

    /**
     * Compare if constraint cv is equal to this constraint.
     *
     * @param *cv Pointer to the constraint to be compared with this one.
     * @return true If both constraints are equal.
     * @return false Otherwise.
     */
    virtual bool equal(ABA_CONVAR *cv);

    /**
     * This method returns the name of this constraint.
     *
     * @return Pointer to the name of this constraint
     */
    virtual const char* name();

    /**
     * This method returns the hashkey of the constraint.
     *
     * @return The hashkey of the constraint.
     */
    virtual unsigned hashKey();


private:

    // -------------------------------------------------------------------------
    // ------------- M e t h o d s  ( p r i v a t e ) --------------------------
    // -------------------------------------------------------------------------

    /**
     * This method calculates the haskey of the constraint.
     */
    void calculateHashKey();


    // -------------------------------------------------------------------------
    // ------------- V a r i a b l e s  ( p r i v a t e ) ----------------------
    // -------------------------------------------------------------------------

    /**
     * Store all vertices which have a multiplier != 0 in this constraint.
     */
    vector<int> nodes;

    /**
     * Store the coefficients of the nodes.
     */
    vector<int> coefficients;

    /**
     * Store the ride hand side of that inequality.
     */ 
    int itsRhs;

    /**
     * Pointer to the graph.
     */
    Graph *itsGraph;

    /**
     * The hashkey of this constraint.
     */
    unsigned thisHashKey;
};

#endif




