////////////////////////////////////// c++ /////////////////////////////////////
//
//  Module           : Graph.cpp
//  Description      : Manage the graph of a problem instance.
//  Author           : Steffen Rebennack
//  Email            : srebenac@ix.urz.uni-heidelberg.de;
//                     steffen.rebennack@web.de
//  Copyright        : (C) 2006 by the University of Heidelberg
//  Created on       : Wed Apr 05 10:57:14 2006
//  Last modified by : -
//  Last modified on : -
//  Update count     : 0
//
////////////////////////////////////////////////////////////////////////////////
//
//  Date	Name		Changes/Extensions
//  ----	----		------------------
//
////////////////////////////////////////////////////////////////////////////////

#include "Graph.hh"
#include <iostream>

using namespace std;



// -----------------------------------------------------------------------------
// --------------- M e t h o d s  ( p u b l i c ) ------------------------------
// -----------------------------------------------------------------------------


// -----------------------------------------------------------------------------
// C o n s t r u c t o r
//
// Store given information. Initialize the variables.
// -----------------------------------------------------------------------------
Graph::Graph(ABA_MASTER *master, int nodeNumber, int edgeNumber, bool weighted,
	     char *fileName):
    master_(master),
    numberOfNodes_(nodeNumber),
    numberOfEdges_(edgeNumber),
    weighted_(weighted),
    fixedVariablesFirstLP(0),
    fixedVariables(0)
{
    adjacentList.resize(nodeNumber);
    if (weighted) {
        // Allocate weightsOfNodes.
	weightsOfNodes = new double[nodeNumber];
    }
    calculateDensity();
    
    sprintf( theFileName, "%s", fileName );
}


// -----------------------------------------------------------------------------
// D e s t r u c t o r
// -----------------------------------------------------------------------------
Graph::~Graph() {

    if (weighted_) {
        delete [] weightsOfNodes;
    }
    for (int i=adjacentList.size()-1; i>=0; i--) {
        adjacentList[i].clear();
    }
}


// -----------------------------------------------------------------------------
// p u s h A d j a z e n t N o d e s
// -----------------------------------------------------------------------------
void Graph::pushAdjacentNodes(int rightNode, int leftNode) {
    adjacentList[rightNode].push_back(leftNode);
    adjacentList[leftNode].push_back(rightNode);
}


// -----------------------------------------------------------------------------
// p u s h W e i g h t
//
// Store the weight of node i.
// -----------------------------------------------------------------------------
void Graph::pushWeight(int i, double weight) {
    weightsOfNodes[i] = weight;
}


// -----------------------------------------------------------------------------
// s o r t A d j a z e n s L i s t
//
// Sort each adjacent list.
// -----------------------------------------------------------------------------
void Graph::sortAdjacentList() {

    for (int i=0; i<numberOfNodes_; i++) {
        sort(adjacentList[i].begin(), adjacentList[i].end());
    }
}


// -----------------------------------------------------------------------------
// n u m b e r O f N o de s
// -----------------------------------------------------------------------------
int Graph::numberOfNodes() const {
    return numberOfNodes_;
}


// -----------------------------------------------------------------------------
// n u m b e r O f E d g e s
// -----------------------------------------------------------------------------
int Graph::numberOfEdges() const {
    return numberOfEdges_;
}


// -----------------------------------------------------------------------------
// g e t A d j a c e n t L i s t
// -----------------------------------------------------------------------------
vector<int> *Graph::getAdjacentList(int i) {
    return &adjacentList[i];
}


// -----------------------------------------------------------------------------
// a d j a c e n t L i s t S i z e
//
// This function returns the number of adjacent nodes to node i.
// -----------------------------------------------------------------------------
int Graph::adjacentListSize(int i) const {
    return adjacentList[i].size();
}


// -----------------------------------------------------------------------------
// a d j a c e n t L i s t E l e m e n t
//
// This function returns the j-th element of adjacent list of node i.
// -----------------------------------------------------------------------------
int Graph::adjacentListElement(int i, int j) const {
    return adjacentList[i][j];
}


// -----------------------------------------------------------------------------
// i s M e m b e r O f A d j a c e n t L i s t
//
// This function checks if a node is member of the adjacent list of the
// node with index index. The adjacent list is sorted and therfore we can use
// information from a previous search and so we start by index begin.
// -----------------------------------------------------------------------------
bool Graph::isMemberOfAdjacentList(int index, int &begin, int node) const {

    bool found = false;
    int  size = adjacentList[index].size();

    while((begin < size) && (node > adjacentList[index][begin])) {
	begin++;
    }
    if ((begin < size) && (node == adjacentList[index][begin])) {
	found = true;
	begin++;
    }

    return found;
}


// -----------------------------------------------------------------------------
// w e i g h t e d
// -----------------------------------------------------------------------------
bool Graph::weighted() const {
    return weighted_;
}


// -----------------------------------------------------------------------------
// w e i g h t
// -----------------------------------------------------------------------------
double Graph::weight(int i) const {
    return weightsOfNodes[i];
}


// -----------------------------------------------------------------------------
// i n i t i a l i z e T r a n s l a t o r
// -----------------------------------------------------------------------------
void Graph::initializeTranslator() {

    translateNodes.resize(numberOfNodes_);

    for (int i=0; i<numberOfNodes_; i++) {
	translateNodes[i] = i;
    }

}


// -----------------------------------------------------------------------------
// t r a n s l a t e N o d e
//
// Get translation of node 'index'
// -----------------------------------------------------------------------------
int Graph::translateNode(int index) const {
    return translateNodes[index];

}


// -----------------------------------------------------------------------------
// d e l e t e N o d e s F r o m G r a p h
//
// Delete all nodes of vector deleteNodes from the adjacent list. Because
// each edge is stored twice we have to delete each node two times.
// The clique has to be sorted.
// Update the translator.
// -----------------------------------------------------------------------------
void Graph::deleteNodesFromGraph(const vector<int> *deleteNodes) {

    const int NUM_NODES = deleteNodes->size();
    int  i, j, k;    		// Loop counter.
    int  adjacentListSize;	// Size of an adjacent list.
    int  vertex;		// A node.
    bool found;			// Flag which shows if a node is member of
    				// vector nodesOfClique.
    int size;

    //
    // Update adjacent list.
    //
    // Go through the adjacent list of all nodes of the vector and delete
    // node deleteNodes[i] from its adjacent list
    for (i=0; i<NUM_NODES; i++) {
	adjacentListSize = adjacentList[(*deleteNodes)[i]].size();
        for (j=0; j<adjacentListSize; j++) {
            vertex = adjacentList[(*deleteNodes)[i]][j];

	    k = 0;
	    // Find index of vertex deleteNodes[i] in the adjacent list.
	    if (!isMemberOfAdjacentList(vertex, k, (*deleteNodes)[i])) {
		master_->err() << "ERROR in Graph::deleteNodesFrom";
		master_->err() << "Graph:\n";
		master_->err() << "Vertex " << (*deleteNodes)[i];
		master_->err() << " could not be found in the adjacent list of";
		master_->err() << " node " << vertex << " ." << endl;
		exit(master_->Fatal);
            }
	    k--;
	    // Remove this element
	    adjacentList[vertex].erase(adjacentList[vertex].begin() +k);
	    numberOfEdges_--;

        }
        adjacentList[(*deleteNodes)[i]].clear();
    }

    //
    // Remove the nodes from the graph.
    //
    for (i=NUM_NODES-1; i>=0; i--) {

	adjacentList.erase(adjacentList.begin() + (*deleteNodes)[i]);
	numberOfNodes_--;

	for (j=(*deleteNodes)[i]; j<translateNodes.size()-1; j++) {
	    translateNodes[j] = translateNodes[j+1];
	}
	translateNodes.pop_back();

	// Update all nodes of the graph and of the list of nodes to be checked.
	for (j=0; j<numberOfNodes_; j++) {
	    size = adjacentList[j].size();
	    for (k=0; k<size; k++) {
	        if (adjacentList[j][k] > (*deleteNodes)[i]) {
		    adjacentList[j][k] -= 1;
		}
	    }

	}

    }

    calculateDensity();
//    output();

}// void deleteNodesFromGraph(..)


// -----------------------------------------------------------------------------
// d e l e t e N o d e s F r o m G r a p h
//
// Delete all nodes of vector deleteNodes from the adjacent list. Because
// each edge is stored twice we have to delete each node from two times.
// This function returns all nodes which have an updates neighborhood.
// -----------------------------------------------------------------------------
void Graph::deleteNodesFromGraph(const vector<int> *deleteNodes,
				 vector<int> *visitedNodes) {

    const int NUM_NODES = deleteNodes->size();

    int  i, j, k;    		// Loop counter.
    int  adjacentListSize;	// Size of an adjacent list.
    int  vertex;		// A node.
    int size;

    //
    // Update adjacent list.
    //
    // Go through the adjacent list of all nodes of the vector and delete
    // node deleteNodes[i] from its adjacent list
    for (i=0; i<NUM_NODES; i++) {
	adjacentListSize = adjacentList[(*deleteNodes)[i]].size();
        for (j=0; j<adjacentListSize; j++) {
            vertex = adjacentList[(*deleteNodes)[i]][j];
	    // Do something only if this is NOT a member of the vector deleteNodes.

	    k = 0;
	    while ((k < (NUM_NODES-1)) && ((*deleteNodes)[k] != vertex)) {
		k++;
	    }

	    if((*deleteNodes)[k] == vertex) {
	        continue;
	    }

	    k = 0;
	    // Find index of vertex deleteNodes[i] in the adjacent list.
	    if (!isMemberOfAdjacentList(vertex, k, (*deleteNodes)[i])) {
		master_->err() << "ERROR in Graph::deleteNodesFrom";
		master_->err() << "Graph:\n";
		master_->err() << "Vertex " << (*deleteNodes)[i];
		master_->err() << " could not be found in the adjacent list of";
		master_->err() << " node " << vertex << " ." << endl;
		exit(master_->Fatal);
            }
	    k--;
	    // Remove this element
	    adjacentList[vertex].erase(adjacentList[vertex].begin() +k);
	    k = 0;
	    while ((k < visitedNodes->size())
	           && ((*visitedNodes)[k] != vertex)
		  ) {
                k++;
	    }
	    if (k == visitedNodes->size()) {
	        visitedNodes->push_back(vertex);
	    }
	    numberOfEdges_--;

        }
        adjacentList[(*deleteNodes)[i]].clear();
    }
    numberOfEdges_ -= (NUM_NODES * (NUM_NODES - 1)) / 2;


    //
    // Remove the nodes from the graph.
    //
    for (i=NUM_NODES-1; i>=0; i--) {

	adjacentList.erase(adjacentList.begin() + (*deleteNodes)[i]);
	numberOfNodes_--;

	for (j=(*deleteNodes)[i]; j<translateNodes.size()-1; j++) {
	    translateNodes[j] = translateNodes[j+1];
	}
	translateNodes.pop_back();

	// Update all nodes of the graph and of the list of nodes to be checked.
	for (j=0; j<numberOfNodes_; j++) {
	    size = adjacentList[j].size();
	    for (k=0; k<size; k++) {
	        if (adjacentList[j][k] > (*deleteNodes)[i]) {
		    adjacentList[j][k] -= 1;
		}
	    }

	}
	size = visitedNodes->size();
	for (j=0; j<size; j++) {
	    if ((*visitedNodes)[j] > (*deleteNodes)[i]) {
		(*visitedNodes)[j] -= 1;
	    }
	}


    }

    fixedVariables += NUM_NODES;

    calculateDensity();
//    output();

}// void deleteNodesFromGraph(..)


// -----------------------------------------------------------------------------
// f i x V a r i a b l e s F i r s t L P
//
// Update counter.
// -----------------------------------------------------------------------------
void Graph::fixVariablesFirstLP(int numberOfFixedNodes) {
    fixedVariables += numberOfFixedNodes;
    fixedVariablesFirstLP = numberOfFixedNodes;
}


// -----------------------------------------------------------------------------
// a l l V a r i a b l e s F i x e d
//
// This function checks if all variables have been fixed.
// -----------------------------------------------------------------------------
bool Graph::allVariablesFixed() const {
    return (fixedVariables == numberOfNodes_);
}


// -----------------------------------------------------------------------------
// g e t F i l e N a m e
// -----------------------------------------------------------------------------
char *Graph::getFileName() {
    return theFileName;
}


// -----------------------------------------------------------------------------
// g e t D e n s i t y
// -----------------------------------------------------------------------------
double Graph::getDensity() const {
    return density;
}



// -----------------------------------------------------------------------------
// p r i n t S t a t i s t i c
//
// Print some information about the fixing of variables.
// -----------------------------------------------------------------------------
void Graph::printStatistic() const{

    master_->out() << "  Number of fixed variables according to LP\t:  ";
    master_->out() << fixedVariablesFirstLP;
    master_->out() << "\n  Number of fixed variables according to clique\t:  ";
    master_->out() << fixedVariables - fixedVariablesFirstLP;

}// end: void printStatistic()


// -----------------------------------------------------------------------------
// o u t p u t
//
// Print some information about the graph itself.
// -----------------------------------------------------------------------------
void Graph::output() const {

    master_->out()  << "\nGraph:\n";
    master_->out()  << "------------------------------------------------------";
    master_->out()  << "\nnNodes:\t" << numberOfNodes_;
    master_->out()  << "\nnEdges:\t" << numberOfEdges_;
    master_->out()  << "\nDensity:\t" << density;
    master_->out()  << "\n\nEdges:\n";
    for (int i=0; i<numberOfNodes_; i++) {
	master_->out()  << "  node[" << i << "]:\t";
	for (int j=0; j<adjacentList[i].size(); j++) {
	    master_->out()  << adjacentList[i][j] << " ";
	}
	master_->out()  << endl;
    }
    master_->out()  << "------------------------------------------------------";
    master_->out()  << endl << endl;
}



// -----------------------------------------------------------------------------
// --------------- M e t h o d s  ( p r i v a t e ) ----------------------------
// -----------------------------------------------------------------------------


// -----------------------------------------------------------------------------
// c a l c u l a t e D e n s i t y
// -----------------------------------------------------------------------------
void Graph::calculateDensity() {
    if (numberOfNodes_ == 0 || numberOfNodes_ == 1) {
	density = -1;
    }
    else {
        density = (1.0*numberOfEdges_) / numberOfNodes_;
        density = density / (numberOfNodes_ -1);
        density *= 2;
    }
}


