////////////////////////////////////// c++ /////////////////////////////////////
//
//  Module           : Graph.hh
//  Description      : Manage the graph of a problem instance.
//  Author           : Steffen Rebennack
//  Email            : srebenac@ix.urz.uni-heidelberg.de;
//                     steffen.rebennack@web.de
//  Copyright        : (C) 2006 by the University of Heidelberg
//  Created on       : Wed Apr 05 10:56:36 2006
//  Last modified by : -
//  Last modified on : -
//  Update count     : 0
//
////////////////////////////////////////////////////////////////////////////////
//
//  Date	Name		Changes/Extensions
//  ----	----		------------------
//
////////////////////////////////////////////////////////////////////////////////


#ifndef GRAPH_HH
#define GRAPH_HH

#include "abacus/master.h"
#include <vector>

using namespace std;



/**
 * This class stores all necessary information about the problem instance:
 * - number of nodes
 * - number of edges
 * - all adjazent nodes
 * - weights of node if they exist
 * The update of the graph due to fixing of variables is also managed by this
 * class.
 */
class Graph {

public:

    // -------------------------------------------------------------------------
    // ------------- M e t h o d s  ( p u b l i c ) ----------------------------
    // -------------------------------------------------------------------------

    /**
     * Constructor.
     *
     * @param *master Pointer to the master of the problem.
     * @param nodeNumber Number of nodes of the problem instance.
     * @param edgeNumber Number of edges of the problem instance.
     * @param weighted If the problem has weights than this variable is true,
     *		       otherwise it is false.
     * @param *fileName Pointer storing the file name where all computed
     *                  inequalities are stored.
     */
    Graph(ABA_MASTER *master, int nodeNumber, int edgeNumber, bool weighted,
	  char *fileName);

    /**
     * Destructor.
     */
    ~Graph();

    /**
     * Store adjacent nodes.
     *
     * @param rightNode One end node of the edge.
     * @param leftNode The other end node of the edge.
     * @return void
     */
    void pushAdjacentNodes(int rightNode, int leftNode);

    /**
     * Set weight of a node.
     *
     * @param i Index of node.
     * @param weight Weight of node i.
     * @return void
     */
    void pushWeight(int i, double weight);

    /**
     * Sort each adjazens list.
     *
     * @param void
     * @return void
     */
    void sortAdjacentList();

    /**
     * Get number of nodes.
     *
     * @param void
     * @return Number of nodes.
     */
    int numberOfNodes() const;

    /**
     * Get number of edges.
     *
     * @param void
     * @return Number of edges.
     */
    int numberOfEdges() const;

    /**
     * Get adjacent list.
     *
     * @param i Index of a node.
     * @return Adjazens list of node i.
     */
    vector<int> *getAdjacentList(int i);

    /**
     * Get size of adjacent list.
     *
     * @param i Index of a node.
     * @return Size of adjazens list of node i.
     */
    int adjacentListSize(int i) const;

    /**
     * Get Element of adjacent list.
     *
     * @param i Index of a node.
     * @param j Index for the adjacent list of node i.
     * @return j-th element of adjacent list of node i.
     */
    int adjacentListElement(int i, int j) const;

    /**
     * Check if two vertices are adjacent.
     *
     * @param index Index of a node.
     * @param &begin Reference to an index we start the search in the adjacent
     *		     list.
     * @param node A vertex.
     * @return true If node is adjacent to vertex index.
     * @return false Otherwise.
     */
    bool isMemberOfAdjacentList(int index, int &begin, int node) const;

    /**
     * Get type of problem.
     *
     * @param void
     * @return true If the problem is weighted.
     * @return false Otherwise.
     */
    bool weighted() const;

    /**
     * Get weight of node.
     *
     * @param i Index od node.
     * @return Weight of node i.
     */
    double weight(int i) const;

    /**
     * Copy the current nodes in translator.
     *
     * @param void
     * @return void
     */
    void initializeTranslator();

    /**
     * Translate a node.
     *
     * @param index Index of a node.
     * @return Translation of node.
     */
    int translateNode(int index) const;

    /**
     * Delete all nodes of the given vector from the graph.
     *
     * @param *nodesOfClique Nodes of a clique.
     * @return void
     */
    void deleteNodesFromGraph(const vector<int> *nodesOfClique);

    /**
     * Delete all nodes from the graph and return all nodes with
     * changed neighborhood.
     *
     * @param *nodesOfClique Nodes of a clique.
     * @return void
     */
    void deleteNodesFromGraph(const vector<int> *deleteNodes,
			      vector<int> *visitedNodes);

    /**
     * Update counter for fixed nodes.
     *
     * @param numberOfFixedNodes Number of nodes which have been fixed.
     * @return void
     */
    void fixVariablesFirstLP(int numberOfFixedNodes);



    /**
     * Have all variables been fixed?
     *
     * @param void
     * @return true If all variables have been fixed.
     * @return false Otherwise
     */
    bool allVariablesFixed() const;

    /**
     * Get name of the file where all computed inequalities are stored.
     *
     * @param void
     * @return Pointer to a char storing the file name.
     */
    char *getFileName();

    /**
     * Get density of the graph.
     *
     * @param void
     * @return Density of the graph.
     */
    double getDensity() const;

    /**
     * Print statistic about fixed nodes.
     *
     * @param void
     * @return void
     */
    void printStatistic() const;

    /**
     * Print some information about the problem instance.
     *
     * @param void
     * @return void
     */
    void output() const;


private:

    // -------------------------------------------------------------------------
    // ------------- M e t h o d s  ( p r i v a t e ) --------------------------
    // -------------------------------------------------------------------------

    /**
     * Calculate the density of the graph.
     */
    void calculateDensity();


    // -------------------------------------------------------------------------
    // ------------- V a r i a b l e s  ( p r i v a t e ) ----------------------
    // -------------------------------------------------------------------------

    /**
     * Pointer to the master of the problem.
     */
    ABA_MASTER *master_;

    /**
     * Number of nodes of the graph.
     */
    int numberOfNodes_;

    /**
     * Number of edges of the graph.
     */
    int numberOfEdges_;

    /**
     * A two dimensional vector to store for each node all its adjazent nodes.
     */
    vector< vector<int> > adjacentList;

    /**
     * If this variable is true the problem has weights; otherwise there are no
     * weights for the nodes.
     */
    bool weighted_;

    /**
     * Store the weights of each node.
     */
    double *weightsOfNodes;

    /**
     * Store the density of the graph.
     */
    double density;

    /**
     * Number of fixed variables of the first LP.
     */
    int fixedVariablesFirstLP;

    /**
     * Store the number of all fixed variables.
     */
    int fixedVariables;

    /**
     * Store the index of the new nodes that they can be translated to the
     * nodes before parts of the graph have been fixed.
     */
    vector<int> translateNodes;
    
    /**
     * Store the file name where all inequalsities are stored.
     */
    char theFileName[256];

};// end: class Graph

#endif


