////////////////////////////////////// c++ /////////////////////////////////////
//
//  Module           : ImprovementHeuristic.cpp
//  Description      :
//  Author           : Steffen Rebennack
//  Email            : srebenac@ix.urz.uni-heidelberg.de;
//                     steffen.rebennack@web.de
//  Copyright        : (C) 2006 by the University of Heidelberg
//  Created on       : Wed Apr 05 13:55:54 2006
//  Last modified by : -
//  Last modified on : -
//  Update count     : 0
//
////////////////////////////////////////////////////////////////////////////////
//
//  Date	Name		Changes/Extensions
//  ----	----		------------------
//
////////////////////////////////////////////////////////////////////////////////


#include "ImprovementHeuristic.hh"
#include "Graph.hh"
#include "StableSetMaster.hh"
#include "StableSetStatistics.hh"



// -----------------------------------------------------------------------------
// --------------- M e t h o d s  ( p u b l i c ) ------------------------------
// -----------------------------------------------------------------------------


// -----------------------------------------------------------------------------
// C o n s t r u c t o r
// -----------------------------------------------------------------------------
ImprovementHeuristic::ImprovementHeuristic(ABA_MASTER *master, Graph *theGraph,
					   StableSetStatistics *theStatistics):
    master_(master),
    itsGraph(theGraph),
    itsStatistics(theStatistics)
{
}


// -----------------------------------------------------------------------------
// D e s t r u c t o r
// -----------------------------------------------------------------------------
ImprovementHeuristic::~ImprovementHeuristic() {
}


// -----------------------------------------------------------------------------
// i m p r o v e
//
// Start with a stable set. Check for three nodes of the stable set if there
// are two path of length 3 which connect two of them. If there is such a path
// with the property that no node of the path is adjacent to one node of the
// current stable set then you can improve the stable set.
// -----------------------------------------------------------------------------
int ImprovementHeuristic::improve(double &value) {

    clock_t t1 = clock();       // Get current time.

    int i, j, k;	// Loop counter.
    int cnt;		// Counter.
    int cnt2;
    int vertex;		// A node.
    int indexOfVertex;	// Index of node vertex.
    int indexOfNode;
    int node;
    int index;
    int nodeOfPathOne;
    int nodeOfPathTwo;
    int v11, v12, v13;  // Three nodes of a path.
    int v21, v22, v23;  // Three nodes of a path.
    int beginOne;
    int beginTwo;
    int stack;
    int member;
    int status = 0;	// Return value;
    double timeLimit = stableSetMaster()->m_timeLimitForImprovementHeuristic;
    vector<int> *solution = stableSetMaster()->getSolution();
    if (solution->size() <= 2) {
	return status;
    }

    //
    // Improve until time limit is reached.
    //
    while (timeLimit > (double)(clock() - t1)/CLOCKS_PER_SEC) {

	// Choose randomly two nodes to start from.
    	indexOfVertex = stableSetMaster()->getRandomNumber(solution->size());
        vertex = (*solution)[indexOfVertex];
	index  = stableSetMaster()->getRandomNumber(solution->size());
	if (index == indexOfVertex) {
	    index++;
	    index = index % solution->size();
	}
	node = index +1;
	node = node % solution->size();
	if (node == indexOfVertex) {
	    node++;
	    node = node % solution->size();
	}
	cnt  = 0;
	v11 = -1;
	v12 = -1;
	while((cnt < 2) && (timeLimit > (double)(clock() - t1)/CLOCKS_PER_SEC)) {
	    if (node == index) {
	        break;
	    }
            // Find path from 'vertex' to '(*solution)[node]' with length 3.
	    if (calculateDistance(vertex, (*solution)[node], v11, v12, v21,
	    			  v22)) {
	        cnt++;
		if (cnt == 1) {
		    v13 = (*solution)[node];
		}
		else {
		    v23 = (*solution)[node];
		}
	    }
	    node++;
	    node = node % solution->size();
	    if (node == indexOfVertex) {
		node++;
	        node = node % solution->size();
	    }
	}
	
	//
	// Two paths have been computed. Try to improve the stable set.
	//
	if (cnt == 2) {

	    cnt2 = 0;
	    // Solution is allways sorted!
	    beginOne = 0;
	    beginTwo = 0;
	    for (j=0; j<solution->size(); j++) {
		stack = (*solution)[j];
		if (stack != vertex) {
		    if(itsGraph->isMemberOfAdjacentList(v11, beginOne, stack)) {
		        cnt2++;
			member = stack;
		    }
		    else {
			if(itsGraph->isMemberOfAdjacentList(v21, beginTwo,
			 				    stack)) {
			    cnt2++;
			    member = stack;
			}
		    }
		}
		if (cnt2 > 1) {
		    break;
		}
	    }
 
	    //
	    // If cnt = 0 the new stable set will have a greater size.
	    //
	    if (cnt2 < 2) {
		// Update solution.
		solution->erase(solution->begin() + indexOfVertex);
		if (cnt2 == 1) {
		    k = 0;
		    while ((*solution)[k] != member) {
			k++;
		    }
		    solution->erase(solution->begin() + k);
		}

		solution->push_back(v11);
		solution->push_back(v21);
		sort(solution->begin(), solution->end());
		testStableSet(solution);
		if (cnt2 == 0) {
                    itsStatistics->improvementHeuristics();
		}
	    }
	}

    }

    value = 1.0*solution->size();

    // Update bounds.
    if (master_->betterPrimal(value)) {
        // Heuristic found better solution.
        master_->out() << "\nImprovement of heuristic was successful:\n";
        master_->out() << "\tNew value:\t" << value << endl;
        
        // Update lower bound (best known feasible solution)
        // (this is done automatically by ABACUS, too).
        master_->primalBound(value);
   
        // Update statistics.
        itsStatistics->solutionFound(2);

	status = 1;
    }

    // Return with status.
    return status;

}// int improve(..)



// -----------------------------------------------------------------------------
// -------------- M e t h o d s  ( p r i v a t e ) -----------------------------
// -----------------------------------------------------------------------------


// -----------------------------------------------------------------------------
// c a l c u l a t e D i s t a n c e
//
// Find path from node 'start' to node 'end' with length 3. Store the nodes
// of the path in n11, n12, n21 or n22.
// -----------------------------------------------------------------------------
bool ImprovementHeuristic::calculateDistance(int start, int end, int &n11,
					     int &n12, int &n21, int &n22)
					     const {

    int i, j, k, l;  // Loop counter.
    int startSize = itsGraph->adjacentListSize(start);
    int endSize   = itsGraph->adjacentListSize(end);
    int indexOne  = ((StableSetMaster *)master_)->getRandomNumber(startSize);
    int indexTwo;
    int vertex;
    int node;
    int index;
    vector<int> *solution = ((StableSetMaster*)master_)->getSolution();

    //
    // Check all adjacent nodes to the start node.
    //
    for (i=0; i<startSize; i++) {
        vertex   = itsGraph->adjacentListElement(start, indexOne);
	indexTwo = ((StableSetMaster *)master_)->getRandomNumber(endSize);
	index = 0;
        //
        // Check all adjacent nodes to the end node.
        //
	for (j=0; j<endSize; j++) {
	    node = itsGraph->adjacentListElement(end, indexTwo);
	    l = 0;
	    if(itsGraph->isMemberOfAdjacentList(vertex, index, node)
	       && (node != n11) && (node != n12)
	       && (vertex != n11) && (vertex != n12)
	       && ((n11 == -1) || !itsGraph->isMemberOfAdjacentList(n11, l,
	       						 vertex))
	      ) {
	    	// Look for vertex in stable set solution.
		k = 0;
		while ((k < (solution->size()-1))
		       && ((*solution)[k] < vertex)
		      ) {
		    k++;
		}
		if ((*solution)[k] == vertex) {
		    continue;
		}
		k = 0;
		while ((k < (solution->size()-1))
		       && ((*solution)[k] < node)
		      ) {
		    k++;
		}
		if ((*solution)[k] == node) {
		    continue;
		}
		if (n11 == -1) {
		    n11 = vertex;
		    n12 = node;
		}
		else {
		    n21 = vertex;
		    n22 = node;
		}

		return true;
	    }
	    indexTwo++;
	    indexTwo = indexTwo % endSize;
	    if (indexTwo == 0) {
		index = 0;
	    }
	}
	indexOne++;
	indexOne = indexOne % startSize;
    }

    return false;
}// bool calculateDistance(..)


// -----------------------------------------------------------------------------
// t e s t S t a b l e S e t
// -----------------------------------------------------------------------------
void ImprovementHeuristic::testStableSet(vector<int> *stableSet) {

    int i, j, l;  // Loop counter

    for (i=1; i<stableSet->size(); i++) {
	for (j=(i-1); j>=0; j--) {
	    l = 0;
	    if (itsGraph->isMemberOfAdjacentList((*stableSet)[i], l,
	    					        (*stableSet)[j])) {
		master_->err() << "ImprovementHeuristic::testStableSet:\t";
		master_->err() << "The constructed vector is no stable set!\n";
		master_->err() << "Solution:\t";
		for (l=0; l<stableSet->size(); l++) {
		    master_->err() << (*stableSet)[l] << " ";
		}
		master_->err() << endl;
		exit(master_->Fatal);
	    }// if
	}
    }// for
}


// -----------------------------------------------------------------------------
// s t a b l e S e t M a s t e r
//
// This function returns a pointer to the corresponding object of the class
// StableSetMaster.
// -----------------------------------------------------------------------------
StableSetMaster *ImprovementHeuristic::stableSetMaster() {
   return ((StableSetMaster *)master_);
}








