////////////////////////////////////// c++ /////////////////////////////////////
//
//  Module           : ImprovementHeuristic.hh
//  Description      :
//  Author           : Steffen Rebennack
//  Email            : srebenac@ix.urz.uni-heidelberg.de;
//                     steffen.rebennack@web.de
//  Copyright        : (C) 2006 by the University of Heidelberg
//  Created on       : Wed Apr 05 13:56:24 2006
//  Last modified by : -
//  Last modified on : -
//  Update count     : 0
//
////////////////////////////////////////////////////////////////////////////////
//
//  Date	Name		Changes/Extensions
//  ----	----		------------------
//
////////////////////////////////////////////////////////////////////////////////

#ifndef IMPROVEMENT_HEURISTIC_HH
#define IMPROVEMENT_HEURISTIC_HH
#include <vector>

using namespace std;

class ABA_MASTER;
class Graph;
class StableSetMaster;
class StableSetStatistics;



/**
 * This class provides a heuristic to improve a stable set from a current
 * stable set.
 */
class ImprovementHeuristic {

public:

    // -------------------------------------------------------------------------
    // ------------- M e t h o d s  ( p u b l i c ) ----------------------------
    // -------------------------------------------------------------------------

    /**
     * Constructor.
     *
     * @param *master Pointer to the master of the problem.
     * @param *theGraph Pointer to the graph storing all information about the
     *			problem instance.
     * @param *theStatistics Pointer to an object managing the statistics.
     */
    ImprovementHeuristic(ABA_MASTER *master, Graph *theGraph,
			 StableSetStatistics *theStatistics);

    /**
     * Destructor.
     */
    ~ImprovementHeuristic();

    /**
     * Try to improve a stable set.
     *
     * @param &value Reference to the size of the best stable set found so far.
     * @return 1 The stable set could be improved.
     * @return 0 There was no improvement.
     */
    int improve(double &value);


private:

    // -------------------------------------------------------------------------
    // ------------- M e t h o d s  ( p r i v a t e ) --------------------------
    // -------------------------------------------------------------------------
    
    /**
     * Find path from node 'start' to node 'end' with length 3.
     *
     * @param start Node to start from.
     * @param end Node in which the path ends.
     * @param &v11 Reference to a node which is part of the path.
     * @param &v12 Reference to a node which is part of the path.
     * @param &v21 Reference to a node which is part of the path.
     * @param &v22 Reference to a node which is part of the path.
     * @return True A path could be generated.
     * @return False There is no such path.
     */

    bool calculateDistance(int start, int end, int &v11, int &v12, int &v21,
    			   int &v22) const;

    /**
     * Test if a set of nodes is a stable set.
     *
     * @param *stableSet Pointer to a vector storing the stable set to be
     *			 checked.
     * @return void
     */
    void testStableSet(vector<int> *stableSet);

    /**
     * The function stableSetMaster().
     *
     * @param void
     * @return Pointer to the corresponding object of the class
     *         StableSetMaster.
     */
    StableSetMaster *stableSetMaster();


    // -------------------------------------------------------------------------
    // ------------- V a r i a b l e s  ( p r i v a t e ) ----------------------
    // -------------------------------------------------------------------------

    /**
     * Pointer to the master of the problem.
     */
    ABA_MASTER *master_;

    /**
     * Pointer to the graph of the problem.
     */
    Graph *itsGraph;

    /**
     * Pointer to the object managing the statistics.
     */
    StableSetStatistics *itsStatistics;

};

#endif


