////////////////////////////////////// c++ /////////////////////////////////////
//
//  Module           : LocalCutSeparator.cpp
//  Description      : Separate a local cut.
//  Author           : Steffen Rebennack
//  Email            : srebenac@ix.urz.uni-heidelberg.de;
//                     steffen.rebennack@web.de
//  Copyright        : (C) 2006 by the University of Heidelberg
//  Created on       : Wed Apr 05 15:13:59 2006
//  Last modified by : -
//  Last modified on : -
//  Update count     : 0
//
////////////////////////////////////////////////////////////////////////////////
//
//  Date	Name		Changes/Extensions
//  ----	----		------------------
//
////////////////////////////////////////////////////////////////////////////////

#include "LocalCutSeparator.hh"
#include "abacus/array.h"
#include "abacus/row.h"
#include "abacus/cplexif.h"
#include "GeneralConstraint.hh"
#include "Graph.hh"
#include "MaxCliquesSeparator.hh"
#include "StableSetLPSolution.hh"
#include "StableSetMaster.hh"

#include <vector>

using namespace std;


//
// The struct NodesAndWeights2 stores an index of a node and a weight.
//
struct NodesAndWeights2 {
    int    inode;         // index of node
    double wweight;       // weighted weight of node
};// struct NodesAndWeights2


//
// This operator is used by the sort function in lowerBoundHeuristic to get
// a sorted vector of variables with increasing weight.
//
bool operator<(const NodesAndWeights2 &a, const NodesAndWeights2 &b) {
        return a.wweight > b.wweight;
}



// -----------------------------------------------------------------------------
// --------------- M e t h o d s  ( p u b l i c ) ------------------------------
// -----------------------------------------------------------------------------


// -----------------------------------------------------------------------------
// C o n s t r u c t o r
//
// The main part of this constructor is the initialization of the abstract
// template class ABA_SEPARATOR. The second argument is true because we do
// not want double constraints in the buffer and the nuber 500 states the
// maximla number of cuttings planes which are stored.
// -----------------------------------------------------------------------------
LocalCutSeparator::LocalCutSeparator(ABA_MASTER *master,
    			StableSetLPSolution *solution, Graph *theGraph):
    ABA_SEPARATOR<ABA_CONSTRAINT, ABA_VARIABLE> (solution, true, 500),
    master_(master),
    lpSolution(solution),
    itsGraph(theGraph)
{
}


// -----------------------------------------------------------------------------
// D e s t r u c t o r
// -----------------------------------------------------------------------------
LocalCutSeparator::~LocalCutSeparator()
{
    for (int i=0; i<adjacentList.size(); i++) {
	adjacentList[i].clear();
    }
}


// -----------------------------------------------------------------------------
// s e p a r a t e
//
// This separation routine computes a subgraph and with help of a LP it creates
// a cuttinf plane. This is called local cut. Only one cut will be generated!
// -----------------------------------------------------------------------------
void LocalCutSeparator::separate() {

    //
    // Constrcut a small graph. The graph has to be connected and the number of
    // Stable Set solutions should not exceed 10.000. The reason is that the
    // number of Stable Set solutions is the number of constraints in a LP.
    // The function getPartitialGraph() returns true if a connected graph could
    // be found. Otherwise we will stop this separation.
    //
    if (getPartitialGraph()) {

        //
        // Enumerate all Stable Sets of the small graph.
        //

	int   i, j, k;		// Loop variables;
	int   size;
	bool  allFound = false; // If this is true, all Stable Sets have been
				// computed.
	bool  noStableSet;	// Indicates whether a 0-1 sequence is a Stable
				// Set or not.
	// Store the last computed 0-1 sequence, even if it is no Stable Set.
	short *oldSolution = new short[numberOfNodes];
	// Store a new Stable Set.
	short *newSolution = new short[numberOfNodes];
	// Store all computed Stable Sets.
	ABA_BUFFER<short*> allSolutions(master_, numberOfNodes * numberOfNodes);

	// Compute the first Stable Set wich is the 0 sequence and only the
	// last vertex is set to 1.
	// -> Only the last vertex is in the Stable Set.
	for (i=0; i<numberOfNodes-1; i++) {
	    oldSolution[i] = 0;
	    newSolution[i] = 0;
	}
	oldSolution[numberOfNodes -1] = 1;
	newSolution[numberOfNodes -1] = 1;
	// Store this Stable Set.
	allSolutions.push(newSolution);



	//
	// Compute all Stable Set.
	// The algorithm computes in each while-loop one 0-1 sequence. This
	// sequence is unique determined by the last 0-1 sequence.
	// To be sure that this sequence is a Stable Set we have to check it.
	// If the check is positive we store the new solution.
	// The while-loop is finished, ater we have computed all Stable Sets.
	//
	while (!allFound) {

	    // Allocate memory for the new solution.
	    newSolution = new short[numberOfNodes];
	    // i is the last vertex.
	    i = numberOfNodes - 1;
	    // There are two cases:
	    // 1.) The last vertex has value 0.
	    // 2.) The last vertex has value 1.
	    if (oldSolution[i] == 0) {
	        // This is the first case.
		// The next sequence differs from the last one only in the
		// last element: It is set to 1.
	        newSolution[i] = 1;
		oldSolution[i] = 1;
		i--;
		// Copy all the other values.
	    	while (i >= 0) {
		    newSolution[i] = oldSolution[i];
		    i--;
	    	}// while
	    }// if
	    else {

		// The second case.
		// We search for the next vertex which has value 0.
		// The search is done backwards.
		while (oldSolution[i] == 1) {
		    // Reset all the vertecies to value 0.
	            newSolution[i] = 0;
		    oldSolution[i] = 0;
		    i--;
		    // If i is equal to -1 we know that the last 0-1 sequence
		    // was the hole graph. In this case we can stop because
		    // we have computed all Stable Sets if we add the empty set.
		    if (i == -1) {
			allFound = true;
			// Add the empty set to the set of Stable Set solutions.
			if (allSolutions.full()) {
		    	    allSolutions.realloc(allSolutions.size() + 1);
			}
			allSolutions.push(newSolution);
			break;
		    }
	        } // while
		if (allFound) {
		    continue;
		}
		// The new solution selects vertex i into the Stable Set.
		// All vertices which are greater than i are not in the set.
		newSolution[i] = 1;
		oldSolution[i] = 1;
		i--;
	        while (i >= 0) {
		    newSolution[i] = oldSolution[i];
		    i--;
	        }
	    }// else

	    // At this point we have computed a new 0-1 sequence.
	    // The next step is to check if this sequence is a Stable Set.

	    noStableSet = false;
	    // Check each pair of vertecies of the Stable Set if they are
	    // adjazent.
	    for (i=0; i<numberOfNodes; i++) {
		for (j=i+1; j<numberOfNodes; j++) {
		    k = 0;
		    // Look if the two nodes are in the set.
		    if ((newSolution[i] == 1) && (newSolution[j] == 1)) {
		    	// Check if node i and node j are adjazent.
			// We exploit that the adjazens list is upward sorted.
			size = adjacentList[i].size();
    			while((k < size) && (j > adjacentList[i][k])) {
			    k++;
    			}// while
			// This index k is decisive.
    			if ((k < size) && (j == adjacentList[i][k])) {
			    // -> This is no Stable Set because node i and j
			    //    are adjazent.
			    noStableSet = true;
			    // Update the 0-1 sequence.
			    // It can easily be seen that there is no 0-1
			    // sequence which is a Stable Set and have the first
			    // j vertecies in common with this sequence.
			    // Therefore we can jumpt to the set which has all
			    // ones after the j vertex.
			    for (k=j+1; k<numberOfNodes; k++) {
				oldSolution[k] = 1;
			    }// for
			    break;
			}// if
		    }// if
		}// for
		if (noStableSet) {
		    break;
		}// if
	    }// for
	    if (noStableSet) {
		delete [] newSolution;
	    }// if
	    else {
	    	// The sequence is a Stable Set and therefore we store it.
		// Check if we have enough memory. If not we have to allocate
		// some additional.
	        if (allSolutions.full()) {
		    allSolutions.realloc(allSolutions.size()*2);
	        }// if
	        allSolutions.push(newSolution);
	    }// else
	}// while

        // Clean up.
	delete [] oldSolution;



	//
	// Create and solve LP.
	//


	// The number of constraints of the LP is the number of Stable Set
	// solutions.
	int numberOfColumns = allSolutions.number();
	// The number of non zero elements in a constraint.
	int numberOfNonZeros = 0;
	// Right hand side of a constraint.
	double rhs;
	// Solution value of the last LP for each node.
	// Be careful with the indizes. They have not to be equal to the one
	// of the small graph!
	double *solutionValue = lpSolution->zVal();
	// Inner point of the Stable Set polytop.
	double innerPoint[numberOfNodes];
	// Coefficint to calculate the inner point.
	double coeff = 1.0 / allSolutions.number();
	// A pointer to an ABACUS constraint.
	ABA_ROW *constraint;
	// Buffer the non zero elements of a constraint.
	ABA_BUFFER<int> nonZeroElementsBuf(master_, numberOfColumns);
	// Store the coefficients of each buffered vertex.
	ABA_BUFFER<double> coefficientsBuf(master_, numberOfColumns);
	// Store the vertecies of a constraint...
	ABA_ARRAY<int> *nonZeroElements;
	// ...and their coefficients.
	ABA_ARRAY<double> *coefficients;
	// The coefficients of the target function. They are initialized
	// with value 1.
	ABA_ARRAY<double> objectiveFunctionCoefficient(master_,
						       numberOfColumns, 1);
    	// Lower bound of each variable is 0.
        ABA_ARRAY<double> lowerBound(master_, numberOfColumns, 0);
        // There is no upper bound for the variables.
        ABA_ARRAY<double> upperBound(master_, numberOfColumns, 1000000);
	// Store all constraints.
	// There are number of nodes constraints:
    	// For each component of the solution vector one.
	ABA_ARRAY<ABA_ROW*> allConstraints(master_, numberOfNodes);
	

	// Construct interior point as a convex combination of all Stable Sets.
	for (i=0; i<numberOfNodes; i++) {
	    innerPoint[i] = 0;
	}
	for (i=0; i<allSolutions.number()-1; i++) {
            for (j=0; j<numberOfNodes; j++) {
	        innerPoint[j] += coeff * allSolutions[i][j];
	    }
        }

	// Construct all constraints.
	for (i=0; i<numberOfNodes; i++) {
	    for (j=0; j<numberOfColumns; j++) {
		if (allSolutions[j][i] - innerPoint[i] != 0) {
		    nonZeroElementsBuf.push(j);
		    coefficientsBuf.push(allSolutions[j][i] - innerPoint[i]);
		    numberOfNonZeros++;
		}// if
	    }// for

	    nonZeroElements = new ABA_ARRAY<int>(master_, nonZeroElementsBuf);
	    coefficients    = new ABA_ARRAY<double>(master_, coefficientsBuf);
	    nonZeroElementsBuf.clear();
	    coefficientsBuf.clear();

	    // Compute the right hand side of this constraint.
	    rhs = solutionValue[nodes[i]] - innerPoint[i];

	    // The new constraint.
	    constraint = new ABA_ROW(master_,
				     numberOfNonZeros,
				     (*nonZeroElements),
				     (*coefficients),
				     ABA_CSENSE::Equal,
				     rhs);
	    // Store this constraint.
	    allConstraints[i] = constraint;
	}// for


	//
	// Optimize
	//
	ABA_CPLEXIF lp(master_ );
	lp.initialize(ABA_OPTSENSE::Min,
		      numberOfNodes,
		      numberOfNodes,
		      numberOfColumns,
		      numberOfColumns,
		      objectiveFunctionCoefficient,
		      lowerBound,
		      upperBound,
		      allConstraints);

	// Optimize and check if optimal solution could be computed.
	if (!(ABA_LP::Optimal == lp.optimize(ABA_LP::Primal))) {
	    master_->err() << "ERROR in LocalCutSeparator::separate()." << endl;
	    master_->err() << "LP could not be dolved to optimality." << endl;
	    exit(master_->Fatal);
	}

	double eps = master_->machineEps(); // Machine precision.

	//
	// Compute cut.
	//
	if (lp.value() > (1 + eps)) {

	    double a;			// Value of a dual variable
	    double b;			// Value of a dual variable.
	    double gcd;			// Greates common divisor of two double.
	    int    dualMult;		// A multiplicator.
	    int    dualVariable[numberOfNodes];	// Dual variables.
	    int    rhsConstraint;	// Right hand side of the cut
	    				// constraint.
	    vector<int> vertices;	// Vertices of the cut constraint.
	    vector<int> dualCoeff;	// Coefficients of the variables of the
	    				// cut constraints.


	    i = 0;
    	    a = lp.yVal(i);
	    dualVariable[i] = -1;
    	    while (((-eps < a) && (a < eps))
		   || (a <= -100000000)
		   || (a >=  100000000)
		  ) {
	        dualVariable[i] = 0;
	        i++;
	        if (i == numberOfNodes) {
		    master_->err() << "ERROR in LocalCutSeparator::separate().";
		    master_->err() << "\nAll dual variables are zero." << endl;
		    exit(master_->Fatal);
	        }
		a = lp.yVal(i);
		dualVariable[i] = -1;
            }// while

	    i++;

	    b = lp.yVal(i);
	    dualVariable[i] = -1;
	    while (((-eps < b) && (b < eps))
		   || (b <= -100000000)
		   || (b >=  100000000)
		  ) {
	        dualVariable[i] = 0;
	        i++;
		if (i == numberOfNodes) {
		    break;
		}
		b = lp.yVal(i);
		dualVariable[i] = -1;
	    }// while

	    gcd = calculateGCD(a, b);

	    i++;
	    while (i < numberOfNodes) {
		a = lp.yVal(i);
		if (((-eps < a) && (a < eps))
		    || (a <= -100000000)
		    || (a >=  100000000)
		   ) {
		    dualVariable[i] = 0;
		}
		else {
		    gcd = calculateGCD(a, gcd);
		    dualVariable[i] = -1;
		}
		i++;
	    }// while

	    for (i=0; i<numberOfNodes; i++) {
		if (dualVariable[i] == -1) {
		    a = lp.yVal(i) / gcd;
		    dualMult = (int) ceil(a);
		    if ((dualMult - a) > 0.001) {
		        dualMult -= 1;
		    }
		    dualVariable[i] = dualMult;
	        }
	    }// for

	    // Calculate rhs.
	    rhs = 1.0 / gcd;
	    for (i=0; i<numberOfNodes; i++) {
	        rhs += innerPoint[i] * dualVariable[i];
	    }

	    rhsConstraint = (int)floor(rhs + 0.001);

	    // Vector vertices is sorted beacuse vector nodes is sorted.
	    for (i=0; i<numberOfNodes; i++) {
	        if (dualVariable[i] != 0) {
		    vertices.push_back(nodes[i]);
		    dualCoeff.push_back(dualVariable[i]);
	        }
	    }// for

	    GeneralConstraint *generalConstr;
	    generalConstr = new GeneralConstraint(master_,
	    					  &vertices,
						  &dualCoeff,
						  rhsConstraint,
						  itsGraph);
	    cutFound(generalConstr);

        }

    	// Clean up.
	for (i=0; i<allSolutions.number(); i++) {
	    delete [] allSolutions[i];
	}


    }// if (getPartitialGraph())
    else {
        master_->out() << "Local cut separation: Subgraph could ";
	master_->out() << "not be computed." << endl;
    }

}// void separate()



// -----------------------------------------------------------------------------
// --------------- M e t h o d s  ( p r i v a t e ) ----------------------------
// -----------------------------------------------------------------------------


// -----------------------------------------------------------------------------
// g e t P a r t i t i a l G r a p h
//
// Construct a small graph where the local cut can be computed.
// -----------------------------------------------------------------------------
bool LocalCutSeparator::getPartitialGraph() {

    int i, j, k;
    double eps = master_->machineEps(); // Machine precision.
    double *solution = lpSolution->zVal();
    int    graphSize = itsGraph->numberOfNodes();
    vector<NodesAndWeights2> fractionalValue; // Save all fractional variables.

    // Get all fractional variables.
    for (i=0; i<graphSize; i++ ) {
        if ((solution[i] < -eps)
	    || (solution[i] > 1 + eps)
	    || ((solution[i] < 1 -eps) && (solution[i] > eps))
	   ) {
            // This variable is fractional.
            NodesAndWeights2 a;
	    a.inode   = i;
	    a.wweight = solution[i];
            fractionalValue.push_back(a);
	}
    }

    // Sort the fractional variables:
    // the value with the highest weight is the last in this vector.
    sort(fractionalValue.begin(), fractionalValue.end());
    nodes.clear();

    int sizeOfGraph  = fractionalValue.size() < MAX_NUMBER_OF_NODES
    		       ? fractionalValue.size()
		       : MAX_NUMBER_OF_NODES;
    bool found = false;
    int cnt;
    int maxNum;
    int indexOne;
    int indexTwo;
    int node = fractionalValue.back().inode;
    nodes.push_back(node);
    fractionalValue.pop_back();

    //
    // Increase until the partitial graph is big enough.
    //
    while (nodes.size() < sizeOfGraph) {
        if (found) {
	    break;
	}

    	cnt    = 0;
    	found  = false;
	maxNum = stableSetMaster()->getRandomNumber(10) < 4 ? 1 : 2;
	indexOne = stableSetMaster()->getRandomNumber(fractionalValue.size());

	for (i=0; i<fractionalValue.size(); i++) {
	    k = 0;
	    indexTwo = (i + indexOne) % fractionalValue.size();
	    if (itsGraph->isMemberOfAdjacentList(node, k, // 
					fractionalValue[indexTwo].inode)) {
		nodes.push_back(fractionalValue[indexTwo].inode);
		fractionalValue.erase(fractionalValue.begin() + indexTwo);
		cnt++;
		if (cnt == maxNum) {
		    break;
		}
	    }
	}
	if (cnt == 0) {
	    if (nodes.size() == 1) {
	        if (fractionalValue.size() < MIN_NUMBER_OF_NODES) {
		    return false;
		}
		nodes.pop_back();
		node = fractionalValue.back().inode;
		nodes.push_back(node);
    		fractionalValue.pop_back();
		// Update max size.
		sizeOfGraph = fractionalValue.size() < MAX_NUMBER_OF_NODES
			      ? fractionalValue.size()
			      : MAX_NUMBER_OF_NODES;
	    }
	    else {
	        // Get index of current node.
		k = nodes.size() - 1;
		while (k > 0) {
		    if (node == nodes[k]) {
			break;
		    }
		    k--;
		}

		if ((k == 0) && (node != nodes[0])) {
		    master_->err() << "LocalCutSeparator::";
		    master_->err() << "getPartitialGraph()" << endl;
		    master_->err() << "node could not be found" << endl;
		    exit(master_->Fatal);
		}
		if (k != 0) {
		    node = nodes[k-1];
		}
		else { // k == 0

		    if (nodes.size() >= MIN_NUMBER_OF_NODES) {
			found = true;
		    }
		    else {
		        if (fractionalValue.size() < MIN_NUMBER_OF_NODES) {
			    sizeOfGraph = fractionalValue.size()
					  < MAX_NUMBER_OF_NODES
			    		  ? fractionalValue.size()
					  : MAX_NUMBER_OF_NODES;
			    return false;
			}
			else {
			    nodes.clear();
			    node = fractionalValue.back().inode;
    			    nodes.push_back(node);
    			    fractionalValue.pop_back();
			}
		    }
		}
	    }
	}
	else {
	    node = nodes.back();
	}
    }

    // A partitional graph could be found.
    // Create the adjacent list and store the size of the graph.
    numberOfNodes = nodes.size();
    sort(nodes.begin(), nodes.end());
    adjacentList.resize(numberOfNodes);
    for (i=0; i<numberOfNodes; i++) {
        k = 0;
	for (j=i+1; j<numberOfNodes; j++) {
	    if (itsGraph->isMemberOfAdjacentList(nodes[i], k, nodes[j])) {
		adjacentList[i].push_back(j);
		adjacentList[j].push_back(i);
	    }
	}
    }

    return true;

}// bool getPartitialGraph()


// -----------------------------------------------------------------------------
// c a l c u l a t e G C D
// -----------------------------------------------------------------------------
double LocalCutSeparator::calculateGCD(double a, double b) {

    double eps = master_->machineEps(); // Machine precision.
    while ((a > (b + eps)) || (a < (b - eps))) {
	if (a > b) {
	    a -= b;
	}
	else {
	    b -= a;
	}
    }

    return a;
}


// -----------------------------------------------------------------------------
// s t a b l e S e t M a s t e r
//
// This function returns a pointer to the corresponding object of the class
// StableSetMaster.
// -----------------------------------------------------------------------------
StableSetMaster *LocalCutSeparator::stableSetMaster() {
    return (StableSetMaster *) master_;
}







