////////////////////////////////////// c++ /////////////////////////////////////
//
//  Module           : LowerBoundHeuristic.cpp
//  Description      : Rounding heuristic for the maximum stable set problem.
//  Author           : Steffen Rebennack
//  Email            : srebenac@ix.urz.uni-heidelberg.de;
//                     steffen.rebennack@web.de
//  Copyright        : (C) 2006 by the University of Heidelberg
//  Created on       : Wed Apr 05 14:44:10 2006
//  Last modified by : -
//  Last modified on : -
//  Update count     : 0
//
////////////////////////////////////////////////////////////////////////////////
//
//  Date	Name		Changes/Extensions
//  ----	----		------------------
//
////////////////////////////////////////////////////////////////////////////////


#include "LowerBoundHeuristic.hh"
#include "Graph.hh"
#include "StableSetMaster.hh"
#include "StableSetStatistics.hh"



//
// The struct NodesAndWeights stores an index of a node and a weight.
//
struct NodesAndWeights {
    int    inode;         // index of node
    double wweight;       // weighted weight of node
};// struct NodesAndWeights


//
// This operator is used by the sort function in lowerBoundHeuristic to get
// a sorted vector of variables with increasing weight.
//
bool operator<(const NodesAndWeights &a, const NodesAndWeights &b) {
        return a.wweight > b.wweight;
}



// -----------------------------------------------------------------------------
// --------------- M e t h o d s  ( p u b l i c ) ------------------------------
// -----------------------------------------------------------------------------


// -----------------------------------------------------------------------------
// C o n s t r u c t o r
// -----------------------------------------------------------------------------
LowerBoundHeuristic::LowerBoundHeuristic(ABA_MASTER *master, Graph *theGraph,
					 StableSetStatistics *theStatistics):
    master_(master),
    itsGraph(theGraph),
    itsStatistics(theStatistics)
{
}


// -----------------------------------------------------------------------------
// D e s t r u c t o r
// -----------------------------------------------------------------------------
LowerBoundHeuristic::~LowerBoundHeuristic() {
}


// -----------------------------------------------------------------------------
// c o m p u t e S o l u t i o n
//
// The function findSolution() is a heuristic to compute a Stable Set
// from a given solution vector of a fractional LP solution. It uses the
// structure NodesAndWeights with its special order.
//
// The heuristic works in the following way:
// - First fix all variables which have binary values.
// - Sort all variables which are not binary by the solution value multiplied
//   with its weight (if there are no weights than the weight is set to 1).
// - Fix the variable with the highest value to 1 and fix all its adjazent
//   nodes to 0. Do this last step untill all variables are binary.
//
// The heuristic stores the solution in the arry integerSolution and value is
// the corresponding weight of the STable Set
// -----------------------------------------------------------------------------
int LowerBoundHeuristic::computeSolution(double &value, double *xVal_, int level) {

    int i, j, a; 			// Loop counter.
    int size;				// Loop variable.
    int numberOfNodes = itsGraph->numberOfNodes(); // Problem size.
    int stack; 				// Storage.
    int node;  				// A node.
    int index;
    int max;
    double eps = 0.000001; // Machine precision.
    vector<NodesAndWeights> fractionalValue; // Save all fractional variables.
    double weight = 0;
    int integerSolution[numberOfNodes];
    int status = 0;

    //
    // Parameter
    //
    const int    ITERATION_NUMBER = level==1 ? 15 : 2;
    const int    RANDOM_NUMBER    = 100;
    const double ADD              = 0.001;

    for (a=0; a<ITERATION_NUMBER; a++) {

        weight = 0;

        // Initialize the solution array.
        for (i=0; i<numberOfNodes; i++) {
            integerSolution[i] = -1;
        }

        // Fix all binary variables and if the value is 1 fix all adjazent nodes
        // to value 0. If a variable is not binary store it together with its
        // weight in the vector fractionalValue.
        for (i=0; i<numberOfNodes; i++ ) {
            if ((xVal_[i] < eps) && (xVal_[i] > -eps)) {
                // the value is zero
                integerSolution[i] = 0;
            }
            else {
    	        if ((xVal_[i] > 1 - eps) && (xVal_[i] < 1 + eps)) {
                    if (integerSolution[i] == 0) {
		        continue;
		    }
    		    // Its value is one: fix all adjazent nodes.
                    integerSolution[i] = 1;
                    weight += itsGraph->weighted() ? itsGraph->weight(i) : 1;
                    for (j=0; j<itsGraph->adjacentListSize(i); j++ ) {
    		        stack = itsGraph->adjacentListElement(i, j);
    	                integerSolution[stack] = 0;
	            }
                }
                else {
                    // A fractional variable.
                    NodesAndWeights a;
	            a.inode   = i;
		    a.wweight = itsGraph->weighted()
		                ? itsGraph->weight(i)* xVal_[i]
			        : xVal_[i];
		    a.wweight  += ADD * //
		   ((StableSetMaster *)master_)->getRandomNumber(RANDOM_NUMBER);
                    fractionalValue.push_back(a);
	        }
            }
        }

        // Sort the fractional variables:
        // the value with the highest weight is the last in this vector.
        sort(fractionalValue.begin(), fractionalValue.end());

        // Fix all variables untill this vector is empty.
        while (!fractionalValue.empty()) {
	    size = fractionalValue.size();
	    max = 512;
	    stack = ((StableSetMaster *)master_)->getRandomNumber(1024);

	    index = size -1;
	    while (stack < max) {
	        max = max / 2;
	        if (index == 0) {
		    break;
	        }
	        index--;
	    }
            stack = fractionalValue[index].inode;

            if (integerSolution[stack] == -1) {
                integerSolution[stack] = 1;
                weight += itsGraph->weighted() ? itsGraph->weight(i) : 1;

	        size = itsGraph->adjacentListSize(stack);
                for (i=0; i<size; i++) {
	            node = itsGraph->adjacentListElement(stack, i);

                    if(integerSolution[node] == 1) {
                        master_->err() << "StableSetSUB::lowerBoundHeuristic: ";
                        master_->err() << "Setting of variable " << node;
                        master_->err() << " to value 0 failed." << endl;
                        exit(master_->Fatal);
		    }
                    integerSolution[node] = 0;
	        }
                fractionalValue.erase(fractionalValue.begin() + index);
            }
            else {
                fractionalValue.erase(fractionalValue.begin() + index);
            }
        }// while

        // Update.
        if (master_->primalBound() <= weight) {
	    vector<int> *solution = ((StableSetMaster*)master_)->getSolution();
	    solution->clear();
	    // Update solution.
 	    for (i=0; i<numberOfNodes; i++) {
	        if (integerSolution[i] == 1) {
	            solution->push_back(i);
	        }
	    }
	    testStableSet(solution);
	    if (master_->primalBound() < weight) {
	        master_->out() << "\nHeuristic found better feasible ";
                master_->out() << "solution with value " << weight;
                master_->out() << "." << endl;

	        // Update lower bound (best known feasible solution)
                // (this is done automatically by ABACUS, too).
	        value = weight;
                master_->primalBound(value);
	    
                // Update statistics.
	        itsStatistics->roundingHeuristics();
	        itsStatistics->solutionFound(1);

	        status = 1;
	    }
        }// if
    }// for

    return status;

}// end: void lowerBoundHeuristic(..)



// -----------------------------------------------------------------------------
// -------------- M e t h o d s  ( p r i v a t e ) -----------------------------
// -----------------------------------------------------------------------------


// -----------------------------------------------------------------------------
// t e s t S t a b l e S e t
// -----------------------------------------------------------------------------
void LowerBoundHeuristic::testStableSet(vector<int> *stableSet) {

    int i, j, l;

    for (i=1; i<stableSet->size(); i++) {
	for (j=(i-1); j>=0; j--) {
	    l = 0;
	    if (itsGraph->isMemberOfAdjacentList((*stableSet)[i], l,
	    					 (*stableSet)[j])) {
		master_->err() << "LowerBoundHeuristic::testStableSet:\t";
		master_->err() << "The constructed svector is no stable set!\n";
		master_->err() << "Solution:\t";
		for (l=0; l<stableSet->size(); l++) {
		    master_->err() << (*stableSet)[l] << " ";
		}
		master_->err() << endl;
		exit(master_->Fatal);
	    }// if
	}
    }// for
}


