////////////////////////////////////// c++ /////////////////////////////////////
//
//  Module           : LowerBoundHeuristic.hh
//  Description      : Rounding heuristic for the maximum stable set problem.
//  Author           : Steffen Rebennack
//  Email            : srebenac@ix.urz.uni-heidelberg.de;
//                     steffen.rebennack@web.de
//  Copyright        : (C) 2006 by the University of Heidelberg
//  Created on       : Wed Apr 05 14:32:54 2006
//  Last modified by : -
//  Last modified on : - 
//  Update count     : 0
//
////////////////////////////////////////////////////////////////////////////////
//
//  Date	Name		Changes/Extensions
//  ----	----		------------------
//
////////////////////////////////////////////////////////////////////////////////

#ifndef LOWER_BOUND_HEURISTIC_HH
#define LOWER_BOUND_HEURISTIC_HH

#include <vector>

using namespace std;

class ABA_MASTER;
class Graph;
class StableSetStatistics;



/**
 * This class provides a heuristic to compute a stable set from a fractional
 * LP solution.
 */
class LowerBoundHeuristic {

public:

    // -------------------------------------------------------------------------
    // ------------- M e t h o d s  ( p u b l i c ) ----------------------------
    // -------------------------------------------------------------------------

    /**
     * Constructor.
     *
     * @param *master Pointer to the master of the problem.
     * @param *theGraph Pointer to the graph storing all information about the
     *			problem instance.
     * @param *theStatistics Pointer to an object managing the statistics.
     */
    LowerBoundHeuristic(ABA_MASTER *master, Graph *theGraph,
			StableSetStatistics *theStatistics);

    /**
     * Destructor.
     */
    ~LowerBoundHeuristic();

    /**
     * Heuristic to compute a Stable Set.
     *
     * @param &value Reference to the weight of the best stable set found so
     *               far.
     * @param *xVal_ Pointer to the current LP solution value.
     * @param level Level of the subprobelm in the tree.
     * return 1 A stable set with higher weight was computed.
     * return 0 There was no better stable set found.
     */
    int computeSolution(double &value, double *xVal_, int level);


private:

    // -------------------------------------------------------------------------
    // ------------- M e t h o d s  ( p r i v a t e ) --------------------------
    // -------------------------------------------------------------------------
  
    /**
     * Test if a set of nodes is a stable set.
     *
     * @param *stableSet Pointer to a vector storing the stable set to be
     *			 checked.
     * @return void
     */
  
    void testStableSet(vector<int> *stableSet);


    // -------------------------------------------------------------------------
    // ------------- V a r i a b l e s  ( p r i v a t e ) ----------------------
    // -------------------------------------------------------------------------

    /**
     * Pointer to the master of the problem.
     */
    ABA_MASTER *master_;

    /**
     * Pointer to the graph of the problem.
     */
    Graph *itsGraph;

    /**
     * Pointer to an object storing the statistics.
     */    
    StableSetStatistics *itsStatistics;

};


#endif


