////////////////////////////////////// c++ /////////////////////////////////////
//
//  Module           : MaxCliquesSeparator.cpp
//  Description      : Exact clique separation.
//  Author           : Steffen Rebennack
//  Email            : srebenac@ix.urz.uni-heidelberg.de;
//                     steffen.rebennack@web.de
//  Copyright        : (C) 2006 by the University of Heidelberg
//  Created on       : Wed Apr 05 14:49:52 2006
//  Last modified by : -
//  Last modified on : -
//  Update count     : 0
//
////////////////////////////////////////////////////////////////////////////////
//
//  Date	Name		Changes/Extensions
//  ----	----		------------------
//
////////////////////////////////////////////////////////////////////////////////


#include "MaxCliquesSeparator.hh"
#include "Clique.hh"
#include "CliqueConstraint.hh"
#include "Graph.hh"
#include "StableSetLPSolution.hh"
#include "StableSetMaster.hh"



// -----------------------------------------------------------------------------
// --------------- M e t h o d s  ( p u b l i c ) ------------------------------
// -----------------------------------------------------------------------------


// -----------------------------------------------------------------------------
// C o n s t r u c t o r
//
// The main part of this constructor is the initialization of the abstract
// template class ABA_SEPARATOR. The second argument is true because we do
// not want double constraints in the buffer and the nuber 500 states the
// maximla number of cuttings planes which are stored.
// -----------------------------------------------------------------------------
MaxCliquesSeparator::MaxCliquesSeparator(ABA_MASTER *master,
    			StableSetLPSolution *solution, Graph *theGraph,
			Clique *clique):
    ABA_SEPARATOR<ABA_CONSTRAINT, ABA_VARIABLE> (solution, true, 500),
    master_(master),
    lpSolution(solution),
    itsGraph(theGraph),
    maxCliques(clique)
{
}


// -----------------------------------------------------------------------------
// D e s t r u c t o r
// -----------------------------------------------------------------------------
MaxCliquesSeparator::~MaxCliquesSeparator() {
}


// -----------------------------------------------------------------------------
// s e p a r a t e
//
// This function initializes the exact separation of the max clique
// inequalities. It constructs the last maximal clique found and calls the
// function getNextClique(..) untill the timelimit is reached or all maximal
// cliques have been computed. If one clique is found this function adds the
// coresponding inequality only if the clique is a new one and the weight of
// the clique is greater than 1 plus a tolerance.
// -----------------------------------------------------------------------------
void MaxCliquesSeparator::separate() {

    clock_t t1 = clock();       // Get current time.

    int    i;			// Loop variable.
    int    cliqueSize;		// Size of a clique.
    int    *lastClique;		// Store a clique.
    bool   maxCliqueFound;	// true if a maximal clique has been computed.
    bool   added;		// true if a clique could be added.
    bool   fixed;		// true if the variables of the clique could
    				// be fixed.
    double weight;		// Weight of a clique.
    double *solutionValue = lpSolution->zVal();     // LP solution
    double minViolation = stableSetMaster()->m_minAbsViolationClique;
    double timeLimit = stableSetMaster()->m_timeLimitForMaxCliqueSep;
    vector<int> nodesOfClique;  // Store the nodes of a clique.
    CliqueConstraint *cliqueConstr;	// Pointer to clique constraint.


    // Fill vector nodesOfClique with last found max clique.
    lastClique = maxCliques->getLastMaxClique();
    if (lastClique != NULL) {
	cliqueSize = lastClique[0];
	nodesOfClique.resize(cliqueSize);
	for (i=0; i<cliqueSize; i++) {
	    nodesOfClique[i] = lastClique[i+1];
	}
    }
    else {
        // No cliques found, yet.
        nodesOfClique.push_back(-1);
    }

    // Produce new maximal cliques untill time limit is reached
    // or all cliques are computed.
    while ((!maxCliques->allMaximalCliquesFound())
	   && (timeLimit > (double)(clock() - t1)/CLOCKS_PER_SEC)
	  ) {
	// Compute next maximal clique.
	maxCliqueFound = getNextClique(&nodesOfClique);
	if (!maxCliqueFound) {
	    continue;
	}


	// Store this clique if it is not part of a greater one and non of
	// its variables could be fixed!
	added = maxCliques->addMaxClique(&nodesOfClique);
	if (!added) {
	    continue;
	}

	// This clique is a new maximal clique and therefore we
	// compute the weight of it.

	weight = 0;
	for (i=0; i<nodesOfClique.size(); i++) {
	    weight += solutionValue[nodesOfClique[i]];
	}

	if (weight >= 1 + minViolation) {
	    // This clique constraint is violated.
	    // So we have a new clique costraint to be added.
	    cliqueConstr = new CliqueConstraint(master_, &nodesOfClique, itsGraph);
	    cutFound(cliqueConstr);
	}
    }// end: while(..)


} // end: void separate()



// -----------------------------------------------------------------------------
// --------------- M e t h o d s  ( p r i v a t e ) ----------------------------
// -----------------------------------------------------------------------------


// -----------------------------------------------------------------------------
// g e t N e x t C l i q u e
//
// This function computes the next maximal clique in the following way:
// 1.) Delete last node of clique 'nodesOfClique'
//     (if it has only one member: increase this node)
// 2.) Find all adjacent nodes to the vertecies of the clique and store them in
//     vector 'adjacentNodes'.
// 3.) If there are no adjacent nodes -> goto 1.)
//     Else increase clique until there are no more adjacent nodes.
// The function returns true, if a max clique inequality could be found;
// else the returned value is false. In the latter case all maximal cliques
// are found.
// -----------------------------------------------------------------------------
bool MaxCliquesSeparator::getNextClique(vector<int> *nodesOfClique) const {

    int  i, j;  	   // Loop counter.
    int  lastNode;  	   // Save node which has been deleted from the clique.
    int  node;		   // A node.
    int  vertex;	   // A vertex.
    int  nodeOfClique;	   // Node of the clique.
    int  minIndex;         // Index of node with the minimal size of its
    			   // adjacent list.
    int  minValue = -1;    // Size of the adjacent list of the node
    			   // corresponding to minIndex
    int  adjacentListSize; // Size of an adjacent list.
    int  numberOfNodes = itsGraph->numberOfNodes();
    int  cliqueSize = nodesOfClique->size();
    int  begin[cliqueSize];
    vector<int> adjacentNodes;  // Store all adjacent nodes of the clique
    				// 'nodesOfClique'.
    bool found;		   // Flag which shows if something is found.

    for (i=0; i<cliqueSize; i++) {
	begin[i] = 0;
    }


    // Delete last element of vector nodesOfClique.
    if (cliqueSize == 1) {
        // Only one element in this vector
        // -> increase this node to search for new cliques.
        lastNode = (*nodesOfClique)[0];
        if (lastNode < (numberOfNodes - 1)) {
            (*nodesOfClique)[0] = lastNode + 1;
	}
        else {
            maxCliques->setAllCliquesFound();
	    return false;
        }
    }
    else {
        lastNode = nodesOfClique->back();
        nodesOfClique->pop_back();
	--cliqueSize;
    }

     // Get element with minimal number of adjacent nodes.
    for (i=0; i<cliqueSize; i++) {
        adjacentListSize = itsGraph->adjacentListSize((*nodesOfClique)[i]);
        if ((adjacentListSize < minValue) || (minValue == -1)) {
            minValue = adjacentListSize;
            minIndex = i;
        }
    }
    nodeOfClique = (*nodesOfClique)[minIndex];

    // Compute all adjacent nodes to the nodes of the clique stored in
    // nodesOfClique.
    adjacentListSize = itsGraph->adjacentListSize(nodeOfClique);
    for (i=0; i<adjacentListSize; i++ ) {
	// Ignore all nodes, which have allready been checked.
	vertex = itsGraph->adjacentListElement(nodeOfClique, i);
        if((vertex > (*nodesOfClique)[0]) && (vertex > lastNode)) {
            // Check for all other nodes:
	    // Is vertex adjacent to all other nodes of the clique?
	    found = true;
	    j = 0;
	    while ((j < cliqueSize) && found) {
		node = (*nodesOfClique)[j];
		if((minIndex != j) && (vertex != node)) {
		    found = itsGraph->isMemberOfAdjacentList(node, begin[j],
		    						 vertex);
		}
		j++;
	    }

	    if (found || (cliqueSize == 1)) {
		adjacentNodes.push_back(vertex);
	    }
	}// end: if
    }// end: for

    if (adjacentNodes.size() == 0) {
    	// No max clique possible.
        return getNextClique(nodesOfClique);
    }
    else {
    	// Increase clique untill there are no more adjacent nodes.
	while (true) {
	    // Add the first node of the adjacent list to the clique.
	    node = adjacentNodes[0];
	    nodesOfClique->push_back(node);
	    adjacentNodes.erase(adjacentNodes.begin());
	    j = 0;

            // Update adjacent nodes:
            // find out for each node of the adjacent list if it is also
	    // adjacent to the new added node.
	    for (i=0; i<adjacentNodes.size(); i++) {
		// Find node adjacentNodes[i] in adjacent list of node.
		if(!itsGraph->isMemberOfAdjacentList(node, j,
						     adjacentNodes[i])) {
		    // This node is not a memebr of the adjacent list
		    // -> delete this node.
		    adjacentNodes.erase(adjacentNodes.begin() + i);
		    i--;
		}
	    }

	    // If vector adjacentNodes is empty than there are no more
	    // adjacent nodes in the graph to the nodes of the clique of
	    // vector nodesOfClique. You have to check if the clique
    	    // is great enough - all cliques with size 2 are ignored.
	    // The check if the clique is a new one or only a subset of a
	    // greater clique is done by the function separate().
	    if (adjacentNodes.empty() && nodesOfClique->size() > 2) {
		// New maximal clique found.
		return true;
	    }
	    else {
		if(adjacentNodes.empty()) {
		    return getNextClique(nodesOfClique);
		}
	    }
	}
    }// else

}// end: bool getNextClique(..)


// -----------------------------------------------------------------------------
// s t a b l e S e t M a s t e r
//
// This function returns a pointer to the corresponding object of the class
// StableSetMaster.
// -----------------------------------------------------------------------------
StableSetMaster *MaxCliquesSeparator::stableSetMaster() {
   return (StableSetMaster *)master_;
}


