///////////////////////////////////// c++ /////////////////////////////////////
//
//  Module           : ModkSeparator.cpp
//  Description      :
//  Author           : Lars Fricke
//  Email            : lars.fricke@iwr.uni-heidelberg.de
//  Copyright        : University of Heidelberg
//  Created on       : Tue Nov 8 2005
//  Last modified by : Rebennack
//  Last modified on : Wed Apr 05 15:51:14 2006
//  Update count     : 2
//
////////////////////////////////////////////////////////////////////////////////
//
//  Date	Name		Changes/Extensions
//  Apr 04	Rebennack	Disable output; Clean up memory.
//  Apr 05      Rebennack	Fitting to programming guides.
//
////////////////////////////////////////////////////////////////////////////////

#include "ModkSeparator.defs"
#include "ModkSeparator.hh"
#include <time.h>
#include <stdio.h>
#include <math.h>
#include <stdlib.h>


// -----------------------------------------------------------------------------
// C o n s t r u c t o r
// 
// Needs the constraints with slack zero.
// -----------------------------------------------------------------------------
ModkSeparator::ModkSeparator(int modfac, int nConstraints, int nVariables,
			     int* nNonZeros, int** NonZeroIndices,
			     double** NonZeroCoeff, double* rhs,
			     double* lpSolution):
    ordGen(false),
    constraintsGen(false),
    congruenceSystemGen(0),
    colsGen(false),
    generatedCutsGen(false)
{
        K = modfac;
	// Order of elements

	if( K == 2 ){
	  Ord = new int[2];
	  ordGen = true;
	  Ord[0]=0;
	  Ord[1]=1;
        }
  	else if( K == 3 ){
  	  Ord = new int[3];
	  ordGen = true;
  	  Ord[0]=0;
  	  Ord[1]=1;
  	  Ord[2]=2;
  	}
  	else if( K == 5 ){
  	  Ord = new int[5];
	  ordGen = true;
  	  Ord[0]=0;
  	  Ord[1]=1;
  	  Ord[2]=3;
  	  Ord[3]=2;
  	  Ord[4]=4;
  	}
  	else if( K == 7 ){
  	  Ord = new int[7];
	  ordGen = true;
  	  Ord[0] = 0;
  	  Ord[1] = 1;
  	  Ord[2] = 4;
  	  Ord[3] = 5;
  	  Ord[4] = 2;
  	  Ord[5] = 3;
  	  Ord[6] = 6;
 	}
 	else{
 	  printf("%d is not a supported modulo factor\n", K);
 	  exit(1);
 	}


 	if(IMPORTANTVARIABLES){
          int curVar = 0;

          // Analyse variable distribution
          // Reset Count
          int* count = new int[nVariables];
          for(int i=0; i<nVariables; i++)
            count[i] = 0; 
          // Count variable occurences
          for(int i=0; i<nConstraints; i++)
            for(int j=0; j<nNonZeros[i]; j++)
                count[NonZeroIndices[i][j]]++;
          // Greedy - take variable with most occurences            
          for(int i=0; i<nVariables; i++){
            if(count[i] > count[curVar]){
              curVar = i;
            } 
          }
          // Start with the variable with most occurences

	  constraintsGen = true;
          NonZeroI = new int*[nConstraints]; 
          NonZeroC = new double*[nConstraints];   
	  nNonZ = new int[nConstraints];
          RHS = new double[nConstraints];
          
          int maxCon = nVariables + MAXGAP;
          int* taken = new int[nConstraints];
          for(int i=0; i<nConstraints; i++)
            taken[i] = 0;
          bool doBreak = false;
          bool doContinue = false;
          nCon = 0;
          int nConOld = nCon;
 	
          while(1){
            nConOld = nCon;
            //printf("Searching for occurences of variable %d\n", curVar);
            for(int i=0; i<nConstraints; i++){
              if(nCon == maxCon){
                //printf("Maximum number of constrains reached: %d\n", nCon);
                doBreak = true;
                break;
              }
              if(taken[i] == 0){
                for(int j=0; j<nNonZeros[i]; j++){
                  if(NonZeroIndices[i][j] == curVar){
                    //printf("Added constraint %d\n", i);
                    // Add constraint to pool
                    taken[i] = 1;
                    nNonZ[nCon] = nNonZeros[i];
                    NonZeroI[nCon] = NonZeroIndices[i];
                    NonZeroC[nCon] = NonZeroCoeff[i];
                    RHS[nCon] = rhs[i];
                    nCon++;
                  } // end if
                } // end for
              } // end if
            } // end for               
            if(doBreak) // break initiated 
              break; // break while        

            if(nConOld == nCon){ // No new constraints found
              doContinue = false;
              for(int i=0; i<nVariables; i++){
                if(count[i] != -1){
                  //printf("Changed variable to %d\n", i);
                  curVar = i;	// Take the next available variable
                  count[curVar] = -1;
                  doContinue = true;
                  break;     // and start new search
                }
              }
              if(doContinue)
                continue;
              else{
                //printf("Nothing more to find\n");
                break; // No variable found so lets stop the search
              }
            }
            // Reset Count
            count[curVar] = -1;
            for(int i=0; i<nVariables; i++){
              if(count[i] != -1)
                count[i] = 0;
            }
            // Count variable occurences
            for(int i=0; i<nCon; i++){
              for(int j=0; j<nNonZ[i]; j++){
                if(count[NonZeroI[i][j]] >= 0)
                  count[NonZeroI[i][j]]++;
              }
            }
            // Greedy - take variable with most occurences            
            
            for(int i=0; i<nVariables; i++){
              if(count[i] > count[curVar]){
                curVar = i;
              } 
            } 
                 
              
          } // end while
	
  	  // Clean up.
	  delete [] count;
	  delete [] taken;	

 	} // end if IMPORTANTVARIABLES
 	else{
          // Save pointers to important stuff
          //printf("Using all constraints\n");
          NonZeroI = NonZeroIndices;
          NonZeroC = NonZeroCoeff;
          RHS = rhs;
          nNonZ = nNonZeros;
          nCon = nConstraints;
        }
        
        nVar = nVariables;
        lpSol = lpSolution;

	nRows = nVar + 1; // + 1 for original rhs
        nCols = nCon + 1; // + 1 for congruence rhs
	
	// Allocate memory for the congruence system
	
	CongruenceSystem = new int*[nRows];
        congruenceSystemGen = nRows;

	for(int i=0; i<=nVar; i++)
		CongruenceSystem[i] = new int[nCols];

	// Create the mod-K congruence system
	
	int entry=0;

	for(int i=0; i<nCon; i++){
		for(int j=0; j<nVar; j++){
			entry = 0;
			for(int k=0; k<(int) nNonZ[i]; k++){
				if(NonZeroI[i][k] == j){ // is non-zero
		entry = (K + (((int) NonZeroC[i][k]) % K)) % K; // modulo-K
				}
			}
			CongruenceSystem[j][i] = entry;
		}	
	}

	for(int i=0; i<nRows-1; i++)
		CongruenceSystem[i][nCols-1] = 0;
	CongruenceSystem[nRows-1][nCols-1] = K-1;

	for(int i=0; i<nCon; i++){
		CongruenceSystem[nVar][i] = (K + (((int) RHS[i]) % K)) % K; 
	}
}// constructor
	

// -----------------------------------------------------------------------------
// D e s t r u c t o r
// -----------------------------------------------------------------------------
ModkSeparator::~ModkSeparator()
{
    // Clean up.
    if (ordGen) {
        delete [] Ord;
    }
    if (constraintsGen) {
        // Do not delete the entries.
	// They are only pointers to a object and not copied!
	delete [] NonZeroI;
    	delete [] NonZeroC;
    	delete [] nNonZ;
	delete [] RHS;
    }
    if (congruenceSystemGen) {
	for (int i=0; i<congruenceSystemGen; i++) {
	    delete [] CongruenceSystem[i];
//	    CongruenceSystem[i] = NULL; 
	}
	delete [] CongruenceSystem;
    }
    if (colsGen) {
        delete [] Cols;
    }
    if (generatedCutsGen) {
	for (int i=0; i<nGen; i++) {
	    delete [] GeneratedCuts[i];
	}
	delete [] GeneratedCuts;
    }
}


// -----------------------------------------------------------------------------
// s e p a r a t e
// -----------------------------------------------------------------------------
int ModkSeparator::separate(int nMaxCuts)
{    
        time_t begin = time(NULL);

	bool doBreak = false;
	int* swap;
	int swap2=0;
	int idSize= ((nRows) < (nCols)) ? (nRows) : (nCols);
	int factor = 0;

	Cols = new int[nCols];
        colsGen = true;
	for(int i=0; i<nCols; i++)
		Cols[i] = i;	

    /**
     * Randomize columns
     */
        if(RANDOMIZE){
          srand(time(NULL));
          for (int i=0; i<nCols; i++) {
            int r = rand() % nCols;  // generate a random position
            int temp = Cols[i]; Cols[i] = Cols[r]; Cols[r] = temp;
          }
        }

	swap = CongruenceSystem[0];
	CongruenceSystem[0] = CongruenceSystem[nRows-1];
	CongruenceSystem[nRows-1] = swap;

	
	// Gaussian Elimination on the congruence system (forwards)
	
	int scalFac = 0;
	
	for(int i=0; (i<nRows) && (i<nCols-1); i++){ // for every row i
		// Check if there is anything greater than 1
		if( CongruenceSystem[i][Cols[i]] > 1 ){
			scalFac = Ord[CongruenceSystem[i][Cols[i]]];
			// If so, multiply row with Ord[N] so that a
		        // 1 is on the diagonal
			for(int j=i; j<nCols; j++){    // for every column geq i
CongruenceSystem[i][Cols[j]] = (scalFac*CongruenceSystem[i][Cols[j]]) % K; 
			}
		}
				
		// Pivot search
                // there is a diagonal 0
		if( CongruenceSystem[i][Cols[i]] == 0 ){         
		    // Row pivot search
		    for(int k=i+1; k<nRows; k++){ 
                        // so search in the other rows             
		 	if( CongruenceSystem[k][Cols[i]] != 0 ){
                            // there is a non-zero  
		            swap = CongruenceSystem[i];     // swap the rows
			    CongruenceSystem[i] = CongruenceSystem[k];
			    CongruenceSystem[k] = swap;
			    break;	
	                    // and break out of the for-loop						
				}
			} // end of row search
		}	
		
                // Still a diagonoal 0 ?
		if( CongruenceSystem[i][Cols[i]] == 0 ){ 
			// Column pivot search
			
			for(int k=i+1; k<nCols-1; k++){
				// there is a non-zero
				if( CongruenceSystem[i][Cols[k]] != 0 ){ 
					swap2 = Cols[k];
   				        // swap the index vector entries
					Cols[k] = Cols[i];
					Cols[i] = swap2;
					break;
					// and break out of for-loop						
				} 	
			} 
		} //end of column search

		// Full pivot search			
		if( CongruenceSystem[i][Cols[i]] == 0 ){
		        // Still a diagonal 0 ?? Jesus tapdancing christ
			doBreak = false;
			for(int k=i+1; k<nRows; k++){		 
			// We now have to search the lower right quadrant
			// for any 1
				for(int l=i+1; l<nCols-1; l++){
					if( CongruenceSystem[k][Cols[l]] != 0){
				    // Finally !
				   swap = CongruenceSystem[i];
			           // swap the rows
				   CongruenceSystem[i] = CongruenceSystem[k];
			 	   CongruenceSystem[k] = swap;
				   swap2 = Cols[i];					
				   // and then swap the index vector entries
				   Cols[i] = Cols[l];
				   Cols[l] = swap2;	
				   doBreak = true;
				   // break out of the for-loop    
				   break;             
					}
					// break out of the for-loop
					if(doBreak) break;     
				}
			}
		}  // end of full pivot search
	  	// end of pivot search

		// Check if there is anything greater than 1
		if( CongruenceSystem[i][Cols[i]] > 1 ){
			// If so, multiply the row with Ord[N]
			scalFac = Ord[CongruenceSystem[i][Cols[i]]];
			for(int j=i; j<nCols; j++){    
				// for every column geq i
CongruenceSystem[i][Cols[j]] = (scalFac*CongruenceSystem[i][Cols[j]]) % K;
			}
		}		

		// NOW THERE IS 1 IN THE DIAGONAL OR NOTHING 

		if( CongruenceSystem[i][Cols[i]] == 1 ){ 
		// is there a diagonal 1 ?	
		// Elimination
			for(int k=i+1; k<nRows; k++){					
			    if( CongruenceSystem[k][Cols[i]] != 0 ){
	// there is a non-zero, so we substract row i from row k (x-times)
				factor = CongruenceSystem[k][Cols[i]];
				for(int j=i; j<nCols; j++){    
				// for every column geq i
CongruenceSystem[k][Cols[j]] = (K + ((CongruenceSystem[k][Cols[j]] - //
factor*CongruenceSystem[i][Cols[j]])%K))%K; // modulo-K
					}
				}
			}
		}
		else{  // no pivot found, elimination is over
			idSize = i;	
			break; // break out of for-loop
		}
	} // enf of gaussian elimination


	// Backward Elimination

	for(int i=idSize-1; i>=0; i--){
		for(int k=i-1; k>=0; k--){
			if( CongruenceSystem[k][Cols[i]] != 0 ){
				factor = CongruenceSystem[k][Cols[i]];
				for(int j=i; j<nCols; j++){   
				 // for every column geq i
CongruenceSystem[k][Cols[j]] = (K + ((CongruenceSystem[k][Cols[j]] - //
factor*CongruenceSystem[i][Cols[j]])%K))%K; // modulo-K
				}
			}
		}
	}

	// Check if the Congruence System has a solution
	
	for(int i=idSize; i<nRows; i++)
		if(CongruenceSystem[i][nCols-1] != 0){
			// BAD BAD BAD
			// The congruence system has no solution
			return 0;
		}

	// Allocate memory for internal cut pool

	GeneratedCuts = new int*[nMaxCuts];
	generatedCutsGen = true;

	nGen = 0;

        //	
	// GENERATE CUTS
	//

	int checksum = 0;
	bool feasible = false;


	// The basic solution

	GeneratedCuts[nGen] = new int[nCols-1];
	generatedCutsGen = true;
	for(int l=0; l<nCols-1; l++){
		if(l<idSize)
			GeneratedCuts[nGen][Cols[l]] =
				 CongruenceSystem[l][Cols[nCols-1]];
		else
			GeneratedCuts[nGen][Cols[l]] = 0;
	}
	nGen++;

	// Additional solutions

	bool cutGood = true;
	

	int curSum;

	time_t START = time(NULL);
	time_t END;
	for(int i=1; i<pow(2,nCols-idSize); i++){
          END = time(NULL);
          if(nGen == nMaxCuts || difftime(END,START) > TIMEOUT)
            break;
          
          GeneratedCuts[nGen] = new int[nCols-1];
          for(int l=0; l<nCols-1; l++)
            GeneratedCuts[nGen][l] = 0;
                                   
          for(int j=0; j<idSize; j++){
            
            curSum = 0;
            for(int k=idSize; k<nCols-1; k++){
              if( (i & (1 << (k-idSize))) ){
                curSum = curSum + CongruenceSystem[j][Cols[k]];
                GeneratedCuts[nGen][Cols[k]] = 1;
              } 
           }

GeneratedCuts[nGen][Cols[j]] = (K + ((CongruenceSystem[j][Cols[nCols-1]] - curSum)%K )) %K;
          }
          cutGood = true;

/**      
  * Maximum density heuristic
  */
  
    
          if(MAXDENSITY){
            int nNZ = 0;

            double* Cut = new double[nVar];
            
            for(int j=0; j<nVar; j++)
              Cut[j] = 0; // Just in case theres junk in the memory

            double Multiplier = 1.0/((double) K);

            for(int l=0; l<nCon; l++){
              if(GeneratedCuts[nGen][l] != 0){ // Column i takes part
		for(int j=0; j<nNonZ[l]; j++){
		  Cut[NonZeroI[l][j]] = Cut[NonZeroI[l][j]] + //
			Multiplier*GeneratedCuts[nGen][l]*NonZeroC[l][j]; 
		}
              }
            }

            for(int l=0; l<nVar; l++){
              if(Cut[l] != 0)
                nNZ++;
            }
            
            if(nNZ > MAXDENSITY){
              cutGood = false;
            }
	    
 	    // Clean up
	    delete [] Cut;
          }

          if(cutGood)
            nGen++;
        }

	// END OF CUT GENERATION

	time_t end = time(NULL);

	// END OF SEPARATION

	nCutsLeft = nGen;
	return nGen;

}// separate(..)


// -----------------------------------------------------------------------------
// g e t C u t
// -----------------------------------------------------------------------------
int* ModkSeparator::getCut() {

	int nNonZerosCut=0; 
	int k=-1;
	double* Cut = new double[nVar];
	double CutRHS = 0;
	int* sparseCut;

	double entry=0;
	
	if(nCutsLeft){
		nCutsLeft--;
		int cI = nGen-(nCutsLeft+1);
	
		// Construct cut from the congruence system solutions
	
		for(int j=0; j<nVar; j++)
			Cut[j] = 0; // Just in case theres junk in the memory

		double Multiplier = 1.0/((double) K);
		

		for(int i=0; i<nCon; i++){

			if(GeneratedCuts[cI][i] != 0){ // Column i takes part
				for(int j=0; j<nNonZ[i]; j++){
Cut[NonZeroI[i][j]] = Cut[NonZeroI[i][j]] +//
 Multiplier*GeneratedCuts[cI][i]*NonZeroC[i][j]; 
				}
				CutRHS = CutRHS + Multiplier*GeneratedCuts[cI][i]*RHS[i]; 
			}
		}
	
              
                double lhs=0;

		// Convert cut to sparse format
		
		for(int i=0; i<nVar; i++)
			if(Cut[i] != 0)
				nNonZerosCut++;
		sparseCut = new int[(nNonZerosCut*2)+2];
		sparseCut[0] = nNonZerosCut;
		for(int i=0; i<nVar; i++){
			if(Cut[i] != 0){				
				lhs = lhs + ((int) Cut[i])*lpSol[i];
				k++;
				sparseCut[k+1] = i;
				sparseCut[k+1+nNonZerosCut] = (int) Cut[i];
			}
		}
		sparseCut[(2*nNonZerosCut)+1] = (int) floor(CutRHS);
	
	        // Clean up.
		delete [] Cut;

		return sparseCut;
	}
	else {
	    // Clean up.
	    delete [] Cut;	

		return NULL;
	}
}// getCut()



// -----------------------------------------------------------------------------
// --------------- M e t h o d s  ( p r i v a t e ) ----------------------------
// -----------------------------------------------------------------------------


// -----------------------------------------------------------------------------
// h w e i g h t
// -----------------------------------------------------------------------------
unsigned int ModkSeparator::hweight(unsigned int w)
{
    unsigned int res = (w & 0x55555555) + ((w >> 1) & 0x55555555);
    res = (res & 0x33333333) + ((res >> 2) & 0x33333333);
    res = (res & 0x0F0F0F0F) + ((res >> 4) & 0x0F0F0F0F);
    res = (res & 0x00FF00FF) + ((res >> 8) & 0x00FF00FF);
    return (res & 0x0000FFFF) + ((res >> 16) & 0x0000FFFF);
}


