////////////////////////////////////// c++ /////////////////////////////////////
//
//  Module           : ModkSeparator.hh
//  Description      :
//  Author           : Lars Fricke
//  Email            : lars.fricke@iwr.uni-heidelberg.de
//  Copyright        : University of Heidelberg
//  Created on       : Tue Nov 8 2005
//  Last modified by : Rebennack
//  Last modified on : Wed Apr 05 15:59:06 2006
//  Update count     : 1
//
////////////////////////////////////////////////////////////////////////////////
//
//  Date	Name		Changes/Extensions
//  Apr 05      Rebennack       Fitting programming guides
//
////////////////////////////////////////////////////////////////////////////////

#ifndef MOD_K_SEPARATOR_HH
#define MOD_K_SEPARATOR_HH



/**
 * This class realizes the separation of mod-k cuts. 
 */
class ModkSeparator {

public:

    // -------------------------------------------------------------------------
    // ------------- M e t h o d s  ( p u b l i c ) ----------------------------
    // -------------------------------------------------------------------------

    /**
     * Constructor.
     * 
     * Needs the constraints with slack zero.
     *
     * @param modfac Modulo factor
     * @param nConstraints Number of constraints.
     * @param nVariables Number of variables.
     * @param nNonZeros Array of numbers of non-zero entries in rows.
     * @param NonZeroIndices Matrix of indices of the non-zero entries.
     * @param NonZeroCoeff Matirx of the coefficients of the non-zero entries.
     * @param rhs Right-Hand-side.
     * @param lpSolution The current LP-Solution to be separated.
     */ 
    ModkSeparator(int modfac, int nConstraints, int nVariables, int* nNonZeros,
		  int** NonZeroIndices, double** NonZeroCoeff, double* rhs,
		  double* lpSolution);

    /**
     * Destructor.
     */
    ~ModkSeparator();

    /**
     * Contains the separation procedure.
     * 
     * @param nMaxCuts Number of cuts to be generated at maximum.
     * @return Number of cuts generated.
     */
    int separate(int nMaxCuts);
	
    /**
     * Returns a cut if there is one left in the pool, the NULL-pointer otherwise.
     * 
     * @param void
     * @return A cut if there is one left in the pool, the NULL-pointer otherwise.
     */
    int* getCut();


private:

    // -------------------------------------------------------------------------
    // ------------- M e t h o d s  ( p r i v a t e ) --------------------------
    // -------------------------------------------------------------------------
    
    /**
     * Function hweight.
     *
     * @param w An integer.
     * @returns The hamming-weight of w.
     */
    unsigned int hweight(unsigned int w);


    // -------------------------------------------------------------------------
    // ------------- V a r i a b l e s  ( p r i v a t e ) ----------------------
    // -------------------------------------------------------------------------

    /**
     * Contains the generated cuts (after execution of separate).
     */
    int** GeneratedCuts;
   
    /**
     * Contains the congruence system matrix.
     */
    int** CongruenceSystem;

    /**
     * Number of columns in the congruence system.
     */ 
    int nCols;

    /**
     * Number of rows in the congruence system.
     */
    int nRows;

    /**
     * Number of generated cuts.
     */
     int nGen;

    /**
     * Number of cuts left in pool.
     */
    int nCutsLeft;

    /**
     * Index vector for the columns of the congruence system.
     */
    int* Cols;

    /**
     * mod-K.
     */
    int K;

    /**
     * Order of elements.
     */
    int* Ord;
 	

    int** NonZeroI;
    int* nNonZ;
    double** NonZeroC;
    double* RHS;
    int nVar;
    int nCon;
    double* lpSol;

    /**
     * Indicate if the corresponding object has been generated.
     */
    bool ordGen;

    /**
     * Indicate if the corresponding object has been generated.
     */
    bool constraintsGen;

    /**
     * Indicate if the corresponding object has been generated.
     */
    int  congruenceSystemGen;


    /**
     * Indicate if the corresponding object has been generated.
     */ 
    bool colsGen;

    /**
     * Indicate if the corresponding object has been generated.
     */
    bool generatedCutsGen;

};

#endif /*MODKSEPARATOR_HH_*/

