////////////////////////////////////// c++ /////////////////////////////////////
//
//  Module           : OddCycleConstraint.cpp
//  Description      : Odd cycle constraint.
//  Author           : Steffen Rebennack
//  Email            : srebenac@ix.urz.uni-heidelberg.de;
//                     steffen.rebennack@web.de
//  Copyright        : (C) 2006 by the University of Heidelberg
//  Created on       : Wed Apr 05 15:08:19 2006
//  Last modified by : -
//  Last modified on : -
//  Update count     : 0
//
////////////////////////////////////////////////////////////////////////////////
//
//  Date	Name		Changes/Extensions
//  ----	----		------------------
//
////////////////////////////////////////////////////////////////////////////////

#include "OddCycleConstraint.hh"
#include "abacus/row.h"
#include "Graph.hh"
#include "Node.hh"
#include "StableSetMaster.hh"



// -----------------------------------------------------------------------------
// --------------- M e t h o d s  ( p u b l i c ) ------------------------------
// -----------------------------------------------------------------------------


// -----------------------------------------------------------------------------
// C o n s t r u c t o r
//
// This class is derived from the abstract base class ABA_CONSTRAINT.
//
// Because aa odd cycle inequality is uniquely determined by their nodes, this
// constructor saves all nodes of the odd cycle. The main part is
// the call of the constructor of class ABA_CONSATRINT with the following 
// arguments:
// 1.) A pointer to the corresponding master of the optimization.
// 2.) A pointer to the subproblem associated with this constraint is not
//     necessary because this constraint is globally valid.
// 3.) This constraint is of kind less (or equal).
// 4.) The ride hand side of this constraint is "size of the odd cycle -1 / 2".
// 5.) This constraint can be removed from the active constraint set.
// 6.) This constraint is global valid and not only local.
// 7.) There is no lifting procedure implemented.
// -----------------------------------------------------------------------------
OddCycleConstraint::OddCycleConstraint(ABA_MASTER *master, int oddCycleSize,
                                       int *nodes, Graph *theGraph):
    ABA_CONSTRAINT(master, NULL, ABA_CSENSE::Less, (oddCycleSize-1)*0.5, true,
                   false, false),
    oddCycle(oddCycleSize),
    itsGraph(theGraph)
{

    for (int i=0; i<oddCycleSize; i++) {
        oddCycle[i] = nodes[i];
    }
    sort(oddCycle.begin(), oddCycle.end());

    // Store this inequality that it can be checked.
    ofstream fout(theGraph->getFileName(), ios::app);
    for (int i=0; i<oddCycleSize; i++) {
        fout << "  1 x_" << itsGraph->translateNode(oddCycle[i]);
    }
    fout << "  <  " << (oddCycleSize-1)*0.5 << endl;
    fout.close();

    // Calculate the key.
    calculateHashKey();
}


// -----------------------------------------------------------------------------
// D e s t r u c t o r
// -----------------------------------------------------------------------------
OddCycleConstraint::~OddCycleConstraint() {
}


// -----------------------------------------------------------------------------
// g e n R o w
//
// This function redefines the pure virtual function of the base class
// ABA_CONSTRAINT. It returns the constraint in sprase format. 
// -----------------------------------------------------------------------------
int OddCycleConstraint::genRow(ABA_ACTIVE<ABA_VARIABLE,ABA_CONSTRAINT> *, 
			     ABA_ROW &row)
{
    // Store the indices of the nodes of the odd cycle.
    for (int i=0; i<oddCycle.size(); i++) {
        row.insert(oddCycle[i],1.0);
    }
    // The rhs of the odd cycle..
    row.rhs((oddCycle.size()-1)*0.5);
    // Sense of the inequality.
    row.sense(ABA_CSENSE::Less);

    return row.nnz();
}


// -----------------------------------------------------------------------------
// c o e f f
//
// This function computes the coefficient of the variable var. It redefines the
// pure virtual function of the base class ABA_CONSTRAINT. The coefficient of
// a variable of an odd cycle constraint is 1 if the variable is part of the
// odd cycle and otherwise its coefficient is 0.
// -----------------------------------------------------------------------------
double OddCycleConstraint::coeff(ABA_VARIABLE *var) {

    // cast ABA_VARIABLE* to NODE*
    Node *node = (Node *) var;
    // search for this variable in the clique
    for (int i=0; i<oddCycle.size(); i++) {
        if (node->nodeNumber() == oddCycle[i]) {
            return 1.0;
        }
    }
    return 0.0;
}


// -----------------------------------------------------------------------------
// g e t S i z e
// -----------------------------------------------------------------------------
int OddCycleConstraint::getSize() {
    return oddCycle.size();
}


// -----------------------------------------------------------------------------
// i s I n C o n s t r a i n t
//
// This function checks if a node is member of this constraint. It return the
// value true if the node is a memebr and otherwise false.
// This method is used by the function equal(..).
// -----------------------------------------------------------------------------
bool OddCycleConstraint::isInConstraint(int node) {

    for(int i=0; i<oddCycle.size(); i++) {
        if (oddCycle[i] == node) {
            return true;
	}
    }

    return false;
}


// -----------------------------------------------------------------------------
// e q u a l
//
// This function compares if this constraint is identical with the constraint
// cv in a mathematical sense. This function is required becuase the constraint
// is stored in a pool of the class ABA_NONDUPLPOOL.
// It returns true if the constraints are equal and otherwise false.
// -----------------------------------------------------------------------------
bool OddCycleConstraint::equal(ABA_CONVAR *cv) {

    bool isEqual = true;
    if (thisHashKey != ((OddCycleConstraint *)cv)->hashKey()) {
        isEqual = false;
    }
    if (oddCycle.size() != ((OddCycleConstraint *)cv)->getSize()) {
        isEqual = false;
    }

    for(int i=0; i<oddCycle.size(); i++) {
        if (!((OddCycleConstraint *)cv)->isInConstraint(oddCycle[i])) {
	    isEqual = false;
	}
    }

    return isEqual;
}


// -----------------------------------------------------------------------------
// n a m e
//
// This function returns the name of the constraint. It is required if the
// constraint is stored in a pool of class ABA_NONDUPLPOOL.
// -----------------------------------------------------------------------------
const char* OddCycleConstraint::name() {
    return "OddCycleConstraint";
}


// -----------------------------------------------------------------------------
// h a s h K e y
//
// This method provides a key for the constraint that can be used to insert it
// into a hash table. In this case it is not required that any two items have
// different keys. This function is required to store the constraint in a pool
// of the class ABA_NIONDUPLPOOL.
// -----------------------------------------------------------------------------
unsigned OddCycleConstraint::hashKey() {
    return thisHashKey;
}



// -----------------------------------------------------------------------------
// --------------- M e t h o d s  ( p r i v a t e ) ----------------------------
// -----------------------------------------------------------------------------


// -----------------------------------------------------------------------------
// c a l c u l a t e H a s h K e y
// -----------------------------------------------------------------------------
void OddCycleConstraint::calculateHashKey() {
    thisHashKey = 1;
    for (int i=0; i<oddCycle.size(); i++) {
        thisHashKey *= (oddCycle[i] + 1);// + 5041);
	thisHashKey = thisHashKey % 9973;
    }
}


