////////////////////////////////////// c++ /////////////////////////////////////
//
//  Module           : OddCyclesSeparator.cpp
//  Description      : Separate the odd cycle inequalities.
//  Author           : Steffen Rebennack
//  Email            : srebenac@ix.urz.uni-heidelberg.de;
//                     steffen.rebennack@web.de
//  Copyright        : (C) 2006 by the University of Heidelberg
//  Created on       : Wed Apr 05 15:19:53 2006
//  Last modified by : -
//  Last modified on : -
//  Update count     : 0
//
////////////////////////////////////////////////////////////////////////////////
//
//  Date	Name		Changes/Extensions
//  ----	----		------------------
//
////////////////////////////////////////////////////////////////////////////////

#include "OddCyclesSeparator.hh"
#include "Graph.hh"
#include "OddCycleConstraint.hh"
#include "ShortestPathCalculator.hh"
#include "StableSetLPSolution.hh"
#include "StableSetMaster.hh"



// -----------------------------------------------------------------------------
// --------------- M e t h o d s  ( p u b l i c ) ------------------------------
// -----------------------------------------------------------------------------


// -----------------------------------------------------------------------------
// C o n s t r u c t o r
// -----------------------------------------------------------------------------
OddCyclesSeparator::OddCyclesSeparator(ABA_MASTER *master,
				       StableSetLPSolution *solution,
				       Graph *theGraph, bool flag):
    ABA_SEPARATOR<ABA_CONSTRAINT, ABA_VARIABLE> (solution, true, 300),
    master_(master),
    itsGraph(theGraph),
    lpSolution(solution),
    lifting(flag)
{
}


// -----------------------------------------------------------------------------
// D e s t r u c t o r
// -----------------------------------------------------------------------------
OddCyclesSeparator::~OddCyclesSeparator ()
{
    for(int i=cycles.size()-1; i>=0; i--) {
        delete [] cycles[i];
    }
}


// -----------------------------------------------------------------------------
// s e p a r a t e
//
// This function separates odd cycles. Therefore it constucts an auxiliary
// graph. All the nodes which have value 1 are delted before the graph is
// constructed because an violated odd cycle inequality will not include this
// node. After the construction of this graph a shortest path calculator is
// called to produce shortest paths in the new graph with special cost. This
// path can be constructed to a violated odd cycle constraint if the length of
// the path is smaller than 0.5. After a violated odd cycle constraint could be
// found, this function checks the size of the cycle. A 3-cycle will be stored
// and in the sub problem it will be lifted to a max clique. Otherwise the
// corresponding constraint is generated and added to the buffer cutFound().
// -----------------------------------------------------------------------------
void OddCyclesSeparator::separate() {


    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // Construct the auxiliary graph.
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // There are two different graphs. The small graph is the graph of the
    int    i, j, k, l, m;                         // Loop variables.
    int    numberOfNodes = lpSolution->nVarCon(); // Number of nodes of the
    						  // complete graph.
    int    numberOfNonBinaryNodes = 0;            // Number of nodes which have
    						  // not a binary value.
    vector<int> index(numberOfNodes);		  // Store the index of each
                                                  // node of the new list of
						  // nodes of the auxiliary
						  // graph.
    vector<int> indexOfNonBinaryNode(numberOfNodes);   // This is the translator
    						  // which gives the real index
						  // of the node of the small
						  // graph.
    int    node;                                  // A vertex.
    int    element;                               // Another vertex.
    int    size;                                  // Size of adjazens list:
    						  // complete graph.
    int    sizeOfList;                            // size of adjazens list:
    						  // small graph.
    int    adjazensListSmall[numberOfNodes * (numberOfNodes - 1)];
    int    *adjazensPointer;                      // Pointer to the nodes of the
    						  // adjazens list of the samll
						  // graph.
    int    *adjazensList;                         // Adjazens list of the small
    						  // graph.
    long   numberOfEdges = 0;                     // Number of edges in the
    						  // small graph.
    double *weight = new double[numberOfNodes * (numberOfNodes - 1)];
    double *solutionValue = lpSolution->zVal();   // Solution of LP.
    double *edgeWeight;                           // Weight of the edges of the
    						  // auxiliary graph.
    double eps = master_->machineEps();		  // Precision.


    //
    // First step: get small graph.
    //
    // Check for each variable: if it has not value 0 and 1, save the index and
    // increase counter of the number of non binary nodes.
    for (i=0; i<numberOfNodes; i++) {
        if ((solutionValue[i] > eps)
	    && (solutionValue[i] < (1 - eps))
	   ) {
            index[i] = numberOfNonBinaryNodes;
            indexOfNonBinaryNode[numberOfNonBinaryNodes] = i;
	    numberOfNonBinaryNodes++;
        }
	else {
            index[i] = -1;
	}
    }

    // Add all adjazent nodes but ignore the variables with value 0 or 1.
    adjazensPointer = new int[2*numberOfNonBinaryNodes+1];
    adjazensPointer[0] = 0; // All isolated notes will be deleted.

    for (i=0; i<numberOfNonBinaryNodes; i++) {
        node       = indexOfNonBinaryNode[i];
        size       = itsGraph->adjacentListSize(node);
	sizeOfList = 0;
	for (j=0; j<size; j++) {
	    element = itsGraph->adjacentListElement(node, j);
            if (index[element] != -1) {
                adjazensListSmall[numberOfEdges] = index[element];
		weight[numberOfEdges++] = 0.5 * (1 - solutionValue[node] -
		          solutionValue[indexOfNonBinaryNode[index[element]]]);
		sizeOfList++;
            }
        }
	if (sizeOfList == 0) {
            // Selete isolated nodes.
	    index.erase(index.begin() + indexOfNonBinaryNode[i]);
            indexOfNonBinaryNode.erase(indexOfNonBinaryNode.begin() + i);
	    numberOfNonBinaryNodes--;
	    i--;
	    continue;
	}
        if (i < (numberOfNonBinaryNodes - 1)) {
	    adjazensPointer[i+1] = adjazensPointer[i] + sizeOfList;
	}
    }

    //
    // Second step: construct auxiliary graph.
    //
    adjazensList = new int[2*numberOfEdges];
    edgeWeight   = new double[2*numberOfEdges];

    // Update adjazensPointer.
    for (i=0; i < numberOfNonBinaryNodes; i++) {
        adjazensPointer[i + numberOfNonBinaryNodes] =
	                                     adjazensPointer[i] + numberOfEdges;
    }
    adjazensPointer[2*numberOfNonBinaryNodes] = numberOfEdges*2;

    // Update adjazensList & calculate new edge weights.
    for (i=0; i < numberOfEdges; i++ ) {
        adjazensList[i] = numberOfNonBinaryNodes + adjazensListSmall[i];
	adjazensList[i + numberOfEdges] = adjazensListSmall[i];

        edgeWeight[i] = weight[i] + eps * 100;
	edgeWeight[i + numberOfEdges] = edgeWeight[i];
    }

    // Clean up.
    index.clear();


    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // Construct odd cycles.
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    int    predNode[2*numberOfNonBinaryNodes]; // Solution of the shortest path
    					       // algorithm.
    int    currentNode;                        // A vertex...
    int    smallGraphNode;                     // ...and the corresponding node
    					       // to the small graph
    int    completeGraphNode;                  // ...and the corresponding node
    					       // to the complete graph.
    int    oddCycle[numberOfNonBinaryNodes];
    // Start node of a shortest path.
    int    start = ((StableSetMaster *)master_)->getRandomNumber(
                                                        numberOfNonBinaryNodes);
    double shortestPathValue;                  // Value of a shortest path.
    // Minimum violation of the odd cycle inequality to be added.
    double tolerance = ((StableSetMaster *)master_)->m_minAbsViolationCycle;
    bool   inOddCycle[numberOfNonBinaryNodes]; // Markes each node if it is
    					       // allready in a violated odd
					       // cycle.
    int    marked[numberOfNonBinaryNodes];
    int    marker;
    bool   found;
    double weightOfCycle;
    OddCycleConstraint *oddCycleConstr;        // Pointer to an odd cycle
    					       // constraint.

    // Create ShortestPathCalculator.
    ShortestPathCalculator* shortestPath =
            new ShortestPathCalculator(2*numberOfNonBinaryNodes,
	                               2*numberOfEdges,
				       adjazensPointer,
				       adjazensList,
				       edgeWeight,
				       0.5,
				       eps);

    // Calculate for each node of the small graph a shortest path to its
    // corresponding copy. But the shortest path is not calculted if it is
    // a member of a violated odd cycle inequality which was allready found
    // in this separation routine. This is a heuristical approach.
    for (i=0; i<numberOfNonBinaryNodes; i++) {
        inOddCycle[i] = false;
        marked[i]     = -1;
    }

    j = start;
    for (i=j; i<j+numberOfNonBinaryNodes; i++) {
	start = i % numberOfNonBinaryNodes;
	// Calculate a shortest path only if the node is not member of
	// violated odd cycle which have allready be found.
	if (inOddCycle[start] == true) {
            continue;
	}

	shortestPathValue = shortestPath->calculateShortestPath(
	        start, start + numberOfNonBinaryNodes, predNode);

        // No violated odd cycle inequality found?
	if (shortestPathValue > 0.5 - eps) {
            continue;
	}

	// Construct odd cycle.
	currentNode   = start + numberOfNonBinaryNodes;
	size          = 0;
	weightOfCycle = 0;
        while (currentNode != start) {
	    node                  = predNode[currentNode];
            currentNode           = node;
	    smallGraphNode        = currentNode % numberOfNonBinaryNodes;
            // Check if there is an odd cycle part of this cycle.
	    if (marked[smallGraphNode] != -1) {
	        weightOfCycle = 0;
	        marker        = marked[smallGraphNode];
                // Odd cycle found.
                for (k=0; k<marker; k++) {
                    marked[oddCycle[k]] = -1;
		}
		for ( ; k<size; k++) {
		    node                 = oddCycle[k];
                    oddCycle[k - marker] = node;
		    completeGraphNode    = indexOfNonBinaryNode[node];
		    weightOfCycle       += solutionValue[completeGraphNode];
		}
		// Update size.
		size -= marker;
		break;  // Terminate this while loop.
	    }
	    else {
	        oddCycle[size]         = smallGraphNode;
		marked[smallGraphNode] = size;
                completeGraphNode      = indexOfNonBinaryNode[smallGraphNode];
		weightOfCycle         += solutionValue[completeGraphNode];
		size++;
	    }
	}

	if (weightOfCycle < 0.5*(size-1) + tolerance) {
	    for(k=0; k<size-1; k++) {
		marked[oddCycle[k]] = -1;
	    }
            continue;
	}

	found = true;
	while (found) {
            // Check if this cycle has a chord.
	    for (k=0; k<size-1; k++) {
                node = oddCycle[k];

	        // Is this node adjacent to the others?
                found = false;
	        l = k+1;
	        while (l<size && !found && (l+1)%size!=k) {
	            // Only even differences are possible.
	            if (!(1 & (l-k))) {
                        element = oddCycle[l];
                        m       = adjazensPointer[node];

		        while (m<adjazensPointer[node+1] && !found) {
                            if (adjazensListSmall[m] == element) {
		                // Chord found.
                                found = true;
                                weightOfCycle = 0;
			        // Construct odd cycle.
			        for (m=0; m<k; m++) {
                                    marked[oddCycle[m]] = -1;
		                }
                                for ( ; m<=l; m++) {
			            node              = oddCycle[m];
				    oddCycle[m-k]     = node;
				    completeGraphNode =
				            indexOfNonBinaryNode[node];
				    weightOfCycle     +=
				            solutionValue[completeGraphNode];
			        }
			        for ( ; m<size; m++) {
                                    marked[oddCycle[m]] = -1;
			        }
			        size = l - k + 1;
			        break;
	                       
			    }// if
		            m++;
		        }// while
                        if (found) {
			    break;
                        }
                        if (weightOfCycle < 0.5*(size-1) + tolerance) {
                            break;
	                }
		    }//if
	            l++;
	        }// while
                if (weightOfCycle < 0.5*(size-1) + tolerance) {
                    break;
	        }
                if (found) {
		    break;
                }
	    }// for
            if (weightOfCycle < 0.5*(size-1) + tolerance) {
                break;
	    }
	}// while


	// Store this odd cycle to increase it to a maximal clique in the SUB
	// problem if it has only 3 nodes.
	if (size == 3) {
	    for(k=0; k<3; k++) {
                // Reset marker.
	        marked[oddCycle[k]]     = -1;
	        // Update flags.
	        inOddCycle[oddCycle[k]] = true;
		// Get node number of complete graph.
                completeGraphNode       = indexOfNonBinaryNode[oddCycle[k]];
		// Save this clique and increase it in StableSetSub.
	        clique.push_back(completeGraphNode);
	    }
	}
	else {
	    // If lifting is true, than this odd cycle will be lifted
	    // to some inequality for the complete graph. Therfore
	    // the odd cycle inequality may not be valid for the
	    // reduced graph and we cannot add a odd cycle constraint.
	    if(lifting) {
	        // Store the cycle.
                int *aCycle = new int[size+1];
		// Store the size of the odd cycle.
		aCycle[0] = size+1;
		// Convert each element of the small graph to the element
		// of the hole graph and store it.
		for(k=0; k<size; k++) {
                    // reset marker
	            marked[oddCycle[k]]     = -1;
	            // Update flags.
                    inOddCycle[oddCycle[k]] = true;
		    completeGraphNode       = indexOfNonBinaryNode[oddCycle[k]];
                    aCycle[k+1]             = completeGraphNode;

		}
		cycles.push_back(aCycle);
	    }
	    else {
                // Create new odd cycle constraint and store it with the
		// function cutFound(..). In the coresponding sub problem the
		// constraints will be added to the pool with the call of the
		// function addCons(..).
		for(k=0; k<size; k++) {
		    // Reset marker.
	            marked[oddCycle[k]]     = -1;
                    inOddCycle[oddCycle[k]] = true;
		    completeGraphNode       = indexOfNonBinaryNode[oddCycle[k]];
		    oddCycle[k]             = completeGraphNode;
		}
                oddCycleConstr = new OddCycleConstraint(master_, size, oddCycle,//
							itsGraph);
	        cutFound(oddCycleConstr);

	    }
	}
    }// end: for (i=j; i<j+numberOfNonBinaryNodes; i++) {

    // Clean up.
    // Destroy the object of class ShortestPathCalculator.
    delete shortestPath;
    delete [] adjazensPointer;
    delete [] adjazensList;
    delete [] edgeWeight;
    delete [] weight;

} // end: void separate()


// -----------------------------------------------------------------------------
// g e t S i z e O f C l i q u e
// -----------------------------------------------------------------------------
int OddCyclesSeparator::getSizeOfClique() const {
    return clique.size();
}


// -----------------------------------------------------------------------------
// g e t C l i q u e E l e m e n t
// -----------------------------------------------------------------------------
int OddCyclesSeparator::getCliqueElement(int i) const {
    return clique[i];
}


// -----------------------------------------------------------------------------
// g e t N u m b e r O f O d d C y c l e s
// -----------------------------------------------------------------------------
int OddCyclesSeparator::getNumberOfOddCycles() const {
    return cycles.size();
}


// -----------------------------------------------------------------------------
// g e t O d d C y c l e
// -----------------------------------------------------------------------------
int *OddCyclesSeparator::getOddCycle(int i) {
    return cycles[i];
}



