////////////////////////////////////// c++ /////////////////////////////////////
//
//  Module           : oddCyclesSeparator.hh
//  Description      : Separate the odd cycle inequalities.
//  Author           : Steffen Rebennack
//  Email            : srebenac@ix.urz.uni-heidelberg.de;
//                     steffen.rebennack@web.de
//  Copyright        : (C) 2006 by the University of Heidelberg
//  Created on       : Wed Apr 05 15:13:14 2006
//  Last modified by : -
//  Last modified on : -
//  Update count     : 0
//
////////////////////////////////////////////////////////////////////////////////
//
//  Date	Name		Changes/Extensions
//  ----	----		------------------
//
////////////////////////////////////////////////////////////////////////////////

#ifndef ODD_CYCLES_SEPARATOR_HH
#define ODD_CYCLES_SEPARATOR_HH

#include <abacus/separator.h>
#include <vector>

using namespace std;

class Graph;
class StableSetLPSolution;
class StableSetMaster;



/**
 * This class provides an exact separation procedure for odd cycle constraints.
 */
class OddCyclesSeparator: public ABA_SEPARATOR <ABA_CONSTRAINT, ABA_VARIABLE> {

public:

    // -------------------------------------------------------------------------
    // ------------- M e t h o d s  ( p u b l i c ) ----------------------------
    // -------------------------------------------------------------------------

    /**
     * Constructor.
     *
     * @param *master Pointer to the master of the problem.
     * @param *solution A pointer to the current LP-Solution to be separated.
     * @param *theGraph Pointer to the graph storing all information about the
     *			problem instance.
     * @param flag If its value is true the odd cycles will be lifted.
     */
    OddCyclesSeparator(ABA_MASTER *master, StableSetLPSolution *solution,
		       Graph *theGraph, bool flag);

    /**
     * Destructor.
     */
    ~OddCyclesSeparator();

    /**
     * This method provides the separation procedure itself and overrides the
     * pure virtual function of ABA_SEPARATOR.
     *
     * @param void
     * @return void
     */
    virtual void separate();

    /**
     * Get the size of the vector clique.
     *
     * @param void
     * @return Size of vector clique.
     */
    int getSizeOfClique() const;

    /**
     * Get the i-th element of the vector clique.
     *
     * @param void
     * @return i-th element of vector clique.
     */
    int getCliqueElement(int i) const;

    /**
     * Get the number of odd cycle inequalities found by this separation.
     *
     * @param void
     * @return Size of vector cycles.
     */
    int getNumberOfOddCycles() const;

    /**
     * Get the i-th odd cycle of vector cycle.
     *
     * @param void
     * @return i-th element of vector cycle.
     */
    int *getOddCycle(int i);


private:

    // -------------------------------------------------------------------------
    // ------------- V a r i a b l e s  ( p r i v a t e ) ----------------------
    // -------------------------------------------------------------------------

    /**
     * Pointer to the master of the problem.
     */
    ABA_MASTER *master_;

    /**
     * Pointer to the graph of the problem.
     */
    Graph *itsGraph;

    /**
     * Pointer to the current LP solution.
     */
    StableSetLPSolution *lpSolution;

    /**
     * This is a flag. If its value is true the odd cycles will be lifted and
     * therefore the constraints will be stored in a vector and not in a pool.
     */
    bool lifting;

    /**
     * Store all 3-cycles which have been found as a clique to get a maximal
     * clique.
     */
    vector<int> clique;

    /**
     * In the case of lifting = true save all found odd cycles with length > 3
     * in this vector. The first element of the integer array is the size of 
     * the odd cylce.
     */
    vector<int*> cycles;

};

#endif


