////////////////////////////////////// c++ /////////////////////////////////////
//
//  Module           : Preprocessing.cpp
//  Description      : Preprocessing for the maximum stable set problem.
//  Author           : Steffen Rebennack
//  Email            : srebenac@ix.urz.uni-heidelberg.de;
//                     steffen.rebennack@web.de
//  Copyright        : (C) 2006 by the University of Heidelberg
//  Created on       : 
//  Last modified by : -
//  Last modified on : -
//  Update count     : 0
//
////////////////////////////////////////////////////////////////////////////////
//
//  Date	Name		Changes/Extensions
//  ----	----		------------------
//
////////////////////////////////////////////////////////////////////////////////

#include "Preprocessing.hh"
#include "abacus/row.h"
#include "Graph.hh"
#include "StableSetMaster.hh"
#include "StableSetStatistics.hh"


#include "abacus/array.h"
#include "abacus/cplexif.h"


// -----------------------------------------------------------------------------
// --------------- M e t h o d s  ( p u b l i c ) ------------------------------
// -----------------------------------------------------------------------------

// -----------------------------------------------------------------------------
// C o n s t r u c t o r
//
// -----------------------------------------------------------------------------
Preprocessing::Preprocessing(StableSetMaster *theMaster, Graph *theGraph,
			     StableSetStatistics *theStatistics):
    itsMaster(theMaster),
    itsGraph(theGraph),
    itsStatistics(theStatistics)
{
}


// -----------------------------------------------------------------------------
// D e s t r u c t o r
// -----------------------------------------------------------------------------
Preprocessing::~Preprocessing() {
}


// -----------------------------------------------------------------------------
// p r e p r o c e s s i n g 
//
// Fox nodes according cliques or integer LP values.
// -----------------------------------------------------------------------------
void Preprocessing::preprocessing() {

    bool   fixed = true; // Indicate if a fixing could be done. 
    double *sol;         // Pointer to LP-solution.

    //
    // Fix nodes until no fixing could be done.
    //
    while (fixed) {

        //
	// Call clique fixing.
        //
	fixed = fixVariablesClique();

        //
        // Fixing according to LP.
        //
	if (itsGraph->numberOfNodes()) {
	    sol = solveLP();
	    fixed = fixVariablesEdgeLP(sol) || fixed;
            if (!itsGraph->numberOfNodes()) {
                fixed = false;            
            }

	    delete [] sol;
	}
    }// while

    if (itsGraph->numberOfNodes() == 0) {
        // Solution was found during preprocessing.
	itsStatistics->solutionFound(3);
    }

    itsMaster->sortFixedNodes();

}// end: void preprocessing()



// -----------------------------------------------------------------------------
// --------------- M e t h o d s  ( p r i v a t e ) ----------------------------
// -----------------------------------------------------------------------------


// -----------------------------------------------------------------------------
// f i x V a r i a b l e s C l i q u e
//
// Fix all variables which neighborhodd buils a clique.
// -----------------------------------------------------------------------------
bool Preprocessing::fixVariablesClique() {

    int  i, j, k;          // Loop counter.
    bool fixed = false;    // True if a fixing could be done.
    bool isClique;         // Indicate if this clique can be fixed or not.
    int numberOfNodes = itsGraph->numberOfNodes(); 
    int node;              // A node
    int size;              // A size.
    vector<int> nodesToCheck(numberOfNodes); // Vector storing all nodes which
					     // have to be checked.
    vector<int> fixNodes;  // 

    // At the beginning check each node.
    for (i=0; i<itsGraph->numberOfNodes(); i++) {
	nodesToCheck[i] = i;
    }

    //
    // Try to fix a node until all nodes have been checked 
    // and no more fixing could be done.
    //
    while (nodesToCheck.size() != 0) {
        
        // Get node to check next. 
 	node = nodesToCheck.back();
	nodesToCheck.pop_back();

	// Size of the adjacent list of this node.
	size = itsGraph->adjacentListSize(node);

	isClique = true;

	// Check whether all the neighbours of this node build a clique or not.
	for (i=0; i<size; i++) {

            // The weight of that node has to be greater than the weight of each
            // member of the clique. 
  	    if (itsGraph->weighted() 
                && (itsGraph->weight(node) //
                    < itsGraph->weight(itsGraph->adjacentListElement(node,i)))){
		isClique = false;
		break;
	    }

	    //
            // Check if ALL its neighbors build a clique.
            //  
	    for (j=i+1; j<size; j++) {
	        k = 0;
		if (!itsGraph->isMemberOfAdjacentList( //
			itsGraph->adjacentListElement(node,i),//
                        k, itsGraph->adjacentListElement(node,j))
                   ) {
		    i = size;
		    isClique = false;
		    break;
		}
	    }// for
	}// for


        //
	// If all the neighbours of node 'node' build a clique we can fix it
	// in the following cases:
	// - we have the cardinality stable set problem
	// - if the problem is weighted and node 'node' has the greates
        // weight of all the members of the clique.
        //
	if (isClique) {

	    fixed = true;
	    fixNodes.resize(size + 1);
	    fixNodes[0] = node;

	    for (i=1; i<size + 1; i++) {
		fixNodes[i] = itsGraph->adjacentListElement(node, i-1);
		// Delete this node from the list of nodes to be checked.
		k = 0;
		while ((k < nodesToCheck.size()) && (nodesToCheck[k] != fixNodes[i])
		       ) {
                    k++;
	        }
	        if (nodesToCheck[k] == fixNodes[i]) {
	            nodesToCheck.erase(nodesToCheck.begin() + k);
	        }

	    }
	    sort(fixNodes.begin(), fixNodes.end());

	    // Fix this node to value one.
	    itsMaster->pushFixedNode(itsGraph->translateNode(node));

	    // Delete all the nodes of that clique from the graph
            // and update the nodes to be checked.
	    itsGraph->deleteNodesFromGraph(&fixNodes, &nodesToCheck);
	}
    }

    // Return status.
    return (fixed && (itsGraph->numberOfNodes() != 0));

}// fixVariablesClique()


// -----------------------------------------------------------------------------
// s o l v e L P
//
// Create and solve the initial LP. This linear program constains only the
// edge inequalities and the non-negativity constraints. 
// Cplex is used to determine the optimum which is return as a pointer. 
// -----------------------------------------------------------------------------
double *Preprocessing::solveLP() {

    //
    // Create first LP
    //

    // The number of constraints of the LP is the number of edges.
    int i,j;
    int size;
    int numberOfConstraints = itsGraph->numberOfEdges();
    int numberOfVariables   = itsGraph->numberOfNodes();

    ABA_ROW *constraint;
    // Buffer the non zero elements of a constraint.
    ABA_BUFFER<int> nonZeroElementsBuf(itsMaster, 2);
    // Store the vertices of a constraint...
    ABA_ARRAY<int> *nonZeroElements;
    // ...and their coefficients.
    ABA_ARRAY<double> *coefficients;
    // The coefficients of the target function. They are initialized
    // with value 1.
    ABA_ARRAY<double> objectiveFunctionCoefficient(itsMaster,//
					           numberOfVariables, 1);
    // Lower bound of each variable is 0.
    ABA_ARRAY<double> lowerBound(itsMaster, numberOfVariables, 0);
    // The upper bound for the variable is 1.
    ABA_ARRAY<double> upperBound(itsMaster, numberOfVariables, 1);
    // Store all constraints.
    // There are number of edges constraints.
    ABA_ARRAY<ABA_ROW*> allConstraints(itsMaster, numberOfConstraints);


    //
    // Compute LP
    //

    int cnt = 0;
    int element;
    // Construct all constraints.
    for (i=0; i<numberOfVariables; i++) {
	size = itsGraph->adjacentListSize(i);
	for (j=0; j<size; j++) {
	    element = itsGraph->adjacentListElement(i,j);
	    if (element > i) {
		nonZeroElementsBuf.push(i);
	        nonZeroElementsBuf.push(element);
		nonZeroElements = new ABA_ARRAY<int>(itsMaster,//
						     nonZeroElementsBuf);
	        coefficients    = new ABA_ARRAY<double>(itsMaster, 2, 1);
	        nonZeroElementsBuf.clear();

		// The new constraint.
	        constraint = new ABA_ROW(itsMaster,
					 2,
				         (*nonZeroElements),
				         (*coefficients),
				         ABA_CSENSE::Less,
				         1);
	        // Store this constraint.
	        allConstraints[cnt] = constraint;
		cnt++;
	
		delete nonZeroElements;
		delete coefficients;		

	    }// if
	}// for
    }// for


    //
    // Optimize LP 
    //

    ABA_CPLEXIF lp(itsMaster);
    lp.initialize(ABA_OPTSENSE::Max,
		  numberOfConstraints,
		  numberOfConstraints,
		  numberOfVariables,
		  numberOfVariables,
		  objectiveFunctionCoefficient,
		  lowerBound,
		  upperBound,
		  allConstraints);

    // Optimize and check if optimal solution could be computed.
    if (!(ABA_LP::Optimal == lp.optimize(ABA_LP::Primal))) {
        itsMaster->err() << "ERROR in Preprocessing::solveLP()." << endl;
        itsMaster->err() << "LP could not be dolved to optimality." << endl;
        exit(itsMaster->Fatal);
    }

    double *sol = new double[numberOfVariables];
    for (i=0; i<numberOfVariables; i++) {
	sol[i] = lp.xVal(i);
    }

    // Clean up.
    for (i=numberOfConstraints-1; i>=0; i--) {
	delete allConstraints[i];
	allConstraints[i] = NULL;
    }

    return sol;

}// solveLP



// -----------------------------------------------------------------------------
// f i x V a r i a b l e s F i r s t L P
//
// The LP Relaxation of the stable set polytop with only the trivial- and
// the edge inequalities has a special property. If a variable of the optimal
// solution has value 1 it can be fixed. The function fixVariablesFirstLP()
// uses this attribute and fixes all variables with value 1 and all their
// neighbours to value 0.
// This is independent from the weighted case.
// -----------------------------------------------------------------------------
bool Preprocessing::fixVariablesEdgeLP(double *sol) {

    int  i, j, k;  	  // Loop variables.
    int  numberOfNodes = itsGraph->numberOfNodes(); // Problem size.
    bool fixed = false;   // Indicate if a node could be fixed or not.
    int  node;
    bool isElement;


    vector<int> fixOne;   // Fix the values of the nodes in this array to one.
    vector<int> fixZero;  // Fix the values of the nodes in this array to zero.

    //
    // Get all variables wich have value 1 and save them in the vector fixOne.
    // If a variables with value 1 is found, save all its neighbours in
    // vector fixZero.
    //
    for (i=0; i<numberOfNodes; i++) {
        if ((sol[i] < (1 + itsMaster->machineEps()))
	     && (sol[i] > (1 - itsMaster->machineEps()))
	   ) {
	    fixOne.push_back(i);
	    // Check the neighbourhood.
	    for (j=0; j<itsGraph->adjacentListSize(i); j++) {
	    	node = itsGraph->adjacentListElement(i, j);
		// Check if this node is already an element of the vecor.
		isElement = false;
		for (k=0; k<fixZero.size(); k++) {
		    if (fixZero[k] == node) {
		        isElement = true;
			break;
		    }
		}
		if (!isElement) {
                    fixZero.push_back(node);
		}
	    }
	}
    }

    //
    // Fix all variables of the two vectors fixOne and fixZero.
    //
    if (!fixOne.empty()) {

	// Store all nodes which should be fixed.
	vector<int> fixNodes(fixOne.size() + fixZero.size());
	int size;		// Size of vector fixOne.

	size = fixOne.size();
        for (i=0; i<size; i++) {
	    fixNodes[i] = fixOne[i];
	    // Store all nodes which are fixed to value 1.
	    itsMaster->pushFixedNode(itsGraph->translateNode(fixOne[i]));
        }

        for (i=0; i<fixZero.size(); i++) {
	    fixNodes[size + i] = fixZero[i];
        }
	sort(fixNodes.begin(), fixNodes.end());

	// Delete all fixed nodes from the graph.
	itsGraph->deleteNodesFromGraph(&fixNodes);

	// Update statistics.
	itsGraph->fixVariablesFirstLP(fixNodes.size());
        fixed = true;
    }
    
    // Return status.
    return fixed;

} // end: bool fixVariablesFirstLP(..)










