////////////////////////////////////// c++ /////////////////////////////////////
//
//  Module           : ProjectionNode.hh
//  Description      :  Node of the edge projection tree.
//  Author           : Steffen Rebennack
//  Email            : srebenac@ix.urz.uni-heidelberg.de;
//                     steffen.rebennack@web.de
//  Copyright        : (C) 2006 by the University of L'Aquvila and
//                     University of Heidelberg
//  Created on       : Wed Apr 05 15:29:39 2006
//  Last modified by : -
//  Last modified on : -
//  Update count     : 0
//
////////////////////////////////////////////////////////////////////////////////
//
//  Date	Name		Changes/Extensions
//  ----	----		------------------
//
////////////////////////////////////////////////////////////////////////////////

#ifndef PROJECTION_NODE_HH
#define PROJECTION_NODE_HH

#include <iostream>
#include <vector>

using namespace std;


/**
 * This class represents a node of the tree needed for the edge projection.
 */
class ProjectionNode  {

public:

    // -------------------------------------------------------------------------
    // ------------- M e t h o d s  ( p u b l i c ) ----------------------------
    // -------------------------------------------------------------------------

    /**
     * Constructor.
     *
     * @param theEdge Projection edge.
     * @param *theDeletedNodes Pointer to the nodes which are deleted.
     * @param *theDeletedEdges Pointer to the edges whcih are deleted.
     * @param *theFalseEdges pointer to the false edges.
     * @param *prevNode Pointer to the previous node of the tree.
     */
    ProjectionNode(int *theEdge, vector<int> *theDeletedNodes, 
                   vector<int*> *theDeletedEdges, vector<int*> *theFalseEdges,
                   ProjectionNode *prevNode);
    
    /**
     * Destructor
     */  
    ~ProjectionNode();  
                
    /**
     * Add new node to the tree.
     *
     * @param theEdge Projection edge.
     * @param *theDeletedNodes Pointer to the nodes which are deleted.
     * @param *theDeletedEdges Pointer to the edges whcih are deleted.
     * @param *theFalseEdges pointer to the false edges.
     * @return Pointer to the new node of the tree.
     */
    ProjectionNode* addNode(int theEdge[2], vector<int> *theDeletedNodes,
                            vector<int*> *theDeletedEdges, 
                            vector<int*> *theFalseEdges);

    /**
     * Get an endnode of the projection edge.
     *
     * @param i Index of the node.
     * @return On endnode of the edge. 
     */
    int getEdge(int i);
    
    /**
     * Get all deleted nodes.
     *
     * @param void
     * @return Pointer to a vector storing the nodes which have been deleted.
     */
    vector<int>* getDeletedNodes();
    
    /**
     * Get all deleted edges.
     *
     * @param void
     * @return Pointer to a vector storing the edges which have been deleted.
     */    
    vector<int*>* getDeletedEdges();
                
    /**
     * Get all false edges.
     *
     * @param void
     * @return Pointer to a vector storing the false edges.
     */    
    vector<int*>* getFalseEdges();

    /**
     * Get previous node in the tree.
     *
     * @param void
     * @return Pointer to the previous node of the tree.
     */        
    ProjectionNode* getPreviousNode();


private:

    // -------------------------------------------------------------------------
    // ------------- M e t h o d s  ( p r i v a t e ) --------------------------
    // -------------------------------------------------------------------------


    // -------------------------------------------------------------------------
    // ------------- V a r i a b l e s  ( p r i v a t e ) ----------------------
    // -------------------------------------------------------------------------
        
    /**
     * Store the projection edge according to this node of the tree.
     */
    int edge[2];
    
    /**
     * Store all deleted node according to this node of the tree.
     */
    vector<int> deletedNodes;

    /**
     * Store all deleted edges according to this node of the tree.
     */    
    vector<int*> deletedEdges;
    
    /**
     * Store all false edges according to this node of the tree.
     */
    vector<int*> falseEdges;
    
    /**
     * Pointer to the previous node of the tree.
     */
    ProjectionNode* previousNode;  
      
    /**
     * Store all sons of this node.
     */
    vector< ProjectionNode* > itsNodes;
    
};

#endif

