////////////////////////////////////// c++ /////////////////////////////////////
//
//  Module           : RankConstraint.cpp
//  Description      : Constraint class for a rank inequality.
//  Author           : Steffen Rebennack
//  Email            : srebenac@ix.urz.uni-heidelberg.de;
//                     steffen.rebennack@web.de
//  Copyright        : (C) 2006 by the University of Heidelberg
//  Created on       : Wed Apr 05 16:14:12 2006
//  Last modified by : -
//  Last modified on : -
//  Update count     : 0
//
////////////////////////////////////////////////////////////////////////////////
//
//  Date	Name		Changes/Extensions
//  ----	----		------------------
//
////////////////////////////////////////////////////////////////////////////////


#include "RankConstraint.hh"
#include "abacus/row.h"
#include "Graph.hh"
#include "Node.hh"


// -----------------------------------------------------------------------------
// C o n s t r u c t o r
//
// This class is derived from the abstract base class ABA_CONSTRAINT.
//
// Because a rank constraint is uniquely determined by their nodes, this
// constructor saves all nodes of it. The main part is
// the call of the constructor of class ABA_CONSTRAINT with the following
// arguments:
// 1.) A pointer to the corresponding master of the optimization.
// 2.) A pointer to the subproblem associated with this constraint because the
//     constraints should be only active in this sub problem and its children.
// 3.) This constraint is of kind less (or equal).
// 4.) The ride hand side of this constraint is rhs.
// 5.) This constraint can be removed from the active constraint set.
// 6.) This constraint is global valid and not only local.
// 7.) There is no lifting procedure implemented.
// -----------------------------------------------------------------------------
RankConstraint::RankConstraint(ABA_MASTER *master, vector<int> *nodes, int rhs,
			       Graph *theGraph):
    ABA_CONSTRAINT(master, NULL, ABA_CSENSE::Less, rhs, true, false, false),
    itsNodes(nodes->size()),
    itsGraph(theGraph),
    itsRhs(rhs)
{
    // Store this inequality that it can be checked.
    ofstream fout(theGraph->getFileName(), ios::app);

    for (int i=0; i<nodes->size(); i++) {
        itsNodes[i] = (*nodes)[i];
	fout << "  1 x_" << itsGraph->translateNode(itsNodes[i]);
    }
    fout << "  <  " << itsRhs << endl;
    fout.close();

    // Calculate the key.
    calculateHashKey();
}


// -----------------------------------------------------------------------------
// C o n s t r u c t o r
//
// It is the same as the other constructor but this time the inequality can
// be removed or not: depending on the value of dynamic.
//
// 5.) This constraint can be removed from the active constraint set.
// -----------------------------------------------------------------------------
RankConstraint::RankConstraint(ABA_MASTER *master, vector<int> *nodes, int rhs,
			       bool dynamic, Graph *theGraph):
    ABA_CONSTRAINT(master, NULL, ABA_CSENSE::Less, rhs, dynamic, false, false),
    itsNodes(nodes->size()),
    itsGraph(theGraph),
    itsRhs(rhs)
{
    // Store this inequality that it can be checked.
    ofstream fout(theGraph->getFileName(), ios::app);

    for (int i=0; i<nodes->size(); i++) {
        itsNodes[i] = (*nodes)[i];
	fout << "  1 x_" << itsGraph->translateNode(itsNodes[i]);
    }
    fout << "  <  " << itsRhs << endl;
    fout.close();

    // Calculate the key.
    calculateHashKey();
}


// -----------------------------------------------------------------------------
// D e s t r u c t o r
// -----------------------------------------------------------------------------
RankConstraint::~RankConstraint() {
}


// -----------------------------------------------------------------------------
// g e n R o w
//
// This function redefines the pure virtual function of the base class
// ABA_CONSTRAINT. It returns the constraint in sprase format. 
// -----------------------------------------------------------------------------
int RankConstraint::genRow(ABA_ACTIVE<ABA_VARIABLE,ABA_CONSTRAINT> *, 
			     ABA_ROW &row)
{

    for (int i=0; i<itsNodes.size(); i++) {
        row.insert(itsNodes[i],1.0);
    }
    row.rhs(itsRhs);
    row.sense(ABA_CSENSE::Less);

    return row.nnz();
}


// -----------------------------------------------------------------------------
// c o e f f
// 
// This function computes the coefficient of the variable var. It redefines the
// pure virtual function of the base class ABA_CONSTRAINT. The coefficient of
// a variable of a clique constraint is 1 if the variable is part of the clique 
// and otherwise its coefficient is 0.
// -----------------------------------------------------------------------------
double RankConstraint::coeff(ABA_VARIABLE *var) {

    // Cast ABA_VARIABLE* to NODE*.
    Node *node = (Node *) var;
    // search for this variable in the clique
    for (int i=0; i<itsNodes.size(); i++) {
        if (node->nodeNumber() == itsNodes[i]) {
            return 1.0;
        }
    }
    return 0.0;
}


// -----------------------------------------------------------------------------
// g e t S i z e
// -----------------------------------------------------------------------------
int RankConstraint::getSize() {
    return itsNodes.size();
}


// -----------------------------------------------------------------------------
// i s I n C o n s t r a i n t
//
// This function checks if a node is member of this constraint. It return the
// value true if the node is a memebr and otherwise false.
// This method is used by the function equal(..).
// -----------------------------------------------------------------------------
bool RankConstraint::isInConstraint(int node) {

    for(int i=0; i<itsNodes.size(); i++) {
        if (itsNodes[i] == node) {
            return true;
	}
    }

    return false;
}


// -----------------------------------------------------------------------------
// e q u a l
//
// This function compares if this constraint is identical with the constraint
// cv in a mathematical sense. This function is required becuase the constraint
// is stored in a pool of the class ABA_NONDUPLPOOL.
// It returns true if the constraints are equal and otherwise false.
// -----------------------------------------------------------------------------
bool RankConstraint::equal(ABA_CONVAR *cv) {

    bool isEqual = true;
    if (thisHashKey != ((RankConstraint *)cv)->hashKey()) {
        isEqual = false;
    }
    if (itsNodes.size() != ((RankConstraint *)cv)->getSize()) {
        isEqual = false;
    }

    for(int i=0; i<itsNodes.size(); i++) {
        if (!((RankConstraint *)cv)->isInConstraint(itsNodes[i])) {
	    isEqual = false;
	}
    }

    return isEqual;
}


// -----------------------------------------------------------------------------
// n a m e
//
// This function returns the name of the constraint. It is required if the
// constraint is stored in a pool of class ABA_NONDUPLPOOL.
// -----------------------------------------------------------------------------
const char* RankConstraint::name() {
    return "RankConstraint";
}


// -----------------------------------------------------------------------------
// h a s h K e y
//
// This method provides a key for the constraint that can be used to insert it
// into a hash table. In this case it is not required that any two items have
// different keys. This function is required to store the constraint in a pool
// of the class ABA_NIONDUPLPOOL.
// -----------------------------------------------------------------------------
unsigned RankConstraint::hashKey() {
    return thisHashKey;
}


// -----------------------------------------------------------------------------
// c a l c u l a t e H a s h K e y
// -----------------------------------------------------------------------------
void RankConstraint::calculateHashKey() {

    thisHashKey = 1;
    for (int i=0; i<itsNodes.size(); i++) {
        thisHashKey *= (itsNodes[i] + 1265);
	thisHashKey = thisHashKey % 9973;
    }
}


