////////////////////////////////////// c++ /////////////////////////////////////
//
//  Module           : ShortestPathCalculator.cpp
//  Description      :
//  Author           : Marcus Oswald
//  Email            : Marcus.Oswald@informatik.uni-heidelberg.de
//  Copyright        : University of Heidelberg
//  Created on       : Wed Jul 17 10:39:08 2002
//  Last modified by : rebennack
//  Last modified on : 06.10.2005
//  Update count     : 67
//
////////////////////////////////////////////////////////////////////////////////
//
//  Date	Name		Changes/Extensions
//  ----	----		------------------
//
////////////////////////////////////////////////////////////////////////////////


#include "ShortestPathCalculator.hh"
#include "StableSetSub.hh"



// -----------------------------------------------------------------------------
// --------------- M e t h o d s  ( p u b l i c ) ------------------------------
// -----------------------------------------------------------------------------


// -----------------------------------------------------------------------------
// C o n s t r u c t o r
//
//
// -----------------------------------------------------------------------------
ShortestPathCalculator::ShortestPathCalculator(int nNodes, int nEdges,
					       int* adjPtr, int* adjList,
					       double* edgeWeight, double bound,
					       double tol)
{

  int i, j;	// Loop variables.

  m_nNodes = nNodes;
  m_nEdges = nEdges;

  m_adjPtr = new int[m_nNodes+1];
  for(i=0; i<m_nNodes+1; i++) m_adjPtr[i] = adjPtr[i];

  m_adjList = new int[m_nEdges];
  for(j=0; j<m_nEdges; j++) m_adjList[j] = adjList[j];

  m_edgeWeight = new double[m_nEdges];
  for(j=0; j<m_nEdges; j++) m_edgeWeight[j] = edgeWeight[j];

  m_bound = bound;
  m_tol = tol;

  m_heap  = new int[m_nNodes];
  m_invHeap   = new int[m_nNodes];
  m_nodeLabel = new double[m_nNodes];

}


// -----------------------------------------------------------------------------
// Destructor.
// -----------------------------------------------------------------------------
ShortestPathCalculator::~ShortestPathCalculator() {
  delete[] m_adjPtr;
  delete[] m_adjList;
  delete[] m_edgeWeight;

  delete[] m_heap;
  delete[] m_invHeap;
  delete[] m_nodeLabel;
}


// -----------------------------------------------------------------------------
// c a l c u l a t e S h o r t e s t P a t h
// -----------------------------------------------------------------------------
double ShortestPathCalculator::calculateShortestPath(int startNode, int endNode,
                                                     int* predNode)
{

    int    i, j;			// Loop variables.
    int    currentNode = startNode;
    int    nextNode;
    int    child;
    int    parent;
    double minLabel = 0.0;
    double nextLabel;

    buildHeap(startNode);

    // terminate if bound is violated or if shortest path is found
    while (minLabel < m_bound-m_tol && currentNode!=endNode) {

        // look at each neighbour of current node
        for (j=m_adjPtr[currentNode]; j<m_adjPtr[currentNode+1]; j++) {
            nextNode = m_adjList[j];

            if(m_invHeap[nextNode] < m_heapSize) {
	        nextLabel = minLabel + m_edgeWeight[j];
   	        if(nextLabel < m_nodeLabel[nextNode]) {

	  //changeKey(nextNode,nextLabel);

  	            child = m_invHeap[nextNode];

	            while (child != 0) {
	                parent = (child-1)/2;
	                if(nextLabel < m_nodeLabel[m_heap[parent]]) {
	                    m_heap[child] = m_heap[parent];
	                    m_invHeap[m_heap[parent]] = child;
	                    child = parent;
	                }
	                else break;
	            }
	            m_heap[child]         = nextNode;
	            m_invHeap[nextNode]   = child;
	            m_nodeLabel[nextNode] = nextLabel;

	            predNode[nextNode] = currentNode;
	        }
            }
        }

        currentNode = extractMin();
        minLabel = m_nodeLabel[currentNode];

    }

    if (currentNode == endNode) {
        return minLabel;
    }
    else {
        return m_bound;
    }

} // end double calculateShortestPath(..)



// -----------------------------------------------------------------------------
// --------------- M e t h o d s  ( p r i v a t e ) ----------------------------
// -----------------------------------------------------------------------------


// -----------------------------------------------------------------------------
// b u i l d H e a p
// -----------------------------------------------------------------------------
void ShortestPathCalculator::buildHeap(int startNode)
{
    int i;   // Loop variable.

    m_heapSize = m_nNodes-1;

    for (i=0; i<m_heapSize; i++) {
        m_heap[i]    = i;
        m_invHeap[i] = i;
    }
    m_heap[m_heapSize] = startNode;
    m_heap[startNode] = m_heapSize;
    m_invHeap[m_heapSize] = startNode;
    m_invHeap[startNode] = m_heapSize;

    for (i=0; i<m_nNodes; i++) {
        m_nodeLabel[i] = m_bound;
    }

    //m_nodeLabel[startNode] = 0.0;
}


// -----------------------------------------------------------------------------
// e x t r a c t M i n
// -----------------------------------------------------------------------------
int ShortestPathCalculator::extractMin() {

  m_heapSize--;

  int min = m_heap[0];

  int insertNode = m_heap[m_heapSize];
  double insertLabel = m_nodeLabel[insertNode];

  m_heap[m_heapSize] = min;
  m_invHeap[min] = m_heapSize;

  int parent,firstChild,secondChild,minChild;
  double firstLabel,secondLabel,minLabel;

  parent = 0;
  firstChild = 1;
  while(firstChild<m_heapSize) {
    firstLabel = m_nodeLabel[m_heap[firstChild]];
    secondChild = firstChild+1;
    if(secondChild==m_heapSize) {
      minChild = firstChild;
      minLabel = firstLabel;
    }
    else {
      secondLabel = m_nodeLabel[m_heap[secondChild]];
      if(firstLabel < secondLabel) {
	minChild = firstChild;
	minLabel = firstLabel;
      }
      else {
	minChild = secondChild;
	minLabel = secondLabel;
      }
    }

    if(insertLabel < minLabel) break;
    else {
      m_heap[parent] = m_heap[minChild];
      m_invHeap[m_heap[minChild]] = parent;
      parent = minChild;
      firstChild = 2*parent+1;
    }

  }

  m_heap[parent] = insertNode;
  m_invHeap[insertNode] = parent;

  return min;

}// end: int extractMin()


// -----------------------------------------------------------------------------
// c h a n g e K e y
// -----------------------------------------------------------------------------
void ShortestPathCalculator::changeKey(int node, double label)
{

  int child,parent;

  child = m_invHeap[node];

  while(child != 0) {
    parent = (child-1)/2;
    if(label < m_nodeLabel[m_heap[parent]]) {
      m_heap[child] = m_heap[parent];
      m_invHeap[m_heap[parent]] = child;
      child = parent;
    }
    else break;
  }
  m_heap[child] = node;
  m_invHeap[node] = child;
  m_nodeLabel[node] = label;

}// end: void changeKey(..)


