////////////////////////////////////// c++ /////////////////////////////////////
//
//  Module           : StableSetMaster.cpp
//  Description      : Master of the stable set problem. Root of the Branch
//		       and Cut tree.
//  Author           : Steffen Rebennack
//  Email            : srebenac@ix.urz.uni-heidelberg.de;
//                     steffen.rebennack@web.de
//  Copyright        : (C) 2006 by the University of Heidelberg
//  Created on       : Thu Apr 06 10:39:14 2006
//  Last modified by : -
//  Last modified on : -
//  Update count     : 0
//
////////////////////////////////////////////////////////////////////////////////
//
//  Date	Name		Changes/Extensions
//  ----	----		------------------
//
////////////////////////////////////////////////////////////////////////////////

#include "StableSetMaster.hh"
#include "abacus/nonduplpool.h"
#include "Clique.hh"
#include "CliqueConstraint.hh"
#include "EdgeProjection.hh"
#include "Graph.hh"
#include "Node.hh"
#include "Preprocessing.hh"
#include "StableSetStatistics.hh"
#include "StableSetSub.hh"


#include <cstdlib>                   // rand()
#include <string>

#include <iostream>


extern "C"
{
   #include <stdio.h>
   #include <string.h>
}


#ifdef ABACUS_MATH_CPP
#include <math.h>
#else
extern "C"
{
   #include <math.h>
}
#endif



// -----------------------------------------------------------------------------
// --------------- M e t h o d s  ( p u b l i c ) ------------------------------
// -----------------------------------------------------------------------------


// -----------------------------------------------------------------------------
// C o n s t r u c t o r
//
// Argument problemName is the name of the optimization problem instance. If
// problemName starts with "./" then the input file with the same name is
// searched in the current directory, if it strats with "/" then the absolute
// path name of problem instances is taken, otherwise the input file is
// searched in the directory defined by environment variable StableSetLIB_DIR.
// The input file must have the special StableSetLIB format.
// Argument theInequalitiesFileName is the name of the file where all generated
// cuts are saved that they can be checked.
//
// The constructor calls first the constructor of its base class ABA_MASTER.
// This constructor takes 4 arguments:
// 1.) The name of the problem being solved.
// 2.) With the value true cutting planes are generated with the function
//     ABA_SUB::separate().
// 3.) pricing
// 4.) The sense of optimization.
//
// After the call of ABA_MASTER, the problem name and the name of the file
// for all generated inequalities is read. The next step is to read the input
// graph and to generate the object storing all statistic information.
// -----------------------------------------------------------------------------
StableSetMaster::StableSetMaster(const char *problemName,
				 const char *theInequalitiesFileName,
				 const char *theSmallGraphName):
    ABA_MASTER(problemName, true, false, ABA_OPTSENSE::Max)
{
    
    //
    // Read problem name.
    //
    // Check if problem name is not 0.
    if (problemName == 0) {
        err() << "StableSetMASTER::StableSetMASTER(): ";
        err() << "problem name is missing." << endl;
        exit(Fatal);
    }
    // Determine the complete file name.
    char *fileName;
    if ((problemName[0] == '/')
         || ((strlen(problemName) > 1
             && (problemName[0] == '.')
             && (problemName[1] == '/')))
       ){
        fileName = new char[strlen(problemName) +1];
        // Copy problemName in fileName.
        sprintf( fileName, "%s", problemName );
    }
    else {
        // Find location of StableSetLIB.
        const char *stableSetLib = getenv("StableSetLIB_DIR");
        if (stableSetLib == 0) {
            err() << "StableSetMASTER::StableSetMASTER(): environment ";
            err() << "variable StableSetLIB_DIR not found." << endl;
            exit(Fatal);
        }

        fileName = new char[strlen(stableSetLib) + strlen(problemName) + 2];

        #ifdef ABAUS_VISUAL_CPP
            sprintf(fileName, "%s\\%s", stableSetLib, problemName);
        #else
            sprintf(fileName, "%s/%s", stableSetLib, problemName);
        #endif
    }

    //
    // Read file name to store all generated cuts.
    //
    // Check if inequalities file name is not 0.
    if (theInequalitiesFileName == 0) {
        err() << "StableSetMASTER::StableSetMASTER(): ";
        err() << "inequalities file name is missing." << endl;
        exit(Fatal);
    }
    // Copy outputFileName in outputFile.
    char *inequalitiesFileName = new char[strlen(theInequalitiesFileName) + 1];
    sprintf(inequalitiesFileName, "%s", theInequalitiesFileName);
    // Delete contents of file
    ofstream out(inequalitiesFileName);
    out.close();

    // Copy small graph name.
    sprintf(smallGraphName, "%s", theSmallGraphName);

    //
    // Read input data.
    //
    readStableSetInputData(fileName, inequalitiesFileName);

    //
    // Generate object to store statistics.
    //
    theStatistics = new StableSetStatistics;

    // Clean up.
    delete [] fileName;
    delete [] inequalitiesFileName;

}// end: constructor


// -----------------------------------------------------------------------------
// D e s t r u c t o r
// -----------------------------------------------------------------------------
StableSetMaster::~StableSetMaster() {

    if (itsGraph->numberOfNodes() != 0) {
	delete theProjection;
	delete cycleCutPool_;
	delete cliqueCutPool_;
	delete rankCutPool_;
	delete generalCutPool_;
    }   
    delete myCPUTimer;
    delete itsGraph;
    delete maxCliques;
    delete theStatistics;
}


// -----------------------------------------------------------------------------
// f i r s t S u b
//
// It redefines a pure virtual function of the base class ABA_MASTER. The root
// of the branch-and-bound tree is initialized with an object of the type
// StableSetSUB. The return value is a pointer to the root node of the
// enumeration tree.
// -----------------------------------------------------------------------------
ABA_SUB *StableSetMaster::firstSub() {
    return new StableSetSub(this, itsGraph, maxCliques, theProjection,
			    theStatistics);
}


// -----------------------------------------------------------------------------
// c y c l e C u t P o o l
// -----------------------------------------------------------------------------
ABA_NONDUPLPOOL<ABA_CONSTRAINT, ABA_VARIABLE> *StableSetMaster::cycleCutPool() {
    return cycleCutPool_;
}


// -----------------------------------------------------------------------------
// c l i q u e C u t P o o l
// -----------------------------------------------------------------------------
ABA_NONDUPLPOOL<ABA_CONSTRAINT, ABA_VARIABLE> *StableSetMaster::cliqueCutPool(){
    return cliqueCutPool_;
}


// -----------------------------------------------------------------------------
// r a n k C u t P o o l
// -----------------------------------------------------------------------------
ABA_NONDUPLPOOL<ABA_CONSTRAINT, ABA_VARIABLE> *StableSetMaster::rankCutPool() {
    return rankCutPool_;
}


// -----------------------------------------------------------------------------
// g e n e r a l C u t P o o l
// -----------------------------------------------------------------------------
ABA_NONDUPLPOOL<ABA_CONSTRAINT, ABA_VARIABLE> *StableSetMaster::generalCutPool()
{
    return generalCutPool_;
}


// -----------------------------------------------------------------------------
// g e t S o l u t i o n
//
// This function is used by the improvement heuristic.
// -----------------------------------------------------------------------------
vector<int> *StableSetMaster::getSolution() {
    return &bestSolution;
}


// -----------------------------------------------------------------------------
// p u s h F i x e d N o d e
//
// This function is used by the preprocessor.
// -----------------------------------------------------------------------------
void StableSetMaster::pushFixedNode(int node) {
     fixedInStableSet.push_back(node);
}


// -----------------------------------------------------------------------------
// s o r t F i x e d N o d e s
//
// This function is used by the preprocessor.
// -----------------------------------------------------------------------------
void StableSetMaster::sortFixedNodes() {
    if (!fixedInStableSet.empty()) {
	sort(fixedInStableSet.begin(), fixedInStableSet.end());
    }
}


// -----------------------------------------------------------------------------
// f i x e d N o d e s T o O n e
//
// This function is used to determine the stablity number if the preprocessing
// phase was successful.
// -----------------------------------------------------------------------------
int StableSetMaster::fixedNodesToOne() const{
    return fixedInStableSet.size();
}


// -----------------------------------------------------------------------------
// p r i n t T i m e
// -----------------------------------------------------------------------------
void StableSetMaster::printTime() {
    out() << "Current time:" << myCPUTimer->seconds();
    out() << ":" << myCPUTimer->centiSeconds() % 100 << endl;
}


// -----------------------------------------------------------------------------
// s t a r t T i m e r
// -----------------------------------------------------------------------------
void StableSetMaster::startTimer() {
    myCPUTimer->start();
}


// -----------------------------------------------------------------------------
// s t o p T i m e r
// -----------------------------------------------------------------------------
void StableSetMaster::stopTimer() {
    myCPUTimer->stop();
}


// -----------------------------------------------------------------------------
// g e t S m a l l G r a p h N a m e
// -----------------------------------------------------------------------------
char *StableSetMaster::getSmallGraphName() {
    return smallGraphName;
}


// -----------------------------------------------------------------------------
// g e t R a n d o m N u m b e r
//
// Get an integer random number which is non-negative and smaller than 'bound'.
// -----------------------------------------------------------------------------
int StableSetMaster::getRandomNumber(int bound) const {
    int    aNumber = rand();
    double d       = (double(aNumber)/RAND_MAX);

    return (int)(d*bound);
}


// -----------------------------------------------------------------------------
// o u t p u t
//
// The function output() redefines a virtual dummy function of the base class
// master to output statistics of the run and the best stable set. This function
// is called at the end of the optimization after the output of the ABACUS
// statistics.
// -----------------------------------------------------------------------------
void StableSetMaster::output() {

    // Stop counter.
    myCPUTimer->stop();

    //
    // Preprocessing.
    //
    out() << "\n\nStatistics on preprocessing:\n";
    itsGraph->printStatistic();


    //
    // Output statistics on constraint generation.
    //
    out() << "\n\nStatistics on generated cuts:\n";
    out() << "  Number of generated edge inequalities\t\t:  "; 
    out() << theStatistics->getEdgeInequalitiesCut() << endl;
    out() << "  Number of generated odd cycles\t\t:  ";
    if (m_oddCycleSeparation) {
        out() << theStatistics->getOddCyclesCut() << endl;
        out() << "    Number of lifted odd circles\t\t:  ";
        out() << theStatistics->getLiftedOddCyclesCut() << endl;
    }
    else {
        out() << "off" << endl;
    }
    out() << "  Number of generated cliques\t\t\t:  ";
        if (m_CliqueHeuristicsSeparation) {
            out() << theStatistics->getCliqueHeuristicsCut() << endl;
    }
    else {
        out() << "off" << endl;
    }
    if (m_maxCliqueSeparation) {
        out() << "  Number of generated cliques (exact)\t\t:  ";
        out() << theStatistics->getExactCliquesCut() << endl;
    }

    out() << "  Number of generated rank inequalities\t\t:  ";
    if (m_EdgeProjection) {
        out() << theStatistics->getEdgeProjectionCut() << endl;
    }
    else {
        out() << "off" << endl;
    }
    out() << "  Number of generated mod-k cuts\t\t:  ";
    if (m_ModKCutsSeparation) {
        out() << theStatistics->getModKCut() << endl;
    }
    else {
        out() << "off" << endl;
    }
    out() << "  Number of generated local cuts\t\t:  ";
    if (m_LocalCutSeparation) {
        out() << theStatistics->getLocalCutCut() << endl;
    }
    else {
        out() << "off" << endl;
    }
    out() << "  Number of cuts from pool separation\t\t:  "; 
    if (m_PoolSeparation) {
        out() << theStatistics->getPoolSeparationCut() //
		+ theStatistics->getMaxCliquesMemoryCut() << endl;
    }
    else {
        out() << "off" << endl;
    }
    out() << "  Number of generated large rank cuts\t\t:  ";
    out() << theStatistics->getLargeCut() << endl; 

    //
    // Heuristics.
    //
    out() << "\nStatistics on heuristics:";
    out() << "\n  Number of solutions found by rounding heuristic\t:  ";
    out() << theStatistics->getRoundedSolutions();
    out() << "\n  Number of solutions found by the improvment heuristic\t:  ";
    out() << theStatistics->getImprovedSolutions();
    out() << "\n  Solution was found by\t\t\t\t\t:  ";
    out() << theStatistics->solutionWasFoundBy();

    //
    // Branch and Cut tree.
    //
    out() << "\n\nStatistics on Branch and Cut tree:";  
    out() << "\n  Number of generated subproblems\t:  " << nSub();
    out() << "\n  Number of solved LPs\t\t\t:  " << nLp();
    out() << "\n  Correct Dual Bound\t:  " << theStatistics->getDualBound();
    out() << "\n  Alpha\t\t:  ";
    out() << bestSolution.size() + fixedInStableSet.size();
    out() << "\n  Alpha in root\t:  ";
    out() << theStatistics->getAlphaRoot();
    out() << "\n  CPU Time\t:  ";
    out() << myCPUTimer->seconds() << ":";
    out() << myCPUTimer->centiSeconds() % 100 << endl;
    out() << endl;

    //
    // Output best stable set.
    //   
    if (m_showBestStableSet) {
        if (bestSolution.empty()) {
            out() << "No Stable Set found!" << endl;
	    return;
        }

        out() << "(Maximum) Stable Set:\t";

        if (!fixedInStableSet.empty()) {
	    int cnt_one = 0;
	    int cnt_two = 0;
	    int node_one = fixedInStableSet[cnt_one];
	    int node_two = itsGraph->translateNode(bestSolution[cnt_two]);
	    while (true) {
	        if (node_one < node_two) {
	   	    out() << node_one << " ";
		    cnt_one++;
		    if (cnt_one < fixedInStableSet.size()) {
		        node_one = fixedInStableSet[cnt_one];
		    }
		    else {
		        for (int i=cnt_two; i<bestSolution.size(); i++) {
		    	    out() << itsGraph->translateNode(bestSolution[i]) << " ";
			}
			break;
		    }// else
		}// if
		else {
		    out() << node_two << " ";
		    cnt_two++;
		    if (cnt_two < bestSolution.size()) {
		        node_two = itsGraph->translateNode(bestSolution[cnt_two]);
		    }
		    else {
		        for (int i=cnt_one; i<fixedInStableSet.size(); i++) {
		    	    out() << fixedInStableSet[i] << " ";
			}
		        break;
		    }
	        }// else
   	    }// while
        }// if
	else {
	    for (int i=0; i<bestSolution.size(); i++) {
	        out() << bestSolution[i]  << " ";
            }
            out() << endl;
	}// else
    }// if


}// end: void output()



// -----------------------------------------------------------------------------
// --------------- M e t h o d s  ( p r i v a t e ) ----------------------------
// -----------------------------------------------------------------------------


// -----------------------------------------------------------------------------
// r e a d S t a b l e S e t I n p u t D a t a
//
// This function reads a problem instance in StableSetLIB format and stores all
// this data in the class Graph.
// Argument fileName is the name of the input file
// -----------------------------------------------------------------------------
void StableSetMaster::readStableSetInputData(const char *fileName,
					     char *inequalitiesFileName) {

    int i;               		// Loop counter.
    int numberOfNodes;			// Store number of nodes.
    int numberOfEdges;			// Store number of edges.
    const int MAX_CHAR_PER_LINE = 1024;
    const int MAX_CHAR_PER_FILE_NAME = 256;
    char lineBuf[MAX_CHAR_PER_LINE +1];//
    char weightFileName[MAX_CHAR_PER_FILE_NAME +1];
    bool numberOfNodesFound = false;
    bool numberOfEdgesFound = false;
    bool weightsFound       = false;
    bool edgeSectionFound   = false;
    bool noWeights          = false;


    // Open input file.
    FILE *stableSetFile = fopen(fileName, "r");
    if (stableSetFile == NULL) {
        err() << "StableSetMASTER::readStableSetInputData(): ";
        err() << "stableSetLIB file " << fileName;
        err() << " could not be opened." << endl;
        exit(Fatal);
    }

    // Check input file format and store numberOfNodes and numberOfEdges
    // therefor read all lines of file stableSetFile and store it in lineBuf
    // - but max. 1024 characters per line.
    while (fgets(lineBuf, MAX_CHAR_PER_LINE, stableSetFile)) {
        // Compare lineBuf with key word NNODES.
        if (strncmp( lineBuf, "NNODES", strlen("NNODES") ) == 0) {
            // Save number of nodes.
           if (sscanf( lineBuf, "NNODES : %d", &numberOfNodes ) != 1) {
               err() << "StableSetMASTER::readStableSetInputData(): ";
               err() << "Error when reading numer of nodes (NNODES) of ";
               err() << "problem." << endl;
               exit(Fatal);
           }
           numberOfNodesFound = true;
        }
        else {
            if (strncmp(lineBuf, "NEDGES", strlen("NEDGES")) == 0) {
                // Save number of edges.
                if (sscanf( lineBuf, "NEDGES : %d", &numberOfEdges ) != 1) {
                    err() << "StableSetMASTER::readStableSetInputData(): ";
                    err() << "Error when reading numer of edges (NEDGES) ";
                    err() << "of problem." << endl;
                    exit(Fatal);
                }
                numberOfEdgesFound = true;
            }
            else {
                if (strncmp(lineBuf, "WEIGHTS", strlen("WEIGHTS")) == 0) {
                    // Is this a weighted stable set problem?
	            if (strncmp(lineBuf, "WEIGHTS : NO",
                                strlen("WEIGHTS : NO")) == 0) {
                        noWeights = true;
			weightsFound = true;
	            }
	            else {
			// Save file name which includes all weights.
                        if (sscanf(lineBuf, "WEIGHTS : %s",
                                   &weightFileName) != 1) {
                            err() << "StableSetMASTER::";
                            err() << "readStableSetInputData(): ";
                            err() << "Error when reading name of weight file.";
                            err() << endl;
                            exit(Fatal);
                        }
                        weightsFound = true;
                    }
                }
                else {
                    if (strncmp(lineBuf, "EDGE_SECTION",
                                strlen("EDGE_SECTION")) == 0) {
                        edgeSectionFound = true;
			// All data found -> exit while loop
                        break;
                    }
                }
            }
        }
    }  // end: while (fgets( lineBuf, MAX_CHAR_PER_LINE, stableSetFile)) {

    // Have all required keywords be found in the file?
    if (!numberOfNodesFound) {
        err() << "StableSetMASTER::readStableSetInputData(): ";
        err() << "Keyword NNODES could not be found in file ";
        err() << fileName << "." << endl;
        exit( Fatal );
    }

    if (!numberOfEdgesFound) {
        err() << "StableSetMASTER::readStableSetInputData(): ";
        err() << "Keyword NEDGES could not be found in file ";
        err() << fileName << "." << endl;
        exit( Fatal );
    }

    if (!weightsFound) {
        err() << "StableSetMASTER::readStableSetInputData(): ";
        err() << "Keyword WEIGHTS could not be found in file ";
        err() << fileName << "." << endl;
        exit( Fatal );
    }

    if (!edgeSectionFound) {
        err() << "StableSetMASTER::readStableSetInputData(): ";
        err() << "Keyword EDGE_SECTION could not be found in file ";
        err() << fileName << "." << endl;
        exit( Fatal );
    }

    //
    // Call constructor of graph.
    //
    itsGraph = new Graph(this, numberOfNodes, numberOfEdges, !noWeights,//
			 inequalitiesFileName);
	
    // Get all information of the graph.
    for (i=0; i<numberOfEdges; i++) {
        static int rightNode;
        static int leftNode;
        if (fscanf(stableSetFile, "%d %d",&rightNode, &leftNode) != 2) {
            err() << "StableSetMASTER::readStableSetInputData(): ";
            err() << "Error in edge line " << i+1 << " while reading ";
            err() << "EDGE_SECTION." << endl;
            exit(Fatal);
        }
        itsGraph->pushAdjacentNodes(rightNode, leftNode);

    }
    itsGraph->sortAdjacentList();

    // Close stableSetLIB file.
    if (fclose(stableSetFile)) {
        err() << "StableSetMASTER::readStableSetInputData(): ";
        err() << "Error in closing file " << fileName << "." << endl;
        exit(Fatal);
    }

    // Read coefficients.
    if (!noWeights) {
        bool   weightsSectionFound = false;
	double weight;	// Weight of a node.
        // Find location of StableSetLIB.
        const char *stableSetLib = getenv("StableSetLIB_DIR");
        if (stableSetLib == 0) {
            err() << "StableSetMASTER::readStableSetInputData(): ";
            err() << "Environment variable StableSetLIB_DIR not found.";
            err() << endl;
            exit(Fatal);
        }

        char weightFile[strlen(stableSetLib) + strlen(weightFileName) +2];

        #ifdef ABAUS_VISUAL_CPP
            sprintf(weightFile, "%s\\%s", stableSetLib, weightFileName);
        #else
            sprintf(weightFile, "%s/%s", stableSetLib, weightFileName);
        #endif

        // Open file which includes all weights.
        FILE *stableSetWeightFile = fopen(weightFile, "r");
        if (stableSetWeightFile == NULL) {
            err() << "StableSetMASTER::readStableSetInputData(): ";
            err() << "StableSetLIB file " << weightFile;
            err() << " could not be opened." << endl;
            exit(Fatal);
        }

        // Read all lines of file StableSetFile and store it in lineBuf
        // - but max. 1024 characters per line.
        while (fgets(lineBuf, MAX_CHAR_PER_LINE, stableSetWeightFile)) {
            // Compare 'lineBuf' with key word 'NNODES'.
            if (strncmp(lineBuf, "WEIGHTS_SECTION",
                        strlen("WEIGHTS_SECTION")) == 0) {
                weightsSectionFound = true;
	        break;
	    }
        }

        if (!weightsSectionFound) {
            err() << "StableSetMASTER::readStableSetInputData(): ";
            err() << "Keyword WEIGHTS_SECTION could not be found in file ";
            err() << weightFile << "." << endl;
            exit(Fatal);
        }
        
        // Read all weights.
        for (i=0; i<numberOfNodes; i++) {
            if( fscanf(stableSetWeightFile, "%lf", &weight) == 1) {
		itsGraph->pushWeight(i, weight);
	    }
	    else {
                err() << "StableSetMASTER::readStableSetInputData(): ";
                err() << "Error in weight line " << i+1;
                err() << " while reading WEIGHTS_SECTION." << endl;
                exit(Fatal);
            }
        }

        // Close stableSetWeightFile file.
        if (fclose(stableSetWeightFile)) {
            err() << "StableSetMASTER::readStableSetInputData(): ";
            err() << "Error in closing file " << weightFile << "." << endl;
            exit(Fatal);
        }

        // clean up
        delete stableSetLib;
    }// if (!noWeights) {

    //
    // Initialize clique constraints.
    //
    maxCliques = new Clique(this, itsGraph, numberOfEdges);

    // Output information about problem instance.
//    itsGraph->output();

}// end: void readStableSetInputData(..)


// -----------------------------------------------------------------------------
// i n i t i a l i z e O p t i m i z a t i o n
//
// This function defines a virtual dummy function of the class ABA_MASTER. 
// It starts the preprocessing phase and computes the first LP. This LP
// contains a clique covering and the trivial inequalities.
// After that is done, all cut pools are initialized.
// -----------------------------------------------------------------------------
void StableSetMaster::initializeOptimization() {

    int i, j;  						// Loop counter.

    // Output a banner.
    out() << "A Maximum Stable Set Solver." << endl;
    out() << "Copyright (c) Universitaet Heidelberg." << endl << endl;
    out() << "Program Version: 1.0 " << endl;
    out() << "Last update: Fri Apr 07 2006" << endl << endl;

    // Set timer.
    myCPUTimer = new ABA_CPUTIMER(this);
    myCPUTimer->start(0);

    itsGraph->initializeTranslator();

    if (itsGraph->numberOfNodes() < 100 
        && (itsGraph->getDensity() < .2 || itsGraph->getDensity() > .8)
       ) {
        // Call preprocessing
        Preprocessing *thePreprocessing;
        thePreprocessing = new Preprocessing(this, itsGraph, theStatistics);
        thePreprocessing->preprocessing();
       
        // Clean up.
        delete thePreprocessing; 
    }


    //
    // Generate new object edge projection.
    //
    int size = itsGraph->numberOfNodes();
    if (size) {
        vector< vector<int> > adjacentList(size);
        for (int i=0; i<size; i++) {
	    adjacentList[i] = (*itsGraph->getAdjacentList(i));
        }

        //cout << "m_storeInequalities:\t" << m_storeInequalities << endl;

	theProjection = new EdgeProjection(&adjacentList);
    }
    else {
	theStatistics->setDualBound(fixedInStableSet.size());
    }

    int numberOfNodes = itsGraph->numberOfNodes();	// Problem size.
    int numberOfEdges = itsGraph->numberOfEdges();	// Number of edges.

    // Check if the preprocessing step was successful.
    if (numberOfNodes == 0) {

	primalBound(fixedInStableSet.size());
	dualBound(fixedInStableSet.size());

	ABA_BUFFER<ABA_VARIABLE *> variables(this, 0);
	ABA_BUFFER<ABA_CONSTRAINT *> edgeConstraints(this, 0);

	initializePools(edgeConstraints, variables, 1, 0);

        return;
    }

    // Generate the variables.
    // Each variable in this stable set problem solver is associated with an
    // vertex of the undirected graph. This variables are created by using the
    // class NODE that is derived from the class ABA_VARIABLE. The objective
    // function coefficient is one in the unweighted case and otherwise the
    // weight itself.
    ABA_BUFFER<ABA_VARIABLE *> variables(this, numberOfNodes);

    // Add all variables with its weight.
    if (itsGraph->weighted()) {
        for (i=0; i<numberOfNodes; i++) {
            variables.push(new Node(this, i, itsGraph->weight(i)));
        }
    }
    else {
        static double DISPLACE = 0.0000000001;
	static int MAX_DISPLAYS_MULT  = (int) 100;
	int factor;
        for (i=0; i<numberOfNodes; i++) {
	    factor = getRandomNumber(MAX_DISPLAYS_MULT);
            variables.push(new Node(this, i, 1 + DISPLACE * factor));
        }
    }


    // Construct clique covering.
    ABA_BUFFER<ABA_CONSTRAINT *> cliqueConstraints(this, numberOfEdges);
    vector<int> theNodes(itsGraph->numberOfNodes());
    vector<int> adjacentNodes;
    vector<int> clique(2);
    int index;
    int node;
    int vertex;
    int begin;
    int numberOfCliques = 0;


    for (i=0; i<itsGraph->numberOfNodes(); i++) {
	theNodes[i] = i;	
    }

    ofstream fout(itsGraph->getFileName(), ios::app);
    fout << "* Clique covering:" << endl;
    fout.close();

    while (!theNodes.empty()) {
        clique[0] = theNodes.back();
	theNodes.pop_back();
	
	adjacentNodes = (*itsGraph->getAdjacentList(clique[0]));
	
	index = getRandomNumber(adjacentNodes.size());
	clique[1] = adjacentNodes[index];
	adjacentNodes.erase(adjacentNodes.begin() + index);
	
	for (i=adjacentNodes.size()-1; i>=0; i--) {
	    node = adjacentNodes[i];
            size = clique.size();
	    begin = 0;
	    if (!itsGraph->isMemberOfAdjacentList(clique[1], begin, node)) {
	   	adjacentNodes.erase(adjacentNodes.begin() + i);
	    }
	}// for

	while (!adjacentNodes.empty()) {
	    index = getRandomNumber(adjacentNodes.size());
	    vertex = adjacentNodes[index];
	    adjacentNodes.erase(adjacentNodes.begin() + index);
	    clique.push_back(vertex);
	    
	    // Update adjacent list.
	    for (i=adjacentNodes.size()-1; i>=0; i--) {
		node = adjacentNodes[i];
		begin = 0;
		if (!itsGraph->isMemberOfAdjacentList(vertex, begin, node)) {
		    adjacentNodes.erase(adjacentNodes.begin() + i);
		}
	    }
	}// while

	// Maximal clique found.
	// Delete all nodes from set theNodes.
	// Store the clique.
	sort(clique.begin(), clique.end());

	for (i=clique.size()-1; i>=0; i--) {
	    j = 0;	    
	    while (j<theNodes.size()) {
		if (clique[i] <= theNodes[j]) {
		    break;		
		}
		j++;	
	    }
	    if (j == theNodes.size()) {
		continue;	
	    }
	    if (clique[i] == theNodes[j]) {
		theNodes.erase(theNodes.begin() + j);
	    }
	}// for

	cliqueConstraints.push(new CliqueConstraint(this, &clique, itsGraph));
	numberOfCliques++;

	clique.clear();
	clique.resize(2);

    }// while


    cliqueConstraints.realloc(numberOfCliques);

    // Initialize the pools.
    // The first argument inserts the constraints in the constraint pool and the
    // second argument inserts the variables in the variable pool. Because there
    // are no more variables added, the number of variables is fixed to
    // numerOfNodes. The last argument is the size of the pool for the
    // cutting planes. It is zero because we have for each type an own cutting
    // plane.
    initializePools(cliqueConstraints, variables, numberOfNodes, 0);

     
    // The cut pool for the cycle constraints. The size of the pool is fixed.
    // The pool is of type non_dupl which means that no equal constraints are
    // stored twice.
    //
    // Size is updated automatically.
    //
    cycleCutPool_  = numberOfEdges < 100
    		     ? new ABA_NONDUPLPOOL<ABA_CONSTRAINT, ABA_VARIABLE>(
    					this, numberOfEdges*numberOfEdges, true)
		     : new ABA_NONDUPLPOOL<ABA_CONSTRAINT, ABA_VARIABLE>(
    					this, 25000, true);

    // The cut pool for the clique constraints. The size of the pool is fixed.
    // The pool is of type non_dupl which means that no equal constraints are
    // stored twice.
    //
    // Size is fixed.
    //
    cliqueCutPool_ = numberOfEdges < 100
    		     ? new ABA_NONDUPLPOOL<ABA_CONSTRAINT, ABA_VARIABLE>(
    					this, numberOfEdges*numberOfEdges, false)
		     : new ABA_NONDUPLPOOL<ABA_CONSTRAINT, ABA_VARIABLE>(
    					this, 15000, false);

    // The cut pool for the rank constraints. The size of the pool is dynamic.
    // The pool is of type non_dupl which means that no equal constraints are
    // stored twice.
    //
    // Size is fixed.
    //
    rankCutPool_ = numberOfEdges < 100
    		     ? new ABA_NONDUPLPOOL<ABA_CONSTRAINT, ABA_VARIABLE>(
    					this, numberOfEdges*numberOfEdges, false)
		     : new ABA_NONDUPLPOOL<ABA_CONSTRAINT, ABA_VARIABLE>(
    					this, 15000, false);



    // The cut pool for the local cuts. The size of the pool is fixed.
    // The pool is of type non_dupl which means that no equal constraints are
    // stored twice.
    generalCutPool_ = numberOfEdges < 100
    		      ? new ABA_NONDUPLPOOL<ABA_CONSTRAINT, ABA_VARIABLE>(
    					this, numberOfEdges*numberOfEdges, true)
		      : new ABA_NONDUPLPOOL<ABA_CONSTRAINT, ABA_VARIABLE>(
    					this, 5000, true);


    // Initialize the primal bound.
    // In this solver there is no heuristic for finding a solution without a
    // fractional solution and therfore the lower bound is zero.
    primalBound(0);


}// end: void initializeOptimization()


// -----------------------------------------------------------------------------
// i n i t i a l i z e P a r a m e t e r s
//
// This function redefines a virtual dummy function of the class ABA_MASTER. It
// is called at the beginning of the optimization.
// -----------------------------------------------------------------------------
void StableSetMaster::initializeParameters() {

    int  status;

    // The function readParameters() reads the file .myparameters that must be
    // in the ABACUS parameter file format. All parameters together with their
    // values are added to an internal paramter table from which they can be
    // read with the function getParameter().
    readParameters(".myparameters");

    status = getParameter("ShowBestStableSet", m_showBestStableSet);
    if (status) {
        err() << "StableSetMASTER::initializeParameters(): ";
        err() << "Parameter \"ShowBestStableSet\" could not be found ";
	err() << "in configuration file \".myparameters.\"" << endl;
        exit(Fatal);
    }

    status = getParameter("TimeLimitImprovementHeuristic",
    			  m_timeLimitForImprovementHeuristic);
    if (status) {
        err() << "StableSetMASTER::initializeParameters(): ";
        err() << "Parameter \"TimeLimitImprovementHeuristic\" could not";
	err() << " be found ";
	err() << "in configuration file \".myparameters.\"" << endl;
        exit(Fatal);
    }

    status = getParameter("TimeLimit", m_timeLimitForMaxCliqueSep);
    if (status) {
        err() << "StableSetMASTER::initializeParameters(): ";
        err() << "Parameter \"TimeLimit\" could not be found ";
	err() << "in configuration file \".myparameters.\"" << endl;
        exit(Fatal);
    }

    status = getParameter("MinimumViolationOfOddCycle", m_minAbsViolationCycle);
    if (status) {
        err() << "StableSetMASTER::initializeParameters(): ";
        err() << "Parameter \"MinimumViolationOfOddCycle\" could not be found ";
	err() << "in configuration file \".myparameters.\"" << endl;
        exit(Fatal);
    }

    status = getParameter("MinimumViolationOfClique", m_minAbsViolationClique);
    if (status) {
        err() << "StableSetMASTER::initializeParameters(): ";
        err() << "Parameter \"MinimumViolationOfClique\" could not be found ";
	err() << "in configuration file \".myparameters.\"" << endl;
        exit(Fatal);
    }

    status = getParameter("PoolSeparation", m_PoolSeparation);
    if (status) {
        err() << "StableSetMASTER::initializeParameters(): ";
        err() << "Parameter \"PoolSeparation\" could not be found ";
	err() << "in configuration file \".myparameters.\"" << endl;
        exit(Fatal);
    }


    status = getParameter("OddCycleSeparation", m_oddCycleSeparation);
    if (status) {
        err() << "StableSetMASTER::initializeParameters(): ";
        err() << "Parameter \"oddCycleSeparation\" could not be found ";
	err() << "in configuration file \".myparameters.\"" << endl;
        exit(Fatal);
    }

    status = getParameter("CliqueHeuristicsSeparation", m_CliqueHeuristicsSeparation);
    if (status) {
        err() << "StableSetMASTER::initializeParameters(): ";
        err() << "Parameter \"CliqueHeuristicsSeparation\" could not be found ";
	err() << "in configuration file \".myparameters.\"" << endl;
        exit(Fatal);
    }

    status = getParameter("EdgeProjection", m_EdgeProjection);
    if (status) {
        err() << "StableSetMASTER::initializeParameters(): ";
        err() << "Parameter \"EdgeProjection\" could not be found ";
	err() << "in configuration file \".myparameters.\"" << endl;
        exit(Fatal);
    }

    status = getParameter("ModKCutsSeparation", m_ModKCutsSeparation);
    if (status) {
        err() << "StableSetMASTER::initializeParameters(): ";
        err() << "Parameter \"ModKCutsSeparation\" could not be found ";
	err() << "in configuration file \".myparameters.\"" << endl;
        exit(Fatal);
    }


    status = getParameter("LocalCutSeparation", m_LocalCutSeparation);
    if (status) {
        err() << "StableSetMASTER::initializeParameters(): ";
        err() << "Parameter \"LocalCutSeparation\" could not be found ";
	err() << "in configuration file \".myparameters.\"" << endl;
        exit(Fatal);
    }

    status = getParameter("MaxCliqueSeparation", m_maxCliqueSeparation);
    if (status) {
        err() << "StableSetMASTER::initializeParameters(): ";
        err() << "Parameter \"MaxCliqueSeparation\" could not be found ";
	err() << "in configuration file \".myparameters.\"" << endl;
        exit(Fatal);
    }

}// end: void initializeParameters()


