////////////////////////////////////// c++ /////////////////////////////////////
//
//  Module           : StableSetMaster.hh
//  Description      : Master of the stable set problem. Root of the Branch
//		       and Cut tree.
//  Author           : Steffen Rebennack
//  Email            : srebenac@ix.urz.uni-heidelberg.de;
//                     steffen.rebennack@web.de
//  Copyright        : (C) 2006 by the University of Heidelberg
//  Created on       : Thu Apr 06 10:44:03 2006
//  Last modified by : -
//  Last modified on : -
//  Update count     : 0
//
////////////////////////////////////////////////////////////////////////////////
//
//  Date	Name		Changes/Extensions
//  ----	----		------------------
//
////////////////////////////////////////////////////////////////////////////////


#ifndef STABLE_SET_MASTER_HH
#define STABLE_SET_MASTER_HH

#include "abacus/master.h"
#include <vector>

using namespace std;


class Clique;
class EdgeProjection;
class Graph;
class StableSetStatistics;



/**
 * The class StableSetMaster is derived from the abstract class ABA_MASTER.
 * Its main purpose is:
 * - the inizialization of the pools with edge constraints and node variables,
 * - read and storage of input data (graph and paramters)
 * - memorization of the best stable set.
 */
class StableSetMaster: public ABA_MASTER {

public:

    // -------------------------------------------------------------------------
    // ------------- M e t h o d s  ( p u b l i c ) ----------------------------
    // -------------------------------------------------------------------------

    /**
     * Constructor.
     *
     * @param *problemName Pointer to the name of the optimization problem
     *			   instance.
     * @param *theInequalitiesFileName Pointer to the name of the file where all
     *                                 generated cuts are stored.
     * @param *theSmallGraphName Store the name of a small graph. This is used
     *                           by the rank inequality separation.
     */
    StableSetMaster(const char *problemName, const char *theInequalitiesFileName,
                    const char *theSmallGraphName);

    /**
     * Destructor.
     */
    virtual ~StableSetMaster();
    
    /**
     * Initialize the root of the branch-and-bound tree.
     *
     * @param void
     * @return Pointer to the root node of the enumeration tree.
     */
    virtual ABA_SUB *firstSub();

    /**
     * Get pool for the cycle constraints.
     *
     * @param void
     * @return Pointer to the cut pool for the cycle constraints.
     */
    ABA_NONDUPLPOOL<ABA_CONSTRAINT, ABA_VARIABLE> *cycleCutPool();

    /**
     * Get pool for the clique constraints.
     *
     * @param void
     * @return Pointer to the cut pool forthe clique constraints.
     */
    ABA_NONDUPLPOOL<ABA_CONSTRAINT, ABA_VARIABLE> *cliqueCutPool();

    /**
     * Get pool for the rank constraints.
     *
     * @param void
     * @return Pointer to the cut pool for the rank cuts.
     */
    ABA_NONDUPLPOOL<ABA_CONSTRAINT, ABA_VARIABLE> *rankCutPool();

    /**
     * Get pool for the general constraints.
     *
     * @param void
     * @return Pointer to the cut pool for the local cuts.
     */
    ABA_NONDUPLPOOL<ABA_CONSTRAINT, ABA_VARIABLE> *generalCutPool();

    /**
     * Get currently best solution.
     *
     * @param void
     * @return Pointer to a  vector storing the best solution found.
     */
    vector<int> *getSolution();

    /**
     * Store a node which has been fixed.
     *
     * @param node Index of a node.
     * @return void
     */
    void pushFixedNode(int node);

    /**
     * Sort the vector of the fixed nodes.
     *
     * @param void
     * @return void
     */
    void sortFixedNodes();

    /**
     * Get number of nodes which have been FIXED to value 1 in the
     * preprocessing phase.
     *
     * @param void
     * @return Number of fixed nodes to value 1.
     */
    int fixedNodesToOne() const;

    /**
     * Print CPU time.
     *
     * @param void
     * @return void
     */    
    void printTime();

    /**
     * Start CPU-timer.
     *
     * @param void
     * @return void
     */   
    void startTimer();

    /**
     * Stop CPU-timer.
     *
     * @param void
     * @return void
     */   
    void stopTimer();

    /**
     * Get name of the small graph.
     *
     * @param void
     * @return void
     */   
    char *getSmallGraphName(); 

    /**
     * Get a random number.
     *
     * @param bound Upper bound for the random number.
     * @return Random number y with property: y integer and 0 <= y < bound.
     */
    int getRandomNumber(int bound) const;

    /**
     * Output statistics at the end of the program.
     *
     * @param void
     * @return void
     */
    virtual void output();


    // -------------------------------------------------------------------------
    // -------------- V a r i a b l e s  ( p u b l i c ) -----------------------
    // -------------------------------------------------------------------------

    /**
     * Parameter managed by file .myparameters:
     * Time limit for the improvement heuristic.
     */
    double m_timeLimitForImprovementHeuristic; 

    /**
     * Parameter managed by file .myparameters:
     * Time limit for the exact max clique separation.
     */
    double m_timeLimitForMaxCliqueSep;

    /**
     * Parameter managed by file .myparameters:
     * Absolute value which every cycle constraint must violate to be added.
     */
    double m_minAbsViolationCycle;

    /**
     * Parameter managed by file .myparameters:
     * Absolute value which every clique constraint must violate to be added.
     */
    double m_minAbsViolationClique;

    /**
     * Parameter managed by file .myparameters:
     * If odd cycle separation is vished the value is true.
     */
    bool m_PoolSeparation;

    /**
     * Parameter managed by file .myparameters:
     * If pool separation is wished the value is true.
     */
    bool m_oddCycleSeparation;

    /**
     * Parameter managed by file .myparameters:
     * Separate the cliques randomly or not.
     */
    bool m_CliqueHeuristicsSeparation;

    /**
     * Parameter managed by file .myparameters:
     * Use edge projection or not.
     */
    bool m_EdgeProjection;

    /**
     * Parameter managed by file .myparameters:
     * Separate with mod-k or not.
     */
    bool m_ModKCutsSeparation;

    /**
     * Parameter managed by file .myparameters:
     * Use local cut separation or not.
     */
    bool m_LocalCutSeparation;

    /**
     * Parameter managed by file .myparameters:
     * If exact max clique separation is vished the value is true.
     */
    bool m_maxCliqueSeparation;

    /**
     * Parameter managed by file .myparameters:
     * If this flag is true than an optimal stable set is printed.
     */
    bool m_showBestStableSet;


private:

    // -------------------------------------------------------------------------
    // ------------- M e t h o d s  ( p r i v a t e ) --------------------------
    // -------------------------------------------------------------------------

    /**
     * This function reads a problem instance in StableSetLIB format.
     *
     * @param fileName The name of the input file.
     * @param *theInequalitiesFileName Pointer to the name of the file where all
     *                                 generated cuts are stored.
     */
    void readStableSetInputData(const char *fileName,
			        char *inequalitiesFileName);

    /**
     * This function initializes the optimization.
     * After the preprocessing, the cut pools are initialized and the first
     * clique covering is produced for the first LP. 
     *
     * @param void
     * @return void
     */
    virtual void initializeOptimization();

    /**
     * Read parameters of file .myparameters.
     *
     * @param void
     * @return void
     */
    virtual void initializeParameters();


    // -------------------------------------------------------------------------
    // ------------- V a r i a b l e s  ( p r i v a t e ) ----------------------
    // -------------------------------------------------------------------------

    /**
     * The corresponding graph of the problem instance.
     */
    Graph *itsGraph;

    /**
     * Manage the storage of all max cliques.
     */
    Clique *maxCliques;

    /**
     * This object handles the edge projection.
     */
    EdgeProjection *theProjection;

    /**
     * Pointer to an object managing the statistics.
     */
    StableSetStatistics *theStatistics;

    /**
     * Store the fixed nodes according to preprocessing.
     */
    vector<int> fixedInStableSet;

    /**
     * Store best stable set.
     */
    vector<int> bestSolution;
  
    /**
     * Small graph name.
     */ 
    char smallGraphName[256];
  
    /**
     * Cut pool for the odd cycle constraints.
     */
    ABA_NONDUPLPOOL<ABA_CONSTRAINT, ABA_VARIABLE> *cycleCutPool_;

    /**
     * Cut pool for the max clique constraints.
     */
    ABA_NONDUPLPOOL<ABA_CONSTRAINT, ABA_VARIABLE> *cliqueCutPool_;

    /**
     * Cut pool for the rank inequalities.
     */
    ABA_NONDUPLPOOL<ABA_CONSTRAINT, ABA_VARIABLE> *rankCutPool_;

    /**
     * Cut pool for general constraints.
     */
    ABA_NONDUPLPOOL<ABA_CONSTRAINT, ABA_VARIABLE> *generalCutPool_;

    /**
     * Abacus timer to get the CPU time.
     */
    ABA_CPUTIMER *myCPUTimer;

};// end: class StableSetMaster

#endif


