///////////////////////////////////// c++ /////////////////////////////////////
//
//  Module           : StableSetModKSeparator.cpp
//  Description      : Mod-k separation.
//  Author           : Hanna Seitz, Steffen Rebennack
//  Email            : Hanna.Seitz@informatik.uni-heidelberg.de;
//                     srebenac@ix.urz.uni-heidelberg.de;
//                     steffen.rebennack@web.de
//  Copyright        : University of Heidelberg
//  Created on       : Thu Apr 06 11:00:09 2006
//  Last modified by : -
//  Last modified on : -
//  Update count     : 0
//
////////////////////////////////////////////////////////////////////////////////
//
//  Date	Name		Changes/Extensions
//  ----	----		------------------
//
////////////////////////////////////////////////////////////////////////////////

#include <abacus/constraint.h>
#include <abacus/variable.h>
#include "abacus/row.h"

#include "StableSetModKSeparator.hh"
#include "GeneralConstraint.hh"
#include "Graph.hh"
#include "ModkSeparator.hh"
#include "StableSetLPSolution.hh"
#include "StableSetMaster.hh"

#include <vector>

using namespace std;



// -----------------------------------------------------------------------------
// --------------- M e t h o d s  ( p u b l i c ) ------------------------------
// -----------------------------------------------------------------------------


// -----------------------------------------------------------------------------
// C o n s t r u c t o r
// -----------------------------------------------------------------------------
StableSetModKSeparator::StableSetModKSeparator(StableSetLPSolution *solution,
					       Graph *theGraph):
    ABA_SEPARATOR<ABA_CONSTRAINT, ABA_VARIABLE> (solution, true, 10000),
    lpSolution(solution),
    itsGraph(theGraph)
{
    nVariables = theGraph->numberOfNodes();
}


// -----------------------------------------------------------------------------
// D e s t r u c t o r
// -----------------------------------------------------------------------------
StableSetModKSeparator::~StableSetModKSeparator() {
}


// -----------------------------------------------------------------------------
// s e p a r a t e
//
// Separate mod-2, mod-3, mod-5 and mod-7 cuts. Use as input the clique
// constraints which have slack 0 and eliminate all nodes of that clique
// which have LP value 0.
// -----------------------------------------------------------------------------
void StableSetModKSeparator::separate() {

    //
    // Variables.
    //
    // Stable set master.
    StableSetMaster *master = lpSolution->stableSetMaster();
    // Current sub problem.
    ABA_SUB *sub = lpSolution->sub();
    // Pointer to the pool which contains the variables to be added.
    ABA_STANDARDPOOL<ABA_CONSTRAINT,ABA_VARIABLE> *poolPointer;
    poolPointer = master->cliqueCutPool();
    // Number of conatraints in the pool.
    int nConstraints = poolPointer->number();
    // Loop counter;
    int i, j, a;
    // Number of conatrsints to be added to the separator.
    int nSlack0Constr = 0;
    // Tolerance.
    static double eps = 0.0000001;
    // Store constraints to be added.
    // Overestimate size.
    int *nNonZeros = new int[nConstraints +nVariables];
    int **nonZeroIndices = new int*[nConstraints +nVariables];
    double **nonZeroCoeffs = new double*[nConstraints +nVariables];
    double *rhs = new double[nConstraints +nVariables];
    // This will contain the information of the constraints.
    ABA_ROW *row = new ABA_ROW(master, nVariables);
    // Pointer to a constraint.
    ABA_CONSTRAINT *constrPointer;
    vector<int> indexVector;
    int numberOfVariablesInConstr;
    int index;
    double *solutionVal = lpSolution->zVal();
    // The k's which are used for the mod-k separation.
    int theKs[] = {2, 3, 5, 7}; 
    int numberOfKs = 4;

    //
    // Construct all inequalities for the mod-k-separator.
    //
    for (i=0; i<nConstraints; i++){
        // Get next constraint.
	constrPointer = (ABA_CONSTRAINT *)(( poolPointer->slot(i) )->conVar());
	if (constrPointer == NULL) {
	    continue;
	}

	// Select only constraints which have slack zero.
  	if (constrPointer->slack(sub->actVar(), lpSolution->zVal()) < -eps
	   || constrPointer->slack(sub->actVar(), lpSolution->zVal()) > +eps) {
	    continue;
	}

	// Store inequality.
	rhs[nSlack0Constr] = constrPointer->rhs(); // = 1;
	numberOfVariablesInConstr = constrPointer->genRow(sub->actVar(), *row);

	for (j=0; j<numberOfVariablesInConstr; j++) {
	    index = row->support(j);
	    if (solutionVal[index] >= eps) {
		indexVector.push_back(index);
	    }
	}
	if (indexVector.size() == 0) {
	    cout << "\n\n\n\nALARM!!!\n\n\n\n" << endl;
	    int abcd;
	    cin >> abcd;
	}
	nNonZeros[nSlack0Constr] = indexVector.size();
	nonZeroIndices[nSlack0Constr] = new int[nNonZeros[nSlack0Constr]];
	nonZeroCoeffs[nSlack0Constr]  = new double[nNonZeros[nSlack0Constr]];
	for (j=0; j<indexVector.size(); j++) {
	    nonZeroIndices[nSlack0Constr][j] = indexVector[j];
	    nonZeroCoeffs[nSlack0Constr][j] = row->coeff(j);
	}
	nSlack0Constr++;

        // Clean up.
	row->clear();
	indexVector.clear();
    }

    // Check the trivial inequalities.
    for (i=0; i<nVariables; i++) {
	if (solutionVal[i] <= eps) {
            nNonZeros[nSlack0Constr] = 1;
	    nonZeroIndices[nSlack0Constr] = new int[1];
	    nonZeroIndices[nSlack0Constr][0] = i;
	    nonZeroCoeffs[nSlack0Constr] = new double[1];
	    nonZeroCoeffs[nSlack0Constr][0] = -1;
	    rhs[nSlack0Constr] = 0;
	    nSlack0Constr++;
	}
	else {
	    if (solutionVal[i] >= 1-eps) {
		nNonZeros[nSlack0Constr] = 1;
	        nonZeroIndices[nSlack0Constr] = new int[1];
	        nonZeroIndices[nSlack0Constr][0] = i;
	        nonZeroCoeffs[nSlack0Constr] = new double[1];
	        nonZeroCoeffs[nSlack0Constr][0] = 1;
	        rhs[nSlack0Constr] = 1;
	        nSlack0Constr++;
	    }
	}
    }// for


    //
    // Check if there were inequalities.
    //
    if (nSlack0Constr == 0) {
	cout << "\n\t\tNo constraints with slack 0." << endl;
        
        return;
    }

    //
    // Call Mod-k separation routine.
    //

    for (a=0; a<numberOfKs; a++) {    
    
        // Generate object.
	ModkSeparator *modkSeparator;
	modkSeparator = new ModkSeparator(theKs[a], nSlack0Constr, nVariables,
					  nNonZeros,
			      		  nonZeroIndices, nonZeroCoeffs, rhs,
			      		  lpSolution->zVal());
	// Separate.
	int num = modkSeparator->separate( nVariables );
     
        ofstream fout(itsGraph->getFileName(), ios::app);
        fout << "\n* Mod :" << theKs[a] << endl;
        fout.close();

	// Store solution in abacus.
	if (num) {
	    // This array stores the information of the generated cut.
	    int *cutInfo; //= new int[3*nVariables];
	    // Store the inequalities: lhs * coeff <= rhs.
	    vector<int> lhs;
	    vector<int> coeff;
	    int rhsCut;
	    int cutSize;	// Number of nodes in the inequality.

	    // Get all inequalities.
	    for (i=0; i<num; i++) {
	    	// Get next inequality from separator.
	        cutInfo = modkSeparator->getCut();
	
	        if (cutInfo == NULL || cutInfo[0] == 0) {
    		    if (cutInfo != NULL) {
			delete [] cutInfo;
    		    }
		    continue;
		}

	        // Construct inequality.
	        cutSize = cutInfo[0];
		lhs.resize(cutSize);
		coeff.resize(cutSize);
	        for (j=0; j<cutSize; j++) {
		    lhs[j] = cutInfo[j+1];
		}
	        for (j=0; j<cutSize; j++) {
		    coeff[j] = cutInfo[j +1 +cutSize];
		}
	        rhsCut = cutInfo[1 +cutSize +cutSize];

	        // Generate new inequality.
	    	GeneralConstraint *generalConstr;
		generalConstr = new GeneralConstraint(master, &lhs, &coeff,
						      rhsCut, itsGraph);
		// Add cut to buffer.
		cutFound(generalConstr);
		
	        // Clean up.
	        delete[] cutInfo;
	    }// for
	   
        }// if
    
        // Output number of generated cuts.
        lpSolution->stableSetMaster()->out() << "\tMod " << theKs[a] << ":\t";
        lpSolution->stableSetMaster()->out() << num << endl;
	
        // Clean up.
	delete modkSeparator;

    }// for


    // Clean up.
    delete row;
    for (i=0; i<nSlack0Constr; i++){
	delete[] nonZeroIndices[i];
	delete[] nonZeroCoeffs[i];
    }
    delete[] nNonZeros;
    delete[] rhs;
    delete[] nonZeroIndices;
    delete[] nonZeroCoeffs;

}// end separate()


// -----------------------------------------------------------------------------
// a p p l y R a n k i n g
// -----------------------------------------------------------------------------
void StableSetModKSeparator::applyRanking(ABA_BUFFER<double> *rank) {

    // Nothing to do if buffer is empty.
    if (cutBuffer().number() == 0) {
	return;
    }

    //
    // Variables.
    //
    StableSetMaster *master = lpSolution->stableSetMaster();
    ABA_SUB *sub = lpSolution->sub();
    ABA_CONSTRAINT *constrPointer;
    ABA_POOLSLOT<ABA_CONSTRAINT, ABA_VARIABLE> *constrSlot;
    int i, j; // Loop varibels.
    double slack, nNonZeroCoeffs, rankValue;

    ABA_STANDARDPOOL<ABA_VARIABLE, ABA_CONSTRAINT> *varPool;
    ABA_POOLSLOT<ABA_VARIABLE, ABA_CONSTRAINT> *varSlot;
    varPool = master->varPool();
    ABA_VARIABLE *varPointer;
    double constrCoeff, objectiveCoeff, scaleCoeff = 0, scaleObj = 0;

    //
    // Rank the inequalities.
    //
    for (i=0; i<cutBuffer().number(); i++){
	constrPointer = cutBuffer()[i];
	rankValue = 0;
	for(j=0; j<nVariables; j++){
	    varPointer = (varPool->slot(j))->conVar();
	    objectiveCoeff = varPointer->obj();
	    if (objectiveCoeff > 0.999999 && objectiveCoeff < 1.000001) {
		objectiveCoeff = 1;
	    }
	    if(objectiveCoeff < -0.999999 && objectiveCoeff > -1.000001) {
		objectiveCoeff = -1;
	    }
	    constrCoeff = constrPointer->coeff(varPointer);
            /* Skalarprodukt berechnen */
	    rankValue += (objectiveCoeff * constrCoeff);
	    /* Fuer das spaetere Normieren */ 
	    scaleCoeff += (constrCoeff * constrCoeff);
	    scaleObj += (objectiveCoeff* objectiveCoeff);
	    
	}
        /* Normieren des Skalarproduktes */
	scaleCoeff = sqrt(scaleCoeff);
	scaleObj = sqrt(scaleObj);
	rankValue = rankValue / (scaleCoeff * scaleObj); 
	/* Kleine Winkel sind am besten. 
	   Aber fuer ABACUS je groesser, desto besser,daher */
	rankValue = 1 - rankValue;
	rank->push(rankValue);
    }
 
    return;

}// end applyRanking(..)


