////////////////////////////////////// c++ /////////////////////////////////////
//
//  Module           : StableSetStatistics.hh
//  Description      : Statistics.
//  Author           : Steffen Rebennack
//  Email            : srebenac@ix.urz.uni-heidelberg.de;
//                     steffen.rebennack@web.de
//  Copyright        : (C) 2006 by the University of Heidelberg
//  Created on       : Thu Apr 06 11:37:54 2006
//  Last modified by : -
//  Last modified on : -
//  Update count     : 0
//
////////////////////////////////////////////////////////////////////////////////
//
//  Date	Name		Changes/Extensions
//  ----	----		------------------
//
////////////////////////////////////////////////////////////////////////////////

#ifndef StableSetStatistics_HH
#define StableSetStatistics_HH



/**
 * This class provides the statistics for the stable set solver.
 */
class StableSetStatistics  {

public:

    // -------------------------------------------------------------------------
    // ------------- M e t h o d s  ( p u b l i c ) ----------------------------
    // -------------------------------------------------------------------------

    /**
     * Constructor.
     */
    StableSetStatistics();

    /**
     * Destructor
     */  
    ~StableSetStatistics();

    /**
     * Increase counter for the computed edge inequalities.
     *
     * @param num Increase counter with that number.
     * @return void
     */
    void edgeInequalitiesCut(int num);

    /**
     * Increase number of cuts from pool separation.
     *
     * @param num Increase counter with that number.
     * @return void
     */
    void poolSeparationCut(int num);

    /**
     * Increase generated maximal cliques from memory.
     *
     * @param num Increase counter with that number.
     * @return void
     */
    void maxCliquesMemoryCut(int num);

    /**
     * Increase number of odd cyles found.
     *
     * @param num Increase counter with that number.
     * @return void
     */
    void oddCyclesCut(int num);

    /**
     * Increase number of lifted odd cyles (triangles).
     *
     * @param num Increase counter with that number.
     * @return void
     */
    void liftedOddCyclesCut(int num);

    /**
     * Increase number of computed maximal cliques (exact).
     *
     * @param num Increase counter with that number.
     * @return void
     */
    void exactCliquesCut(int num);

    /**
     * Increase number of computed maximal cliques (heuristics).
     *
     * @param num Increase counter with that number.
     * @return void
     */
    void cliqueHeuristicsCut(int num);

    /**
     * Increase number of computed cuts according to edge projection.
     *
     * @param num Increase counter with that number.
     * @return void
     */
    void edgeProjectionCut(int num);

    /**
     * Increase number of local cuts.
     *
     * @param num Increase counter with that number.
     * @return void
     */
    void localCutCut(int num);

    /**
     * Increase number of computed mod-k cuts.
     *
     * @param num Increase counter with that number.
     * @return void
     */
    void modKCut(int num);

    /**
     * Increase number of computed large cuts (rank inequalities).
     *
     * @param num Increase counter with that number.
     * @return void
     */
    void largeCutCut(int num);

    /**
     * Dual bound of the root.
     *
     * @param val Dual bound.
     * @return void
     */
    void setDualBound(double val);

    /**
     * Increase counter (by value one) for the solutions found
     * by the rounding heuristics.
     *
     * @param void.
     * @return void
     */
    void roundingHeuristics();

    /**
     * Increase counter (by value one) for the solutions found
     * by the improvement heuristics.
     *
     * @param void.
     * @return void
     */
    void improvementHeuristics();

    /**
     * Stable set solution was found by ...
     *
     * @param index Indicate who has found the solution.
     * @return void
     */
    void solutionFound(int index);

    /**
     * Store the weight of the best stable set computed in the root.
     *
     * @param alpha Weight of a stable set.
     * @return void
     */
    void setAlphaRoot(int alpha);

    /**
     * Get number.
     *
     * @param void
     * @return Number of computed edge inequalities.
     */
    int getEdgeInequalitiesCut() const;

    /**
     * Get number.
     *
     * @param void
     * @return Number of computed inequalities from pool.
     */
    int getPoolSeparationCut() const;

    /**
     * Get number.
     *
     * @param void
     * @return Number of computed cliques from memory.
     */
    int getMaxCliquesMemoryCut() const;

    /**
     * Get number.
     *
     * @param void
     * @return Number of computed odd cycles.
     */
    int getOddCyclesCut() const;

    /**
     * Get number.
     *
     * @param void
     * @return Number of lifted odd cycles.
     */
    int getLiftedOddCyclesCut() const;

    /**
     * Get number.
     *
     * @param void
     * @return Number of computed cliques due to exact separation.
     */
    int getExactCliquesCut() const;

    /**
     * Get number.
     *
     * @param void
     * @return Number of computed clique inequalities.
     */
    int getCliqueHeuristicsCut() const;

    /**
     * Get number.
     *
     * @param void
     * @return Number of computed cuts from edge projection.
     */
    int getEdgeProjectionCut() const;

    /**
     * Get number.
     *
     * @param void
     * @return Number of computed local cuts.
     */
    int getLocalCutCut() const;

    /**
     * Get number.
     *
     * @param void
     * @return Number of computed mod-k cuts.
     */
    int getModKCut() const;

    /**
     * Get number.
     *
     * @param void
     * @return Number of computed large rank inequalities.
     */
    int getLargeCut() const;

    /**
     * Get number.
     *
     * @param void
     * @return Number of solutions found by rounding heuristics.
     */
    int getRoundedSolutions() const;

    /**
     * Get number.
     *
     * @param void
     * @return Number of improved solutions.
     */
    int getImprovedSolutions() const;

    /**
     * Get number.
     *
     * @param void
     * @return Maximal weight of a stable set computed in the root.
     */
    int getAlphaRoot() const;

    /**
     * Get number.
     *
     * @param void
     * @return Get dual bound of the root.
     */
    double getDualBound() const;

    /**
     * Get number.
     *
     * @param void
     * @return Who found the solution.
     */
    char *solutionWasFoundBy() const; 


private:

    // -------------------------------------------------------------------------
    // ------------- M e t h o d s  ( p r i v a t e ) --------------------------
    // -------------------------------------------------------------------------


    // -------------------------------------------------------------------------
    // ------------- V a r i a b l e s  ( p r i v a t e ) ----------------------
    // -------------------------------------------------------------------------
        
    /**
     * Number of computed edge inequalities.
     */
    int edgeInequalities;

    /**
     * Number of cuts separated from pool.
     */
    int poolSeparationCuts;

    /**
     * Number of cuts from clique memorization.
     */
    int maxCliquesMemoryCuts;

    /**
     * Number of odd cyles.
     */
    int oddCyclesCuts;

    /**
     * Number of lifted odd cycles.
     */
    int liftedOddCyclesCuts;

    /**
     * Number of computed cliques (exact separation).
     */
    int exactCliquesCuts;

    /**
     * Number of computed cliques (heuristics).
     */
    int cliqueHeuristicsCuts;

    /**
     * Number of cuts from edge projection.
     */
    int edgeProjectionCuts;

    /**
     * Number of local cuts.
     */
    int localCutCuts;

    /**
     * Number of mod-k cuts.
     */
    int modKCuts;

    /**
     * Number of large cuts.
     */
    int largeCuts;

    /**
     * Number of soltions found by the rounding heuristics.
     */
    int roundedSolutions;

    /**
     * Number of solution computed by improvement heuristics.
     */
    int improvedSolutions;

    /**
     * Solution was found by.
     */
    int solutionFoundBy;

    /**
     * Best weight of a stable set in the root of the Branch and Cut tree.
     */
    int alphaRoot;

    /**
     * Dual bound of the root.
     */
    double dualBoundOfRoot;

};

#endif
