////////////////////////////////////// c++ /////////////////////////////////////
//
//  Module           : StableSetSub.cpp
//  Description      : The nodes of the Branchand Cut tree.
//  Author           : Steffen Rebennack
//  Email            : srebenac@ix.urz.uni-heidelberg.de;
//                     steffen.rebennack@web.de
//  Copyright        : (C) 2006 by the University of Heidelberg
//  Created on       : Thu Apr 06 13:40:40 2006
//  Last modified by : -
//  Last modified on : - 
//  Update count     : 0
//
////////////////////////////////////////////////////////////////////////////////
//
//  Date	Name		Changes/Extensions
//  ----	----		------------------
//
////////////////////////////////////////////////////////////////////////////////


#include "StableSetSub.hh"
#include "abacus/lpsub.h"
#include "abacus/row.h"
#include "Clique.hh"
#include "CliqueConstraint.hh"
#include "CliqueHeuristicsSeparator.hh"
#include "EdgeProjection.hh"
#include "GeneralConstraint.hh"
#include "Graph.hh"
#include "ImprovementHeuristic.hh"
#include "LocalCutSeparator.hh"
#include "LowerBoundHeuristic.hh"
#include "MaxCliquesSeparator.hh"
#include "Node.hh"
#include "OddCyclesSeparator.hh"
#include "RankConstraint.hh"
#include "StableSetBranchRule.hh"
#include "StableSetLPSolution.hh"
#include "StableSetMaster.hh"
#include "StableSetModKSeparator.hh"
#include "StableSetStatistics.hh"



//
// The struct NodesAndDegree is used to store a node with its degree.
// It is used by the branching rule.
//
struct NodesAndDegree {
    int    inode;        // index of node
    double degree;       // degree of a node
};// struct


//
// This operator is used by the sort a set of nodes according to its degree.
//
bool operator<(const NodesAndDegree &a, const NodesAndDegree &b) {
        return a.degree > b.degree;
}



// -----------------------------------------------------------------------------
// --------------- M e t h o d s  ( p u b l i c ) ------------------------------
// -----------------------------------------------------------------------------


// -----------------------------------------------------------------------------
// C o n s t r u c t o r
//
// This is the constructor for the root node of the enumeration tree. Calling
// the constructor of the base class ABA_SUB with the arguments:
// 1.) A pointer to the corresponding master of the problem.
// 2.) 10% additional memory allocated for constraints.
// 3.) No additional storage for variables
// 4.) 50% additional space for the nonzeros of the constraint matrix in the
//     LP-solver should be allocated.
// -----------------------------------------------------------------------------
StableSetSub::StableSetSub(ABA_MASTER *master, Graph *theGraph,
			   Clique *theClique, EdgeProjection *anEdgeProjection,
			   StableSetStatistics *aStatistics):
   ABA_SUB(master, 10, 0, 50),
   itsGraph(theGraph),
   maxCliques(theClique),
   theEdgeProjection(anEdgeProjection),
   theStatistics(aStatistics),
   violatedEdge(false)
{
}


// -----------------------------------------------------------------------------
// C o n s t r u c t o r
//
// This is the constructor for a son of an existing node. The constructor of the
// ABACUS class ABA_SUB takes three arguments: a pointer to the master of the
// problem, a pointer to the father of the enumeration tree and the rule
// defining the subspace of solution space associated with this sub porblem.
// -----------------------------------------------------------------------------
StableSetSub::StableSetSub(ABA_MASTER *master, ABA_SUB *father,
                           ABA_BRANCHRULE *branchRule, Graph *theGraph,
			   Clique *theClique, EdgeProjection *anEdgeProjection,
			   StableSetStatistics *aStatistics):
   ABA_SUB(master, father, branchRule),
   itsGraph(theGraph),
   maxCliques(theClique),
   theEdgeProjection(anEdgeProjection),
   theStatistics(aStatistics)
{
}


// -----------------------------------------------------------------------------
// D e s t r u c t o r
// -----------------------------------------------------------------------------
StableSetSub::~StableSetSub() {
}


// -----------------------------------------------------------------------------
// f e a s i b l e
//
// This function redefines a pure virtual function of ABACUS. It checks for the
// feasibility of a solution of the LP-relaxation. A solution indicates a 
// Stable Set if all edge constraints are satisfied and all variables are 
// binary. If a solution is feasible, than this function returns true. If in
// addition the value of the primal bound is smaller than the value of this
// feasible solution, the solution and the value of the primal bound are
// updated. The update of the primal bound is done automatically by ABACUS.
// As all the edge constraints are relaxed and not fixed in the cut pool, one
// has to check them prior to the update of the solution.
// -----------------------------------------------------------------------------
bool StableSetSub::feasible() {

    bool feasible;  // True: the LP solution is feasible; else false.

    //
    // Check all the edge inequalities.
    //
    if (true) {
        static int MAX_NUM = (int) itsGraph->numberOfNodes() / 2;
        static double EPS = 0.000001;
	
        int newCuts = 0;
	int numPrevious = 0;
	int num = 0;
        int number = -1;

   	int i, j, k;
        int numberOfNodes = itsGraph->numberOfNodes();
	vector<int> adjacentList;
	double * solutionVal = xVal_;
	vector<int> clique(2);
	CliqueConstraint *cliqueConstr;
	// Add all violated maximal cliques.
        ABA_BUFFER< ABA_CONSTRAINT* > constraint(master_, MAX_NUM);
	int size;
	int vertex;
	int node;
	int index;
	double val;
	int begin;

	violatedEdge = false;

	// Store all inequalities in file 'storeInequalitiesFile' that they
        // can be checked if they are valid.
        ofstream fout(itsGraph->getFileName(), ios::app);
	fout << "\n* Edge inequalities:" << endl;

	for (i=0; i<numberOfNodes; i++) {
	    if (solutionVal[i] <= EPS) {
		continue;
	    }
	    adjacentList = (*itsGraph->getAdjacentList(i));
	    size = itsGraph->adjacentListSize(i);
	    for (j=size-1; j>=0; j--) {
	        val = solutionVal[i] + solutionVal[adjacentList[j]];

		if ( (val > (1 + EPS)) && (val < 3)) {
		    // Violated edge inequality found.
		    vertex = adjacentList[j];
		    clique[0] = i;
		    clique[1] = vertex;
		    adjacentList.erase(adjacentList.begin() + j);
		    // Lift to maximal clique.
		    // Update adjacentList.
		    for (k=adjacentList.size()-1; k>=0; k--) {
			begin = 0;
			if (!itsGraph->isMemberOfAdjacentList(vertex, begin, adjacentList[k])) {
			    adjacentList.erase(adjacentList.begin() + k);
			}
		    }

		    while (!adjacentList.empty()) {
			index = stableSetMaster()->getRandomNumber(adjacentList.size());
	    		vertex = adjacentList[index];
	    		clique.push_back(vertex);
	    		adjacentList.erase(adjacentList.begin() + index);
	    		// Update set 'adjacentList'
	    		for (k=adjacentList.size()-1; k>=0; k--) {
			    node = adjacentList[k];
			    begin = 0;
			    if (!itsGraph->isMemberOfAdjacentList(vertex, begin, node)) {
		    	        adjacentList.erase(adjacentList.begin() + k);
			    }
			}
	    	    }// while

		    // Add this inequality.
		    // New constraint.
		    sort(clique.begin(), clique.end());
		    cliqueConstr = new CliqueConstraint(master_, &clique, itsGraph);
	            constraint.push(cliqueConstr);
		    newCuts++;

		    // Clean up
		    clique.clear();
		    clique.resize(2);
		    adjacentList = (*itsGraph->getAdjacentList(i));

		    if (newCuts == MAX_NUM) {
			break;
		    }
		}
	    }// for
	    if (newCuts == MAX_NUM) {
		break;
	    }
	}

	fout.close();

	if (newCuts) {
	    // Add cuts to the pool.
	    numPrevious = stableSetMaster()->cliqueCutPool()->number();
            number = addCons(constraint, stableSetMaster()->cliqueCutPool());
	    number = stableSetMaster()->cliqueCutPool()->number();
	    number -= numPrevious;
	    num = number >= 0 ? number : num;
	    violatedEdge = true;	    

	    // Output
	    master_->out() << "\tEdge inequalities:\t" << num << endl;

	    // Update statistics.
            theStatistics->edgeInequalitiesCut(num);

	    return false;
	}
	else {
	    master_->out() << "\tAll edge inequalities are valid" << endl; 
	}
    }// if


    // The function integerFeasible() checks if the solution of the
    // LP-relaxation is primally feasible. This means that all variables are
    // binary and all constraints of the LP relaxation are satisfied.
    feasible = integerFeasible();

    // ABACUS does not update the solution - so you do.
    if (feasible) {
        // LP solution is incedence vector of a Stable Set
        // If the value of the Stable Set is higher than the best known then
        // we update the solution. xVal_ stores the values of the variables of
        // the LP solution.
	if (lp_->value() >= (master_->primalBound() + .5)) {

	    int i;
	    int numberOfNodes = itsGraph->numberOfNodes();	// Problem size.
	    double OneMinusEps = 1.0 - master_->machineEps();	// Precision.
	    vector<int> *bestSolution = stableSetMaster()->getSolution();
	    bestSolution->clear();

	    for (i=0; i<numberOfNodes; i++) {
       		// xVal[i] == 1?
       		if ((xVal_[i] > OneMinusEps) && (xVal_[i] < 2)) {
		    bestSolution->push_back(i);
       		}
		else {
		    if ((xVal_[i] > master_->machineEps())
		        || (xVal_[i] < -master_->machineEps())) {
			master_->err() << "StableSetMASTER::update():";
			master_->err() << " x is not incidence vector ";
			master_->err() << "of a Stable Set " << endl;
			exit(master_->Fatal);
		    }
		}
            }

	    master_->out() << "\nSolution was integer feasible with weight:\t";
            master_->out() << (int)lp_->value() << "." << endl;

	    // Update lower bound (best known feasible solution)
            // (this is done automatically by ABACUS, too).
	    master_->primalBound((int)lp_->value());
	    theStatistics->solutionFound(0);
	}
    }

    return feasible;

}// end: bool feasible()


// -----------------------------------------------------------------------------
// g e n e r a t e S o n
//
// This function redefindes a pure virtual function of the base class ABA_SUB.
// While the function stableSetMaster::firstSub() initializes the root node of
// the branch-and-bound tree with a problem specific subproblem, this function
// generates a problem specific son of a subproblem.
// -----------------------------------------------------------------------------
ABA_SUB *StableSetSub::generateSon(ABA_BRANCHRULE *rule) {
    return new StableSetSub(master_, this, rule, itsGraph, maxCliques,
    			    theEdgeProjection,
			    theStatistics);
}


// -----------------------------------------------------------------------------
// i m p r o v e
//
// This is a redefinition of the pure virtual ABACUS function improve(). It is
// called after every solved LP-relaxation. 
// This function uses two heuristics:
// ~ Construction of a stabel set via a rounding heuristics from a fractional
//   LP solution. This is done by the lowerBoundHeuristic.
// ~ Improve a stable set.
// If one of the heuristics computes a better solution than the
// best known one, this solution is stored and the lower bound is updated. In
// this case the function return value 1. If no better solution is found it
// will return 0.
// -----------------------------------------------------------------------------
int StableSetSub::improve(double &primalValue) {

    int status = 0;  // This stores the return value.
    double value;

    LowerBoundHeuristic *roundingHeuristic;
    roundingHeuristic = new LowerBoundHeuristic(master_, itsGraph,
						theStatistics),
    // Call heuristic.
    status = roundingHeuristic->computeSolution(value, xVal_, level());
 	
    // Clean up.
    delete roundingHeuristic;       

    if (!itsGraph->weighted()) {
        // Generate object.
        ImprovementHeuristic *improveHeuristic;
        improveHeuristic = new ImprovementHeuristic(master_, itsGraph,
							 theStatistics);
	     
        // Improve current solution.
        status = (improveHeuristic->improve(value) || status);
  
	// Clean up.
	delete improveHeuristic;
    }

    if (status) {
       primalValue = value;
   }
   
   return status;
}// end: int improve(..)


// -----------------------------------------------------------------------------
// s e p a r a t e
//
// This function redefines a pure virtual function of ABACUS. It separates the
// following inequalities (in this order):
// ~ pool separation
// ~ clique separation of memory (only with exact clique separation)
// ~ odd cycle separation and lifting of trinagles
// ~ cliques (heuristic)
// ~ edge projection
// ~ mod-k cuts
// ~ local cuts
// ~ exact clique separation
// ~ large rank inequality
//
// Remark: The check if all computed inequalities are valid is done
//         at the end of this program in StableSetMain.
// Remark: The edge inequalities are separated in the function feasible().
//
// -----------------------------------------------------------------------------
int StableSetSub::separate() {
 
    int i, j, k;		    // Loop counter.
    int numberOfGeneratedCuts = 0;  // Store the number of added cuts.
    int newCuts = 0;
    int num;			    // Number of cuts.
    int number = -1;
    int numPrevious;
    int error;
    static int largeCutCnt = 0;

    // Parameter.
    static int MIN_NUMBER_OF_CUTS = (int) itsGraph->numberOfNodes() / 10;
    static double EPS = 0.000001;
 
 
    // Create LP solution instance: needed for the separation classes.
    StableSetLPSolution *currentLPSolution;
    currentLPSolution = new StableSetLPSolution(master_, this,
    						itsGraph->numberOfNodes(),
    						xVal_);

      
    
    //  
    // If the edge inequalities are not satisfied separate cliques and check pool.
    //
    if (violatedEdge) {
	violatedEdge = false;

        //
        // Heuristic clique separation.
        //
        if (stableSetMaster()->m_CliqueHeuristicsSeparation) {

        	ofstream fout(itsGraph->getFileName(), ios::app);
        	fout << "\n* Clique heuristics:" << endl;
        	fout.close();

           	// Create new object.
        	CliqueHeuristicsSeparator *cliqueHeuristicsSep;
		cliqueHeuristicsSep = new CliqueHeuristicsSeparator(
						       currentLPSolution,
						       itsGraph);
		// Separate.
		cliqueHeuristicsSep->separate();

		// Add inequalities.
		numPrevious = stableSetMaster()->cliqueCutPool()->number();
        	number = addCons(cliqueHeuristicsSep->cutBuffer(),
		      stableSetMaster()->cliqueCutPool());
		number = stableSetMaster()->cliqueCutPool()->number();
		number -= numPrevious;
        	num = number >= 0 ? number : num;
		newCuts += num;
		numberOfGeneratedCuts =+ newCuts;

		// Output.
        	master_->out() << "\tCliques:\t\t" << num << endl;

		// Update statistics.
		theStatistics->cliqueHeuristicsCut(num);

		// Clean up.
		delete cliqueHeuristicsSep;
                delete currentLPSolution;
    	}


        //
        // Constraint pool separation.
        //
        if (stableSetMaster()->m_PoolSeparation) {

                number = constraintPoolSeparation(0,
    					 stableSetMaster()->cycleCutPool());
       		number += constraintPoolSeparation(0,
    					 stableSetMaster()->cliqueCutPool());
		number += constraintPoolSeparation(0,
    					 stableSetMaster()->rankCutPool());
//    		number += constraintPoolSeparation(0,
//				         stableSetMaster()->generalCutPool());
            
		// Update statistics.
        	theStatistics->poolSeparationCut(number);
    	}
	
        // Return.    
	return newCuts;    
    } 
    
     
   
    //
    // Odd cycle separation.
    //
    if (stableSetMaster()->m_oddCycleSeparation) {

        ofstream fout(itsGraph->getFileName(), ios::app);
        fout << "\n* Odd Cycles:" << endl;

	// Create new object.
        OddCyclesSeparator *oddCyclesSep;
	oddCyclesSep = new OddCyclesSeparator(master_, currentLPSolution,
					      itsGraph, false);
	// Separate.
	oddCyclesSep->separate();

	fout.close();
	
	// Add inequalities.
	// Add cuts to the pool.
	numPrevious = stableSetMaster()->cycleCutPool()->number();
        number = addCons(oddCyclesSep->cutBuffer(),
		      stableSetMaster()->cycleCutPool());
	number = stableSetMaster()->cycleCutPool()->number();
	number -= numPrevious;
        num = number >= 0 ? number : num;
	numberOfGeneratedCuts += num;
	newCuts += num;

	// Output.
        master_->out() << "\tOdd cycles:\t\t" << num << endl;

	// Update statistics.
	theStatistics->oddCyclesCut(num);

	// If the odd cycle separation found 3-cycles than they are saved
	// so we can increase them to max cliques.
        if (oddCyclesSep->getSizeOfClique() != 0) {

            ofstream fout(itsGraph->getFileName(), ios::app);
            fout << "\n* Lifted triangles:" << endl;
            fout.close();

            num = increaseToMaximalClique(oddCyclesSep);
	    numberOfGeneratedCuts += num;
	    newCuts += num;

	    // Output.
	    master_->out() << "\tLifted odd cycles:\t" << num << endl;

	    // Update statistics.
	    theStatistics->liftedOddCyclesCut(num);
	}

	// Clean up.
	delete oddCyclesSep;
    }


    //
    // Heuristic clique separation.
    //
    if (stableSetMaster()->m_CliqueHeuristicsSeparation) {

        ofstream fout(itsGraph->getFileName(), ios::app);
        fout << "\n* Clique heuristics:" << endl;
        fout.close();

        // Create new object.
        CliqueHeuristicsSeparator *cliqueHeuristicsSep;
	cliqueHeuristicsSep = new CliqueHeuristicsSeparator(
						currentLPSolution,
						itsGraph);
	// Separate.
	cliqueHeuristicsSep->separate();

	// Add inequalities.
	numPrevious = stableSetMaster()->cliqueCutPool()->number();
        number = addCons(cliqueHeuristicsSep->cutBuffer(),
		      stableSetMaster()->cliqueCutPool());
	number = stableSetMaster()->cliqueCutPool()->number();
	number -= numPrevious;
        num = number >= 0 ? number : num;
	numberOfGeneratedCuts += num;
	newCuts += num;

	// Output.
        master_->out() << "\tCliques:\t\t" << num << endl;

	// Update statistics.
	theStatistics->cliqueHeuristicsCut(num);

	// Clean up.
	delete cliqueHeuristicsSep;
    }

    if (newCuts >= MIN_NUMBER_OF_CUTS) {
	// Clean up.
        delete currentLPSolution;
	// End separation, if enough cuts are generated.
	return numberOfGeneratedCuts;
    }



    //
    // Edge projection.
    //
    if (stableSetMaster()->m_EdgeProjection) {

        ofstream fout(itsGraph->getFileName(), ios::app);
        fout << "\n* Edge projection:" << endl;
        fout.close();

        int cnt = 0;
	bool goon = true;
        while (goon) {
	
  	    // Separate.
            error = theEdgeProjection->separate(xVal_);

	    if (error) {
	        master_->err() << "\nERROR:\n";
	        master_->err() << "StableSetSub::separate():";
	        master_->err() << " There was a problem with the edge";
	        master_->err() << " projection.\n" << endl;
	        master_->err() << "Program aborted." << endl;
	        exit(master_->Fatal);
	    }

            // Get inequalities
            vector< vector<int>* >* inequalities = theEdgeProjection->getLHS();
            vector<int> *rhs = theEdgeProjection->getRHS();

 	    if (inequalities == NULL) {
	        num = 0;
	    }
	    else {
	        // The inequality type.
	        RankConstraint *rankConstr;
                // Add this inequalities.
	        ABA_BUFFER< ABA_CONSTRAINT* > constraint(master_,
	    					     inequalities->size());
                // Create new inequalities for ABACUS
                for (i=0; i<inequalities->size(); i++) {
        	    rankConstr = new RankConstraint(master_,
	    			                    (*inequalities)[i],
				                    (*rhs)[i],
						    itsGraph);
       		    constraint.push(rankConstr);
    	        }// for

	        // Add new inequalities to the pool.
	        numPrevious = stableSetMaster()->rankCutPool()->number();
                number = addCons(constraint, stableSetMaster()->rankCutPool());
	        number = stableSetMaster()->rankCutPool()->number();
	        number -= numPrevious;
	        num = number >= 0 ? number : num;
	        numberOfGeneratedCuts += num;
	        newCuts += num;
	    }// else

	    // Output
	    master_->out() << "\tEdge projection:\t" << num << endl;

	    // Update statistics.
            theStatistics->edgeProjectionCut(num);
	    
	    cnt++;
	    if (level() == 1 && (cnt >= 10 || newCuts >= 4*MIN_NUMBER_OF_CUTS)) {
	        goon = false;
	    }
	    if (level() != 1 && (newCuts >= MIN_NUMBER_OF_CUTS || cnt >= 5)) {
		goon = false;
	    }
	    
        }// while
    }//  if
    if (newCuts >= MIN_NUMBER_OF_CUTS) {
	// Clean up.
        delete currentLPSolution;
	// End separation, if enough cuts are generated.
	return numberOfGeneratedCuts;
    }


    //
    // Constraint pool separation.
    //
    if (stableSetMaster()->m_PoolSeparation) {

        numberOfGeneratedCuts += constraintPoolSeparation(0,
    					 stableSetMaster()->cycleCutPool());
        numberOfGeneratedCuts += constraintPoolSeparation(0,
    					 stableSetMaster()->cliqueCutPool());
	numberOfGeneratedCuts += constraintPoolSeparation(0,
    					 stableSetMaster()->rankCutPool());
        numberOfGeneratedCuts += constraintPoolSeparation(0,
				        stableSetMaster()->generalCutPool());

	// Update statistics.
        theStatistics->poolSeparationCut(numberOfGeneratedCuts);
    }
    if (newCuts > 0) {
	// Clean up.
        delete currentLPSolution;
	// End separation, if enough cuts are generated.
	return numberOfGeneratedCuts;
    }

    //
    //  Maximally violated Mod k.
    //
    if (stableSetMaster()->m_ModKCutsSeparation) {

    	static int MAX_GEN_NR = 100000;

	// Vorbereitungen fuer das Ranking der separierten Constraints
        ABA_STANDARDPOOL<ABA_CONSTRAINT,ABA_VARIABLE> *poolPointer;
        poolPointer = stableSetMaster()->cliqueCutPool();
        ABA_BUFFER<double> *rank = new ABA_BUFFER<double>(master_, MAX_GEN_NR);
        ABA_BUFFER<bool> *keepInPool = new ABA_BUFFER<bool>(master_,MAX_GEN_NR);
        /* If (*keepInPool)[i] is true the i-th added constraint will even
        be stored in the pool if it is not added to the active constraints.*/
        for (int i=0; i<keepInPool->size();i++) {
	    (*keepInPool)[i] = false;
        }
        // Create new object to separate the mod k cuts.
        StableSetModKSeparator *stableSetModKSeparator;
	stableSetModKSeparator = new StableSetModKSeparator(currentLPSolution,
					      itsGraph);

	// Separate.
        stableSetModKSeparator->separate();

	// Rank the inequalities.
//	stableSetModKSeparator->applyRanking(rank);

	// Add the cuts to the pool.
	numPrevious = stableSetMaster()->generalCutPool()->number();
	number = addCons(stableSetModKSeparator->cutBuffer(),
		      stableSetMaster()->generalCutPool());//, keepInPool, rank);
	number = stableSetMaster()->generalCutPool()->number();
	number -= numPrevious;
        num = number >= 0 ? number : num;
	numberOfGeneratedCuts += num;
	newCuts += num;

	// Update statistics.
	theStatistics->modKCut(num);

	// Clean up.
        delete rank;
        delete keepInPool;
        delete stableSetModKSeparator;
    }
    if (newCuts != 0) {
	// Clean up.
        delete currentLPSolution;
	// End separation, if enough cuts are generated.
	return numberOfGeneratedCuts;
    }


    //
    // Local cuts.
    //
    if (stableSetMaster()->m_LocalCutSeparation) {

        ofstream fout(itsGraph->getFileName(), ios::app);
        fout << "\n* Local Cut:" << endl;
        fout.close();

	// Create separator
	LocalCutSeparator *localCuts;
	localCuts = new LocalCutSeparator(master_, currentLPSolution, itsGraph);

	// Separate.
	localCuts->separate();

	// Add the cuts to the pool.
	numPrevious = stableSetMaster()->generalCutPool()->number();
	number = addCons(localCuts->cutBuffer(),
		      stableSetMaster()->generalCutPool());
	number = stableSetMaster()->generalCutPool()->number();
	number -= numPrevious;
        num = number >= 0 ? number : num;
	numberOfGeneratedCuts += num;
	newCuts += num;

	// Output.
	master_->out() << "\tLocal cut:\t" << num << endl;

	// Update statistics.
	theStatistics->localCutCut(1);

	// Clean up.
	delete localCuts;
    }
    if (newCuts != 0) {
	// Clean up.
        delete currentLPSolution;
	// End separation, if enough cuts are generated.
	return numberOfGeneratedCuts;
    }


    //
    // Check all computed cliques.
    //
    if (stableSetMaster()->m_maxCliqueSeparation) {

        // Store the cliques to be violated.
	vector<int*> cliques;

	// Check inequalities.
	num = maxCliques->checkMaxCliques(&cliques, xVal_);

	// Add new constraints.
	if (num != 0) {

            CliqueConstraint *cliqueConstr;
            vector<int> nodesOfClique;

            ofstream fout(itsGraph->getFileName(), ios::app);
	    fout << "\n* Cliques from pool:" << endl;

	    // Add all violated maximal cliques.
            ABA_BUFFER< ABA_CONSTRAINT* > constraint(master_, num);

	    for (i=0; i<num; i++) {
	        nodesOfClique.resize(cliques[i][0]);
	        for (j=0; j<cliques[i][0]; j++) {
		    nodesOfClique[j] = cliques[i][j+1];
	        }
		// New constraint.
		cliqueConstr = new CliqueConstraint(master_, &nodesOfClique,//
						    itsGraph);
	        constraint.push(cliqueConstr);

		// Clean up.
	        nodesOfClique.clear();
	    }

	    // Add cuts to the pool.
	    numPrevious = stableSetMaster()->cliqueCutPool()->number();
            number = addCons(constraint, stableSetMaster()->cliqueCutPool());
	    number = stableSetMaster()->cliqueCutPool()->number();
	    number -= numPrevious;
	    num = number > 0 ? number : num;
	    numberOfGeneratedCuts += num; 
	    newCuts += num;

            fout.close();

	    // Output
	    master_->out() << "\tMaximal cliques from memory:\t" << num << endl;

	    // Update statistics.
            theStatistics->maxCliquesMemoryCut(num);

	}// if
    }


    //
    // Exact clique separation.
    //
    if ((stableSetMaster()->m_maxCliqueSeparation)
        && !(maxCliques->allMaximalCliquesFound())
       ) {

        ofstream fout(itsGraph->getFileName(), ios::app);
        fout << "\n* Exact max clique:" << endl;
        fout.close();
        
        // Create max clique separator.
	MaxCliquesSeparator *maxCliquesSep;
	maxCliquesSep = new MaxCliquesSeparator(master_, currentLPSolution,
						itsGraph, maxCliques);
	// Separate.
	maxCliquesSep->separate();

	// Add the cuts to the pool.
	numPrevious = stableSetMaster()->cliqueCutPool()->number();
	num = addCons(maxCliquesSep->cutBuffer(),
		      stableSetMaster()->cliqueCutPool());
	num = stableSetMaster()->cliqueCutPool()->number();
	num -= numPrevious;
	numberOfGeneratedCuts += num;

	// Output
	master_->out() << "\tExact cliques:\t" << num << endl;

	// Update statistics.
	theStatistics->exactCliquesCut(num);

	// CLean up.
	delete maxCliquesSep;
    }


    //
    // Large rank inequality.
    //
    int MAX_NUMER_OF_LARGE_CUTS = itsGraph->numberOfNodes() / 5;
    if (itsGraph->numberOfNodes() > 100 && level()==1
        && largeCutCnt < MAX_NUMER_OF_LARGE_CUTS) {
	largeCutCnt++;

        master_->out() << "Compute large inequality." << endl;

	// Separate.
        error = theEdgeProjection->computeLargeInequality(xVal_, stableSetMaster());

	if (error) {
	    master_->err() << "\nERROR:\n";
	    master_->err() << "StableSetSub::compteLargeInequality():";
	    master_->err() << " There was a problem with the edge";
	    master_->err() << " projection.\n" << endl;
	    master_->err() << "Program aborted." << endl;
	    exit(master_->Fatal);
	}

        // Get inequalities
        vector< vector<int>* >* inequalities = theEdgeProjection->getLHS();
        vector<int> *rhs = theEdgeProjection->getRHS();

        ofstream fout(itsGraph->getFileName(), ios::app);
        fout << "\n* Large Rank Inequality:";
	fout.close();

	if (inequalities == NULL) {
	    num = 0;
	}
	else {	
	    
            ABA_BUFFER< ABA_CONSTRAINT* > constraint(master_, 1);
            
            // New constraint.
	    RankConstraint *rankConstr;
	    rankConstr = new RankConstraint(master_, (*inequalities)[0],//
					    (*rhs)[0], true, itsGraph);
	    constraint.push(rankConstr);

	    // Add cuts to the pool.
	    numPrevious = stableSetMaster()->rankCutPool()->number();
            number = addCons(constraint, stableSetMaster()->rankCutPool());
	    number = stableSetMaster()->rankCutPool()->number();
	    number -= numPrevious;
            num = number >= 0 ? number : num; 
	    numberOfGeneratedCuts += num;
	    newCuts += num;

	    // Update statistics.
	    theStatistics->largeCutCut(1);    
	}
    }

    // Clean up.
    delete currentLPSolution;

    // Return the number of computed inequalities.
    return numberOfGeneratedCuts;

}// end: int separate()


// -----------------------------------------------------------------------------
// g e n e r a t e B r a n c h i n g R u l e s
//
// Branch according to Bals and Yu, 1986. Use a covering of inequalities from
// the pool to get alpha.
// -----------------------------------------------------------------------------
int StableSetSub::generateBranchRules(ABA_BUFFER<ABA_BRANCHRULE*> &rules) {

    int i, j, k;
    int itsDegree;
    int size = itsGraph->numberOfNodes();
    int alpha = (int) master_->lowerBound();
    
    vector<int> freeVariables;
    vector<int> L;

    vector<NodesAndDegree> storeNodes;
    NodesAndDegree a;
    ABA_FSVARSTAT * status;

    if (level() == 1) {
	// Store dual bound.
	theStatistics->setDualBound(lp()->value());
	// Store weight of best stable set found so far.
        theStatistics->setAlphaRoot((int) master_->primalBound());
        // Change tailing off.
        master_->tailOffNLp(2);
        master_->tailOffPercent(0.01);
    }


    //
    // Get all nodes which are NOT fixed and calculate alpha tilde.
    //
    for (i=0; i<size; i++) {
        status = fsVarStat(i);
	if (status->status() == 0) {
	    freeVariables.push_back(i);
	}
	else {
	    if (status->status() == 3 || status->status() == 6) {
	        alpha -= 1;
	    }
	}
    }

    
    
    if (freeVariables.empty()) {
    cout << "No free variables" << endl;
	return 1;
    }

    //
    // Calculate degree of each node and sort.
    //
    vector<int> * list;
    for (i=freeVariables.size()-1; i>=0; i--) {
    	itsDegree = 0;
	list = itsGraph->getAdjacentList(freeVariables[i]);
	for (j=itsGraph->adjacentListSize(freeVariables[i])-1; j>=0; j--) {
	    k=0;
	    while (k<freeVariables.size()) {
	        if ((*list)[j] <= freeVariables[k]) {
	  	    break;
		}
		k++;
	    }
	    if (k < freeVariables.size()) {
		if ((*list)[j] == freeVariables[k]) {
		    itsDegree++;
		}
	    }
	}

	a.inode = freeVariables[i];
	a.degree = itsDegree;
	storeNodes.push_back(a);
    }// for

    // Sort in (non-)ascending order of degree.
    sort(storeNodes.begin(), storeNodes.end());
        

    //
    //  Construct from the computed inequalities of the pool a good covering.
    //
    ABA_STANDARDPOOL<ABA_CONSTRAINT,ABA_VARIABLE> *poolPointer;
    int nConstraints;
    ABA_CONSTRAINT *constrPointer;
    ABA_ROW *row = new ABA_ROW(stableSetMaster(), itsGraph->numberOfNodes());
    vector<int> W;	
    double bestWeight;
    double itsWeight;
    int constrSlot;
    int constrType;
    int index;
    int numberOfVariablesInConstr;
    int chiBar = 0;
    int nodeToCover;
    
    //
    // Make it big...
    //
    while ((chiBar < alpha) && !storeNodes.empty()) {
	bestWeight = -1;
	constrSlot = -1;
	constrType = -1;
	
	nodeToCover = storeNodes.back().inode;
	storeNodes.pop_back();
	W.push_back(nodeToCover);

	
	// Check first the clique Cut pool.
	poolPointer = stableSetMaster()->cliqueCutPool();
        // Number of conatraints in the pool.
        nConstraints = poolPointer->number();

        for (i=0; i<nConstraints; i++){
            // Get next constraint.
	    constrPointer = (ABA_CONSTRAINT *)(( poolPointer->slot(i) )->conVar());
	    if (constrPointer == NULL) {
	        continue;
	    }
            if (((int) constrPointer->rhs() + chiBar) > alpha ) {
	         continue;
	    } 
	    
	    // Get information of this inequality.
  	    itsWeight = 0;    
	    numberOfVariablesInConstr = constrPointer->genRow(actVar(), *row);
	        
	    // Check if the node to cover is in that inequality.
	    j = 0;
	    while (j<numberOfVariablesInConstr) {
	        if (nodeToCover <= row->support(j)) {
		    break;
		}
		j++;
	    }
	    if (j==numberOfVariablesInConstr) {
	        row->clear();
		continue;	    
	    }
	    if (nodeToCover != row->support(j)) {
	        row->clear();
		continue;
	    }

	    for (j=0; j<numberOfVariablesInConstr; j++) {
	        index = row->support(j);
	        // Check if this node is free.
		for (k=0; k<storeNodes.size(); k++) {
		    if (storeNodes[k].inode == index) {
			itsWeight++;
			break;
		    }
		}
	    }
	    row->clear();

 	    if (itsWeight > bestWeight) {
		bestWeight = itsWeight;
		constrSlot = i;
		constrType = 0;
	    }
            
	}
	
	// Check next the rank cut pool.
	poolPointer = stableSetMaster()->rankCutPool();
        // Number of conatraints in the pool.
        nConstraints = poolPointer->number();

        for (i=0; i<nConstraints; i++){
            // Get next constraint.
	    constrPointer = (ABA_CONSTRAINT *)(( poolPointer->slot(i) )->conVar());
	    if (constrPointer == NULL) {
	        continue;
	    }
            if (((int) constrPointer->rhs() + chiBar) > alpha ) {
	         continue;
	    } 
	    
    	    // Get information of this inequality.
  	    itsWeight = 0;    
	    numberOfVariablesInConstr = constrPointer->genRow(actVar(), *row);

	    
	    // Check if the node to cover is in that inequality.
	    j = 0;
	    while (j<numberOfVariablesInConstr) {
	        if (nodeToCover <= row->support(j)) {
		    break;
		}
		j++;
	    }
	    if (j==numberOfVariablesInConstr) {
	        row->clear();
		continue;	    
	    }
	    if (nodeToCover != row->support(j)) {
	        row->clear();
		continue;
	    }
	    
	    


	    for (j=0; j<numberOfVariablesInConstr; j++) {
	        index = row->support(j);
	        // Check if this node is free.
		for (k=0; k<storeNodes.size(); k++) {
		    if (storeNodes[k].inode==index) {
			itsWeight++;
			break;
		    }
		}
	    }
	    row->clear();
	    
 	    itsWeight = itsWeight / constrPointer->rhs();	    

 	    if (itsWeight > bestWeight) {
		bestWeight = itsWeight;
		constrSlot = i;
		constrType = 1;
	    }
	}
	
	
        if (bestWeight == -1) {
	    W.push_back(nodeToCover);
	    chiBar += 1;
 	    break;
	}	

		
	// Get this best inequality.
	if (constrType == 0) {
	    poolPointer = stableSetMaster()->cliqueCutPool();	
	}
	else {
	    poolPointer = stableSetMaster()->rankCutPool();
	} 
	constrPointer = (ABA_CONSTRAINT *)(( poolPointer->slot(constrSlot) )->conVar());
	
	numberOfVariablesInConstr = constrPointer->genRow(actVar(), *row);  
	for (j=0; j<numberOfVariablesInConstr; j++) {
	
            index = row->support(j);
	    if (index == nodeToCover) {
		continue;
	    } 
	    // Check if this node is not contained in W.
	    k = 0;
	    while (k<W.size()) {
		if (W[k] == index) {
		    break;
		}	
		k++;
	    }	 
	    if (k == W.size()) {
		W.push_back(index);
		// Delete this node of the list of free nodes.
		storeNodes.size();
	    	for (k=0; k<storeNodes.size(); k++) {
		    if (storeNodes[k].inode == index) {
		        storeNodes.erase(storeNodes.begin() + k);
		        break;
		    }
	        }
	    }
	}
	row->clear();
        
        chiBar += (int) constrPointer->rhs();
    }// while()


    if (chiBar > alpha) {
        cout << "Branching was wrong!" << endl;
	exit(master_->Fatal);
    }

    //
    // Generate subproblems to branch on.
    //
    int node;
    int vertex;
    StableSetBranchRule *newBranchRule;
    int numberOfSubProblems = storeNodes.size();
    master_->out() << "Number of sub problems:\t" << numberOfSubProblems << endl;
    rules.realloc(numberOfSubProblems);
    
    
    for (i=numberOfSubProblems-1; i>=0; i--) {
	node = storeNodes[i].inode;

	// L is neighborhood of node i and set v_j for j > i;
	for (j=0; j<i; j++) {
	    L.push_back(storeNodes[j].inode);
	}
	for (j=0; j<itsGraph->adjacentListSize(node); j++) {
	    vertex = itsGraph->adjacentListElement(node, j);
	    // Check if this node is already contained in the set.
	    for (k=0; k<L.size(); k++) {
		if (vertex == L[k]) {
		    break;
		}
		if (k == L.size()-1) {
		    L.push_back(vertex);
		}
	    }
	    if (L.empty()) {
		L.push_back(vertex);
	    }
	}

	// Generate new sub problem.
	newBranchRule = new StableSetBranchRule(master_, node, &L);
	// Store it.
	rules.push(newBranchRule);
	//CLean up.
	L.clear();

    }
  
    // Clean up.  
    delete row;

    return 0;

}// generateBranchingRules(..)


// -----------------------------------------------------------------------------
// s e t V e r t e x
//
// Set a node to a particular value. This is used when a subproblem is 
// generated in the branching process.
// -----------------------------------------------------------------------------
void StableSetSub::setVertex(int vertex, bool lowerOrUpperBound) {

    int status;
    bool newValue;

    // The function 'set' is a pure virual function of ABACUS. It sets
    // the variable vertex to its upper bound. set() returns value 1 if a
    // contradiction is found and 0 otherwise.
    switch(lowerOrUpperBound) {
    case 0: 
	    status = set(vertex, ABA_FSVARSTAT::SetToLowerBound, newValue);
	    break;

    case 1: status = set(vertex, ABA_FSVARSTAT::SetToUpperBound, newValue);
	    break;
    }
    if (status == 1) {
	master_->err() << "StableSetSub::setVertex(..): ";
	master_->err() << "Setting of variable " << vertex;
	master_->err() << " to value " << lowerOrUpperBound;
	master_->err() << " failed." << endl;
	exit(master_->Fatal);
    }
}



// -----------------------------------------------------------------------------
// --------------- M e t h o d s  ( p r i v a t e ) ----------------------------
// -----------------------------------------------------------------------------


// -----------------------------------------------------------------------------
// i n c r e a s e T o M a x i m a l C l i q u e
//
// This function increases all 3-cycles found by the odd cycle separation to
// maximal cliques. To do this it first constructs the 3-cycles from the data
// of the odd cycle separator and finds out all adjazent nodes to thsi cycle.
// After this the algorithm adds step by step a node and updates the list of
// adjazent nodes untill no more nodes are left.
// -----------------------------------------------------------------------------
int StableSetSub::increaseToMaximalClique(OddCyclesSeparator *oddCyclesSep) {

    int  i, j, k;	// Loop counter.
    int  min;		// Minimal size of an adjazens list.
    int  index;		// Index of the adjazens list with minimal size.
    int  size = oddCyclesSep->getSizeOfClique();
    int  node;		// Store a vertex.
    int  begin[3];	// Remember the indizes to begin the search for the
    			// adjazent nodes.
    bool found;		// Flag to show if something is found or not.
    bool newNode;	// The value is true if a new node is found.
    bool fixed;		// This is true if all variables could be fixed.
    vector<int> clique(3);     // Store the clique.
    vector<int> adjazentNodes; // List of adjazent nodes to the vector clique.
    ABA_BUFFER<ABA_CONSTRAINT*> constraint(master_, (int)size/3);
    CliqueConstraint *cliqueConstr;


    // Get all nodes of the cycles and construct the 3-cycles out of them.
    for (i=0; i<size; i++) {
        node = oddCyclesSep->getCliqueElement(i);

        if ((i+1) % 3 != 0) {
	    clique[i % 3] = node;
	}
	else {
	    // The clique stores now a 3-cycle.
	    clique[2] = node;

	    sort(clique.begin(), clique.end());

	    // Set the indizes to zero.
            for (j=0; j<3; j++) {
		begin[j] = 0;
            }


	    // Get node with smallest size of adjazens list.
	    min   = itsGraph->adjacentListSize(clique[0]);
	    index = 0;
	    for (j=1; j<3; j++) {
		if (itsGraph->adjacentListSize(clique[j]) < min) {
		    min   = itsGraph->adjacentListSize(clique[j]);
		    index = j;
		}
	    }

	    // Get all adjazent nodes to the node of the 3-cycle with the
	    // smallest number of adjazent nodes. Store this nodes only
	    // if they are also adjazent to the two other nodes of the 3-cycle.
	    for (j=0; j<min; j++) {
		// Get node of adjazens list.
		node = itsGraph->adjacentListElement(clique[index], j);

		// Check if it is not one node of the 3-cycle.
		newNode = true;
		for(k=0; k<3; k++) {
		    if ((k != index) && (node == clique[k])) {
		        newNode = false;
			break;
		    }
		}

		if (newNode) {
		    found = true;
		    // Now we know that this node is not a member of the
		    // 3-cycle and so we check if it is also a neighbour of
		    // the two other nodes of the clique.
		    for(k=0; k<3; k++) {
		        if(k == index) {
			    continue;
			}
			if(!itsGraph->isMemberOfAdjacentList(clique[k],
							 begin[k], node)) {
			    found = false;
			    break;
			}
		    }
		    if(found) {
			adjazentNodes.push_back(node);
		    }
		}
	    }

            // Increase the 3-cycle to a maximal clique
	    // -> "sequential lifting".
	    while (!adjazentNodes.empty()) {

		// Add new node: the last one of the list of adjazent nodes.
		clique.push_back(adjazentNodes.back());
		// Delete last element of this list.
		adjazentNodes.erase(adjazentNodes.end() -1);
		k = 0;
		// Update adjazent list.
		for (j=0; j<adjazentNodes.size(); j++) {
		    index = clique.back();
		    node  = adjazentNodes[j];
		    if (!itsGraph->isMemberOfAdjacentList(index, k, node)) {
			adjazentNodes.erase(adjazentNodes.begin() + j);
			j--;
		    }
		}
	    }// while

	    // Sort the new max clique.
	    sort(clique.begin(), clique.end());

	    // New maximal clique found.
	    cliqueConstr = new CliqueConstraint(master_, &clique, itsGraph);
	    constraint.push(cliqueConstr);

	    clique.clear();
	    clique.resize(3);
	}
    }
    int num = 0;
    int number = -1;
    int numPrevious;

    if (constraint.number() != 0) {
	// add constraints to pool
 	numPrevious = stableSetMaster()->cliqueCutPool()->number();
	number = addCons(constraint, stableSetMaster()->cliqueCutPool());
	number = stableSetMaster()->cliqueCutPool()->number();
	number -= numPrevious;
        num = number > 0 ? number : num;
    }

    return num;

}// end: int increaseToMaximalClique(..)


// -----------------------------------------------------------------------------
// s t a b l e S e t M a s t e r
//
// This function returns a pointer to the corresponding object of the class
// StableSetMaster.
// -----------------------------------------------------------------------------
StableSetMaster *StableSetSub::stableSetMaster() {
   return (StableSetMaster *) master_;
}

