////////////////////////////////////// c++ /////////////////////////////////////
//
//  Module           : StableSetSub.hh
//  Description      : The nodes of the Branchand Cut tree.
//  Author           : Steffen Rebennack
//  Email            : srebenac@ix.urz.uni-heidelberg.de;
//                     steffen.rebennack@web.de
//  Copyright        : (C) 2006 by the University of Heidelberg
//  Created on       : Thu Apr 06 13:39:46 2006
//  Last modified by : -
//  Last modified on : -
//  Update count     : 0
//
////////////////////////////////////////////////////////////////////////////////
//
//  Date	Name		Changes/Extensions
//  ----	----		------------------
//
////////////////////////////////////////////////////////////////////////////////


#ifndef STABLE_SET_SUB_HH
#define STABLE_SET_SUB_HH

#include "abacus/sub.h"
#include <vector>

using namespace std;

class Clique;
class EdgeProjection;
class Graph;
class MaxCliquesSeparator;
class OddCyclesSeparator;
class StableSetBranchRule;
class StableSetMaster;
class StableSetStatistics;



/**
 * The class StableSetSub is derived from the abstract class ABA_SUB and
 * implements the problem specific functions for the optimization of a sub
 * problem - a node of the branch-and-bound tree. In particular, this is the
 * feasibility test for the solution of the LP-relaxation and the separation
 * of cutting planes.
 */
class StableSetSub: public ABA_SUB {

public:

    // -------------------------------------------------------------------------
    // ------------- M e t h o d s  ( p u b l i c ) ----------------------------
    // -------------------------------------------------------------------------
    /**
     * Constructor of the root node of the enumeration tree.
     *
     * @param *master Pointer to the corresponding master of the optimization.
     * @param *theGraph Pointer to the grpah of the problem.
     * @param *theClique Pointer to the class Clique of the master of the
     *                   problem.
     * @param *anEdgeProjection Pointer to the class computing the edge
     *                          projection.
     * @param *aStatistic Pointer to the statistic of the problem.
     */
    StableSetSub(ABA_MASTER *master, Graph *theGraph, Clique *theClique,
		 EdgeProjection *anEdgeProjection,
		 StableSetStatistics *aStatistic);

    /**
     * Constructor for a son of an existing node.
     *
     * @param *master Pointer to the corrsponding master of the optimization.
     * @param *father Pointer to the father in the eumeration tree.
     * @param *branchRule The rule defining the subspace of the solution space
     *                    associated with this node.
     * @param *theGraph Pointer to the grpah of the problem.
     * @param *theClique Pointer to the class Clique of the master of the
     *                   problem.
     * @param *anEdgeProjection Pointer to the class computing the edge
     *                          projection.
     * @param *aStatistic Pointer to the statistic of the problem.
     */
    StableSetSub(ABA_MASTER *master, ABA_SUB *father,
                 ABA_BRANCHRULE *branchRule, Graph *theGraph, Clique *theClique,
		 EdgeProjection *anEdgeProjection,
		 StableSetStatistics *aStatistic);

    /**
     * Destructor.
     */
    virtual ~StableSetSub();

    /**
     * Feasibility check of a LP solution.
     *
     * @return true If the LP solution is incidence vector of a Stable Set
     * @return false Otherwise.
     */
    virtual bool feasible();

    /**
     * It generates a problem specific son of a sub problem.
     *
     * @param *rule The branching rule for the generation of the son.
     * @return Pointer to a son of this subproblem, which is generated
     *         according to the branching rule rule.
     */
    virtual ABA_SUB *generateSon(ABA_BRANCHRULE *rule);

    /**
     * Primal heuristic.
     *
     * @param primalValue Holds the value of the feasible solution, if a better
     *                    solution is found by the heuristic.
     * @return 1 If no better solution is found.
     * @return 0 Otherwise.
     */
    virtual int improve(double &primalValue);

    /**
     * Separation of inequalities.
     *
     * @param void
     * @return The number of generated inequalities.
     */
    virtual int separate();

    /**
     * Generate all sons in this node of the tree.
     *
     * @param &rules Reference to an ABACUS buffer storing all open suproblems
     * @return 1 There are no more subproblems.
     * @return 0 All subproblems have been generated and stored.
     */
    virtual int generateBranchRules(ABA_BUFFER<ABA_BRANCHRULE*> &rules);

    /**
     * Set a vertex to a particular value.
     *
     * @param vertex The node to be set.
     * @param i The value to which the node is set.
     * @return void
     */
    void setVertex(int vertex, bool i);


private:

    // -------------------------------------------------------------------------
    // ------------- M e t h o d s  ( p r i v a t e ) --------------------------
    // -------------------------------------------------------------------------

    /**
     * Increase 3-cycles to a maximal cliques.
     *
     * @param *oddCyclesSep Pointer to the odd cycle separator.
     * @return Number of added constraints.
     */
    int increaseToMaximalClique(OddCyclesSeparator *oddCyclesSep);

    /**
     * The function stableSetMaster().
     *
     * @param void
     * @return Pointer to the corresponding object of the class
     *         StableSetMaster.
     */
    StableSetMaster *stableSetMaster();

    
    // -------------------------------------------------------------------------
    // ------------- V a r i a b l e s  ( p r i v a t e ) ----------------------
    // -------------------------------------------------------------------------

    /**
     * Pointer to the corresponding graph of the problem instance of class
     * StableSetMaster.
     */
    Graph *itsGraph;

    /**
     * Pointer to the maxCliques of the class StableSetMaster.
     */
    Clique *maxCliques;
 
    /**
     * Pointer to the edge projection of the StableSetMaster.
     */
    EdgeProjection *theEdgeProjection;
    
    /**
     * Pointer to an object managing the statistics.
     */
    StableSetStatistics *theStatistics;

    /**
     * Indicate whether an egde inequality is violated or not.
     */
    bool violatedEdge;

};

#endif


